/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis;

import org.activityinfo.analysis.map.viewModel.MapViewModel;
import org.activityinfo.analysis.pivot.viewModel.PivotViewModel;
import org.activityinfo.analysis.pivot.viewModel.ReferencedForm;
import org.activityinfo.model.analysis.AnalysisModel;
import org.activityinfo.model.analysis.AnalysisModelVisitor;
import org.activityinfo.model.analysis.ReportSources;
import org.activityinfo.model.analysis.TableAnalysisModel;
import org.activityinfo.model.analysis.map.MapModel;
import org.activityinfo.model.analysis.pivot.PivotModel;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.permission.PermissionOracle;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.observable.Observable;
import org.activityinfo.promise.Maybe;
import org.activityinfo.store.query.shared.FormSource;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

public class ReportPermissionComputer {

    private final FormSource formSource;

    public ReportPermissionComputer(FormSource formSource) {
        this.formSource = formSource;
    }

    private Observable<Set<ReferencedForm>> computeReferencedForms(ReportSources sources) {
        return FormForestComputer.compute(formSource, sources).transform(formForest -> formForest.getReferencedForms());
    }

    public Observable<Boolean> publishable(ReportSources sources) {
        return computeReferencedForms(sources)
            .join(forms -> computePublishable(forms));
    }

    public Observable<Boolean> computePublishable(Set<ReferencedForm> forms) {

        List<Observable<Boolean>> publishable = new ArrayList<>();
        for (ReferencedForm form : forms) {
            publishable.add(formSource.getDatabase(form.getDatabaseId()).transform(d -> computePublishable(d, form)));
        }

        return Observable.flatten(publishable).transform(list -> list.stream().allMatch(p -> p));
    }

    private static boolean computePublishable(Maybe<VisibleDatabaseTree> maybeDatabase, ReferencedForm form) {
        return maybeDatabase
            .transform(database -> PermissionOracle.canPublishForm(database, form.getFormId()))
            .or(false);
    }


    public Observable<Set<ResourceId>> whichDatabasesCanBeSharedWith(ReportSources sources) {
        return computeReferencedForms(sources)
            .transform(forms -> forms.stream().map(f -> f.getDatabaseId()).collect(Collectors.toSet()))
            .join(databases -> Observable.flatJoin(databases, databaseId -> formSource.getDatabase(databaseId)))
            .transform(trees -> trees
                    .stream()
                    .filter(t -> t.isVisible())
                    .map(t -> t.get())
                    .filter(t -> PermissionOracle.canShareReports(t))
                    .map(t -> t.getDatabaseId())
                    .collect(Collectors.toSet()));
    }


}
