/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis;

import org.activityinfo.analysis.pivot.viewModel.ReferencedForm;
import org.activityinfo.model.analysis.ReportGrantSet;
import org.activityinfo.model.analysis.ReportPermission;
import org.activityinfo.model.database.Role;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.permission.PermissionOracle;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.observable.Observable;
import org.activityinfo.promise.Maybe;
import org.activityinfo.store.query.shared.FormSource;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

public class ShareReportViewModel {

    private final Observable<List<VisibleDatabaseTree>> sourceDatabases;
    private final ResourceId reportId;
    private final Observable<Boolean> published;
    private final Observable<ReportGrantSet> grantSet;
    private final Observable<List<ReportRoleGrantItem>> roles;
    private final Observable<Boolean> publishable;
    private final Observable<Boolean> shared;
    private final Observable<Boolean> shareable;

    public ShareReportViewModel(FormSource source, ResourceId reportId, Observable<Boolean> published, Observable<ReportGrantSet> grantSet, Observable<Set<ReferencedForm>> referencedForms) {
        this.reportId = reportId;
        this.published = published;
        this.grantSet = grantSet;
        this.publishable = referencedForms
            .join(forms -> computePublishable(source, forms));
        this.sourceDatabases = referencedForms
            .transform(forms -> forms.stream().map(ReferencedForm::getDatabaseId).collect(Collectors.toSet()))
            .cacheIfEqual()
            .join(databases -> computeSourceDatabases(source, databases));


        this.shared = grantSet.transform(set -> set.isShared());

        this.roles = Observable.transform(sourceDatabases, grantSet, this::computeRoleList);
        this.shareable = roles.transform(list -> list.stream().anyMatch(r -> r.isAllowed()));
    }

    public Observable<Boolean> getShareable() {
        return shareable;
    }

    public ResourceId getReportId() {
        return reportId;
    }

    public Observable<Boolean> getPublishable() {
        return publishable;
    }

    public Observable<Boolean> getShared() {
        return shared;
    }

    public Observable<Boolean> getPublished() {
        return published;
    }

    private Observable<List<VisibleDatabaseTree>> computeSourceDatabases(FormSource source, Set<ResourceId> databases) {
        return Observable.flatJoin(databases, databaseId -> source.getDatabase(databaseId))
            .transform(trees -> {
                List<VisibleDatabaseTree> visible = new ArrayList<>();
                for (Maybe<VisibleDatabaseTree> tree : trees) {
                    tree.ifVisible(visible::add);
                }
                return visible;
            });
    }

    public static Observable<Boolean> computePublishable(FormSource source, Set<ReferencedForm> forms) {

        List<Observable<Boolean>> publishable = new ArrayList<>();
        for (ReferencedForm form : forms) {
            publishable.add(source.getDatabase(form.getDatabaseId()).transform(d -> computePublishable(d, form)));
        }

        return Observable.flatten(publishable).transform(list -> list.stream().allMatch(p -> p));
    }

    private static boolean computePublishable(Maybe<VisibleDatabaseTree> maybeDatabase, ReferencedForm form) {
        return maybeDatabase
            .transform(database -> PermissionOracle.canPublishForm(database, form.getFormId()))
            .or(false);
    }

    private List<ReportRoleGrantItem> computeRoleList(List<VisibleDatabaseTree> sourceDatabases, ReportGrantSet grantSet) {

        List<ReportRoleGrantItem> list = new ArrayList<>();

        for (VisibleDatabaseTree database : sourceDatabases) {
            boolean allowed = PermissionOracle.canShareReports(database);
            for (Role role : database.getRoles()) {
                ReportPermission permission = grantSet.findPermission(database.getDatabaseId(), role.getId());
                list.add(new ReportRoleGrantItem(database, role, permission, allowed));
            }
        }

        return list;
    }

    public Observable<List<ReportRoleGrantItem>> getRoles() {
        return roles;
    }
}
