/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formula.*;
import org.activityinfo.model.formula.functions.*;
import org.activityinfo.model.formula.resolved.ResolvedCallNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.number.Quantity;
import org.activityinfo.model.type.primitive.TextValue;
import org.activityinfo.model.type.time.EpiWeekType;
import org.activityinfo.model.type.time.FortnightType;
import org.activityinfo.model.type.time.LocalDate;
import org.activityinfo.model.type.time.MonthType;

import java.util.List;
import java.util.Optional;

public class ComparisonPredicate implements RegularPredicate {
    private final ResolvedNode field;
    private final String fieldLabel;
    private final ComparisonOperator operator;
    private final ColumnView value;
    private final FormulaType type;

    public ComparisonPredicate(ResolvedNode field, String fieldLabel, ComparisonOperator operator, ConstantNode value) {
        this.field = field;
        this.fieldLabel = fieldLabel;
        this.operator = operator;
        this.value = value.getValue();
        this.type = value.getType();
    }

    public ComparisonPredicate(ResolvedNode field, String fieldLabel, ComparisonOperator operator, LocalDate value) {
        this.field = field;
        this.fieldLabel = fieldLabel;
        this.operator = operator;
        this.value = ColumnViews.of(value);
        this.type = FormulaType.LOCAL_DATE;
    }

    @Override
    public String getFieldLabel() {
        return fieldLabel;
    }

    @Override
    public ResolvedNode getField() {
        return field;
    }

    @Override
    public PredicateType getPredicateType() {
        if (operator == EqualFunction.INSTANCE) {
            return PredicateType.IS;
        }
        if (operator == NotEqualFunction.INSTANCE) {
            return  PredicateType.ISNOT;
        }
        if (operator == GreaterFunction.INSTANCE) {
            return getPredicateTypeForGreaterFunction();
        }
        if (operator == GreaterOrEqualFunction.INSTANCE) {
            return PredicateType.GREATEROREQUALTO;
        }
        if (operator == LessFunction.INSTANCE) {
            return getPredicateTypeForLessFunction();
        }
        if (operator == LessOrEqualFunction.INSTANCE) {
            return PredicateType.LESSOREQUALTO;
        }
        throw new UnsupportedOperationException();
    }

    private PredicateType getPredicateTypeForLessFunction() {
        if(type == FormulaType.LOCAL_DATE) {
            return PredicateType.BEFORE;
        }
        if (field.getFieldType().isPresent()) {
            if (field.getFieldType().get() instanceof MonthType ||
                field.getFieldType().get() instanceof EpiWeekType ||
                field.getFieldType().get() instanceof FortnightType) {
                return PredicateType.BEFORE;
            }
        }
        return PredicateType.LESSTHAN;
    }

    private PredicateType getPredicateTypeForGreaterFunction() {
        if (type == FormulaType.LOCAL_DATE) {
            return PredicateType.AFTER;
        }
        if (field.getFieldType().isPresent()) {
            if (field.getFieldType().get() instanceof MonthType ||
                field.getFieldType().get() instanceof EpiWeekType ||
                field.getFieldType().get() instanceof FortnightType) {
                return PredicateType.AFTER;
            }
        }
        return PredicateType.GREATERTHAN;
    }

    @Override
    public Optional<FieldValue> getArgument() {
        switch (type) {
            case STRING:
                return Optional.of(TextValue.valueOf(value.getString(0)));
            case NUMBER:
                double quantity = value.getDouble(0);
                return Optional.of(new Quantity(quantity));

            case LOCAL_DATE:
                return Optional.of(LocalDate.parse(value.getString(0)));

        }
        throw new UnsupportedOperationException();

    }


    public ComparisonOperator getOperator() {
        return operator;
    }

    public ConstantNode getValue() {
        return new ConstantNode(type, value);
    }


    public static boolean match(FieldMatcher fieldMatcher, ResolvedNode node, List<FilterPredicate> predicates) {
        ComparisonPredicate comparison = match(fieldMatcher, node);
        if(comparison == null) {
            return false;
        } else {
            predicates.add(comparison);
            return true;
        }
    }

    public static ComparisonPredicate match(FieldMatcher fieldMatcher, ResolvedNode node) {
        if (!(node instanceof ResolvedCallNode)) {
            return null;
        }
        ResolvedCallNode callNode = (ResolvedCallNode) node;
        if (!(callNode.getFunction() instanceof ComparisonOperator)) {
            return null;
        }

        if (fieldMatcher.isField(callNode.getArgument(0))) {
            if (callNode.getArgument(1) instanceof ConstantNode) {

                return new ComparisonPredicate(
                    callNode.getArgument(0),
                    fieldMatcher.labelField(callNode.getArgument(0)),
                    (ComparisonOperator) callNode.getFunction(),
                    ((ConstantNode) callNode.getArgument(1)));
            }

            Optional<LocalDate> dateValue = Formulas.localDateConstant(callNode.getArgument(1).toFormula());
            if(dateValue.isPresent()) {
                return new ComparisonPredicate(
                    callNode.getArgument(0),
                    fieldMatcher.labelField(callNode.getArgument(0)),
                    (ComparisonOperator) callNode.getFunction(),
                    dateValue.get());
            }
        }

        return null;
    }

    @Override
    public <T> T accept(PredicateVisitor<T> visitor) {
        return visitor.visitComparison(this);
    }

    @Override
    public FormulaNode toFormula() {
        if(type == FormulaType.LOCAL_DATE) {
            return new FunctionCallNode(operator, field.toFormula(), Formulas.date(value.getLocalDate(0)));
        } else {
            return new FunctionCallNode(operator, field.toFormula(), new ConstantNode(type, value));
        }
    }
}
