/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FunctionCallNode;
import org.activityinfo.model.formula.resolved.*;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumValue;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

public class EnumPredicate implements RegularPredicate {
    private final ResolvedNode field;
    private final String fieldLabel;
    private final PredicateType predicateType;
    private final EnumItem item;


    public EnumPredicate(ResolvedNode field, String fieldLabel, EnumItem item, PredicateType predicateType) {
        this.item = item;
        this.field = field;
        this.fieldLabel = fieldLabel;
        this.predicateType = predicateType;
    }

    @Override
    public <T> T accept(PredicateVisitor<T> visitor) {
        return visitor.visitEnum(this);
    }

    @Override
    public FormulaNode toFormula() {
        return new CompoundExpr(field.toFormula(), item.getId());
    }

    @Override
    public String getFieldLabel() {
        return fieldLabel;
    }

    @Override
    public ResolvedNode getField() {
        return field;
    }

    @Override
    public PredicateType getPredicateType() {
        return predicateType;
    }

    @Override
    public Optional<FieldValue> getArgument() {
        return Optional.of(new EnumValue(item.getId()));
    }

    public EnumItem getItem() {
        return item;
    }

    public ResourceId getItemId() {
        return item.getId();
    }

    public static boolean match(FieldMatcher fieldMatcher, ResolvedNode node, List<? super EnumPredicate> predicates) {

        if (match(PredicateType.IS, fieldMatcher, node, predicates)) {
            return true;
        }
        ResolvedCallNode negation = Matching.isNegated(node);
        if(negation != null) {
            return match(PredicateType.ISNOT, fieldMatcher, negation.getArgument(0), predicates);
        }
        return false;
    }

    private static boolean match(PredicateType type, FieldMatcher fieldMatcher, ResolvedNode node, List<? super EnumPredicate> predicates) {

        if(!(node instanceof ResolvedReference)) {
            return false;
        }

        ResolvedReference reference = (ResolvedReference) node;
        EnumFieldExtractor extractor = new EnumFieldExtractor();
        ReferenceNode fieldRef = reference.getRoot().accept(extractor);

        if(extractor.enumItem == null) {
            return false;
        }

        // Reconstruct the formula for the field, not the enum reference

        FormulaNode ast = node.toFormula();
        FormulaNode fieldFormula;
        if (ast instanceof CompoundExpr) {
            fieldFormula = ((CompoundExpr) ast).getValue();
        } else {
            fieldFormula = ((FunctionCallNode) ast).getArgument(0);
        }

        ResolvedReference field = new ResolvedReference(fieldFormula, fieldRef);

        predicates.add(new EnumPredicate(field, fieldMatcher.labelField(field), extractor.enumItem.getEnumItem(), type));
        return true;
    }

    private static class EnumFieldExtractor implements ReferenceVisitor<ReferenceNode> {

        private EnumItemReference enumItem = null;


        @Override
        public ReferenceNode visitRecordId(RecordIdNode node) {
            return node;
        }

        @Override
        public ReferenceNode visitFormId(FormIdNode node) {
            return node;
        }

        @Override
        public ReferenceNode visitLastEditTime(LastEditTimeNode node) {
            return node;
        }

        @Override
        public ReferenceNode visitField(FieldReferenceNode node) {
            return node;
        }

        @Override
        public ReferenceNode visitCalculatedField(CalculatedFieldNode field) {
            return field;
        }

        @Override
        public ReferenceNode visitEnumItem(EnumItemReference enumItemReference) {
            this.enumItem = enumItemReference;
            return enumItemReference.getField();
        }

        @Override
        public ReferenceNode visitJoin(JoinQueryNode node) {
            return new JoinQueryNode(
                node.getLeft(),
                node.getRightFormId(),
                node.getRight().accept(this));
        }

        @Override
        public ReferenceNode visitUnion(UnionNode node) {
            return new UnionNode(node.getChildren()
                .stream()
                .map(child -> child.accept(this))
                .collect(Collectors.toList()));
        }

        @Override
        public ReferenceNode visitAggregation(AggregateNode node) {
            return node;
        }
    }


}
