/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.Formulas;
import org.activityinfo.model.formula.GroupNode;
import org.activityinfo.model.formula.functions.OrFunction;
import org.activityinfo.model.formula.resolved.ResolvedCallNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.enumerated.EnumItem;

import java.util.*;
import java.util.stream.Collectors;

public class EnumSetPredicate implements RegularPredicate {

    private final ResolvedNode field;
    private final String fieldLabel;
    private final Set<EnumItem> items;

    public EnumSetPredicate(ResolvedNode field, String fieldLabel, Set<EnumItem> items) {
        assert fieldLabel != null;
        this.field = field;
        this.fieldLabel = fieldLabel;
        this.items = items;
    }

    @Override
    public String getFieldLabel() {
        return fieldLabel;
    }

    @Override
    public ResolvedNode getField() {
        return field;
    }

    @Override
    public PredicateType getPredicateType() {
        return PredicateType.IS;
    }

    @Override
    public Optional<FieldValue> getArgument() {
        return Optional.empty();
    }

    public String getItemString() {
        return items.stream().map(s -> s.getLabel()).sorted().collect(Collectors.joining(", "));
    }

    public Set<ResourceId> getItemSet() {
        return items.stream().map(s -> s.getId()).collect(Collectors.toSet());
    }

    @Override
    public <T> T accept(PredicateVisitor<T> visitor) {
        return visitor.visitEnumSet(this);
    }

    @Override
    public FormulaNode toFormula() {
        return toFormula(field, items.stream().map(i -> i.getId()).collect(Collectors.toSet()));
    }

    public static boolean match(FieldMatcher fieldMatcher, ResolvedNode node, List<FilterPredicate> predicates) {

        List<EnumPredicate> enumPredicates = new ArrayList<>();

        if(!collectReferences(fieldMatcher, node, enumPredicates)) {
            return false;
        }

        Iterator<EnumPredicate> it = enumPredicates.iterator();
        EnumPredicate first = it.next();

        ResolvedNode fieldNode = first.getField();

        // Make sure they are the same field.
        // That is, we want to accept A.a || A.b || A.c, but reject A.a || B.x || C.y
        Set<EnumItem> items = new HashSet<>();
        items.add(first.getItem());

        while(it.hasNext()) {
            EnumPredicate next = it.next();
            if(!next.getField().toFormula().equals(first.getField().toFormula())) {
                return false;
            }
            items.add(next.getItem());
        }

        predicates.add(new EnumSetPredicate(fieldNode, fieldMatcher.labelField(fieldNode), items));

        return true;
    }

    private static boolean collectReferences(FieldMatcher matcher, ResolvedNode node, List<EnumPredicate> predicates) {
        if(EnumPredicate.match(matcher, node, predicates)) {
            return true;
        }

        ResolvedCallNode or = Matching.isBinaryCall(node, OrFunction.INSTANCE);
        if(or == null) {
            return false;
        }

        if(!collectReferences(matcher, or.getArgument(0), predicates)) {
            return false;
        }
        if(!collectReferences(matcher, or.getArgument(1), predicates)) {
            return false;
        }

        return true;
    }

    public static FormulaNode toFormula(ResolvedNode field, Set<ResourceId> itemSet) {
        FormulaNode fieldFormula = field.toFormula();
        List<FormulaNode> conditions = itemSet.stream()
            .map(item -> new CompoundExpr(fieldFormula, item))
            .collect(Collectors.toList());

        if(conditions.size() > 1) {
            return new GroupNode(Formulas.anyTrue(conditions));
        } else {
            return Formulas.anyTrue(conditions);
        }
    }

}

