/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FunctionCallNode;
import org.activityinfo.model.formula.functions.GreaterOrEqualFunction;
import org.activityinfo.model.formula.functions.LessOrEqualFunction;
import org.activityinfo.model.formula.resolved.ResolvedNode;

import java.util.Optional;

/**
 * Describes a range that has an (inclusive) upper bound, a (inclusive) lower bound, both, or neither.
 *
 */
public class FieldBounds {


    public enum Endpoint {
        LOWER,
        UPPER
    }

    private ResolvedNode field;
    private FormulaNode lower;
    private FormulaNode upper;


    public FieldBounds(ResolvedNode field) {
        this.field = field;
    }

    public FieldBounds(RegularPredicate predicate) {
        this.field = predicate.getField();

        if(predicate instanceof RangePredicate) {
            RangePredicate rangePredicate = (RangePredicate) predicate;
            this.lower = rangePredicate.getLower();
            this.upper = rangePredicate.getUpper();

        } else if(predicate instanceof ComparisonPredicate) {
            ComparisonPredicate comparison = (ComparisonPredicate) predicate;
            if(comparison.getOperator() == LessOrEqualFunction.INSTANCE) {
                this.upper = comparison.getValue();
            } else if(comparison.getOperator() == GreaterOrEqualFunction.INSTANCE) {
                this.lower = comparison.getValue();
            }
        }
    }


    private FieldBounds(ResolvedNode field, FormulaNode lower, FormulaNode upper) {
        this.field = field;
        this.lower = lower;
        this.upper = upper;
    }

    public Optional<FormulaNode> getLower() {
        return Optional.ofNullable(lower);
    }

    public Optional<FormulaNode> getUpper() {
        return Optional.ofNullable(upper);
    }

    public FieldBounds update(Optional<FormulaNode> value, Endpoint endpoint) {
        FormulaNode lower = this.lower;
        FormulaNode upper = this.upper;
        if(endpoint == Endpoint.LOWER) {
            lower = value.orElse(null);
        } else {
            upper = value.orElse(null);
        }
        return new FieldBounds(field, lower, upper);
    }

    public Optional<FormulaNode> toFilterFormula() {

        if(lower != null && upper != null) {
            return Optional.of(RangePredicate.toFormula(field.toFormula(), lower, upper));
        }

        if(lower != null) {
            return Optional.of(new FunctionCallNode(GreaterOrEqualFunction.INSTANCE, field.toFormula(), lower));
        }

        if(upper != null) {
            return Optional.of(new FunctionCallNode(LessOrEqualFunction.INSTANCE, field.toFormula(), upper));
        }

        return Optional.empty();
    }
}
