/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.functions.AndFunction;
import org.activityinfo.model.formula.functions.FormulaFunction;
import org.activityinfo.model.formula.functions.OrFunction;
import org.activityinfo.model.formula.resolved.FormulaResolver;
import org.activityinfo.model.formula.resolved.ResolvedCallNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class FilterParser {

    private final FieldMatcher fieldMatcher;
    private boolean parseDisjunctions = true;
    private boolean parseRanges = true;
    private FormulaFunction formulaFunction;

    public FilterParser(FieldMatcher fieldMatcher) {
        this.fieldMatcher = fieldMatcher;
    }

    public FilterParser() {
        this.fieldMatcher = new SimpleFieldMatcher();
    }

    public void setParseDisjunctions(boolean parseDisjunctions) {
        this.parseDisjunctions = parseDisjunctions;
    }

    public void setParseRanges(boolean parseRanges) {
        this.parseRanges = parseRanges;
    }

    public static FilterViewModel parse(FormTree formTree, String formula) {
        return parse(formTree, Optional.of(formula));
    }

    public static FilterViewModel parse(FormTree formTree, Optional<String> formula) {
        return parse(formTree, new SimpleFieldMatcher(), formula);
    }

    public static FilterViewModel parse(FormTree formTree, FieldMatcher fieldMatcher, Optional<String> formula) {
        return formula.map(f -> {
            ResolvedNode root = FormulaResolver.resolve(formTree, f);
            FilterParser parser = new FilterParser(fieldMatcher);
            return parser.parse(root);
        })
        .orElse(FilterViewModel.EMPTY);
    }

    public FilterViewModel parse(ResolvedNode root) {
        List<FilterPredicate> predicates = new ArrayList<>();
        parse(root, predicates);
        return new FilterViewModel(root, predicates);
    }

    private void parse(ResolvedNode node, List<FilterPredicate> predicates) {
        if(!node.isValid()) {
            return;
        }
        if(parseRanges && RangePredicate.match(fieldMatcher, node, predicates)) {
            return;
        }
        if(ComparisonPredicate.match(fieldMatcher, node, predicates)) {
            return;
        }
        if(BlankPredicate.match(fieldMatcher, node, predicates)) {
            return;
        }
        if(StringPredicate.match(fieldMatcher, node, predicates)) {
            return;
        }
        if(EnumPredicate.match(fieldMatcher, node, predicates)) {
            return;
        }
        if(parseRanges && EnumSetPredicate.match(fieldMatcher, node, predicates)) {
            return;
        }

        if(matchConjuction(node, predicates)) {
            return;
        }
        predicates.add(new IrregularPredicate(node));
    }

    private boolean matchConjuction(ResolvedNode node, List<FilterPredicate> predicates) {
        if(!(node instanceof ResolvedCallNode)) {
            return false;
        }
        ResolvedCallNode callNode = (ResolvedCallNode) node;
        if(callNode.getFunction() == AndFunction.INSTANCE) {
            if (formulaFunction == null) {
                formulaFunction = AndFunction.INSTANCE;
            }
            if (formulaFunction == AndFunction.INSTANCE) {
                parse(callNode.getArgument(0), predicates);
                parse(callNode.getArgument(1), predicates);
                return true;
            } else {
                return false;
            }
        }
        if(parseDisjunctions && callNode.getFunction() == OrFunction.INSTANCE) {
            if (formulaFunction == null) {
                formulaFunction = OrFunction.INSTANCE;
            }
            if (formulaFunction == OrFunction.INSTANCE) {
                parse(callNode.getArgument(0), predicates);
                parse(callNode.getArgument(1), predicates);
                return true;
            } else {
                return false;
            }
        }
        return false;
    }


}


