/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.Formulas;
import org.activityinfo.model.formula.functions.FormulaFunction;
import org.activityinfo.model.formula.functions.OrFunction;
import org.activityinfo.model.formula.resolved.FormulaResolver;
import org.activityinfo.model.formula.resolved.ResolvedCallNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;

import java.util.*;
import java.util.stream.Collectors;

/**
 * In the table, display the current filter as a series of predicates that are "and-ed" together to
 * form the filter.
 *
 * There are two sorts of bubbles:
 * <ul>
 *     <li>Regular predicates, which are valid and include exactly one FieldReference</li>
 *     <li>Arbitrary predicates, which are valid but cannot be reduced to a regular predicate</li>
 * </ul>
 *
 * <p>We also remove predicates that may have become invalid because a field was deleted. This are not
 * displayed at all or included in the filter.</p>
 *
 */
public class FilterViewModel {

    public static final FilterViewModel EMPTY = new FilterViewModel();

    private List<FilterPredicate> predicates;
    private Set<ResolvedNode> filteredColumns;
    private FormulaFunction formulaFunction;


    private FilterViewModel() {
        this.predicates = Collections.emptyList();
        this.filteredColumns = Collections.emptySet();
    }

    public FilterViewModel(ResolvedNode node, List<FilterPredicate> predicates) {
        if (node.isValid()) {
            if (!(node instanceof ResolvedCallNode)) {
                this.formulaFunction = OrFunction.INSTANCE;
            } else {
                this.formulaFunction = ((ResolvedCallNode) node).getFunction();
            }
        }
        this.predicates = predicates;
        this.filteredColumns = new HashSet<>();
    }

    public static FilterViewModel parse(FormTree formTree, FieldMatcher fieldMatcher, Optional<String> formula) {
        return formula.map(f -> {
            ResolvedNode root = FormulaResolver.resolve(formTree, f);
            FilterParser parser = new FilterParser(fieldMatcher);
            parser.setParseDisjunctions(false);

            return parser.parse(root);
        })
        .orElse(FilterViewModel.EMPTY);
    }

    public static FilterViewModel parseWithoutRanges(FormTree formTree, FieldMatcher fieldMatcher, ResolvedNode formula) {
        FilterParser parser = new FilterParser(fieldMatcher);
        parser.setParseRanges(false);
        return parser.parse(formula);
    }

    public boolean isActive() {
        return !predicates.isEmpty();
    }

    public List<FilterPredicate> getPredicates() {
        return predicates;
    }

    public FormulaFunction getFormulaFunction() {
        return formulaFunction;
    }


    /**
     * The effective formula, if there is one, omitting any invalid predicates.
     */
    public Optional<FormulaNode> getEffectiveFormula() {
        return all(predicates.stream().map(FilterPredicate::toFormula).collect(Collectors.toList()));
    }

    /**
     * Composes a new filter formula without the given the predicate
     */
    public Optional<FormulaNode> removePredicate(FilterPredicate condition) {
        List<FormulaNode> retained = predicates.stream()
                .filter(p -> p != condition)
                .map(FilterPredicate::toFormula)
                .collect(Collectors.toList());

        return all(retained);
    }

    private Optional<FormulaNode> all(List<FormulaNode> retained) {
        if(retained.isEmpty()) {
            return Optional.empty();
        } else {
            return Optional.of(Formulas.allTrue(retained));
        }
    }

    /**
     * Composes a new filter formula, updating the predicate for a given field. If there is already
     * a {@link RegularPredicate} for that field, it will be replaced.
     */
    public Optional<FormulaNode> updateFieldPredicate(ResolvedNode field, Optional<FormulaNode> updatedCondition) {
        List<FormulaNode> retained = new ArrayList<>();
        boolean replaced = false;
        for (FilterPredicate predicate : predicates) {
            if(isRegularPredicateFor(predicate, field)) {
                if(!replaced) {
                    updatedCondition.ifPresent(c -> {
                        retained.add(c);
                    });
                }
                replaced = true;
            } else {
                retained.add(predicate.toFormula());
            }
        }
        if(!replaced) {
            updatedCondition.ifPresent(n -> {
                retained.add(n);
            });
        }
        return all(retained);
    }

    /**
     * Returns {@code true} if the given {@code predicate} is a {@link RegularPredicate}
     * for the given {@code field}
     */
    private boolean isRegularPredicateFor(FilterPredicate predicate, ResolvedNode field) {
        if(predicate instanceof RegularPredicate) {
            RegularPredicate regularPredicate = (RegularPredicate) predicate;
            if(regularPredicate.getField().toFormula().equals(field.toFormula())) {
                return true;
            }
        }
        return false;
    }

    public Optional<RegularPredicate> findFieldPredicate(ResolvedNode field) {
        for (FilterPredicate predicate : predicates) {
            if(isRegularPredicateFor(predicate, field)) {
                return Optional.of((RegularPredicate)predicate);
            }
        }
        return Optional.empty();
    }

    /**
     * Looks for a {@link RegularPredicate} for the given {@code field} that forms a bounds, either two-ended
     * such as {@link RangePredicate}, or one-ended with {@link ComparisonPredicate} with the operator &gt;= or
     * &lt;=.
     *
     */
    public FieldBounds findFieldBounds(ResolvedNode field) {
        return findFieldPredicate(field)
                .map(FieldBounds::new)
                .orElseGet(() -> new FieldBounds(field));
    }
}
