/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formula.ConstantNode;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FunctionCallNode;
import org.activityinfo.model.formula.functions.AndFunction;
import org.activityinfo.model.formula.functions.GreaterOrEqualFunction;
import org.activityinfo.model.formula.functions.LessOrEqualFunction;
import org.activityinfo.model.formula.resolved.ResolvedCallNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.time.LocalDate;

import java.util.List;
import java.util.Optional;

public class RangePredicate implements RegularPredicate {

    private final ResolvedNode field;
    private final String fieldLabel;
    private final ConstantNode lower;
    private final ConstantNode upper;

    public RangePredicate(ResolvedNode field, String fieldLabel, ConstantNode lower, ConstantNode upper) {
        this.field = field;
        this.fieldLabel = fieldLabel;
        this.lower = lower;
        this.upper = upper;
    }

    @Override
    public String getFieldLabel() {
        return fieldLabel;
    }

    @Override
    public ResolvedNode getField() {
        return field;
    }

    @Override
    public PredicateType getPredicateType() {
        return PredicateType.BETWEEN;
    }

    @Override
    public Optional<FieldValue> getArgument() {
        throw new UnsupportedOperationException();
    }

    public ConstantNode getLower() {
        return lower;
    }

    public ConstantNode getUpper() {
        return upper;
    }


    public LocalDate getLowerDate() {
        return LocalDate.parse(lower.getStringValue());
    }

    public LocalDate getUpperDate() {
        return LocalDate.parse(upper.getStringValue());
    }

    public static boolean match(FieldMatcher fieldMatcher, ResolvedNode node, List<FilterPredicate> predicates) {
        if(!(node instanceof ResolvedCallNode)) {
            return false;
        }
        ResolvedCallNode conjunction = (ResolvedCallNode) node;
        if(conjunction.getFunction() != AndFunction.INSTANCE) {
            return false;
        }

        ComparisonPredicate x = ComparisonPredicate.match(fieldMatcher, conjunction.getArgument(0));
        ComparisonPredicate y = ComparisonPredicate.match(fieldMatcher, conjunction.getArgument(1));

        if(x == null || y == null || !x.getField().toFormula().equals(y.getField().toFormula())) {
            return false;
        }

        ConstantNode lower;
        ConstantNode upper;

        if(x.getOperator() == LessOrEqualFunction.INSTANCE && y.getOperator() == GreaterOrEqualFunction.INSTANCE) {
            lower = y.getValue();
            upper = x.getValue();
        } else if(x.getOperator() == GreaterOrEqualFunction.INSTANCE && y.getOperator() == LessOrEqualFunction.INSTANCE) {
            lower = x.getValue();
            upper = y.getValue();
        } else {
            return false;
        }

        predicates.add(new RangePredicate(x.getField(), x.getFieldLabel(), lower, upper));

        return true;
    }


    @Override
    public <T> T accept(PredicateVisitor<T> visitor) {
        return visitor.visitRange(this);
    }

    @Override
    public FormulaNode toFormula() {
        return toFormula(field.toFormula(), lower, upper);
    }

    public static FormulaNode toFormula(FormulaNode field, FormulaNode lower, FormulaNode upper) {
        return new FunctionCallNode(AndFunction.INSTANCE,
                new FunctionCallNode(GreaterOrEqualFunction.INSTANCE, field, lower),
                new FunctionCallNode(LessOrEqualFunction.INSTANCE, field, upper));
    }

}
