/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.filter;

import org.activityinfo.model.formula.ConstantNode;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FunctionCallNode;
import org.activityinfo.model.formula.functions.*;
import org.activityinfo.model.formula.resolved.ResolvedCallNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.primitive.TextValue;

import java.util.List;
import java.util.Optional;

public class StringPredicate implements RegularPredicate {
    private final ResolvedNode field;
    private final String fieldLabel;
    private final String string;
    private final PredicateType predicateType;

    public StringPredicate(ResolvedNode field, String fieldLabel, String string, PredicateType predicateType) {
        this.string = string;
        this.field = field;
        this.fieldLabel = fieldLabel;
        this.predicateType = predicateType;
    }


    @Override
    public String getFieldLabel() {
        return fieldLabel;
    }

    @Override
    public ResolvedNode getField() {
        return field;
    }

    @Override
    public PredicateType getPredicateType() {
        return predicateType;
    }

    @Override
    public Optional<FieldValue> getArgument() {
        return Optional.of(TextValue.valueOf(string));
    }

    @Override
    public <T> T accept(PredicateVisitor<T> visitor) {
        return visitor.visitContainsString(this);
    }

    @Override
    public FormulaNode toFormula() {
        return toFormula(field.toFormula(), string);
    }

    public String getString() {
        return string;
    }

    public static FormulaNode toFormula(FormulaNode field, String string) {
        return new FunctionCallNode(IsNumberFunction.INSTANCE,
                new FunctionCallNode(SearchFunction.INSTANCE, new ConstantNode(string), field));
    }

    public static boolean match(FieldMatcher fieldMatcher, ResolvedNode node, List<FilterPredicate> predicates) {
        if(!(node instanceof ResolvedCallNode)) {
            return false;
        }
        ResolvedCallNode rootCall = (ResolvedCallNode) node;
        if (rootCall.getFunction() == EqualFunction.INSTANCE) {
            return matchStartsWithEndsWith(fieldMatcher, rootCall, predicates);
        }
        if(rootCall.getFunction() != IsNumberFunction.INSTANCE || rootCall.getArgumentCount() != 1) {
            return false;
        }
        if(!(rootCall.getArgument(0) instanceof ResolvedCallNode)) {
            return false;
        }
        ResolvedCallNode search = (ResolvedCallNode) rootCall.getArgument(0);
        if(search.getFunction() != SearchFunction.INSTANCE || search.getArgumentCount() != 2) {
            return false;
        }
        if(!fieldMatcher.isField(search.getArgument(1))) {
            return false;
        }
        ResolvedNode field = search.getArgument(1);

        if(!(search.getArgument(0) instanceof ConstantNode)) {
            return false;
        }
        ConstantNode string = (ConstantNode) search.getArgument(0);
        String stringValue = string.getStringValue();

        predicates.add(new StringPredicate(field, fieldMatcher.labelField(field), stringValue, PredicateType.CONTAINS));

        return true;
    }

    private static boolean matchStartsWithEndsWith(FieldMatcher fieldMatcher, ResolvedCallNode node, List<FilterPredicate> predicates) {
        if (!(node.getArgument(0) instanceof ResolvedCallNode)) {
            return false;
        }
        if (!(node.getArgument(1) instanceof ConstantNode)) {
            return false;
        }
        ResolvedCallNode a = ((ResolvedCallNode) node.getArgument(0));
        PredicateType predicateType;
        if (a.getFunction() == LeftFunction.INSTANCE) {
            predicateType = PredicateType.STARTSWITH;
        } else if (a.getFunction() == RightFunction.INSTANCE) {
            predicateType = PredicateType.ENDSWITH;
        } else {
            return false;
        }

        ResolvedNode field = a.getArgument(0);
        if (!fieldMatcher.isField(field)) {
            return false;
        }

        String b = ((ConstantNode) node.getArgument(1)).getStringValue();

        if (!a.getArgument(1).equals(new ConstantNode(b.length()))) {
            return false;
        }

        predicates.add(new StringPredicate(field, fieldMatcher.labelField(field), b, predicateType));
        return true;
    }
}
