/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.form;

import org.activityinfo.model.database.RecordLock;
import org.activityinfo.model.database.RecordLockSet;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.*;
import org.activityinfo.model.formula.functions.AndFunction;
import org.activityinfo.model.formula.functions.GreaterOrEqualFunction;
import org.activityinfo.model.formula.functions.LessOrEqualFunction;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.time.*;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

public class LockFormulaBuilder {
    private final FormTree formTree;
    private final RecordLockSet lockSet;
    private List<FormulaNode> conditions;

    public static FormulaNode build(VisibleDatabaseTree databaseTree, FormTree formTree) {
        return new LockFormulaBuilder(formTree, databaseTree.getEffectiveLocks(formTree.getRootFormId())).build();
    }

    public static FormulaNode build(RecordLockSet lockSet, FormTree formTree) {
        return new LockFormulaBuilder(formTree, lockSet).build();
    }

    private LockFormulaBuilder(FormTree formTree, RecordLockSet locks) {
        this.formTree = formTree;
        this.lockSet = locks;
    }

    public FormulaNode build() {

        if(lockSet.isEmpty()) {
            return ConstantNode.FALSE;
        }
        conditions = new ArrayList<>();

        collectConditions(formTree.getRootFormClass());

        return Formulas.anyTrue(conditions);
    }

    private void collectConditions(FormClass formClass) {
        for (FormField field : formClass.getFields()) {
            if (field.isKey()) {
                if (field.getType() instanceof LocalDateType) {
                    addConditions(formClass.getId(), field, Formulas::date);
                } else if (field.getType() instanceof MonthType) {
                    addConditions(formClass.getId(), field, d -> new ConstantNode(d.getMonth().toString()));
                } else if (field.getType() instanceof FortnightType) {
                    addConditions(formClass.getId(), field, d -> new ConstantNode((((FortnightValue) FortnightType.INSTANCE.containingDate(d)).toString())));
                } else if (field.getType() instanceof EpiWeekType) {
                    addConditions(formClass.getId(), field, d -> new ConstantNode(EpiWeek.weekOf(d).toString()));
                } else if (field.getType() instanceof YearType) {
                    addConditions(formClass.getId(), field, d -> new ConstantNode(d.getYear()));
                }
            }
        }
        if (formClass.getParentFormId().isPresent()) {
            formClass = formTree.getFormClass(formClass.getParentFormId().get());
            collectConditions(formClass);
        }
    }

    private void addConditions(ResourceId formId, FormField field, Function<LocalDate, FormulaNode> converter) {
        FormulaNode fieldSymbol = new SymbolNode(field.getId());
        Map<ResourceId, List<RecordLock>> resourceLockMap = lockSet.getResourceLockMap();
        for (ResourceId resourceId : resourceLockMap.keySet()) {
            // add conditions for the locks applied to the form, to the parent folder and the database
            if (resourceId.equals(formId) || !formTree.getFormClassIfPresent(resourceId).isPresent()) {
                List<RecordLock> locks = resourceLockMap.get(resourceId);
                addConditions(converter, fieldSymbol, locks);
            }
        }
    }

    private void addConditions(Function<LocalDate, FormulaNode> converter, FormulaNode fieldSymbol, List<RecordLock> locks) {
        for (RecordLock lock : locks) {
            FormulaNode startDate = converter.apply(lock.getDateRange().getStartDate());
            FormulaNode endDate = converter.apply(lock.getDateRange().getEndDate());

            FormulaNode lower = new FunctionCallNode(GreaterOrEqualFunction.INSTANCE, fieldSymbol, startDate);
            FormulaNode upper = new FunctionCallNode(LessOrEqualFunction.INSTANCE, fieldSymbol, endDate);

            FormulaNode condition = new GroupNode(new FunctionCallNode(AndFunction.INSTANCE, lower, upper));

            conditions.add(condition);
        }
    }
}
