/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.longFormat;

import com.google.common.base.Strings;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import org.activityinfo.model.analysis.pivot.*;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.expr.CalculatedFieldType;
import org.activityinfo.model.type.geo.GeoPointType;
import org.activityinfo.model.type.number.QuantityType;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class LongFormatTableBuilder {

    private LongFormatTableBuilder() {
    }

    public static boolean measureFilter(FormField field) {
        return field.getType() instanceof QuantityType || field.getType() instanceof CalculatedFieldType;
    }

    /**
     * Return true if the field should be used as a dimension.
     */
    public static boolean dimensionFilter(FormField field) {
        return field.getType().accept(DimensionFilter.INSTANCE);
    }

    public static boolean referenceFilter(FormField field) {
        return field.getType() instanceof ReferenceType;
    }

    public static LongFormatModel build(List<FormTree> formScope, Optional<FormulaNode> filter) {
        return new LongFormatModel(
                formScope,
                extractAllMeasures(formScope),
                extractAllDimensions(formScope),
                filter);
    }

    private static List<MeasureModel> extractAllMeasures(List<FormTree> formScope) {
        return formScope.stream()
                .map(FormTree::getRootFormClass)
                .flatMap(LongFormatTableBuilder::extractFormMeasures)
                .collect(Collectors.toList());
    }

    private static Stream<MeasureModel> extractFormMeasures(FormClass form) {
        return form.getFields().stream()
                .filter(LongFormatTableBuilder::measureFilter)
                .map(measureField -> new MeasureModelBuilder()
                        .id(ResourceId.generateCuid())
                        .formId(form.getId())
                        .label(measureField.getLabel())
                        .formula(measureField.getId().asString())
                        .code(Strings.nullToEmpty(measureField.getCode()))
                        .statistics(Statistic.SUM)
                        .build());
    }

    private static List<DimensionModel> extractAllDimensions(List<FormTree> formScope) {
        Multimap<String,DimensionMapping> dimensionGroups = formScope.stream()
                .map(FormTree::getRootFields)
                .flatMap(List::stream)
                .flatMap(LongFormatTableBuilder::mapDimensions)
                .collect(Multimaps.toMultimap(LongFormatColumn::getLabel, LongFormatColumn::getMapping, ArrayListMultimap::create));
        return buildModels(dimensionGroups);
    }

    private static Stream<LongFormatColumn> mapDimensions(FormTree.Node node) {
        if (!isInUserForm(node) && node.isReference() && !node.isSubForm()) {
            return node.getChildren().stream().flatMap(LongFormatTableBuilder::mapDimensions);
        }
        if (dimensionFilter(node.getField())) {
            return Stream.of(mapDimension(node));
        }
        if (node.getField().getType() instanceof GeoPointType) {
            return mapGeoDimension(node);
        }
        return Stream.empty();
    }

    private static boolean isInUserForm(FormTree.Node node) {
        if(VisibleDatabaseTree.isUserForm(node.getDefiningFormClass().getId())) {
            return true;
        } else {
            return false;
        }
    }

    private static LongFormatColumn mapDimension(FormTree.Node node) {
        return new LongFormatColumn(label(node), map(node));
    }

    private static Stream<LongFormatColumn> mapGeoDimension(FormTree.Node node) {
        return Stream.of(
                new LongFormatColumn(label(node, GeoPointType.LATITUDE), geoMap(node, GeoPointType.LATITUDE)),
                new LongFormatColumn(label(node, GeoPointType.LONGITUDE), geoMap(node, GeoPointType.LONGITUDE)));
    }

    private static String label(FormTree.Node node, String suffix) {
        return suffix == null
                ? label(node)
                : label(node) + "." + suffix;
    }

    private static String label(FormTree.Node node) {
        if (node.isRoot()) {
            return node.fieldLabel();
        } else if (node.getParent().isRoot() && node.getParent().getDefiningFormClass().isSubForm()) {
            return node.fieldLabel();
        } else {
            return node.formFieldLabel();
        }
    }

    private static DimensionMapping map(FormTree.Node node) {
        return new DimensionMapping(node.getRootFormClass().getId(), node.getPath().toString());
    }

    private static DimensionMapping geoMap(FormTree.Node node, String pointAxis) {
        return new DimensionMapping(node.getRootFormClass().getId(), node.getPath().toString() + "." + pointAxis);
    }

    private static List<DimensionModel> buildModels(Multimap<String, DimensionMapping> dimensionGroups) {
        return dimensionGroups.asMap().entrySet().stream()
                .sorted(Comparator.comparing(Map.Entry::getKey))
                .map(dimensionGroup -> new DimensionModelBuilder()
                        .id(ResourceId.generateCuid())
                        .label(dimensionGroup.getKey())
                        .mappings(dimensionGroup.getValue())
                        .missingIncluded(true)
                        .axis(Axis.ROW)
                        .build())
                .collect(Collectors.toList());
    }

}
