/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.longFormat;

import com.google.common.base.Charsets;
import org.activityinfo.io.csv.TableWriter;
import org.activityinfo.model.query.ColumnView;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;

/**
 * Writes out the long format export in a UTF-8 encoded text format intended
 * to be consumed by API clients.
 *
 * <p>Rows are delimited with a single new-line character, and fields are seperated using the
 * Unicode 'unit seperator' character.</p>
 */
public class LongFormatTextWriter implements TableWriter  {

    private static final char DELIMITER = '\u001F';

    private static final char LINE_ENDING = '\n';

    private final Writer writer;

    public LongFormatTextWriter(OutputStream outputStream) throws IOException {
        this.writer = new OutputStreamWriter(outputStream, Charsets.UTF_8);
    }

    @Override
    public void setColumnWidth(int column, int width) {
        // NOOP
    }

    @Override
    public void setColumnFormat(int column, String pattern) {
        // NOOP
    }

    @Override
    public void writeLine(Object... columns) throws IOException {
        for (int i = 0; i != columns.length; ++i) {
            if (i > 0) {
                writer.append(DELIMITER);
            }
            Object val = columns[i];
            if (val != null) {
                if (val instanceof String) {
                    writeString((String) val);
                } else if(val instanceof Double) {
                    double number = (double) val;
                    writeDouble(number);
                } else {
                    writer.append(val.toString());
                }
            }
        }
        writer.append(LINE_ENDING);
    }

    public void writeString(String val) throws IOException {
        if(ColumnView.isTruncated(val)) {
            writer.append(val.subSequence(0, val.length() - 1));
            writer.append("...");
        } else {
            writer.append(val);
        }
    }

    public void writeDouble(double number) throws IOException {
        if(Double.isFinite(number)) {
            if(number == Math.floor(number)) {
                writer.append(Integer.toString((int) number));
            } else {
                writer.append(Double.toString(number));
            }
        }
    }

    @Override
    public String toString() {
        return writer.toString();
    }

    @Override
    public void close() throws IOException {
        writer.close();
    }
}
