/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.longFormat;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.io.csv.TableWriter;
import org.activityinfo.model.analysis.pivot.DimensionModel;
import org.activityinfo.model.analysis.pivot.MeasureModel;
import org.activityinfo.model.database.ResourceType;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.query.*;
import org.activityinfo.model.resource.ResourceId;

import java.io.IOException;
import java.time.Instant;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.Optional;
import java.util.function.Function;

public class LongFormatWriter {

    public static final int NUM_BASE_COLUMN_COUNT = 13;
    public static final int FIRST_DIMENSION_INDEX = 9;
    public static final int COLUMN_WIDTH = 20;

    private final TableWriter writer;
    private boolean includeBlanks;
    private final Function<QueryModel, ColumnSet> queryExecutor;

    private int fieldIdColumnIndex;
    private int fieldCodeColumnIndex;
    private int fieldColumnIndex;
    private int valueColumnIndex;
    private int recordColumnIndex;
    private int lastEditTimeColumnIndex;
    private final ZoneOffset offset;

    public LongFormatWriter(TableWriter tableWriter, boolean includeBlanks, Function<QueryModel, ColumnSet> queryExecutor, ZoneOffset offset) {
        this.writer = tableWriter;
        this.includeBlanks = includeBlanks;
        this.queryExecutor = queryExecutor;
        this.offset = offset;
    }

    public void write(VisibleDatabaseTree database, LongFormatModel model) throws IOException {

        String[] headers = new String[columnCount(model)];

        int columnIndex = 0;
        headers[columnIndex++] = "Database";
        headers[columnIndex++] = "DatabaseId";
        headers[columnIndex++] = "Folder";
        headers[columnIndex++] = "Form";
        headers[columnIndex++] = "FormId";
        headers[columnIndex++] = "ParentForm";
        headers[columnIndex++] = "ParentFormId";

        recordColumnIndex = columnIndex++;
        lastEditTimeColumnIndex = columnIndex++;

        headers[recordColumnIndex] = I18N.CONSTANTS.recordIdAlternate();;
        headers[lastEditTimeColumnIndex] = I18N.CONSTANTS.lastEditTimeAlternate();

        for (DimensionModel column : model.getColumns()) {
            headers[columnIndex++] = column.getLabel();
        }

        fieldIdColumnIndex = columnIndex++;
        fieldCodeColumnIndex = columnIndex++;
        fieldColumnIndex = columnIndex++;
        valueColumnIndex = columnIndex++;

        headers[fieldIdColumnIndex] = I18N.CONSTANTS.quantityFieldId();
        headers[fieldCodeColumnIndex] = I18N.CONSTANTS.quantityFieldCode();
        headers[fieldColumnIndex] = I18N.CONSTANTS.quantityField();
        headers[valueColumnIndex] = I18N.CONSTANTS.value();

        for (int i = 0; i < headers.length; i++) {
            writer.setColumnWidth(i, COLUMN_WIDTH);
        }

        writer.writeLine((Object[]) headers);
        writer.setColumnFormat(lastEditTimeColumnIndex, "yyyy-mm-dd hh:mm:ss");

        for (FormTree formTree : model.getForms()) {
            try {
                writeForm(database, formTree, model);
            } catch (Exception e) {
                throw new RuntimeException("Exception while writing form " + formTree.getRootFormId(), e);
            }
        }
    }

    public int columnCount(LongFormatModel model) {
        return NUM_BASE_COLUMN_COUNT + model.getColumns().size();
    }

    private void writeForm(VisibleDatabaseTree database, FormTree formTree, LongFormatModel model) throws IOException {

        FormClass formClass = formTree.getRootFormClass();

        Optional<FormClass> parentForm = formClass
            .getParentFormId()
            .flatMap(formTree::getFormClassIfPresent);
        
        Object[] buffer = new Object[columnCount(model)];
        buffer[0] = database.getLabel();
        buffer[1] = database.getDatabaseId().asString();
        buffer[2] = findFolder(database, formClass.getId());
        buffer[3] = formClass.getLabel();
        buffer[4] = formClass.getId().asString();
        buffer[5] = parentForm.map(p -> p.getLabel()).orElse("");
        buffer[6] = parentForm.map(p -> p.getId().asString()).orElse("");

        QueryModel queryModel = buildQuery(formClass, model);

        ColumnSet columnSet = queryExecutor.apply(queryModel);

        ColumnView idView = columnSet.getColumnView("id");
        ColumnView timeView = columnSet.getColumnView("lastEditTime");
        ColumnView[] dimensionViews = buildDimensionViewArray(model, columnSet);

        for (MeasureModel measure : model.getMeasures()) {
            if (measure.getFormId().equals(formClass.getId())) {
                buffer[fieldIdColumnIndex] = measure.getFormula();
                buffer[fieldCodeColumnIndex] = measure.getCode();
                buffer[fieldColumnIndex] = measure.getLabel();

                ColumnView measureView = columnSet.getColumnView(measure.getId());

                for (int rowIndex = 0; rowIndex < columnSet.getNumRows(); rowIndex++) {
                    double value = measureView.getDouble(rowIndex);
                    if (!Double.isNaN(value) || includeBlanks) {
                        buffer[recordColumnIndex] = idView.getString(rowIndex);
                        buffer[lastEditTimeColumnIndex] = getLocalDate(timeView.getDouble(rowIndex));
                        buffer[valueColumnIndex] = Double.isNaN(value) ? null : value;

                        for (int dimensionIndex = 0; dimensionIndex < dimensionViews.length; dimensionIndex++) {
                            buffer[FIRST_DIMENSION_INDEX + dimensionIndex] = dimensionViews[dimensionIndex].getString(rowIndex);
                        }
                        writer.writeLine(buffer);
                    }
                }
            }
        }
    }

    private ColumnView[] buildDimensionViewArray(LongFormatModel model, ColumnSet columnSet) {
        ColumnView[] dimensionViews = new ColumnView[model.getColumns().size()];
        for (int i = 0; i < dimensionViews.length; i++) {
            ColumnView columnView = columnSet.getColumnView(model.getColumns().get(i).getId());
            if(columnView == null) {
                columnView = new BlankColumnView(ColumnType.STRING, columnSet.getNumRows());
            }
            dimensionViews[i] = columnView;
        }
        return dimensionViews;
    }

    private String findFolder(VisibleDatabaseTree database, ResourceId formId) {
        return database.getResource(formId)
            .map(r -> r.getParentId())
            .flatMap(parentId -> database.getResource(parentId))
            .map(parent -> {
                if(parent.getType() == ResourceType.FOLDER) {
                    return parent.getLabel();
                } else {
                    return findFolder(database, parent.getId());
                }
            })
            .orElse("");
    }

    private QueryModel buildQuery(FormClass formClass, LongFormatModel model) {

        QueryModel queryModel = new QueryModel(formClass.getId());
        queryModel.selectRecordId().as("id");
        queryModel.selectLastEditTime().as("lastEditTime");
        queryModel.setFilter(model.getFilter());

        for (DimensionModel column : model.getColumns()) {
            column.getMappings().stream().filter(m -> m.getFormId().equals(formClass.getId())).findAny().ifPresent(mapping -> {
                queryModel.selectExpr(mapping.getFormula()).as(column.getId());
            });
        }

        for (MeasureModel measure : model.getMeasures()) {
            if(measure.getFormId().equals(formClass.getId())) {
                queryModel.selectExpr(measure.getFormula()).as(measure.getId());
            }
        }
        return queryModel;
    }

    ZonedDateTime getLocalDate(double seconds) {
        if(Double.isNaN(seconds)){
            return  null;
        }
        Instant instant = Instant.ofEpochSecond((long) seconds);
        ZonedDateTime dateTime = instant.atZone(offset);
        return dateTime;
    }
}
