/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.map.viewModel;

import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FunctionCallNode;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.formula.functions.AverageFunction;
import org.activityinfo.model.formula.functions.BoundingBoxFunction;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.geo.GeoAreaType;
import org.activityinfo.model.type.geo.GeoPointType;

import java.util.*;

public class GeographySource {
    private static final SymbolNode PARENT_SYMBOL = new SymbolNode(ColumnModel.PARENT_SYMBOL);

    private final FormField field;
    private final FormulaNode formula;
    private final String label;
    private FormulaNode latitude;
    private FormulaNode longitude;

    public GeographySource(FormField field, FormulaNode formula, String label) {
        this.field = field;
        this.formula = formula;
        this.label = label;
        if(field.getType() instanceof GeoPointType) {
            latitude = new CompoundExpr(formula, "latitude");
            longitude = new CompoundExpr(formula, "longitude");
        } else {
            latitude =  new FunctionCallNode(AverageFunction.INSTANCE,
                new FunctionCallNode(BoundingBoxFunction.YMIN, formula),
                new FunctionCallNode(BoundingBoxFunction.YMAX, formula));

            longitude =  new FunctionCallNode(AverageFunction.INSTANCE,
                new FunctionCallNode(BoundingBoxFunction.XMIN, formula),
                new FunctionCallNode(BoundingBoxFunction.XMAX, formula));
        }
    }

    public String getLabel() {
        return label;
    }

    public FormulaNode getLatitudeFormula() {
        return latitude;
    }

    public FormulaNode getLongitudeFormula() {
        return longitude;
    }

    public static List<GeographySource> findSources(FormTree formTree) {
        List<GeographySource> list = new ArrayList<>();
        Set<ResourceId> visitedForms = new HashSet<>();
        collectSources(formTree, formTree.getRootFormClass(), Optional.empty(), list, visitedForms);
        return list;
    }

    public FormulaNode getFormula() {
        return formula;
    }

    public String getFormulaString() {
        return formula.deparse();
    }

    private static void collectSources(FormTree formTree, FormClass form, Optional<FormulaNode> prefix, List<GeographySource> list, Set<ResourceId> visitedForms) {

        visitedForms.add(form.getId());

        for (FormField field : form.getFields()) {
            if(field.getType() instanceof GeoPointType) {
                list.add(new GeographySource(field, join(prefix, field.getId()), label(formTree, form, field)));
            }
            if(field.getType() instanceof GeoAreaType) {
                list.add(new GeographySource(field, join(prefix, field.getId()), label(formTree, form, field)));
            }
        }

        form.getParentFormId().ifPresent(parentId -> {
            collectSources(formTree, parentId, visitedForms, join(prefix, PARENT_SYMBOL), list);
        });


        for (FormField field : form.getFields()) {
            if(field.getType() instanceof ReferenceType) {
                ReferenceType referenceType = (ReferenceType) field.getType();
                if(referenceType.getRange().size() == 1) {
                    ResourceId referencedFormId = referenceType.getRange().iterator().next();
                    collectSources(formTree, referencedFormId, visitedForms, join(prefix, new SymbolNode(field.getId())), list);
                }
            }
        }
    }

    private static void collectSources(FormTree formTree, ResourceId formId, Set<ResourceId> visitedForms, FormulaNode prefix, List<GeographySource> list) {

        // Avoid death by circular reference
        if(!visitedForms.add(formId)) {
            return;
        }

        formTree
            .getFormClassIfPresent(formId)
            .ifPresent(parentForm -> collectSources(formTree, parentForm, Optional.of(prefix), list, visitedForms));
    }

    private static String label(FormTree formTree, FormClass form, FormField field) {
        if (form.getId().equals(formTree.getRootFormId())) {
            return field.getLabel();
        } else {
            return form.getLabel() + " — " + field.getLabel();
        }
    }

    private static FormulaNode join(Optional<FormulaNode> prefix, ResourceId id) {
        return prefix
            .map(prefixFormula -> (FormulaNode)new CompoundExpr(prefixFormula, id))
            .orElseGet(() -> new SymbolNode(id));
    }


    private static FormulaNode join(Optional<FormulaNode> prefix, SymbolNode id) {
        return prefix
            .map(prefixFormula -> (FormulaNode)new CompoundExpr(prefixFormula, id))
            .orElse(id);
    }


}
