/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.map.viewModel;

import org.activityinfo.model.analysis.map.Layer;
import org.activityinfo.model.analysis.map.LayerType;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.Formulas;
import org.activityinfo.model.formula.resolved.FieldReferenceNode;
import org.activityinfo.model.formula.resolved.FormulaResolver;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.formula.resolved.ResolvedReference;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.QueryModel;
import org.activityinfo.observable.Observable;
import org.activityinfo.store.query.shared.FormSource;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

public class LayerViewModel {

    public static final String ID_COLUMN_ID = "$$id";
    public static final String LAT_COLUMN_ID = "$$lat";
    public static final String LNG_COLUMN_ID = "$$lng";
    public static final String LAYER_COLUMN_ID = "$$layer";

    private final Layer layer;
    private final List<GeographySource> geographySources;
    private final Optional<GeographySource> selectedGeographySource;
    private final Optional<String> formLabel;
    private final String sourceLabel;
    private final Optional<FormulaType> formulaType;
    private final LayerType layerType;
    private final Observable<ColumnSet> columnSet;
    private final Observable<LayerResult> layerResult;

    public LayerViewModel(FormTree formTree, Layer layer, FormSource formSource) {
        this.layer = layer;
        this.geographySources = GeographySource.findSources(formTree);
        this.selectedGeographySource = findSelectedSource(geographySources, layer.getGeographyFormula());
        this.columnSet = selectedGeographySource
                .map(source -> composeLayerMapQuery(formTree, layer, source))
                .map(query -> formSource.query(query))
                .orElseGet(() -> Observable.just(new ColumnSet(0, Collections.emptyMap())));

        ResolvedNode formula = FormulaResolver.resolve(formTree, layer.getValueFormula());
        this.formLabel = Optional.of(formTree.getRootFormClass().getLabel());
        this.sourceLabel = computeSourceLabel(formula);
        this.formulaType = formula.getResolvedType();
        this.layerType = layer.getType();

        this.layerResult = columnSet.transform(columnSet ->
            new LayerResult(layer, columnSet, formLabel, formulaType, layerType, sourceLabel));
    }

    public LayerViewModel(Layer layer) {
        this.layer = layer;
        this.geographySources = Collections.emptyList();
        this.formLabel = Optional.empty();
        this.sourceLabel = "";
        this.formulaType = Optional.empty();
        this.layerType = layer.getType();
        this.selectedGeographySource = Optional.empty();
        this.columnSet = Observable.just(ColumnSet.EMPTY);
        this.layerResult = Observable.just(new LayerResult(layer, ColumnSet.EMPTY,
                Optional.empty(), Optional.empty(), layer.getType(), ""));

    }

    public Optional<FormulaType> getFormulaType() {
        return formulaType;
    }

    public Layer getModel() {
        return layer;
    }

    public List<GeographySource> getGeographySources() {
        return geographySources;
    }

    public Optional<GeographySource> getSelectedGeographySource() {
        return selectedGeographySource;
    }

    public LayerType getLayerType() {
        return layerType;
    }

    private String computeSourceLabel(ResolvedNode formula) {
        if(formula instanceof ResolvedReference) {
            ResolvedReference reference = (ResolvedReference) formula;
            if (reference.getRoot() instanceof FieldReferenceNode) {
                FieldReferenceNode field = (FieldReferenceNode) reference.getRoot();
                return field.getField().getLabel();
            }
        }
        return "";
    }

    public static Optional<GeographySource> findSelectedSource(List<GeographySource> geographySources, String selected) {
        Optional<GeographySource> explicitlySelected = geographySources
                .stream()
                .filter(source -> source.getFormulaString().equals(selected))
                .findFirst();

        if(explicitlySelected.isPresent()) {
            return explicitlySelected;
        } else {
            return geographySources.stream().findFirst();
        }
    }

    private static QueryModel composeLayerMapQuery(FormTree formTree, Layer layer, GeographySource source) {
        QueryModel queryModel = new QueryModel(formTree.getRootFormId());
        queryModel.selectRecordId().as(ID_COLUMN_ID);

        queryModel.addColumn(new ColumnModel(LAT_COLUMN_ID, source.getLatitudeFormula().deparse()));
        queryModel.addColumn(new ColumnModel(LNG_COLUMN_ID, source.getLongitudeFormula().deparse()));
        queryModel.addColumn(new ColumnModel(LAYER_COLUMN_ID, layer.getValueFormula()));

        List<FormulaNode> criteria = new ArrayList<>();

        layer.getFilterFormula().ifPresent(f -> {
            FormulaNode filter = FormulaResolver.resolve(formTree, f).toFormula();
            criteria.add(filter);
        });

        if(!criteria.isEmpty()) {
            queryModel.setFilter(Formulas.allTrue(criteria));
        }

        return queryModel;
    }

    public String getId() {
        return layer.getId();
    }

    public Optional<String> getFormLabel() {
        return formLabel;
    }

    public Observable<LayerResult> getLayerResult() {
        return layerResult;
    }

    public boolean isCustomFormula() {
        return sourceLabel.isEmpty();
    }

    public String getSourceLabel() {
        return sourceLabel;
    }
}
