/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.map.viewModel;

import org.activityinfo.analysis.AnalysisResult;
import org.activityinfo.analysis.AnalysisResultVisitor;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.analysis.pivot.Statistic;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.type.geo.GeoPoint;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class MapResult implements AnalysisResult {

    public static final String TYPE = "map";

    private final List<LayerResult> layers;
    private final String baseMap;
    private final double zoomLevel;
    private final GeoPoint center;
    private final Range proportionalCircleRange;

    @Builder
    public MapResult(List<LayerResult> layers, Range proportionalCircleRange, String baseMap,
                     double zoomLevel, GeoPoint center) {
        this.layers = layers;
        this.baseMap = baseMap;
        this.zoomLevel = zoomLevel;
        this.center = center;
        this.proportionalCircleRange = proportionalCircleRange;
    }

    public MapResult(List<LayerResult> layers, String baseMap, double zoomLevel, GeoPoint center) {
        this.layers = layers;
        this.proportionalCircleRange = MultiZoomRangeComputer.computeProportionalCircleRange(layers);
        this.baseMap = baseMap;
        this.zoomLevel = zoomLevel;
        this.center = center;
    }

    public List<LayerResult> getLayers() {
        return layers;
    }


    /**
     * Get an approximate range of values, for all proportional circle layers,
     * taking into account clustering across a reasonable range of zoom levels.
     */
    public Range getProportionalCircleRange() {
        return proportionalCircleRange;
    }

    @Override
    public <T> T accept(AnalysisResultVisitor<T> visitor) {
        return visitor.visitMap(this);
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = Json.createObject();
        object.put("type", TYPE);
        JsonValue layerResults = Json.createArray();
        layers.forEach(layer ->
                layerResults.add(LayerResultJson.toJson(layer)));
        object.put("layers", layerResults);
        object.put("baseMap", baseMap);
        object.put("zoomLevel", zoomLevel);
        object.put("center", center.toJson());
        object.put("proportionalCircleRange", proportionalCircleRange.toJson());
        return object;

    }

    public static MapResult fromJson(JsonValue jsonValue) {
        List<LayerResult> layerResults = new ArrayList<>();
        JsonValue layersJson = jsonValue.get("layers");
        for (int i = 0; i < layersJson.length(); i++) {
            layerResults.add(LayerResultJson.fromJson(layersJson.get(i)));
        }
        return new MapResult(
                layerResults,
                Range.fromJson(jsonValue.get("proportionalCircleRange")),
                jsonValue.getString("baseMap"),
                jsonValue.getNumber("zoomLevel"),
                GeoPoint.fromJson(jsonValue.get("center")));
    }

    public String getBaseMap() {
        return baseMap;
    }

    public double getZoomLevel() {
        return zoomLevel;
    }

    public GeoPoint getCenter() {
        return center;
    }

    public Map<String, Statistic> getLayerClusteringMap() {
        Map<String, Statistic> layerClusteringMap = new HashMap<>();
        layers.forEach(layerResult ->
                layerClusteringMap.put(layerResult.getLayer().getId(),
                        layerResult.getLayer().getClustering()));

        return layerClusteringMap;
    }
}
