/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.map.viewModel;

import org.activityinfo.analysis.pivot.viewModel.FormForest;
import org.activityinfo.analysis.pivot.viewModel.ReferencedForm;
import org.activityinfo.model.analysis.map.Layer;
import org.activityinfo.model.analysis.map.MapModel;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.observable.Observable;
import org.activityinfo.store.query.shared.FormSource;

import java.util.*;
import java.util.logging.Logger;
import java.util.stream.Collectors;

public class MapViewModel {

    private static final Logger LOGGER = Logger.getLogger(MapViewModel.class.getName());

    private final FormSource formSource;
    private final Observable<MapModel> mapModel;
    private final Observable<FormForest> formForest;
    private final Observable<List<LayerViewModel>> layers;
    private final Observable<List<LayerResult>> layerResults;

    public MapViewModel(FormSource formSource, Observable<FormForest> formForest, Observable<MapModel> mapModel) {
        this.formSource = formSource;
        this.mapModel = mapModel;
        this.formForest = formForest;

        this.layers = mapModel
                .transform(m -> m.getLayers())
                .join(layers -> computeLayerModels(layers));

        this.layerResults = layers.join(this::computeLayerResults);
    }

    private Observable<List<LayerViewModel>> computeLayerModels(List<Layer> layers) {

        LOGGER.info("Recomputing LayerViewModels");

        List<Observable<LayerViewModel>> list = layers.stream()
                .map(layer -> {
                    Observable<Optional<FormTree>> formTree = getFormTree(layer.getFormId());
                    return formTree.transform(treeOpt -> treeOpt
                            .map(tree -> new LayerViewModel(tree, layer, formSource))
                            .orElseGet(() -> new LayerViewModel(layer)));
                })
                .collect(Collectors.toList());

        return Observable.flatten(list);
    }

    private Observable<List<LayerResult>> computeLayerResults(List<LayerViewModel> layerModels) {
        List<Observable<LayerResult>> layerResultList = new ArrayList<>();
        layerModels.forEach(layerModel -> {
            layerResultList.add(layerModel.getLayerResult());
        });
        return Observable.flatten(layerResultList);
    }

    public Observable<FormForest> getFormForest() {
        return formForest;
    }

    public Observable<Optional<FormTree>> getFormTree(ResourceId formId) {
        return formForest.transform(forest -> forest.findTree(formId));
    }

    public Observable<Set<ReferencedForm>> getReferencedForms() {
        return Observable.transform(formForest, mapModel, (f, m) -> {
            Set<ReferencedForm> forms = new HashSet<>();
            for (Layer layer : m.getLayers()) {
                f.getFormClass(layer.getFormId()).ifPresent(tree -> {
                    forms.add(ReferencedForm.of(tree));
                });
            }
            return forms;
        });
    }

    public Observable<List<LayerResult>> getLayerResults() {
        return layerResults;
    }

    public Observable<List<LayerViewModel>> getLayers() {
        return layers;
    }
}
