/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.map.viewModel;

import org.activityinfo.model.analysis.map.LayerType;
import org.activityinfo.model.formula.functions.StatFunction;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.ColumnType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.store.query.shared.HeapsortTandem;

import java.util.List;
import java.util.logging.Logger;

/**
 * Compute a [min, max] range for a set of geographic points over a range
 * of zoom levels.
 *
 * <p>The points will be dynamically clustered as the user zooms and pans the map, so we can't find
 * an exact range, but we find a scale that will cover a reasonable range of zoom layers.
 */
public class MultiZoomRangeComputer {

    private static final Logger LOGGER = Logger.getLogger(MultiZoomRangeComputer.class.getName());

    public static final int MAX_ZOOM_LEVEL = 11;
    public static final int MIN_ZOOM_LEVEL = 2;

    /**
     * The minimum number of clusters. If zooming out results in fewer than this number,
     * we stop zooming out. The principle is that a map that shows fewer than four points is not
     * likely to be useful, so we should not base the scale on a zoom level that exceeds this level
     * of generalization.
     */
    public static final int MIN_CLUSTERS = 4;



    private static class Clustering {
        private final Range range;
        private final int numClusters;

        public Clustering(Range range, int numClusters) {
            this.range = range;
            this.numClusters = numClusters;
        }
    }

    private final int[] tiles;
    private final double[] values;
    private int numPoints;
    private final StatFunction statistic;

    public MultiZoomRangeComputer(ColumnSet results, StatFunction statistic) {

        // Now update the min/max for this range using the given statistic
        this.statistic = statistic;

        int numRows = results.getNumRows();
        ColumnView latitudeView = results.getColumnViewOrBlank(LayerViewModel.LAT_COLUMN_ID, ColumnType.NUMBER);
        ColumnView longitudeView = results.getColumnViewOrBlank(LayerViewModel.LNG_COLUMN_ID, ColumnType.NUMBER);
        ColumnView valueView = results.getColumnViewOrBlank(LayerViewModel.LAYER_COLUMN_ID, ColumnType.NUMBER);

        // Group
        tiles = new int[numRows];
        values = new double[numRows];

        // Compute the tile index of each points at this zoom level. We use the tile index
        // as a crude but fast approximate of clustering.

        SphericalMercator tiling = new SphericalMercator(MAX_ZOOM_LEVEL);

        numPoints = 0;
        for (int i = 0; i < numRows; i++) {
            double latitude = latitudeView.getDouble(i);
            double longitude = longitudeView.getDouble(i);
            double value = valueView.getDouble(i);

            if(!Double.isNaN(latitude) && !Double.isNaN(longitude) &&
                    !Double.isNaN(value)) {
                tiles[numPoints] = encodeTileIndex(
                        tiling.longitudeToTileX(longitude),
                        tiling.latitudeToTileY(latitude));
                values[numPoints] = value;
                numPoints++;
            }
        }
    }

    public MultiZoomRangeComputer(LayerResult result) {
        this(result.getColumnSet(), result.getLayer().getClustering().getFunction());
    }

    public Range compute() {

        Clustering clustering = cluster();
        Range range = clustering.range;
        int zoomLevel = MAX_ZOOM_LEVEL;

        while(clustering.numClusters > MIN_CLUSTERS && zoomLevel > MIN_ZOOM_LEVEL) {
            zoomLevel--;
            zoomOut();
            clustering = cluster();

            LOGGER.info("Zoom Level " + zoomLevel + ": " + clustering.range + ", k = " + clustering.numClusters);

            if(clustering.numClusters > MIN_CLUSTERS) {
                range = range.union(clustering.range);
            }
        }
        return range;
    }

    /**
     * Clusters the points by tile index. Compute the aggregated value for each cluster,
     * and track the min/max of the aggregated cluster value at this zoom level.
     */
    private Clustering cluster() {

        HeapsortTandem.heapsortAscending(tiles, values, numPoints);

        int numTiles = 0;
        int previousTile = 0;
        int previousTileStart = 0;
        double minTileValue = Double.POSITIVE_INFINITY;
        double maxTileValue = Double.NEGATIVE_INFINITY;
        for (int i = 0; i < numPoints; i++) {
            if(i > 0 && previousTile != tiles[i]) {

                // calculate the value of this cluster
                double tileValue = statistic.compute(values, previousTileStart, i);
                previousTileStart = i;
                previousTile = tiles[i];

                if(tileValue > 0.0) {
                    minTileValue = Math.min(tileValue, minTileValue);
                    maxTileValue = Math.max(tileValue, maxTileValue);
                    numTiles++;
                }
            }
        }

        return new Clustering(new Range(minTileValue, maxTileValue), numTiles);
    }

    /**
     * Encodes a tile coordinate into a 32-bit integer value. This only works for tile
     * coordinates at zoom level 16 or lower. The x coordinate is stored in the high 16 bits,
     * the y coordinate at the lower 16 bits.
     */
    private static int encodeTileIndex(int x, int y) {
        assert x < (2 << 16);
        assert y < (2 << 16);

        return (x << 16) | y;
    }

    /**
     * Updates the tile index, computing a new tile coordinate for each point.
     */
    private void zoomOut() {
        for (int i = 0; i < numPoints; i++) {
            tiles[i] = zoomOut(tiles[i]);
        }
    }

    /**
     * Computes the tile index for one zoom level up. In the spherical tiling scheme,
     * the number of tiles decreases by a factor of two at each lower zoom level, so we can
     * just shift bits to the right, which is equivalent to dividing by two.
     */
    private int zoomOut(int encodedTileIndex) {
        int x = (encodedTileIndex >>> 1) & 0xFFFF0000;
        int y = ((encodedTileIndex & 0xFFFF) >>> 1);

        return x | y;
    }

    public static Range computeProportionalCircleRange(List<LayerResult> layerResults) {
        return layerResults.stream()
                .map(MultiZoomRangeComputer::rangeOf)
                .reduce(Range.EMPTY, Range::union);
    }

    private static Range rangeOf(LayerResult result) {
        if(result.getLayerType() == LayerType.PROPORTIONAL_CIRCLE) {
            return new MultiZoomRangeComputer(result).compute();
        } else {
            return Range.EMPTY;
        }
    }
}
