/*
 * ActivityInfo
 * Copyright (C) 2009-2013 UNICEF
 * Copyright (C) 2014-2021 BeDataDriven Groep B.V.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.activityinfo.analysis.pivot.viewModel;

import java.util.Arrays;

public class Breaks {

    private double[] sortedValues;
    private int[] breaks;

    /**
     * @param sortedValues the complete array of sorted data values
     * @param breaks       the indexes of the values within the sorted array that begin new classes
     */
    Breaks(double[] sortedValues, int[] breaks) {
        this.sortedValues = sortedValues;
        this.breaks = breaks;
    }

    /**
     * The Goodness of Variance Fit (GVF) is found by taking the difference
     * between the squared deviations from the array mean (SDAM) and the
     * squared deviations from the class means (SDCM), and dividing by the
     * SDAM
     */
    public double gvf() {
        double sdam = sumOfSquareDeviations(sortedValues);
        double sdcm = 0.0;
        for (int i = 0; i != numClassses(); ++i) {
            sdcm += sumOfSquareDeviations(classList(i));
        }
        return (sdam - sdcm) / sdam;
    }

    private double sumOfSquareDeviations(double[] values) {
        double mean = mean(values);
        double sum = 0.0;
        for (int i = 0; i != values.length; ++i) {
            double sqDev = Math.pow(values[i] - mean, 2);
            sum += sqDev;
        }
        return sum;
    }

    public double[] getValues() {
        return sortedValues;
    }

    private double[] classList(int i) {
        int classStart = (i == 0) ? 0 : breaks[i - 1] + 1;
        int classEnd = breaks[i];
        double list[] = new double[classEnd - classStart + 1];
        for (int j = classStart; j <= classEnd; ++j) {
            list[j - classStart] = sortedValues[j];
        }
        return list;
    }

    /**
     * @param classIndex
     * @return the minimum value (inclusive) of the given class
     */
    public double getClassMin(int classIndex) {
        if (classIndex == 0) {
            return sortedValues[0];
        } else {
            return sortedValues[breaks[classIndex - 1] + 1];
        }
    }

    /**
     * @param classIndex
     * @return the maximum value (inclusive) of the given class
     */
    public double getClassMax(int classIndex) {
        return sortedValues[breaks[classIndex]];
    }

    public int getClassCount(int classIndex) {
        if (classIndex == 0) {
            return breaks[0] + 1;
        } else {
            return breaks[classIndex] - breaks[classIndex - 1];
        }
    }

    private double mean(double[] values) {
        double sum = 0;
        for (int i = 0; i != values.length; ++i) {
            sum += values[i];
        }
        return sum / values.length;
    }

    public int numClassses() {
        return breaks.length;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i != numClassses(); ++i) {
            if (getClassMin(i) == getClassMax(i)) {
                sb.append(getClassMin(i));
            } else {
                sb.append(getClassMin(i)).append(" - ").append(getClassMax(i));
            }
            sb.append(" (" + getClassCount(i) + ")");
            sb.append(" = ").append(Arrays.toString(classList(i)));
            sb.append("\n");
        }
        return sb.toString();
    }

    public int classOf(double value) {
        for (int i = 0; i != numClassses(); ++i) {
            if (value <= getClassMax(i)) {
                return i;
            }
        }
        return numClassses() - 1;
    }
}
