/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.model.analysis.pivot.*;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumType;

import java.util.*;
import java.util.function.Function;

import static java.util.stream.Collectors.toList;

public class EffectiveDimension {

    private final int index;
    private final PivotModel pivotModel;
    private final DimensionModel dimensionModel;
    private final List<EffectiveMapping> effectiveMappings;
    private final String totalLabel;

    public EffectiveDimension(int index, PivotModel model, DimensionModel dimensionModel, List<EffectiveMapping> effectiveMappings) {
        this.index = index;
        this.pivotModel = model;
        this.dimensionModel = dimensionModel;
        this.effectiveMappings = effectiveMappings;
        this.totalLabel = dimensionModel.getTotalLabel().orElse(I18N.CONSTANTS.tableTotal());
    }

    public String getId() {
        return dimensionModel.getId();
    }

    public String getLabel() {
        return dimensionModel.getLabel();
    }

    public DimensionModel getModel() {
        return dimensionModel;
    }

    public String getTotalLabel() {
        return totalLabel;
    }

    public boolean isDate() {
        for (EffectiveMapping effectiveMapping : effectiveMappings) {
            if(effectiveMapping.isDate()) {
                return true;
            }
        }
        return false;
    }

    public boolean isChart(){
        return pivotModel.isChart();
    }

    public boolean isPeriod() {
        for (EffectiveMapping effectiveMapping : effectiveMappings) {
            if(effectiveMapping.isPeriod()) {
                return true;
            }
        }
        return false;
    }

    public List<DateLevel> getSupportedDateLevels() {
        for (EffectiveMapping effectiveMapping : effectiveMappings) {
            if(effectiveMapping.getSupportedDateLevels().size() > 0){
                return effectiveMapping.getSupportedDateLevels();
            }
        }
        return new ArrayList<>();
    }

    public boolean isNumber() {
        for (EffectiveMapping effectiveMapping : effectiveMappings) {
            if(effectiveMapping.getFormulaType() == FormulaType.NUMBER) {
                return true;
            }
        }
        return false;
    }

    private Optional<EnumType> getEnum() {
        for (EffectiveMapping effectiveMapping : effectiveMappings) {
            Optional<EnumType> enumType = effectiveMapping.getEnum();
            if(enumType.isPresent()) {
                return enumType;
            }
        }
        return Optional.empty();
    }

    public Axis getAxis() {
        return dimensionModel.getAxis();
    }



    public int getIndex() {
        return index;
    }

    public Function<Point, String> getCategoryProvider() {
        return (p -> p.getCategory(index));
    }

    public Comparator<String> getCategoryComparator() {
        return new Comparator<String>() {
            @Override
            public int compare(String o1, String o2) {
                Optional<EnumType> enumType = getEnum();
                // FIRST: Total categories always are sorted to the end
                Optional<Integer> totalComparison = compareTotals(o1, o2);
                if(totalComparison.isPresent()){
                    return totalComparison.get();
                }else if(enumType.isPresent()){
                    return getEnumComparator(enumType.get()).compare(o1, o2);
                }else if(isNumber()){
                    return getNumberComparator().compare(o1,o2);
                }else{
                    return getDefaultComparator().compare(o1, o2);
                }
            }
        };
    }

    private Comparator getNumberComparator(){
        return new Comparator<String>() {
            @Override
            public int compare(String o1, String o2) {
                Double a, b;
                try{
                    a = Double.parseDouble(o1);
                }catch (NumberFormatException e){
                    return -1;
                }

                try{
                    b = Double.parseDouble(o2);
                }catch (NumberFormatException e){
                    return 1;
                }

                return a.compareTo(b);
            }
        };
    }

    private Comparator getEnumComparator(EnumType enumType){
        HashMap<String, Integer> orderMap = new HashMap<>();
        List<EnumItem> items = enumType.getItems();
        for (int i = 0; i < items.size(); i++) {
            orderMap.put(items.get(i).getLabel(), i);
        }
        return new Comparator<String>() {
            @Override
            public int compare(String o1, String o2) {
                Integer a = orderMap.get(o1);
                Integer b = orderMap.get(o2);
                if(a == null){
                    return -1;
                }else if(b == null){
                    return 1;
                }else{
                    return a.compareTo(b);
                }
            }
        };
    }

    private Comparator getDefaultComparator(){
        switch (dimensionModel.getId()) {
            case DimensionModel.MEASURE_ID:
                return new CategoryComparator(pivotModel.getMeasures().stream().map(m -> m.getLabel()).collect(toList()));
            case DimensionModel.STATISTIC_ID:
                return new CategoryComparator(Statistic.labels());
            default:
                return new CategoryComparator();
        }
    }

    public List<EffectiveMapping> getEffectiveMappings() {
        return effectiveMappings;
    }

    private static class CategoryComparator implements Comparator<String> {

        private final List<String> explicitOrder;

        public CategoryComparator(List<String> explicitOrder) {
            this.explicitOrder = explicitOrder;
        }

        public CategoryComparator() {
            this.explicitOrder = Collections.emptyList();
        }

        @Override
        public int compare(String a, String b) {
            // Take into account any explicit ordering provided.
            // Explicitly ordered values always proceed unordered values.
            if(!explicitOrder.isEmpty()) {
                int ia = explicitOrder.indexOf(a);
                int ib = explicitOrder.indexOf(b);

                if (ia != -1 && ib != -1) {
                    // both values are explicitly ordered
                    return Integer.compare(ia, ib);
                }
                if (ib != -1) {
                    // a is not explicitly ordered, so follows b
                    return +1;
                }
                if (ia != -1) {
                    // b is not explicity ordered, so follows a
                    return +1;
                }
            }

            // Finally, if neither values are explicitly ordered, use natural (alphabetic) order
            return a.compareTo(b);
        }
    }

    private static Optional<Integer> compareTotals(String a, String b){
        boolean ta = Point.TOTAL.equals(a);
        boolean tb = Point.TOTAL.equals(b);
        if(ta && tb) {
            return Optional.of(0);
        } else if(ta) {
            return Optional.of(+1);
        } else if(tb) {
            return Optional.of(-1);
        }
        return Optional.empty();
    }

}
