/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import com.google.common.base.Functions;
import com.google.common.collect.Ordering;
import org.activityinfo.model.analysis.pivot.DateLevel;
import org.activityinfo.model.analysis.pivot.DimensionMapping;
import org.activityinfo.model.analysis.pivot.DimensionModel;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.formula.functions.date.MonthFunction;
import org.activityinfo.model.formula.functions.date.QuarterFunction;
import org.activityinfo.model.formula.resolved.FormulaResolver;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.ColumnType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.type.Cardinality;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.time.*;

import java.util.*;
import java.util.function.Function;

/**
 * Models a measure that is part of the analysis,
 * based on the user's model as well as metadata
 */
public class EffectiveMapping {
    private final int index;
    private final DimensionModel model;
    private final DimensionMapping mapping;
    private ResolvedNode formula;

    private final Optional<EnumType> multiEnumType;
    private final FormulaType formulaType;

    public EffectiveMapping(FormTree formTree, int index, DimensionModel model, DimensionMapping mapping) {
        this.index = index;
        this.model = model;
        this.mapping = mapping;
        if(this.mapping != null) {
            this.formula = FormulaResolver.resolve(formTree, mapping.getFormula());
            this.multiEnumType = isMultiEnum(formula);
            if(this.multiEnumType.isPresent()) {
                this.formulaType = FormulaType.STRING;
            } else {
                this.formulaType = this.formula.getResolvedType().orElse(FormulaType.STRING);
            }
        } else {
            this.multiEnumType = Optional.empty();
            this.formulaType = FormulaType.STRING;
        }
    }

    /**
     * Determine if this formula refers to a field that is multi-valued, such as enum field with cardinality
     * multiple.
     *
     */
    private static Optional<EnumType> isMultiEnum(ResolvedNode node) {
        return node.getFieldType()
            .filter(type -> type instanceof EnumType && ((EnumType) type).getCardinality() == Cardinality.MULTIPLE)
            .map(type -> (EnumType)type);
    }

    public Optional<EnumType> getEnum() {
        if(formula == null){
            return Optional.empty();
        }else{
            return formula.getFieldType()
                    .filter(type -> type instanceof EnumType)
                    .map(type -> (EnumType)type);
        }
    }

    public String getId() {
        return model.getId();
    }

    public int getIndex() {
        return index;
    }

    /**
     *
     * @return true if this is a "fixed" dimension like Measure or Statistic that is independent of
     * the fields aggregated.
     */
    public boolean isFixed() {
        return model.getId().equals(DimensionModel.MEASURE_ID) ||
               model.getId().equals(DimensionModel.STATISTIC_ID);
    }

    public boolean isMultiValued() {
        return multiEnumType.isPresent();
    }

    /**
     * @return true if this dimension has a single value for each row.
     */
    public boolean isSingleValued() {
        return !isFixed() && !multiEnumType.isPresent();
    }

    public DimensionMapping getMapping() {
        return mapping;
    }

    public boolean isDate() {
        if(this.mapping != null) {
            return formula.getResolvedType()
                .map(t -> t == FormulaType.LOCAL_DATE)
                .orElse(false);
        }
        return false;
    }

    public FormulaType getFormulaType(){
        return formulaType;
    }

    public boolean isPeriod() {
        return isDate() || formula.getFieldType().map(type -> type instanceof PeriodType).orElse(false);
    }

    private PeriodType getPeriodType(){
        assert isPeriod();
        if(isDate()){
            return LocalDateType.INSTANCE;
        }else {
            return (PeriodType) formula.getFieldType().get();
        }
    }
    public List<DateLevel> getSupportedDateLevels() {
        if(isPeriod()) {
            PeriodType type = getPeriodType();
            if(type instanceof LocalDateType){
                return Arrays.asList(DateLevel.values());
            } else if(type instanceof EpiWeekType || type instanceof FortnightType) {
                DateLevel[] levels = {DateLevel.WEEK, DateLevel.MONTH, DateLevel.QUARTER, DateLevel.YEAR, DateLevel.WEEKYEAR, DateLevel.MONTHYEAR, DateLevel.QUARTERYEAR};
                return Arrays.asList(levels);
            } else if(type instanceof MonthType) {
                DateLevel[] levels = {DateLevel.MONTH, DateLevel.QUARTER, DateLevel.YEAR, DateLevel.MONTHYEAR, DateLevel.QUARTERYEAR};
                return Arrays.asList(levels);
            } else if(type instanceof YearType) {
                DateLevel[] levels = {DateLevel.YEAR};
                return Arrays.asList(levels);
            }
        }
        return new ArrayList<>();
    }

    public List<ColumnModel> getRequiredColumns() {

        return multiEnumType.map(type -> getRequiredMultiEnumColumns(type))
            .orElseGet(() -> getRequiredSimpleColumn());
    }

    private List<ColumnModel> getRequiredSimpleColumn() {
        if(this.mapping != null && this.formula.isValid()) {
            ColumnModel columnModel = new ColumnModel();
            columnModel.setId(getColumnId());
            columnModel.setFormula(this.formula.toFormula());
            return Collections.singletonList(columnModel);
        } else {
            return Collections.emptyList();
        }
    }


    private List<ColumnModel> getRequiredMultiEnumColumns(EnumType enumType) {
        List<ColumnModel> columns = new ArrayList<>();
        for (EnumItem enumItem : enumType.getItems()) {
            ColumnModel columnModel = new ColumnModel();
            columnModel.setId(getColumnId(enumItem));
            columnModel.setFormula(new CompoundExpr(
                this.formula.toFormula(),
                new SymbolNode(enumItem.getId())));
            columns.add(columnModel);
        }
        return columns;
    }

    private String getColumnId() {
        return "d" + index;
    }

    private String getColumnId(EnumItem item) {
        return getColumnId() + "_" + item.getId().asString();
    }

    public DimensionReader createReader(ColumnSet columnSet) {

        String missingCategory = computeMissingCategory();

        ColumnView columnView = columnSet.getColumnView(getColumnId());
        if(columnView == null) {
            return row -> missingCategory;
        }

        if(columnView.getType() == ColumnType.NUMBER){
            return row -> {
                double value = columnView.getDouble(row);
                if(Double.isNaN(value)) {
                    return missingCategory;
                }else{
                    return Long.toString(Math.round(value));
                }
            };

        }
        else{
            Function<String, String> map = createMap();
            return row -> {
                String category = columnView.getString(row);
                if(category == null) {
                    return missingCategory;
                }
                return map.apply(category);
            };
        }
    }

    private String computeMissingCategory() {
        if(model.isMissingIncluded()) {
            return getMissingLabel();
        } else {
            // Observations with a missing category will be excluded from the analysis
            return null;
        }
    }

    private String getMissingLabel() {
        return model.getEffectiveMissingLabel();
    }

    private Function<String, String> createMap() {
        if(isPeriod()) {
            PeriodType type = getPeriodType();
            Optional<DateLevel> dateLevel = model.getDateLevel();
            if(dateLevel.isPresent()) {
                DateLevel dateLevel_ = dateLevel.get();
                if(type instanceof LocalDateType) {
                    return createDateMap(dateLevel_);
                } else if(type instanceof EpiWeekType || type instanceof FortnightType) {
                    return createWeekMap(dateLevel_);
                } else if(type instanceof MonthType) {
                    return createMonthMap(dateLevel_);
                }
            }
        }
        return Functions.identity();
    }

    private Function<String, String> createDateMap(DateLevel dateLevel) {
        switch (dateLevel) {
            case WEEK: //YYYYW3
                return EffectiveMapping::week;
            case MONTH: //YYYY-MM
                return EffectiveMapping::month;
            case QUARTER: //YYYYQ3
                return EffectiveMapping::quater;
            case YEAR: //YYYY
                return EffectiveMapping::year;
            case WEEKYEAR: //W32
                return EffectiveMapping::weekOfYear;
            case MONTHYEAR: //MM
                return EffectiveMapping::monthOfYear;
            case QUARTERYEAR: ////Q1
                return EffectiveMapping::quarterOfYear;
            case DATE: //YYYY-MM-dd
            default:
                return Functions.identity();
        }
    }

    private Function<String, String> createMonthMap(DateLevel dateLevel) {
        switch (dateLevel) {
            case QUARTER: //YYYYQ3
                return EffectiveMapping::quater;
            case YEAR: //YYYY
                return EffectiveMapping::year;
            case MONTHYEAR: //MM
                return EffectiveMapping::monthOfYear;
            case QUARTERYEAR: ////Q1
                return EffectiveMapping::quarterOfYear;
            case MONTH: ////YYYY-MM
            default:
                return Functions.identity();
        }
    }

    private Function<String, String> createWeekMap(DateLevel dateLevel) {
        switch (dateLevel) {
            case MONTH: //YYYY-MM
                return date -> {
                    return EpiWeek.parse(date).toMonth().toString();
                };
            case QUARTER: //YYYYQ3
                return date -> {
                    return quater(EpiWeek.parse(date).toMonth().toString());
                };
            case YEAR:
                return EffectiveMapping::year; ////YYYY
            case WEEKYEAR: //W01
                return date -> {
                    return date.substring(4);
                };
            case MONTHYEAR: //MM
                return date -> {
                    return monthOfYear(EpiWeek.parse(date).toMonth().toString());
                };
            case QUARTERYEAR: ////Q1
                return date -> {
                    return quarterOfYear(EpiWeek.parse(date).toMonth().toString());
                };
            case WEEK: //YYYYW32
            default:
                return Functions.identity();
        }
    }

    private static String monthOfYear(String date) {
        return date.substring(5, 7);
    }

    private static String quarterOfYear(String date) {
        int month = MonthFunction.fromIsoString(date);
        return "Q" + QuarterFunction.fromMonth(month);
    }

    private static String year(String date) {
        return date.substring(0, 4);
    }

    private static String month(String date) {
        return date.substring(0, 7);
    }

    private static String quater(String date) {
        return year(date) + quarterOfYear(date);
    }

    private static String week(String date) {
        LocalDate localDate = LocalDate.parse(date);
        EpiWeek  week = EpiWeek.weekOf(localDate);
        return week.toString();
    }

    private static String weekOfYear(String date) {
        LocalDate localDate = LocalDate.parse(date);
        EpiWeek  week = EpiWeek.weekOf(localDate);
        int numWeeks = week.getWeekInYear();
        if(numWeeks < 10){
            return "W0" + numWeeks;
        }else {
            return "W" + numWeeks;
        }
    }

    public MultiDim createMultiDimSet(ColumnSet columnSet) {
        EnumType enumType = multiEnumType.get();

        int numItems = enumType.getItems().size();
        int numCategories;

        if(model.isMissingIncluded()) {
            // If we are included cases with no values set, then consider it
            // a category on its own
            numCategories = numItems + 1;
        } else {
            numCategories = numItems;
        }

        String[] labels = new String[numCategories];
        BitSet[] bitSets = new BitSet[numCategories];
        List<EnumItem> values = enumType.getItems();

        for (int j = 0; j < numItems; j++) {

            BitSet bitSet = new BitSet();
            EnumItem enumItem = values.get(j);
            ColumnView columnView = columnSet.getColumnView(getColumnId(enumItem));

            assert columnView != null;
            assert columnView.getType() == ColumnType.BOOLEAN;

            for (int i = 0; i < columnView.numRows(); i++) {
                if (columnView.getBoolean(i) == ColumnView.TRUE) {
                    bitSet.set(i, true);
                }
            }
            labels[j] = enumItem.getLabel();
            bitSets[j] = bitSet;
        }

        if(model.isMissingIncluded()) {
            labels[numCategories - 1] = getMissingLabel();
            bitSets[numCategories - 1] = computeMissingBitSet(columnSet.getNumRows(), numItems, bitSets);
        }

        return new MultiDim(this.index, labels, bitSets);
    }

    /**
     * Computes the bitSet of cases that have NO values sent in any of the
     * {@code numItems}
     *
     * @param numRows the number of rows.
     * @param numItems the number of enum items
     * @param bitSets an array of bitSets, one for each enum item.
     * @return a bitSet where a bit is set to true for each row that has NO categories set to true.
     */
    private BitSet computeMissingBitSet(int numRows, int numItems, BitSet[] bitSets) {

        // Initialize missing bitSet to 1 for all rows
        BitSet missing = new BitSet(numRows);
        missing.set(0, numRows);

        // Now set missing=false when a category is set.
        for (int i = 0; i < numItems; i++) {
            missing.andNot(bitSets[i]);
        }

        return missing;
    }


    public Comparator<String> getCategoryComparator() {
        return Ordering.natural();
    }


}

