/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.model.analysis.pivot.*;
import org.activityinfo.model.resource.ResourceId;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;

public class EffectiveModel {

    private final DimensionSet dimensionSet;
    private final PivotModel model;
    private final FormForest formForest;

    private final List<EffectiveMeasure> measures = new ArrayList<>();
    private final List<EffectiveFormMeasures> forms = new ArrayList<>();
    private final List<EffectiveDimension> dimensions = new ArrayList<>();

    public EffectiveModel(PivotModel pivotModel, FormForest formForest) {
        this.model = pivotModel.withoutTotals(pivotModel.isChart());
        this.formForest = formForest;

        List<DimensionModel> dimensions = new ArrayList<>(model.getDimensions());

        // If we have multiple measures, than they MUST be labeled with a measure dimension
        if ( isMeasureDimensionNeeded(model) && !isDefined(dimensions, DimensionModel.MEASURE_ID)) {
            dimensions.add(
                new DimensionModelBuilder()
                    .id(DimensionModel.MEASURE_ID)
                    .label(I18N.CONSTANTS.measures())
                    .build());
        }

        // Include a Statistics dimension if required but not added by
        // the user to the model.
        if (model.isMeasureDefinedWithMultipleStatistics() &&
                !isDefined(dimensions, DimensionModel.STATISTIC_ID)) {

            dimensions.add(
                    new DimensionModelBuilder()
                            .id(DimensionModel.STATISTIC_ID)
                            .label(I18N.CONSTANTS.statistic())
                            .build());
        }

        // Create a DimensionSet, which maps each dimension
        // to an integer index.
        this.dimensionSet = new DimensionSet(dimensions);

        // Group measures by form / report
        // form -> [ measure1, measure2 ]
        Multimap<ResourceId, EffectiveMeasure> formMap = HashMultimap.create();

        for (MeasureModel measureModel : model.getMeasures()) {
            EffectiveMeasure measure = formForest.findTree(measureModel.getFormId()).map(tree -> new EffectiveMeasure(measureModel, tree)).orElse(new EffectiveMeasure(measureModel));
            measures.add(measure);
            formMap.put(measureModel.getFormId(), measure);
        }

        DimensionMergeMap dimensionMergeMap = new DimensionMergeMap(formForest);

        //Compute the EffectiveFormMeasures
        for (ResourceId formId :formMap.keySet()) {
            formForest.findTree(formId).ifPresent(tree -> {
                Collection<EffectiveMeasure> measures = formMap.get(formId);
                this.forms.add(new EffectiveFormMeasures(tree, dimensionSet, dimensionMergeMap, measures));
            });
        }

        for (DimensionModel dimensionModel : dimensions) {
            int index = dimensionSet.getIndex(dimensionModel);
            List<EffectiveMapping> effectiveMappings = new ArrayList<>();
            for (EffectiveFormMeasures form : forms) {
                effectiveMappings.add(form.getDimension(index));
            }
            this.dimensions.add(new EffectiveDimension(index, model, dimensionModel, effectiveMappings));
        }
    }

    private boolean isMeasureDimensionNeeded(PivotModel model) {
        return model.getMeasures().size() > 1 || model.getDimensions().isEmpty();
    }

    public FormForest getFormForest() {
        return formForest;
    }

    public PivotModel getModel() {
        return model;
    }

    private boolean isDefined(List<DimensionModel> dimensions, String dimensionId) {
        for (DimensionModel dimension : dimensions) {
            if(dimension.getId().equals(dimensionId)) {
                return true;
            }
        }
        return false;
    }

    public List<EffectiveDimension> getDimensions() {
        return dimensions;
    }

    public DimensionSet getDimensionSet() {
        return dimensionSet;
    }

    public List<EffectiveFormMeasures> getForms() {
        return forms;
    }

    public List<EffectiveMeasure> getMeasures() {
        return measures;
    }

    public List<EffectiveDimension> getDimensions(Axis axis) {
        List<EffectiveDimension> items = new ArrayList<>();
        for (EffectiveDimension dimension : dimensions) {
            if(dimension.getAxis() == axis) {
                items.add(dimension);
            }
        }
        return items;
    }

    public Optional<ResourceId> getMapTemplateId() {
        for (EffectiveDimension dimension : dimensions) {
            if(dimension.getModel().getMapTemplateId().isPresent()) {
                return dimension.getModel().getMapTemplateId();
            }
        }
        return Optional.empty();
    }
}
