/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import com.google.common.collect.Sets;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormMetadata;
import org.activityinfo.model.formTree.FormMetadataProvider;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formTree.FormTreeBuilder;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;

/**
 * Set of selected forms and reports
 */
public class FormForest implements FormMetadataProvider {

    public static final FormForest TEST_EMPTY = new FormForest();

    private final Map<ResourceId, FormMetadata> forms = new HashMap<>();
    private final Map<ResourceId, PivotTableSource> tableMap = new HashMap<>();
    private final Set<ResourceId> rootFormIds = new HashSet<>();

    private FormForest() {
    }

    public FormForest(FormTree tree) {
        assert tree != null;
        rootFormIds.add(tree.getRootFormId());
        index(tree);
    }

    public FormForest(List<FormTree> trees) {
        for (FormTree tree : trees) {
            rootFormIds.add(tree.getRootFormId());
            index(tree);
        }
    }

    public FormForest(List<FormTree> trees, List<Optional<PivotTableSource>> tables) {
        for (FormTree tree : trees) {
            rootFormIds.add(tree.getRootFormId());
            index(tree);
        }
        for (Optional<PivotTableSource> optionalTable : tables) {
            optionalTable.ifPresent(table -> {
                rootFormIds.add(table.getId());
                tableMap.put(table.getId(), table);
                index(table.getFormTree());
            });
        }
    }

    private void index(FormTree formTree) {
        for (FormMetadata form : formTree.getForms()) {
            if(!forms.containsKey(form.getId())) {
                forms.put(form.getId(), form);
            }
        }
    }

    public boolean isEmpty() {
        return forms.isEmpty();
    }

    public Collection<PivotTableSource> getTables() {
        return tableMap.values();
    }


    public Optional<FormClass> getFormClass(ResourceId formId) {
        FormMetadata root = forms.get(formId);
        if(root == null) {
            return Optional.empty();
        }
        return root.getSchemaIfVisible().getIfVisible();
    }

    public Optional<FormTree> findTree(ResourceId formId) {
        FormMetadata root = forms.get(formId);
        if(root == null) {
            return Optional.empty();
        }
        return new FormTreeBuilder(this).queryTree(formId).toMaybe().getIfVisible();
    }


    public Optional<PivotTableSource> findTable(ResourceId reportId) {
        return Optional.ofNullable(tableMap.get(reportId));
    }

    @Override
    public FormMetadata getFormMetadata(ResourceId formId) {
        FormMetadata formMetadata = forms.get(formId);
        if(formMetadata == null) {
            return FormMetadata.notFound(formId);
        }
        return formMetadata;
    }

    public Collection<ResourceId> getRootFormIds() {
        return rootFormIds;
    }

    public Collection<ResourceId> getAllFormIds() {
        return forms.keySet();
    }

    public Collection<ResourceId> getReports() {
        return tableMap.keySet();
    }

    public Set<ReferencedForm> getReferencedForms() {
        Optional<Set<ReferencedForm>> referencedForms = forms.keySet().stream().map(formId -> findTree(formId)).map(formTree ->
                formTree.map(formTree_ -> ReferencedForm.of(formTree_)).orElse(Collections.emptySet())).reduce(Sets::union);
        Set<ReferencedForm> referencedForms_ = referencedForms.orElse(Collections.emptySet());

        Optional<Set<ReferencedForm>> referencedTableForms = tableMap.keySet().stream().map(formId -> findTable(formId)).map(tableSource ->
                tableSource.map(tableSource_ -> tableSource_.getFormTree()).map(formTree_ -> ReferencedForm.of(formTree_)).orElse(Collections.emptySet())).reduce(Sets::union);
        Set<ReferencedForm> referencedTableForms_ = referencedTableForms.orElse(Collections.emptySet());

        return Sets.union(referencedForms_, referencedTableForms_);
    }
}
