/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import org.activityinfo.model.analysis.pivot.DimensionModel;
import org.activityinfo.model.analysis.pivot.Statistic;
import org.activityinfo.model.formula.functions.*;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.store.query.shared.Aggregation;
import org.activityinfo.store.query.shared.HeapsortTandem;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


public class MeasureResultBuilder {

    private EffectiveFormMeasures form;

    private static PercentFormatter PERCENT_FORMATTER = PercentFormatter.get();

    private final EffectiveMeasure measure;

    private final DimensionSet dimensionSet;

    private ColumnSet columns;

    private final int numRows;

    /**
     * List of all points in the multi-dimensional space resulting from this measure.
     */
    private final List<Point> points = new ArrayList<>();

    private final List<Point> percentages = new ArrayList<>();

    private final int measureDimensionIndex;


    /**
     * The index of the "statistic" dimension, or -1 if not requested.
     */
    private final int statisticDimensionIndex;

    /**
     * True if hidden categories should be included
     */
    private final boolean showHidden;

    /**
     *  @param measure the effective model of this measure, fully resolved.
     * @param columns the results of the base query.
     * @param showHidden
     */
    public MeasureResultBuilder(EffectiveFormMeasures form, EffectiveMeasure measure, ColumnSet columns, boolean showHidden) {
        this.measure = measure;
        this.form = form;
        this.dimensionSet = form.getDimensionSet();
        this.columns = columns;
        this.numRows = columns.getNumRows();
        this.measureDimensionIndex = form.getDimensionSet().getIndexByDimensionId(DimensionModel.MEASURE_ID);
        this.statisticDimensionIndex = form.getDimensionSet().getIndexByDimensionId(DimensionModel.STATISTIC_ID);
        this.showHidden = showHidden;
    }

    public void execute() {

        /*
         * The value column contains the numerical values that we will be aggregating
         * by dimension.
         */
        MeasureVector values = new MeasureVector(columns.getColumnView(measure.getModel().getId()));


        /*
         * The first step is to divide the value vector into groups defined by the intersection
         * of all single-valued dimensions.
         *
         * For example, if this analysis includes the "Gender" and "Married" dimensions, which
         * take the categories "Female" and "Male" and "Married" and "Single" respectively, then
         * the GroupMap helps us match the combinations to a group index. For example:
         *
         * Female-Married -> Group 0
         * Male-Married ->   Group 1
         * Female-Single ->  Group 2
         * Male-Single ->    Group 3
         *
         */

        /*
         * Using the GroupMapBuilder, we want to build parallel arrays containing the measure's numerical
         * values in one array, and the group index in the other. Continuing the example above,
         * the result might be:
         *
         * valueArray = [1.5, 34, 10, 1.333, 22, 14]
         * groupArray = [  1,  0,  3,     0,  2, -1]
         *
         * Where -1 means that either Gender or Married was missing for this value, and so the value
         * is discarded.
         */

        GroupMap groupMap = GroupMapBuilder.build(columns, form.getDimensions());

        /*
         * If no groups were found, that means that all measure values have at least one missing dimension
         * and so all values are discarded.
         */
        if(groupMap.getGroupCount() == 0) {
            return;
        }


        MultiDimSet multiDimSet = buildMultiDimSet();


        /*
         * If totals are included for any of the dimensions, we need to repeat the aggregation
         * several times, omitting different dimensions at different iterations.
         */
        for(TotalSubset subset : TotalSubset.set(dimensionSet)) {

            /*
             * Regroup single-valued dimensions
             */
            GroupMap singleValuedDims = groupMap.regroup(subset);

            /*
             * Regroup multi-valued dimensions
             */
            MultiDimSet multiValuedDims = multiDimSet.regroup(subset);

            /*
             * Do the aggregation!
             */
            aggregate(subset, singleValuedDims, multiValuedDims, values);
        }

    }

    private MultiDimSet buildMultiDimSet() {
        List<MultiDim> dimSets = new ArrayList<>();
        for (EffectiveMapping dimension : form.getDimensions()) {
            if(dimension.isMultiValued()) {
                dimSets.add(dimension.createMultiDimSet(columns));
            }
        }
        return new MultiDimSet(dimSets);
    }


    private void aggregate(
        TotalSubset totalSubset,
        GroupMap singleValuedDims,
        MultiDimSet multiValuedDims,
        MeasureVector values) {


        /*
         * For dimensions that can have multiple values, we need to run the aggregation
         * multiple times as a single value can be counted towards multiple categories.
         */
        for (MultiDimCategory multiDimCategory : multiValuedDims.build()) {

            int[] groupArray = singleValuedDims.copyOfGroupArray();

            /*
             * Exclude the values that do not belong to this multiDimCategory
             * by marking them as NaN.
             */
            double[] filteredValues = multiDimCategory.filter(values.getDoubleArray());

            /*
             * Sort the group and value arrays in tandem
             * (We can only do this in the inner loop because we need the original
             *  order for applying the multi-dimensional category filter)
             */

            HeapsortTandem.heapsortDescending(groupArray, filteredValues, filteredValues.length);

            /*
             * Now calculated all the required statistics.
             */
            for (Statistic statistic : measure.getModel().getStatistics()) {
                if(isMeasureValidForStatistic(values, statistic)) {
                    aggregate(totalSubset, singleValuedDims, multiDimCategory, statistic, groupArray, filteredValues);
                }
            }
        }
    }

    private boolean isMeasureValidForStatistic(MeasureVector measureVector, Statistic statistic) {
        switch (statistic) {
            case COUNT:
            case COUNT_DISTINCT:
                return true;

            default:
            case SUM:
            case AVERAGE:
            case MEDIAN:
            case MIN:
            case MAX:
                return measureVector.isNumeric();
        }
    }

    private double[] aggregate(TotalSubset totalSubset,
                               GroupMap singleValuedDims,
                               MultiDimCategory multiDimCategory,
                               Statistic statistic,
                               int[] sortedGroupArray,
                               double[] sortedValueArray) {

        int numSingleValuedGroups = singleValuedDims.getGroupCount();

        double[] aggregatedValues = computeStatistic(statistic, sortedGroupArray, sortedValueArray, numSingleValuedGroups);

        boolean includeTotals = totalSubset.includeTotals();
        boolean includePercentages = statistic == Statistic.SUM && totalSubset.includePercentages();

        for (int i = 0; i < numSingleValuedGroups; i++) {
            String[] group = multiDimCategory.withMultiValuedCategories(singleValuedDims.getGroup(i));
            if(showHidden || isVisible(group)) {
                if (this.measureDimensionIndex != -1) {
                    group[this.measureDimensionIndex] = this.measure.getModel().getLabel();
                }
                if (includeTotals) {
                    points.add(new Point(
                        aggregatedValues[i],
                        format(aggregatedValues[i]),
                        withFixedDimensions(group, statistic)));
                }
                if (includePercentages) {
                    addPercentage(group, aggregatedValues[i]);
                }
            }
        }
        return aggregatedValues;
    }

    private boolean isVisible(String[] group) {
        for (int i = 0; i < group.length; i++) {
            if(i != measureDimensionIndex && i != statisticDimensionIndex) {
                if(!dimensionSet.getDimension(i).isCategoryVisible(group[i])) {
                    return false;
                }
            }
        }
        return true;
    }

    private double[] computeStatistic(Statistic statistic, int[] sortedGroupArray, double[] sortedValueArray, int numSingleValuedGroups) {

        return Aggregation.aggregateSorted(
            statistic.getFunction(),
            sortedGroupArray,
            sortedValueArray,
            sortedValueArray.length,
            numSingleValuedGroups);
    }

    private void addPercentage(String[] group, double denominator) {

        for (Point point : points) {
            if(isNumerator(point, group)) {
                double proportion = point.getValue() / denominator;

                percentages.add(new Point(proportion, PERCENT_FORMATTER.format(proportion),
                    withPercentage(point, "%")));
            }
        }
    }

    private String[] withPercentage(Point point, String percentageLabel) {
        String[] group = new String[dimensionSet.getCount()];
        for (int i = 0; i < group.length; i++) {
            group[i] = point.getCategory(i);
        }
        group[statisticDimensionIndex] = percentageLabel;
        return group;
    }

    private boolean isNumerator(Point point, String[] group) {
        for (int i = 0; i < group.length; i++) {
            if(i != statisticDimensionIndex) {
                String numeratorCategory = point.getCategory(i);
                String denominatorCategory = group[i];
                if(!denominatorCategory.equals(Point.TOTAL) &&
                    !numeratorCategory.equals(denominatorCategory)) {
                    return false;
                }
            }
        }
        return true;
    }

    private String[] withFixedDimensions(String[] group, Statistic statistic) {
        return withFixedDimensions(group, statistic.getLabel());
    }

    private String[] withFixedDimensions(String[] group, String statistic) {
        if(statisticDimensionIndex == -1 && measureDimensionIndex == -1) {
            return group;
        } else {
            String[] groupWithFixed = Arrays.copyOf(group, group.length);
            if(measureDimensionIndex != -1) {
                groupWithFixed[measureDimensionIndex] = measure.getModel().getLabel();
            }
            if(statisticDimensionIndex != -1) {
                groupWithFixed[statisticDimensionIndex] = statistic;
            }
            return groupWithFixed;
        }
    }

    private String format(double value) {
        return Long.toString((long)Math.round(value));
    }

    public List<Point> getResult() {
        List<Point> result = new ArrayList<>();
        result.addAll(points);
        result.addAll(percentages);

        return result;
    }
}
