/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.observable.Observable;
import org.activityinfo.store.query.shared.FormSource;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Logger;

/**
 * The result of the analysis
 */
public class PivotPointSet {

    private static final Logger LOGGER = Logger.getLogger(PivotPointSet.class.getName());

    private DimensionSet dimensionSet;
    private List<Point> points = new ArrayList<>();
    private EffectiveModel effectiveModel;
    private final boolean showHidden;

    public PivotPointSet(EffectiveModel effectiveModel, List<FormResultSet> measureSets, boolean showHidden) {
        this.effectiveModel = effectiveModel;
        this.showHidden = showHidden;
        if (!measureSets.isEmpty()) {
            dimensionSet = measureSets.get(0).getDimensions();
        } else {
            dimensionSet = new DimensionSet();
        }
        for (FormResultSet measureSet : measureSets) {
            points.addAll(measureSet.getPoints());
        }
    }

    public boolean isShowHidden() {
        return showHidden;
    }

    public EffectiveModel getEffectiveModel() {
        return effectiveModel;
    }

    public DimensionSet getDimensionSet() {
        return dimensionSet;
    }

    public List<Point> getPoints() {
        return points;
    }

    public static Observable<PivotPointSet> compute(FormSource formSource, EffectiveModel effectiveModel, boolean showHidden) {

        List<Observable<FormResultSet>> points = new ArrayList<>();

        for (EffectiveFormMeasures form : effectiveModel.getForms()) {
            if(effectiveModel.getFormForest().getReports().contains(form.getFormId())) {
                points.add(computePointsForAllMeasuresInReport(effectiveModel.getFormForest(), form, showHidden));
            } else {
                points.add(computePointsForAllMeasuresInForm(formSource, form, showHidden));
            }
        }
        return Observable.flatten(points).transform(resultSets -> new PivotPointSet(effectiveModel, resultSets, showHidden));
    }


    private static Observable<FormResultSet> computePointsForAllMeasuresInForm(FormSource formSource, EffectiveFormMeasures form, boolean showHidden) {

        Observable<ColumnSet> columnSet = formSource.query(form.composeQuery());
        Observable<FormResultSet> resultSet = columnSet.transform(columns -> {
            return buildFormResultSet(form, showHidden, columns);
        });

        return resultSet;
    }

    private static Observable<FormResultSet> computePointsForAllMeasuresInReport(FormForest formForest, EffectiveFormMeasures form, boolean showHidden) {
        FormResultSet formResultSet = formForest.findTable(form.getFormId())
            .map(table -> buildFormResultSet(form, showHidden, table.resolve(form.composeQuery())))
            .orElseGet(() -> new FormResultSet(form.getDimensionSet(), Collections.emptyList()));

        return Observable.just(formResultSet);
    }

    private static FormResultSet buildFormResultSet(EffectiveFormMeasures form, boolean showHidden, ColumnSet columns) {
        List<Point> points = new ArrayList<>();

        for (EffectiveMeasure measure : form.getMeasures()) {
            MeasureResultBuilder builder = new MeasureResultBuilder(form, measure, columns, showHidden);
            builder.execute();
            points.addAll(builder.getResult());
        }
        return new FormResultSet(form.getDimensionSet(), points);
    }
}
