/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonArrays;
import org.activityinfo.json.JsonValue;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

public class PivotTable {

    private final Node rootRow;
    private final Node rootColumn;

    private final List<Dimension> rowDimensions;
    private final List<Dimension> columnDimensions;

    private final List<Node> rowLeaves;
    private final List<Node> columnLeaves;

    PivotTable(Node rootRow,
               Node rootColumn,
               List<Dimension> rowDimensions,
               List<Dimension> columnDimensions) {
        this.rootRow = rootRow;
        this.rootColumn = rootColumn;
        this.rowDimensions = rowDimensions;
        this.columnDimensions = columnDimensions;
        this.rowLeaves = new ArrayList<>();
        this.columnLeaves = new ArrayList<>();

        depthFirstSearch(rootRow, rowLeaves);
        depthFirstSearch(rootColumn, columnLeaves);
    }


    /**
     * Recursively perform a depth first search to find leaf nodes, and calculate
     * the span and start of each node.
     * @param parent the node to search
     * @param leaves the collection to which leaf nodes will be added
     */
    private void depthFirstSearch(PivotTable.Node parent, List<PivotTable.Node> leaves) {

        if(parent.isLeaf()) {
            parent.span = 1;
            parent.index = leaves.size();
            leaves.add(parent);
        } else {
            for (PivotTable.Node child : parent.children) {
                child.depth = parent.depth + 1;
                depthFirstSearch(child, leaves);
                parent.span += child.span;
            }
            parent.index = parent.children.get(0).index;
        }
    }

    public static PivotTable create(PivotPointSet results) {
        PivotTableBuilder builder = new PivotTableBuilder(results);
        return builder.build();
    }

    public void forEachValue(Consumer<Double> consumer) {
        forEachValue(rootRow, consumer);
    }

    private void forEachValue(PivotTable.Node parent, Consumer<Double> consumer) {
        for (Cell value : parent.cells.values()) {
            consumer.accept(value.getValue());
        }
        for (Node child : parent.children) {
            forEachValue(child, consumer);
        }
    }

    public List<Dimension> getRowDimensions() {
        return rowDimensions;
    }

    public List<Dimension> getColumnDimensions() {
        return columnDimensions;
    }

    public List<Node> getRowLeaves() {
        return rowLeaves;
    }

    public List<Node> getColumnLeaves() {
        return columnLeaves;
    }

    public boolean isEmpty() {
        return rootRow.isEmpty() && rootColumn.isEmpty();
    }

    public Node getRootRow() {
        return rootRow;
    }

    public Node getRootColumn() {
        return rootColumn;
    }

    public int getRowDimensionCount() {
        return rowDimensions.size();
    }

    public int getColumnDimensionCount() {
        return columnDimensions.size();
    }

    public static class Dimension {
        private final String id;
        private final String label;

        public Dimension(String id, String label) {
            this.id = id;
            this.label = label;
        }

        public Dimension(EffectiveDimension effectiveDimension) {
            this.id = effectiveDimension.getId();
            this.label = effectiveDimension.getLabel();
        }

        public String getId() {
            return id;
        }

        public String getLabel() {
            return label;
        }
    }

    public static class Cell {
        private double value;
        private String formattedValue;

        private Cell() {
        }

        Cell(Point point) {
            this.value = point.getValue();
            this.formattedValue = point.getFormattedValue();
        }

        public String getFormattedValue() {
            return formattedValue;
        }

        public double getValue() {
            return value;
        }
    }

    public static class Node {

        String dimensionId;
        int index = -1;
        Node parent;
        String category;
        String categoryLabel;
        int depth;
        int span;
        boolean hidden;

        Map<String, Node> childMap = new HashMap<>();
        private final Map<Node, Cell> cells = new HashMap<>();

        List<Node> children = new ArrayList<>();

        Node() {
        }

        Node(Node parent, String category) {
            this.parent = parent;
            this.category = category;
        }

        public double getCellValue(Node node) {
            Cell cell = cells.get(node);
            if (cell == null) {
                return Double.NaN;
            } else {
                return cell.value;
            }
        }

        public String getDimensionId() {
            return dimensionId;
        }

        public Node getChild(String category) {
            return childMap.get(category);
        }

        void setValue(Node column, Point value) {
            cells.put(column, new Cell(value));
        }

        public Cell getCell(Node column) {
            return cells.get(column);
        }

        public String getCategoryLabel() {
            return categoryLabel;
        }

        public Map<Node, Cell> getCells() {
            return cells;
        }

        public Node getParent() {
            return parent;
        }

        public List<Node> getChildren() {
            return children;
        }

        /**
         * @return a list of this Node and all its descendants
         */
        public List<Node> descendants() {
            List<Node> list = new ArrayList<>();
            collectDescendants(list);
            return list;
        }

        private void collectDescendants(List<Node> list) {
            for (Node child : children) {
                list.add(child);
                child.collectDescendants(list);
            }
        }

        public boolean isLeaf() {
            return children.isEmpty();
        }

        /**
         * Returns the depth of this node in the tree.
         *
         * <p>The root node has a depth of zero, while the categories of the first
         * dimension have a depth of one, and so on...</p>
         */
        public int getDepth() {
            return depth;
        }

        public int getSpan() {
            return span;
        }

        public boolean isEmpty() {
            return cells.isEmpty() && children.isEmpty();
        }

        public int getIndex() {
            return index;
        }

        public boolean isRoot() {
            return parent == null;
        }

        public boolean isHidden() {
            return hidden;
        }

        public String flattenLabel() {
            StringBuilder sb = new StringBuilder();
            Node node = this;
            do {
                if (node.getCategoryLabel() != null) {
                    if (sb.length() != 0) {
                        sb.append(" ");
                    }
                    sb.append(node.getCategoryLabel());
                }
                node = node.getParent();

            } while (node != null);

            if(sb.length() == 0) {
                return I18N.CONSTANTS.value();
            } else {
                return sb.toString();
            }
        }

    }

    public JsonValue toJson() {

        JsonValue nodeArray = Json.createArray();
        Map<Node, Integer> nodeMap = new HashMap<>();

        // First add all nodes to an array and assign them numbers
        addNodes(rootRow, nodeMap, nodeArray);
        addNodes(rootColumn, nodeMap, nodeArray);

        // Now build the connections between the nodes, using their
        // index in the node array as a pointer.
        nodeMap.forEach( (node, nodeId) -> {

            JsonValue nodeObject = nodeArray.get(nodeId);
            if(node.parent != null) {
                nodeObject.put("parent", nodeMap.get(node.getParent()));
            }
            if(!node.children.isEmpty()) {
                JsonValue childArray = Json.createArray();
                for (Node child : node.getChildren()) {
                    childArray.add(Json.create(nodeMap.get(child)));
                }
                nodeObject.put("children", childArray);
            }
            if(!node.cells.isEmpty()) {
                JsonValue cells = Json.createObject();
                node.cells.forEach( (key, cell) -> {
                    JsonValue cellObject = Json.createObject();
                    cellObject.put("value", cell.getValue());
                    cellObject.put("formattedValue", cell.getFormattedValue());

                    cells.put(Integer.toString(nodeMap.get(key)), cellObject);
                });
                nodeObject.put("cells", cells);
            }
        });

        JsonValue object = Json.createObject();
        object.put("nodes", nodeArray);
        object.put("rootRow", nodeMap.get(rootRow));
        object.put("rootColumn", nodeMap.get(rootColumn));
        object.put("rowDimensions", toJson(rowDimensions));
        object.put("columnDimensions", toJson(columnDimensions));

        return object;
    }


    private void addNodes(Node parent, Map<Node, Integer> nodeMap, JsonValue nodeArray) {
        JsonValue nodeObject = Json.createObject();
        if(parent.dimensionId != null) {
            nodeObject.put("dimensionId", parent.dimensionId);
        }
        if(parent.category != null) {
            nodeObject.put("category", parent.category);
            if (!parent.category.equals(parent.categoryLabel)) {
                nodeObject.put("categoryLabel", parent.categoryLabel);
            }
        }

        if(parent.hidden) {
            nodeObject.put("hidden", parent.hidden);
        }

        nodeMap.put(parent, nodeMap.size());
        nodeArray.add(nodeObject);

        for (Node child : parent.children) {
            addNodes(child, nodeMap, nodeArray);
        }
    }

    private JsonValue toJson(List<Dimension> dimensions) {
        JsonValue dimArray = Json.createArray();
        for (Dimension rowDimension : dimensions) {
            JsonValue dimObject = Json.createObject();
            dimObject.put("id", rowDimension.getId());
            dimObject.put("label", rowDimension.getLabel());
            dimArray.add(dimObject);
        }
        return dimArray;
    }

    public static PivotTable fromJson(JsonValue object) {

        JsonValue nodeArray = object.get("nodes");

        // First construct the nodes with their basic properties
        Node[] nodes = new Node[nodeArray.length()];
        for (int i = 0; i < nodes.length; i++) {
            nodes[i] = nodeFromJson(nodeArray.get(i));
        }

        // Now link the nodes together
        for (int i = 0; i < nodes.length; i++) {
            JsonValue nodeObject = nodeArray.get(i);
            Node node = nodes[i];

            if(nodeObject.hasKey("parent")) {
                node.parent = nodes[(int)nodeObject.getNumber("parent")];
            }
            if(nodeObject.hasKey("children")) {
                JsonValue childArray = nodeObject.get("children");
                for (int j = 0; j < childArray.length(); j++) {
                    node.children.add(nodes[(int) childArray.getNumber(j)]);
                }
            }
            if(nodeObject.hasKey("cells")) {
                JsonValue cellMap = nodeObject.get("cells");
                for (String key : cellMap.keys()) {
                    Node nodeKey = nodes[Integer.parseInt(key)];
                    Cell cell = cellFromJson(cellMap.get(key));
                    node.cells.put(nodeKey, cell);
                }
            }
        }

        Node rootRow = nodes[(int) object.getNumber("rootRow")];
        Node rootColumn = nodes[(int) object.getNumber("rootColumn")];
        List<Dimension> rowDimensions = JsonArrays.listFromJson(object.get("rowDimensions"), PivotTable::dimensionsFromJson);
        List<Dimension> colDimensions = JsonArrays.listFromJson(object.get("columnDimensions"), PivotTable::dimensionsFromJson);

        return new PivotTable(rootRow, rootColumn, rowDimensions, colDimensions);
    }

    private static Dimension dimensionsFromJson(JsonValue object) {
        return new Dimension(object.getString("id"), object.getString("label"));
    }

    private static Cell cellFromJson(JsonValue jsonValue) {
        Cell cell = new Cell();
        cell.value = jsonValue.getNumber("value");
        cell.formattedValue = jsonValue.getString("formattedValue");
        return cell;
    }

    private static Node nodeFromJson(JsonValue nodeObject) {
        Node node = new Node();
        node.dimensionId = nodeObject.getString("dimensionId");
        node.category = nodeObject.getString("category");
        if(nodeObject.hasKey("categoryLabel")) {
            node.categoryLabel = nodeObject.getString("categoryLabel");
        } else {
            node.categoryLabel = node.category;
        }
        if(nodeObject.hasKey("hidden")) {
            node.hidden = nodeObject.getBoolean("hidden");
        }

        return node;
    }



}
