/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import org.activityinfo.model.analysis.pivot.Axis;

import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;

public class PivotTableBuilder {
    private final PivotPointSet results;
    private final List<EffectiveDimension> rowDimensions;
    private final List<EffectiveDimension> columnDimensions;

    private final PivotTable.Node rootRow;
    private final PivotTable.Node rootColumn;

    public PivotTableBuilder(PivotPointSet results) {
        this.results = results;

        rootRow = new PivotTable.Node();
        rootColumn = new PivotTable.Node();

        rowDimensions = results.getEffectiveModel().getDimensions(Axis.ROW);
        columnDimensions = results.getEffectiveModel().getDimensions(Axis.COLUMN);

        for (Point point : results.getPoints()) {
            PivotTable.Node column = columnDimensions.isEmpty() ? rootColumn : find(rootColumn,
                columnDimensions.iterator(),
                point);
            PivotTable.Node row = rowDimensions.isEmpty() ? rootRow : find(rootRow,
                rowDimensions.iterator(),
                point);

            row.setValue(column, point);
        }
    }

    public PivotTable build() {
        return new PivotTable(
            rootRow,
            rootColumn,
            rowDimensions.stream().map(PivotTable.Dimension::new).collect(Collectors.toList()),
            columnDimensions.stream().map(PivotTable.Dimension::new).collect(Collectors.toList()));
    }

    private PivotTable.Node find(PivotTable.Node parentNode, Iterator<EffectiveDimension> dimensionIterator, Point point) {

        EffectiveDimension childDimension = dimensionIterator.next();
        Function<Point, String> categoryProvider = childDimension.getCategoryProvider();

        if(parentNode.dimensionId == null) {
            parentNode.dimensionId = childDimension.getId();
        }

        Comparator<String> categoryComparator = childDimension.getCategoryComparator();

        String category = categoryProvider.apply(point);
        PivotTable.Node child = parentNode.getChild(category);
        if (child == null) {
            child = addChild(
                childDimension,
                parentNode,
                category,
                categoryComparator);
        }
        if (dimensionIterator.hasNext()) {
            return find(child, dimensionIterator, point);
        } else {
            return child;
        }
    }

    public PivotTable.Node addChild(EffectiveDimension dimension, PivotTable.Node parent, String category, Comparator<String> comparator) {

        PivotTable.Node child = new PivotTable.Node(parent, category);
        child.hidden = !dimension.getModel().isCategoryVisible(category);


        if(Point.TOTAL.equals(category)) {
            child.categoryLabel = dimension.getTotalLabel();
        } else {
            child.categoryLabel = category;
        }

        parent.childMap.put(category, child);

        if (comparator == null) {
            parent.children.add(child);
        } else {
            insertChildSorted(parent, child, comparator);
        }
        return child;
    }


    private void insertChildSorted(PivotTable.Node parent, PivotTable.Node child, Comparator<String> comparator) {
        for (int i = 0; i != parent.children.size(); ++i) {
            if (comparator.compare(child.category, parent.children.get(i).category) < 0) {
                parent.children.add(i, child);
                return;
            }
        }
        parent.children.add(child);
    }


}
