/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import com.google.common.base.Strings;
import org.activityinfo.i18n.shared.I18N;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Renders a tree-based {@link PivotTable} into a two-dimensional grid.
 */
public class PivotTableRenderer {

    public static final String LINE_ENDING = "\r\n";

    private static class Coord {

        public Coord(int row, int column) {
            this.row = row;
            this.column = column;
        }

        int row;
        int column;

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            Coord coord = (Coord) o;

            if (row != coord.row) return false;
            return column == coord.column;

        }

        @Override
        public int hashCode() {
            int result = row;
            result = 31 * result + column;
            return result;
        }

        public String toString() {
            return "(" + row + ", " + column + ")";
        }
    }

    private final Map<Coord, String> cells = new HashMap<>();
    private final List<PivotTable.Node> leafColumns;
    private int numRows;
    private int numCols;


    public PivotTableRenderer(PivotTable table) {
        this.leafColumns = table.getColumnLeaves();
        int numHeaderCols = table.getRowDimensionCount();
        int numHeaderRows = Math.max(1, table.getColumnDimensions().size());

        addRowDimensionsHeaders(table);

        if(table.getColumnDimensions().isEmpty()) {
            addCell(0, numHeaderCols, I18N.CONSTANTS.value());
        } else {
            addColumnHeaders(table.getRootColumn(), 0, numHeaderCols);
        }

        addRowHeaders(table.getRootRow(), numHeaderRows, 0);

        addValues(table.getRootRow(), numHeaderRows, 0);
    }

    private void addRowDimensionsHeaders(PivotTable table) {
        for (int i = 0; i < table.getRowDimensionCount(); i++) {
            addCell(0, i, table.getRowDimensions().get(i).getLabel());
        }
    }


    private int addColumnHeaders(PivotTable.Node parent, int row, int col) {

        if(!parent.isLeaf()) {
            for (PivotTable.Node child : parent.getChildren()) {
                addCell(row, col, child.getCategoryLabel());
                if(child.isLeaf()) {
                    col++;
                } else {
                    col = addColumnHeaders(child, row+1, col);
                }
            }
        }
        return col;
    }

    public static String renderPlainText(PivotTable table) {
        return new PivotTableRenderer(table).renderPlainText();
    }

    public static String renderDelimited(PivotTable table, String delimiter) {
        return new PivotTableRenderer(table).renderDelimited(delimiter);
    }

    private int addRowHeaders(PivotTable.Node parent, int row, int col) {
        for (PivotTable.Node child : parent.getChildren()) {
            for (int i = 0; i < child.getSpan(); i++) {

                addCell(row + i, col, child.getCategoryLabel());
            }
            if (child.isLeaf()) {
                row++;
            } else {
                row = addRowHeaders(child, row, col + 1);
            }
        }
        return row;
    }


    private int addValues(PivotTable.Node parentRow, int row, int col) {
        if(parentRow.isLeaf()) {
            for (Map.Entry<PivotTable.Node, PivotTable.Cell> entry : parentRow.getCells().entrySet()) {
                int leafIndex = leafColumns.indexOf(entry.getKey());

                addCell(row, col + leafIndex, entry.getValue().getFormattedValue());
            }
            row++;
        } else {
            for (PivotTable.Node child : parentRow.getChildren()) {
                row = addValues(child, row, col+1);
            }
        }
        return row;
    }

    private void addCell(int row, int col, String label) {
        cells.put(new Coord(row, col), label);
        if(row+1 > numRows) {
            numRows = row+1;
        }
        if(col+1 > numCols) {
            numCols = col+1;
        }
    }

    public String renderDelimited(String delimiter) {
        StringBuilder s = new StringBuilder();
        for (int i = 0; i < numRows; i++) {
            for (int j = 0; j < numCols; j++) {
                if(j > 0) {
                    s.append(delimiter);
                }
                String value = cells.get(new Coord(i,j));
                if(value != null) {
                    s.append(maybeQuote(value));
                }
            }
            s.append(LINE_ENDING);
        }

        return s.toString();
    }

    private String maybeQuote(String value) {
        // Remove newlines, we don't want them anyway
        String stripped = value.replaceAll("[\n\r]+", " ");

        if(stripped.indexOf(',') != -1) {
            return "\"" + stripped.replace("\"", "\"\"") + "\"";
        } else {
            return stripped;
        }
    }

    public String renderPlainText() {
        int columnWidths[] = new int[numCols];
        for (Map.Entry<Coord, String> entry : cells.entrySet()) {
            int col = entry.getKey().column;
            int width = entry.getValue().length();
            if(width > columnWidths[col]) {
                columnWidths[col] = width;
            }
        }

        StringBuilder s = new StringBuilder();
        for (int i = 0; i < numRows; i++) {
            for (int j = 0; j < numCols; j++) {
                int width = columnWidths[j] + 3;
                String value = Strings.nullToEmpty(cells.get(new Coord(i,j)));
                s.append(Strings.padEnd(value, width, ' '));
            }
            s.append('\n');
        }

        return s.toString();
    }

}
