/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel;

import org.activityinfo.analysis.FormForestComputer;
import org.activityinfo.model.analysis.Analysis;
import org.activityinfo.model.analysis.Report;
import org.activityinfo.model.analysis.ReportComponent;
import org.activityinfo.model.analysis.pivot.ImmutableLists;
import org.activityinfo.model.analysis.pivot.PivotModel;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormMetadata;
import org.activityinfo.model.formTree.FormMetadataProvider;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formTree.FormTreeBuilder;
import org.activityinfo.model.formula.ConstantNode;
import org.activityinfo.model.formula.functions.ColumnFunction;
import org.activityinfo.model.formula.functions.FormulaFunction;
import org.activityinfo.model.formula.resolved.*;
import org.activityinfo.model.permission.FormPermissions;
import org.activityinfo.model.query.*;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.number.QuantityType;
import org.activityinfo.model.type.primitive.TextType;
import org.activityinfo.observable.Observable;
import org.activityinfo.promise.Maybe;
import org.activityinfo.store.query.shared.FormSource;

import java.util.*;
import java.util.function.Supplier;


/**
 * Represents a Pivot Table that is used as a source for another pivot table.
 *
 */
public class PivotTableSource {

    private ResourceId id;
    private PivotTable table;
    private FormTree formTree;
    private Map<String, Supplier<ColumnView>> columns = new HashMap<>();
    private int numRows;
    private BlankColumnView blankColumn;

    public PivotTableSource(String id, String label, PivotTable table) {
        this.id = ResourceId.valueOf(id);
        this.table = table;
        this.numRows = table.getRowLeaves().size();
        this.blankColumn = new BlankColumnView(ColumnType.STRING, numRows);

        // Create a simple FormTree for this report
        FormClass formClass = new FormClass(ResourceId.valueOf(id));
        formClass.setLabel(label);
        formClass.setDatabaseId(ResourceId.valueOf("REPORT"));
        formClass.setStorage("report");

        for (PivotTable.Dimension rowDimension : table.getRowDimensions()) {
            columns.put(rowDimension.getId(), new DimensionSupplier(rowDimension));
            formClass.addField(ResourceId.valueOf(rowDimension.getId()))
                .setType(TextType.SIMPLE)
                .setLabel(rowDimension.getLabel());
        }
        for (PivotTable.Node columnLeaf : table.getColumnLeaves()) {
            String columnLabel = columnLeaf.flattenLabel();
            columns.put(columnLabel, new ValueSupplier(columnLeaf));
            formClass.addField(ResourceId.valueOf(columnLabel))
                .setType(new QuantityType())
                .setLabel(columnLabel);
        }

        FormTreeBuilder formTreeBuilder = new FormTreeBuilder(new FormMetadataProvider() {
            @Override
            public FormMetadata getFormMetadata(ResourceId formId) {
                if (formId.equals(formClass.getId())) {
                    return FormMetadata.of(formClass, FormPermissions.readonly());
                } else {
                    return FormMetadata.notFound(formId);
                }
            }
        });

        this.formTree = formTreeBuilder.queryTree(formClass.getId());
    }

    public ResourceId getId() {
        return id;
    }

    public PivotTable getTable() {
        return table;
    }

    public FormTree getFormTree() {
        return formTree;
    }

    public static Observable<Optional<PivotTableSource>> compute(FormSource formSource,
                                                         ResourceId reportId) {

        Observable<Maybe<Report>> reportObservable = formSource.getReport(reportId.asString());
        return reportObservable.join(maybeReport -> {
            if (!maybeReport.isVisible()) {
                return Observable.just(Optional.empty());
            }

            Report report = maybeReport.get();

            Optional<String> analysisId = report.getSinglePivotAnalysisId();
            if(!analysisId.isPresent()){
                return Observable.just(Optional.empty());
            }

            Optional<ReportComponent> component = ImmutableLists.get(report.getComponents(), analysisId.get(), c->c.getId());
            if(!component.isPresent()){
                return Observable.just(Optional.empty());
            }
            String analysisLabel = component.get().getContent().orElse("");
            Observable<FormForest> formForest = FormForestComputer.compute(formSource, report.getSources());
            Observable<Maybe<Analysis>> analysis = formSource.getAnalysis(reportId.asString(), analysisId.get());

            return Observable.join(formForest, analysis, (forest_, maybeAnalysis) -> {
                if (!maybeAnalysis.isVisible()) {
                    return Observable.just(Optional.empty());
                }
                Analysis analysis_ = maybeAnalysis.get();
                if (!(analysis_.getModel() instanceof PivotModel)) {
                    return Observable.just(Optional.empty());
                }
                PivotModel pivotModel = (PivotModel) analysis_.getModel();
                PivotViewModel viewModel = new PivotViewModel(formSource, Observable.just(forest_), Observable.just(pivotModel));
                return viewModel.getPivotTable().transform(table -> Optional.of(
                        new PivotTableSource(reportId.asString(), analysisLabel, table)));
            });
        });
    }

    public ColumnSet resolve(QueryModel queryModel) {

        FormulaResolver resolver = new FormulaResolver(formTree);

        Map<String, ColumnView> columns = new HashMap<>();
        for (ColumnModel column : queryModel.getColumns()) {
            ResolvedNode node = resolver.resolve(column.getFormula());
            columns.put(column.getId(), resolve(node));
        }
        return new ColumnSet(numRows, columns);
    }

    private class DimensionSupplier implements Supplier<ColumnView> {
        private final String rowDimensionId;
        private ColumnView result = null;

        public DimensionSupplier(PivotTable.Dimension rowDimension) {
            this.rowDimensionId = rowDimension.getId();
        }

        @Override
        public ColumnView get() {
            if(result == null) {
                String[] values = new String[numRows];
                int rowIndex = 0;
                for (PivotTable.Node rowLeaf : table.getRowLeaves()) {
                    values[rowIndex++] = findCategoryLabel(rowLeaf);
                }
                result = new StringArrayColumnView(values);
            }
            return result;
        }

        private String findCategoryLabel(PivotTable.Node rowNode) {
            while(true) {
                if(rowNode == null || rowNode.getParent() == null) {
                    return "";
                }
                if(rowNode.getParent().getDimensionId().equals(rowDimensionId)) {
                    return rowNode.getCategoryLabel();
                }
                rowNode = rowNode.getParent();
            }
        }
    }

    private class ValueSupplier implements Supplier<ColumnView> {

        private final PivotTable.Node columnLeaf;
        private ColumnView result = null;

        public ValueSupplier(PivotTable.Node columnLeaf) {
            this.columnLeaf = columnLeaf;
        }

        @Override
        public ColumnView get() {
            if(result == null) {
                double[] values = new double[numRows];
                int rowIndex = 0;
                for (PivotTable.Node rowLeaf : table.getRowLeaves()) {
                    values[rowIndex++] = rowLeaf.getCellValue(columnLeaf);
                }
                result = new DoubleArrayColumnView(values);
            }
            return result;
        }
    }

    public ColumnView resolve(ResolvedNode formula) {
        return formula.accept(new ResolvedNodeVisitor<ColumnView>() {
            @Override
            public ColumnView visitConstant(ConstantNode node) {
                return new ConstantColumnView(numRows, node.getValue());
            }

            @Override
            public ColumnView visitCallNode(ResolvedCallNode node) {
                FormulaFunction function = node.getFunction();
                if(!(function instanceof ColumnFunction)) {
                    return blankColumn;
                }
                ColumnFunction columnFunction = (ColumnFunction) function;
                List<ColumnView> arguments = new ArrayList<>();
                for (int i = 0; i < node.getArgumentCount(); i++) {
                    arguments.add(resolve(node.getArgument(i)));
                }
                return columnFunction.apply(numRows, arguments);
            }

            @Override
            public ColumnView visitReferenceNode(ResolvedReference node) {
                return node.getRoot().accept(new ReferenceVisitor<ColumnView>() {
                    @Override
                    public ColumnView visitRecordId(RecordIdNode recordIdNode) {
                        return blankColumn;
                    }

                    @Override
                    public ColumnView visitFormId(FormIdNode formIdNode) {
                        return new ConstantColumnView(numRows, id.asString());
                    }

                    @Override
                    public ColumnView visitLastEditTime(LastEditTimeNode node) {
                        return new ConstantColumnView(numRows, Double.NaN);
                    }

                    @Override
                    public ColumnView visitField(FieldReferenceNode field) {
                        Supplier<ColumnView> supplier = columns.get(field.getField().getId().asString());
                        if(supplier == null) {
                            return blankColumn;
                        }
                        return supplier.get();
                    }

                    @Override
                    public ColumnView visitCalculatedField(CalculatedFieldNode field) {
                        // No calculated fields in pivot tables
                        return blankColumn;
                    }

                    @Override
                    public ColumnView visitEnumItem(EnumItemReference enumItemReference) {
                        // No enum fields in pivot tables
                        return blankColumn;
                    }

                    @Override
                    public ColumnView visitJoin(JoinQueryNode node) {
                        // No reference fields in pivot tables
                        return blankColumn;
                    }

                    @Override
                    public ColumnView visitUnion(UnionNode node) {
                        // No reference fields, so should not occur.
                        return blankColumn;
                    }

                    @Override
                    public ColumnView visitAggregation(AggregateNode node) {
                        // No subform fields, so should not occur
                        return blankColumn;
                    }
                });
            }

            @Override
            public ColumnView visitInvalidReferenceNode(InvalidReferenceNode node) {
                return blankColumn;
            }

            @Override
            public ColumnView visitInvalidCall(InvalidCallNode invalidCallNode) {
                return blankColumn;
            }

            @Override
            public ColumnView visitInvalidArgumentNode(InvalidArgumentNode invalidArgumentNode) {
                return blankColumn;
            }

            @Override
            public ColumnView visitInvalidSyntax(InvalidSyntaxNode node) {
                return blankColumn;
            }
        });
    }



}
