/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.MajorCategoryLabelDescriptor;
import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.i18n.shared.TextDirection;
import org.activityinfo.model.analysis.pivot.Axis;
import org.activityinfo.model.analysis.pivot.VisualizationType;

import java.util.List;
import java.util.Map;

/**
 * Shared layout data for Bar charts and Line charts
 */
public class AxisChartLayout implements MajorCategoryLabelDescriptor, ChartSpace {

    public static final double SCALE_LABEL_SIZE = 14;

    /**
     * The space between bars
     */
    public static final double BAR_MARGIN = 40;

    public static final double AUTO_BAR_WIDTH = 50;

    public static final double MARGIN_LEFT = 25;
    public static final double MARGIN_RIGHT = 25;

    /**
     * Padding between the scale and the grid lines
     */
    public static final double SCALE_PADDING = 23;

    /**
     * The top margin, measured from the maximum scale line to the top of the
     * graphic canvas
     */
    public static final double MARGIN_TOP = 55;

    public static final double CATEGORY_LABEL_MARGIN_TOP = 7;

    public static final double MINOR_CATEGORY_LABEL_LINE_HEIGHT = 22;

    public static final double MINOR_CATEGORY_LABEL_FONT_SIZE = 14;

    public static final double MAJOR_CATEGORY_LABEL_LINE_HEIGHT = 30;

    public static final double MAJOR_CATEGORY_LABEL_FONT_SIZE = 18;

    public static final int MAXIMUM_MINOR_CATEGORY_LABEL_LINES = 3;
    public static final int MAXIMUM_MAJOR_CATEGORY_LABEL_LINES = 5;
    public static final int MINOR_CATEGORY_LABEL_LINES_MARGIN = 8;
    public static final int MAJOR_CATEGORY_LABEL_LINES_MARGIN = 12;

    /**
     * The bottom margin, measured from the lowest label
     */
    public static final double MARGIN_BOTTOM = 38;

    public static final double LEGEND_PREFIX_WIDTH = 45;
    public static final double LEGEND_MAXIMUM_WIDTH = 128;

    private final double legendTextWidth;


    private final PivotTable pivotTable;
    private final double chartHeight;
    private final double chartWidth;
    private final Scale scale;
    private final double barWidth;
    private final double scaleWidth;

    private final Rect plotArea;
    private final Rect scaleArea;
    private final int minorDimensionCount;
    private final int majorDimensionCount;

    private final double majorCategoryWidth;
    private final double majorCategoryLabelTop;
    private final double minorCategoryLabelTop;

    private final double[] majorCategoryLabelHeights;
    private final double[] minorCategoryLabelHeights;
    private final Map<PivotTable.Node, String[]> minorCategoryTextLinesMap;
    private final Map<PivotTable.Node, String[]> majorCategoryTextLinesMap;

    AxisChartLayout(PivotTable pivotTable, VisualizationType type, double chartWidth, double chartHeight, Scale scale, double barWidth,
                    double scaleWidth, double LegendTextWidth, double[] minorCategoryLabelHeights, double[] majorCategoryLabelHeights,
                    Map<PivotTable.Node, String[]> minorCategoryTextLinesMap, Map<PivotTable.Node, String[]> majorCategoryTextLinesMap) {
        this.pivotTable = pivotTable;
        this.legendTextWidth = LegendTextWidth;
        this.chartWidth = chartWidth+getLegendWidth(type);
        this.chartHeight = chartHeight;
        this.scale = scale;
        this.barWidth = barWidth;
        this.scaleWidth = scaleWidth;
        this.minorCategoryLabelHeights = minorCategoryLabelHeights;
        this.majorCategoryLabelHeights = majorCategoryLabelHeights;
        this.minorCategoryTextLinesMap = minorCategoryTextLinesMap;
        this.majorCategoryTextLinesMap = majorCategoryTextLinesMap;

        majorDimensionCount = pivotTable.getRowDimensionCount();
        minorDimensionCount = pivotTable.getColumnDimensionCount();

        majorCategoryWidth = getMajorCategoryWidth(pivotTable, barWidth);

        double categoryLabelHeights =
                CATEGORY_LABEL_MARGIN_TOP +
                sum(minorCategoryLabelHeights) +
                sum(majorCategoryLabelHeights);

        double plotLeft = MARGIN_LEFT + scaleWidth + SCALE_PADDING;
        double scaleLeft = MARGIN_LEFT;
        if(TextDirection.currentLocale().equals(TextDirection.RTL)){
            plotLeft += getLegendWidth(type);
            scaleLeft += getLegendWidth(type);
            chartWidth = this.chartWidth;
        }

        this.plotArea = new Rect(
                plotLeft,
                MARGIN_TOP,
                chartWidth - MARGIN_RIGHT,
                chartHeight - MARGIN_BOTTOM - categoryLabelHeights);


        this.scaleArea = new Rect(
                scaleLeft,
                plotArea.getTop(),
                scaleLeft + scaleWidth,
                plotArea.getBottom());

        minorCategoryLabelTop = plotArea.getBottom() + CATEGORY_LABEL_MARGIN_TOP;
        majorCategoryLabelTop = type.equals(VisualizationType.BARCHART)?(minorCategoryLabelTop
                + sum(minorCategoryLabelHeights)):minorCategoryLabelTop;
    }

    public static double sum(double[] x) {
        double sum = 0;
        for (int i = 0; i < x.length; i++) {
            sum += x[i];
        }
        return sum;
    }

    public static double sum(double[] x, int count) {
        double sum = 0;
        for (int i = 0; i < count; i++) {
            sum += x[i];
        }
        return sum;
    }

    private double getLegendWidth(VisualizationType type){
        if(hasLegend(type)){
            return LEGEND_PREFIX_WIDTH + getLengendTextWidth();
        } else {
            return 0;
        }
    }

    public boolean hasLegend(VisualizationType type){
        return type.equals(VisualizationType.LINECHART) && getMinorCategories().size() > 0;
    }

    @Override
    public double getChartWidth() {
        return chartWidth;
    }

    @Override
    public double getChartHeight() {
        return chartHeight;
    }

    @Override
    public String getViewBox() {
        return "0 0 " + chartWidth + " " + chartHeight;
    }

    public Scale getScale() {
        return scale;
    }

    public double getScaleWidth(){
        return scaleWidth;
    }

    public int getGridLineCount() {
        return scale.getDivisions() + 1;
    }

    public Rect getPlotArea() {
        return plotArea;
    }

    public Rect getScaleArea() {
        return scaleArea;
    }

    public PivotTable getPivotTable() {
        return pivotTable;
    }

    public double valueY(double value) {
        double p = (value - scale.getLower()) / scale.getRange();
        return plotArea.getBottom() - (plotArea.getHeight() * p);
    }

    /**
     * Computes the left coordinate of the major category bar(s)
     * @param index the zero-based index of the category.
     */
    public double getMajorCategoryLeft(int index) {
        return plotArea.getLeft() + BAR_MARGIN + (index * (majorCategoryWidth + BAR_MARGIN));
    }

    public Rect getBarRect(int majorCategoryIndex, int minorCategoryIndex, double value) {
        double left = getMajorCategoryLeft(majorCategoryIndex) + (barWidth * minorCategoryIndex);
        double top = valueY(value);
        double right = left + barWidth;
        double bottom = plotArea.getBottom();

        return new Rect(left, top, right, bottom);
    }

    public double[] getLineCoord(PivotTable.Node majorCategoryLeaf, double value) {
        Rect rect = getMajorCategoryLabelWidthRect(majorCategoryLeaf); //only width necessary, height not pertinent
        double x = rect.getCenterX();
        double y = valueY(value);
        double[] coord =  {x, y};
        return coord;
    }

    @Override
    public Rect getMajorCategoryLabelRect(PivotTable.Node node) {

        int level = (this.majorDimensionCount - node.getDepth());
        Rect widthRect = getMajorCategoryLabelWidthRect(node);
        double left = widthRect.getLeft();
        double top = majorCategoryLabelTop + sum(majorCategoryLabelHeights, level);
        double right = widthRect.getRight();
        double bottom = top + majorCategoryLabelHeights[level];

        return new Rect(left, top, right, bottom);
    }

    private Rect getMajorCategoryLabelWidthRect(PivotTable.Node node) {
        double left = getMajorCategoryLeft(node.getIndex()) - majorCategoryLeftOffset();
        double right = left +
                (node.getSpan() * majorCategoryWidth) +
                ((node.getSpan() - 1) * BAR_MARGIN);

        return new Rect(left, 0, right, 0);
    }

    public Rect getMinorCategoryLabelRect(PivotTable.Node majorCategoryLeaf, PivotTable.Node minorCategory) {

        int level = (this.minorDimensionCount - minorCategory.getDepth());

        double left = getMajorCategoryLeft(majorCategoryLeaf.getIndex()) + (minorCategory.getIndex() * barWidth);
        double top = minorCategoryLabelTop + sum(minorCategoryLabelHeights, level);
        double right = left + (minorCategory.getSpan() * barWidth);
        double bottom = top + minorCategoryLabelHeights[level];

        return new Rect(left, top, right, bottom);
    }

    private static double majorCategoryLeftOffset(){
        return BAR_MARGIN/2 - 5;
    }

    public double getCategoryLabelLineHeight(Axis axis) {
        if(axis.isMajor()) {
            return MAJOR_CATEGORY_LABEL_LINE_HEIGHT;
        } else {
            return MINOR_CATEGORY_LABEL_LINE_HEIGHT;
        }
    }

    public List<PivotTable.Node> getMajorLeafCategories() {
        return getMajorLeafCategories(pivotTable);
    }

    public List<PivotTable.Node> getMinorLeafCategories() {
        return getMinorLeafCategories(pivotTable);
    }

    public List<PivotTable.Node> getMajorCategories() {
        return getMajorCategories(pivotTable);
    }

    public List<PivotTable.Node> getMinorCategories() {
        return getMinorCategories(pivotTable);
    }

    public static List<PivotTable.Node> getMajorLeafCategories(PivotTable pivotTable) {
        return pivotTable.getRowLeaves();
    }

    public static  List<PivotTable.Node> getMinorLeafCategories(PivotTable pivotTable) {
        return pivotTable.getColumnLeaves();
    }

    public static List<PivotTable.Node> getMajorCategories(PivotTable pivotTable) {
        return pivotTable.getRootRow().descendants();
    }

    public static List<PivotTable.Node> getMinorCategories(PivotTable pivotTable) {
        return pivotTable.getRootColumn().descendants();
    }

    public static Rect getMajorCategoryDivider(Rect labelRect, double top, double margin) {
        double x = labelRect.getLeft() - (margin / 2d);
        double bottom = labelRect.getBottom();

        return new Rect(x, top, x, bottom);
    }

    public double getLengendTextWidth(){
        return Math.min(legendTextWidth, LEGEND_MAXIMUM_WIDTH);
    }

    public String[] getMinorTextLines(PivotTable.Node leafNode){
        return minorCategoryTextLinesMap.get(leafNode);
    }

    @Override
    public String[] getMajorTextLines(PivotTable.Node leafNode){
        return majorCategoryTextLinesMap.get(leafNode);
    }

    public static double getMajorCategoryWidth(PivotTable pivotTable, double barWidth){
        return (Math.max(1, getMinorLeafCategories(pivotTable).size()) * barWidth) +  (majorCategoryLeftOffset() * 2);
    }
}
