/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.model.analysis.pivot.VisualizationType;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class AxisChartLayoutBuilder {

    private final PivotTable pivotTable;
    private final Scale scale;
    private final ChartFontContext fontContext;
    private final double scaleWidth;
    private final VisualizationType type;
    private final double[] minorCategoryLabelHeights;
    private final double[] majorCategoryLabelHeights;
    private final Map<PivotTable.Node, String[]> minorCategoryTextLinesMap;
    private final Map<PivotTable.Node, String[]> majorCategoryTextLinesMap;

    private final double chartHeight = 450;
    private double chartWidth;

    private double barWidth;

    public AxisChartLayoutBuilder(PivotTable pivotTable, VisualizationType type, ChartFontContext fontContext) {
        this.fontContext = fontContext;
        this.pivotTable = pivotTable;
        this.type = type;
        this.scale = new NiceChartScaleBuilder().values(pivotTable).build(type);
        this.scaleWidth = computeScaleWidth(scale);
        this.majorCategoryLabelHeights = new double[pivotTable.getRowDimensionCount()];
        this.minorCategoryLabelHeights = new double[pivotTable.getColumnDimensionCount()];
        this.majorCategoryTextLinesMap = new HashMap<>();
        this.minorCategoryTextLinesMap = new HashMap<>();
    }

    private double computeChartWidth(PivotTable pivotTable, double barWidth, double barMargin) {
        int numMajorCategories = AxisChartLayout.getMajorLeafCategories(pivotTable).size();
        int numMinorCategories = AxisChartLayout.getMinorLeafCategories(pivotTable).size();


        double majorCategoryWidth = AxisChartLayout.getMajorCategoryWidth(pivotTable, barWidth);

        double plotAreaWidth = (numMajorCategories * majorCategoryWidth) +
            ((numMajorCategories + 1) * barMargin);

        return AxisChartLayout.MARGIN_LEFT +
            scaleWidth + AxisChartLayout.SCALE_PADDING +
            plotAreaWidth +
            AxisChartLayout.MARGIN_RIGHT;
    }

    private double computeScaleWidth(Scale scale) {

        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.SCALE_LABELS);

        double width = 0;
        for (String label : scale.getLabels()) {
            double labelWidth = font.measureWidth(label);
            if(labelWidth > width) {
                width = labelWidth;
            }
        }
        return width;
    }

    public static double getLengendTextWidth(PivotTable pivotTable, ChartFontContext fontContext){
        List<PivotTable.Node> minorList = AxisChartLayout.getMinorLeafCategories(pivotTable);
        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.MINOR_CATEGORY_LABELS);
        double max = 0;
        for(PivotTable.Node minor : minorList){
            String label = minor.flattenLabel();
            double textWidth = font.measureWidth(label);
            if(textWidth > max){
                max = textWidth;
            }
        }
        return max;
    }

    private double getLengendTextWidth(){
        return getLengendTextWidth(pivotTable, fontContext);
    }

    public AxisChartLayout build() {

        barWidth = AxisChartLayout.AUTO_BAR_WIDTH;
        double barMargin = AxisChartLayout.BAR_MARGIN;
        double chartWidth = computeChartWidth(pivotTable, barWidth, barMargin);
        TextLineOperator minorLabelOperator = new TextLineOperator(fontContext.getFontMetrics(ChartFontStyle.MINOR_CATEGORY_LABELS), barWidth);
        TextLineOperator majorLabelOperator = new TextLineOperator(fontContext.getFontMetrics(ChartFontStyle.MAJOR_CATEGORY_LABELS), AxisChartLayout.getMajorCategoryWidth(pivotTable, barWidth));
        computeCategoryLabelsWrap(AxisChartLayout.getMinorCategories(pivotTable), AxisChartLayout.MAXIMUM_MINOR_CATEGORY_LABEL_LINES,
                AxisChartLayout.MINOR_CATEGORY_LABEL_LINES_MARGIN, pivotTable.getColumnDimensionCount(),
                minorLabelOperator, minorCategoryTextLinesMap, minorCategoryLabelHeights);
        computeCategoryLabelsWrap(AxisChartLayout.getMajorCategories(pivotTable), AxisChartLayout.MAXIMUM_MAJOR_CATEGORY_LABEL_LINES,
                AxisChartLayout.MAJOR_CATEGORY_LABEL_LINES_MARGIN, pivotTable.getRowDimensionCount(),
                majorLabelOperator, majorCategoryTextLinesMap, majorCategoryLabelHeights);

        return new AxisChartLayout(pivotTable, type, chartWidth, chartHeight, scale, barWidth, scaleWidth,getLengendTextWidth(),
                minorCategoryLabelHeights, majorCategoryLabelHeights,
                minorCategoryTextLinesMap, majorCategoryTextLinesMap);
    }

    public static void computeCategoryLabelsWrap(List<PivotTable.Node> leaves, int maxLines,
                                                                 double labelMargin, int dimensionCount,TextLineOperator operator,
                                                                 Map<PivotTable.Node, String[]> textLinesMap, double[] heights) {
        for (PivotTable.Node columnLeaf : leaves) {
            int level = (dimensionCount - columnLeaf.getDepth());
            String[] wrapped = wrap(columnLeaf.getCategoryLabel(), maxLines, operator);
            textLinesMap.put(columnLeaf, wrapped);
            double nodeHeight = (operator.getFontAscent() * wrapped.length) + labelMargin;
            if(heights[level] < nodeHeight){
                heights[level] = nodeHeight;
            }
        }
    }

    private static String[] wrap(String categoryLabel, int maximumNumLines, TextLineOperator operator) {
        List<String> wrappedLines = operator.wrap(categoryLabel);
        String[] lines = new String[Math.min(maximumNumLines, wrappedLines.size())];
        for(int i = 0; i< lines.length-1; i++){
            lines[i] = wrappedLines.get(i);
        }
        String lastLine = wrappedLines.get(lines.length - 1);
        if(wrappedLines.size() > lines.length) {
            lastLine += TextLineOperator.ELIPSES;
            lastLine = operator.truncate(lastLine); //append above might exceed naxWidth
        }
        lines[lines.length - 1] = lastLine;
        return TextLineOperator.trim(lines);
    }


}
