/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.MajorCategoryLabelDescriptor;
import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.analysis.pivot.viewModel.ReportStyle;
import org.activityinfo.ui.vdom.shared.html.SvgTag;
import org.activityinfo.ui.vdom.shared.tree.*;

import java.util.ArrayList;
import java.util.List;

public abstract class GraphChartRenderer {
    public static final String CATEGORY_ATTRIBUTE = "data-category";
    public static final String CATEGORY_DIM_ATTRIBUTE = "data-category-dim";

    protected final AxisChartLayout layout;
    protected final ChartFontContext fontContext;

    protected GraphChartRenderer(AxisChartLayout layout, ChartFontContext fontContext) {
        this.layout = layout;
        this.fontContext = fontContext;
    }

    protected abstract VNode renderChartData();

    protected abstract VNode renderLegend();

    public VNode render(AxisChartLayout layout) {
        VAttrMap svgProps = getChartProperties(layout);

        return new VElement(SvgTag.SVG, svgProps,
                background(layout),
                renderChartData(),
                gridLines(),
                minorCategoryLabels(),
                majorCategoryLabels(layout.getPivotTable(),fontContext, layout, layout.getPlotArea().getBottom(), AxisChartLayout.BAR_MARGIN),
                scaleLabels(),
                renderLegend());
    }

    public static VAttrMap getChartProperties(ChartSpace chartSpace){
        VAttrMap svgProps = VAttr.create();
        svgProps.set("viewBox", chartSpace.getViewBox());
        svgProps.set("width", chartSpace.getChartWidth());
        svgProps.set("height", chartSpace.getChartHeight());
        return svgProps;
    }


    public static VNode background(ChartSpace chartSpace) {
        VAttrMap props = VAttr.create();
        props.set("fill", "white");
        props.set("x", 0);
        props.set("y", 0);
        props.set("width", chartSpace.getChartWidth());
        props.set("height", chartSpace.getChartHeight());

        return new VElement(SvgTag.RECT, props);
    }

    protected VNode gridLines() {

        SvgPathBuilder path = new SvgPathBuilder();

        double value = layout.getScale().getLower();
        for (int i = 0; i < layout.getGridLineCount(); i++) {
            path.moveTo(layout.getPlotArea().getLeft(), layout.valueY(value));
            path.horizontalLineTo(layout.getPlotArea().getRight());
            value += layout.getScale().getUnit();
        }

        VAttrMap pathProps = VAttr.create();
        pathProps.set("stroke", ColorScales.BLACK);
        pathProps.set("stroke-opacity", "0.1");
        pathProps.set("stroke-width", "1");
        pathProps.set("d", path.toString());

        return new VElement(SvgTag.PATH, pathProps);
    }

    public static VNode majorCategoryLabels(PivotTable pivotTable, ChartFontContext fontContext, MajorCategoryLabelDescriptor categoryDescriptor, double dividerTop, double dividerMargin) {
        List<VNode> labels = new ArrayList<>();

        List<PivotTable.Node> categories = AxisChartLayout.getMajorCategories(pivotTable);

        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.MAJOR_CATEGORY_LABELS);

        for (PivotTable.Node category : categories) {
            Rect rect = categoryDescriptor.getMajorCategoryLabelRect(category);
            String[] lines = categoryDescriptor.getMajorTextLines(category);
            labels.add(categoryLabel(rect, category, lines, font.getAscent()));

            if(!category.isLeaf() && category.getIndex() != 0) {
                labels.add(categoryDivider(rect, dividerTop, dividerMargin));
            }
        }

        VAttrMap style = VAttr.create();
        style.set("font", font.getSvgFontString());

        VAttrMap groupProps = VAttr.create();
        groupProps.set("style", style);

        return new VElement(SvgTag.GROUP, groupProps, labels);
    }

    /**
     * Add a vertical line to separate nested categories
     */
    private static VNode categoryDivider(Rect categoryRect, double top, double margin) {

        Rect rect = AxisChartLayout.getMajorCategoryDivider(categoryRect, top, margin);

        VAttrMap props = VAttr.create();
        props.set("x1", rect.getLeft());
        props.set("y1", rect.getTop());
        props.set("x2", rect.getLeft());
        props.set("y2", rect.getBottom());
        props.set("stroke", ColorScales.BLACK);
        props.set("stroke-opacity", "0.1");
        props.set("stroke-width", "1");

        return new VElement(SvgTag.LINE, props);
    }

    public static VNode debugRect(Rect rect) {
        VAttrMap props = VAttr.create();
        props.set("fill", "none");
        props.set("stroke", ColorScales.BLACK);
        props.set("stroke-opacity", "0.1");
        props.set("stroke-width", "1");
        props.set("x", rect.getLeft());
        props.set("y", rect.getTop());
        props.set("width", rect.getWidth());
        props.set("height", rect.getHeight());

        return new VElement(SvgTag.RECT, props);
    }

    protected VNode minorCategoryLabels() {
        List<VNode> labels = new ArrayList<>();

        List<PivotTable.Node> majorLeaves = layout.getMajorLeafCategories();
        List<PivotTable.Node> minorCategories = layout.getMinorCategories();

        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.MINOR_CATEGORY_LABELS);

        for (PivotTable.Node majorLeaf : majorLeaves) {

            for (PivotTable.Node minorCategory : minorCategories) {
                Rect rect = layout.getMinorCategoryLabelRect(majorLeaf, minorCategory);
                String[] lines = layout.getMinorTextLines(minorCategory);
                labels.add(categoryLabel(rect, minorCategory, lines, font.getAscent()));
            }
        }

        VAttrMap style = VAttr.create();
        style.set("font", font.getSvgFontString());

        VAttrMap groupProps = VAttr.create();
        groupProps.set("style", style);

        return new VElement(SvgTag.GROUP, groupProps, labels);
    }

    private static VElement categoryLabel(Rect rect, PivotTable.Node node, String[] lines, double ascent) {
        List<VNode> spans = new ArrayList<>();
        double dy = rect.getTop() + ascent;
        for(String line : lines){
            VAttrMap textLineProps = VAttr.create();
            textLineProps.set("x", rect.getCenterX());
            textLineProps.set("y", dy);
            textLineProps.set("text-anchor", "middle");
            if(node.isHidden()) {
                textLineProps.set("fill-opacity", "0.3");
            }
            if(node.getParent() != null) {
                textLineProps.set(CATEGORY_DIM_ATTRIBUTE, node.getParent().getDimensionId());
                textLineProps.set(CATEGORY_ATTRIBUTE, node.getCategoryLabel());
            }
            spans.add(new VElement(SvgTag.TEXT, textLineProps, new VText(line)));
            dy += ascent;
        }

        VAttrMap labelProps = VAttr.create();
        return new VElement(SvgTag.GROUP, labelProps, spans.stream());
    }


    protected VNode scaleLabels() {

        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.SCALE_LABELS);

        List<VNode> labels = new ArrayList<>();

        double value = layout.getScale().getLower();
        for (int i = 0; i < layout.getGridLineCount(); i++) {
            String label = layout.getScale().getLabel(i);

            VAttrMap labelProps = VAttr.create();
            labelProps.set("x", layout.getScaleArea().getRight());
            labelProps.set("y", layout.valueY(value) + font.getAscent() * 0.5);
            labelProps.set("text-anchor", "end");

            labels.add(new VElement(SvgTag.TEXT, labelProps, new VText(label)));

            value += layout.getScale().getUnit();
        }

        VAttrMap style = VAttr.create();
        style.set("font", font.getSvgFontString());

        VAttrMap groupProps = VAttr.create();
        groupProps.set("style", style);

        return new VElement(SvgTag.GROUP, groupProps, labels);
    }

    public static String graphElementColor(int minorIndex) {
        final String[] CHARTCOLORS = ReportStyle.ACTCTIVITYINFO.getColorScheme(7);
        int colorIndex = minorIndex % CHARTCOLORS.length;
        return "#" + CHARTCOLORS[colorIndex];
    }
}
