/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.i18n.shared.TextDirection;
import org.activityinfo.model.analysis.pivot.VisualizationType;
import org.activityinfo.ui.vdom.shared.html.H;
import org.activityinfo.ui.vdom.shared.html.SvgTag;
import org.activityinfo.ui.vdom.shared.tree.*;

import java.util.ArrayList;
import java.util.List;

public class LineChartRenderer extends GraphChartRenderer {
    final static int LEGENDRADIUS = 5;

    public LineChartRenderer(AxisChartLayout layout, ChartFontContext fontContext) {
        super(layout, fontContext);
    }

    @Override
    protected VNode renderChartData() {
        List<PivotTable.Node> major = layout.getMajorLeafCategories();
        List<PivotTable.Node> minor = layout.getMinorLeafCategories();

        List<VNode> lines = new ArrayList<>();
        for (int j = 0; j < minor.size(); j++) {
            String color = graphElementColor(j);
            SvgPathBuilder path = new SvgPathBuilder();
            for (int i = 0; i < major.size(); i++) {
                double point = major.get(i).getCellValue(minor.get(j));
                if(Double.isFinite(point) ) {
                    double[] coord = layout.getLineCoord(major.get(i), point);
                    if(path.isPathStarting()){
                        path.moveTo(coord);
                    } else {
                        path.lineTo(coord);
                        path.moveTo(path.getCurrentCoord());
                    }

                    //render legend icon on plot point
                    LegendIcon icon = LegendIcon.getIcon(j);
                    double[] dims = icon.getDimensionsFromRadius(LEGENDRADIUS - 1);
                    coord = icon.getCenterCoord(coord[0], coord[1], dims[0], dims[1]);
                    lines.add(icon.render(coord[0], coord[1], LEGENDRADIUS - 1, color));

                }
            }
            VAttrMap pathProps = VAttr.create();
            pathProps.set("stroke", color);
            pathProps.set("stroke-width", "1");
            pathProps.set("d", path.toString());

            lines.add(new VElement(SvgTag.PATH, pathProps));
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), lines);
    }

    @Override
    protected VNode minorCategoryLabels() {
        return H.t("");
    }

    @Override
    protected VNode renderLegend(){
        if(!layout.hasLegend(VisualizationType.LINECHART)){
            return H.t("");
        }
        final int LEFTMARGIN = 14;
        final int LABELMARGIN = 9;
        final int BARWIDTH = 21;
        final double TEXTWIDTH = layout.getLengendTextWidth();
        final double x;
        double y = layout.getPlotArea().getTop();
        if(TextDirection.currentLocale().equals(TextDirection.LTR)){
            x = LEFTMARGIN + layout.getPlotArea().getRight();
        }else {
            x = layout.getScaleArea().getLeft() - layout.getScaleWidth() - LEFTMARGIN;
        }

        List<PivotTable.Node> minor = layout.getMinorLeafCategories();
        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.MINOR_CATEGORY_LABELS);
        ArrayList<VNode> legend = new ArrayList<>();
        TextLineOperator textOperator = new TextLineOperator(font, TEXTWIDTH);
        for(int i = 0; i < minor.size(); i++) {
            String color = graphElementColor(i);

            //render line
            legend.add(renderLine(x, y, BARWIDTH, color));

            //render icon
            LegendIcon icon = LegendIcon.getIcon(i);
            double[] dims = icon.getDimensionsFromRadius(LEGENDRADIUS);
            double[] coord = icon.getCenterCoord(x, y, dims[0], dims[1], BARWIDTH);
            legend.add(icon.render(coord[0], coord[1], LEGENDRADIUS, color));

            double delta;
            if(TextDirection.currentLocale().equals(TextDirection.LTR)){
                delta = BARWIDTH + LABELMARGIN;
            }else{
                delta =  - LABELMARGIN;
            }
            //render text
            VAttrMap labelProps = VAttr.create();
            labelProps.set("x", x + delta);
            labelProps.set("y", y + font.getAscent()/3);
            labelProps.set("font", font.getSvgFontString());


            String label = minor.get(i).flattenLabel();
            legend.add(new VElement(SvgTag.TEXT, labelProps, new VText(textOperator.truncate(label))));
            y += AxisChartLayout.MINOR_CATEGORY_LABEL_LINE_HEIGHT;
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), legend);
    }

    private VNode renderLine(double x, double y, double length, String color ) {
        SvgPathBuilder p = new SvgPathBuilder();
        p.moveTo(x, y).horizontalLineTo(x + length);

        VAttrMap pathProps = VAttr.create();
        pathProps.set("stroke", color);
        pathProps.set("stroke-width", "2");
        pathProps.set("d", p.toString());

        return new VElement(SvgTag.PATH, pathProps);
    }



    enum LegendIcon{
        SQUARE{
            @Override
            VNode render(double x, double y, double radius, String color) {
                double width = getDimensionsFromRadius(radius)[0];
                VAttrMap rectProps = VAttr.create();
                rectProps.set("x", x);
                rectProps.set("y", y);
                rectProps.set("width", width);
                rectProps.set("height", width);
                rectProps.set("fill", color);

                return new VElement(SvgTag.RECT, rectProps);
            }

            @Override
            double[] getDimensionsFromRadius(double radius){
                double width =  Math.sqrt(Math.PI * Math.pow(radius, 2));
                double[] result = {width, width};
                return result;
            }
        },
        TRIANGLE {
            @Override
            VNode render(double x, double y, double radius, String color) {
                double[] dims = getDimensionsFromRadius(radius);
                double width = dims[0];
                double height = dims[1];
                SvgPolygonBuilder p = new SvgPolygonBuilder();
                String points = p
                        .appendCoords(x, y + height)
                        .appendCoords(x + width, y + height)
                        .appendCoords(x + width/2, y)
                        .toString();

                VAttrMap triangleProps = VAttr.create();
                triangleProps.set("points", points);
                triangleProps.set("fill", color);

                return new VElement(SvgTag.POLYGON, triangleProps);
            }

            @Override
            double[] getDimensionsFromRadius(double radius){
                double area = Math.PI * Math.pow(radius, 2);

                //equilateral triangle
                double width = Math.sqrt((area * 4)/Math.sqrt(3));
                double height = (Math.sqrt(3) * width)/2;
                double[] result = {width, height};
                return result;
            }
        },
        CIRCLE{
            @Override
            VNode render(double x, double y, double radius, String color) {
                VAttrMap circleProps = VAttr.create();
                circleProps.set("cx", x);
                circleProps.set("cy", y);
                circleProps.set("r", radius);
                circleProps.set("fill", color);
                return new VElement(SvgTag.CIRCLE, circleProps);
            }

            @Override
            double[] getDimensionsFromRadius(double radius) {
                double[] result = {radius, radius};
                return result;
            }

            @Override
            double[] getCenterCoord(double x, double y, double width, double height) {
                double[] result = {x, y};
                return result;
            }
        };

        static LegendIcon getIcon(int minorIdx){
            LegendIcon[] values = LegendIcon.values();
            int searchIdx = minorIdx % values.length;
            for(LegendIcon icon : values){
                if(icon.ordinal() == searchIdx){
                    return icon;
                }
            }
            return null;
        }
        abstract VNode render(double x, double y, double radius, String color);
        abstract double[] getDimensionsFromRadius(double radius);

        double [] getCenterCoord(double x, double y, double width, double height, double lineWidth) {
            x = x + lineWidth/2;
            return getCenterCoord(x, y, width, height);
        }

        double [] getCenterCoord(double x, double y, double width, double height) {
            x = x - width/2;
            y = y - height/2;
            double [] result = {x, y};
            return result;
        }
    }
}
