/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

/**
 * Prints nice scales for charts
 * taken from
 * https://stackoverflow.com/questions/8506881/nice-label-algorithm-for-charts-with-minimum-ticks
 */
public class NiceChartScale {

    private final double minPoint;
    private final double maxPoint;
    private final int maxTicks;
    private final boolean showMaxTicks;
    private double tickSpacing;
    private double range;
    private double niceMin;
    private double niceMax;


    /**
     * Instantiates a new instance of the NiceScale class.
     *
     * @param max the maximum data point on the axis
     */
    public NiceChartScale(double min, double max, int maxTicks, boolean alwaysShowMaxTicks) {
        this.minPoint = min;
        this.maxPoint = max;
        this.maxTicks = maxTicks;
        this.showMaxTicks = alwaysShowMaxTicks;
        calculate();
    }

    /**
     * Calculate and update values for tick spacing and nice
     * minimum and maximum data points on the axis.
     */
    private void calculate() {
        this.range = niceNum(maxPoint - minPoint, false);
        this.tickSpacing = niceNum(range / (maxTicks - 1), true);
        this.niceMin = Math.floor(minPoint / tickSpacing) * tickSpacing;
        if(showMaxTicks) {
            this.niceMax = this.niceMin + tickSpacing * (maxTicks - 1); // Always display maxticks
        }else{
            this.niceMax = Math.ceil(maxPoint / tickSpacing) * tickSpacing;
        }
    }

    /**
     * Returns a "nice" number approximately equal to range Rounds
     * the number if round = true Takes the ceiling if round = false.
     *
     * @param range the data range
     * @param round whether to round the result
     * @return a "nice" number to be used for the data range
     */
    private double niceNum(double range, boolean round) {
        double exponent; /** exponent of range */
        double fraction; /** fractional part of range */
        double niceFraction; /** nice, rounded fraction */

        exponent = Math.floor(Math.log10(range));
        fraction = range / Math.pow(10, exponent);

        if (round) {
            if (fraction < 1.5)
                niceFraction = 1;
            else if (fraction < 3)
                niceFraction = 2;
            else if (fraction < 7)
                niceFraction = 5;
            else
                niceFraction = 10;
        } else {
            if (fraction <= 1)
                niceFraction = 1;
            else if (fraction <= 2)
                niceFraction = 2;
            else if (fraction <= 5)
                niceFraction = 5;
            else
                niceFraction = 10;
        }

        return niceFraction * Math.pow(10, exponent);
    }

    public double getTickSpacing() {
        return tickSpacing;
    }

    public double getNiceMin() {
        return niceMin;
    }

    public double getNiceMax() {
        return niceMax;
    }

    public int getNumberOfTicks(){
        return (int) Math.round((niceMax-niceMin)/tickSpacing);
    }
}
