/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.model.analysis.pivot.VisualizationType;

import java.text.DecimalFormat;
import java.util.HashSet;
import java.util.function.Function;

/**
 * Constructs a "pretty" scale with round intervals for charts.
 * see {@link NiceChartScale}
 */
public class NiceChartScaleBuilder {
    public static final int MAXTICKS = 10;

    private int decimalCount = 0;
    private int count = 0;
    private static final String INTEGER_PATTERN = "#,##0";

    HashSet<Double> distinctPoints = new HashSet<>();

    private double minPivotPointValue = Double.MAX_VALUE;

    private double maxPivotPointValue = -Double.MAX_VALUE;


    public void value(double x) {
        if (x < minPivotPointValue) {
            minPivotPointValue = x;
        }
        if (x > maxPivotPointValue) {
            maxPivotPointValue = x;
        }
        count++;
        if(!isInteger(x)){
            decimalCount++;
        }
        distinctPoints.add(x);
    }

    private boolean exceedsDecimalThreshold(){
        return decimalCount > (count / 3); //more than a third of the values are decimals
    }

    public NiceChartScaleBuilder values(PivotTable pivotTable) {
        pivotTable.forEachValue(this::value);
        return this;
    }

    public Scale build(VisualizationType type) {
        minPivotPointValue = 0; //always start chart at 0

        if(distinctPoints.size() == 1 && distinctPoints.contains(0d)){
            maxPivotPointValue = 1;
        }
        NiceChartScale graphScale = getPreferredScale();
        double scaleMin = graphScale.getNiceMin();
        double scaleMax = graphScale.getNiceMax();
        double unit = graphScale.getTickSpacing();
        int numDivisions = graphScale.getNumberOfTicks();
        String[] labels = labels(scaleMin, unit, numDivisions);
        return new Scale(scaleMin, scaleMax, unit, numDivisions, labels);
    }

    private NiceChartScale getPreferredScale(){
        int tickCount = MAXTICKS;
        NiceChartScale candidateScale = new NiceChartScale(minPivotPointValue, maxPivotPointValue, tickCount, false);
        if(exceedsDecimalThreshold()){
            return candidateScale;
        }else { //attempt to retrieve a scale with no decimals
            while (!isInteger(candidateScale.getTickSpacing()) && tickCount > 1) {
                candidateScale = new NiceChartScale(minPivotPointValue, maxPivotPointValue, --tickCount, false);
            }
        }
        return candidateScale;
    }

    private String[] labels(double scaleMin, double unit, int ndiv) {

        Function<Double, String> formatter = labelFormatter(unit);

        String[] labels = new String[ndiv + 1];
        double x = scaleMin;
        for (int i = 0; i < labels.length; i++) {
            labels[i] = formatter.apply(x);
            x += unit;
        }

        return labels;
    }

    private static Function<Double, String> labelFormatter(double unit) {
        String pattern = getFormatPattern(unit);
        return x -> new DecimalFormat(pattern).format(x);
    }

    public static boolean isInteger(double unit) {
        return unit == (int)unit;
    }

    private static String getFormatPattern(double unit){
        if(isInteger(unit)){
            return INTEGER_PATTERN;
        }
        String unitScalePattern = INTEGER_PATTERN + ".##########";//defensive programming, shouldn't require that level of precision
        String unitStr = new DecimalFormat(unitScalePattern).format(unit);
        return getFormatPatternSuffix(unitStr.split("\\.")[1].length());
    }

    private static String getFormatPatternSuffix(int n){
        assert(n > 0);
        StringBuilder result = new StringBuilder(n+2)
                .append(INTEGER_PATTERN)
                .append(".");

        for(int i = 0; i < n; i++){
            result.append("0");
        }
        return result.toString();
    }
}
