/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.MajorCategoryLabelDescriptor;
import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.i18n.shared.TextDirection;

import java.util.Map;

/**
 * Shared layout data for Bar charts and Line charts
 */
public class PieChartLayout implements MajorCategoryLabelDescriptor, ChartSpace {

    public static final double PIECHART_RADIUS = 180.5;
    public static final double PIECHART_MARGIN = 40;
    public static final double PIECHART_MARGIN_TOP = 61;
    public static final double LEGEND_PREFIX_WIDTH = 40;
    public static final double LEGEND_MAXIMUM_WIDTH = 128;


    private final double legendTextWidth;


    private final PivotTable pivotTable;
    private final double chartHeight;
    private final double chartWidth;

    private final Rect plotArea;

    private final int majorDimensionCount;

    private final double majorCategoryWidth;
    private final double majorCategoryLabelTop;

    private final double[] majorCategoryLabelHeights;
    private final Map<PivotTable.Node, String[]> majorCategoryTextLinesMap;

    PieChartLayout(PivotTable pivotTable, double chartWidth, double legendChartHeight,
                    double LegendTextWidth, double[] majorCategoryLabelHeights,
                    Map<PivotTable.Node, String[]> majorCategoryTextLinesMap) {
        this.pivotTable = pivotTable;
        this.legendTextWidth = LegendTextWidth;
        this.chartWidth = chartWidth+getLegendWidth();
        this.majorCategoryLabelHeights = majorCategoryLabelHeights;
        this.majorCategoryTextLinesMap = majorCategoryTextLinesMap;

        majorDimensionCount = pivotTable.getRowDimensionCount();

        majorCategoryWidth = getMajorCategoryWidth();

        double categoryLabelHeights =
                AxisChartLayout.CATEGORY_LABEL_MARGIN_TOP +
                        AxisChartLayout.sum(majorCategoryLabelHeights);

        chartHeight = AxisChartLayout.MARGIN_TOP
                + AxisChartLayout.MARGIN_BOTTOM
                + Math.max(categoryLabelHeights+PIECHART_RADIUS*2, legendChartHeight);

        double plotLeft = AxisChartLayout.MARGIN_LEFT;
        if(TextDirection.currentLocale().equals(TextDirection.RTL)){
            plotLeft += getLegendWidth();
        }

        this.plotArea = new Rect(
                plotLeft,
                AxisChartLayout.MARGIN_TOP,
                chartWidth - AxisChartLayout.MARGIN_RIGHT,
                PIECHART_RADIUS * 2 + AxisChartLayout.MARGIN_TOP);

        majorCategoryLabelTop = plotArea.getBottom() + AxisChartLayout.CATEGORY_LABEL_MARGIN_TOP;
    }


    private double getLegendWidth(){
        if(hasLegend()){
            return LEGEND_PREFIX_WIDTH + getLengendTextWidth();
        } else {
            return 0;
        }
    }

    public boolean hasLegend(){
        return AxisChartLayout.getMinorCategories(pivotTable).size() > 0;
    }


    @Override
    public double getChartWidth() {
        return chartWidth;
    }

    @Override
    public double getChartHeight() {
        return chartHeight;
    }

    @Override
    public String getViewBox() {
        return "0 0 " + chartWidth + " " + chartHeight;
    }


    public Rect getPlotArea() {
        return plotArea;
    }

    public PivotTable getPivotTable() {
        return pivotTable;
    }

    /**
     * Computes the left coordinate of the major category bar(s)
     * @param index the zero-based index of the category.
     */
    public double getMajorCategoryLeft(int index) {
        return plotArea.getLeft() + PIECHART_MARGIN + (index * (majorCategoryWidth + PIECHART_MARGIN));
    }

    public Rect getPieRect(int majorCategoryIndex) {
        double left = getMajorCategoryLeft(majorCategoryIndex);
        double top = plotArea.getTop();
        double right = left + PIECHART_RADIUS * 2;
        double bottom = plotArea.getBottom();

        return new Rect(left, top, right, bottom);
    }

    @Override
    public Rect getMajorCategoryLabelRect(PivotTable.Node node) {

        int level = (this.majorDimensionCount - node.getDepth());
        double left = getMajorCategoryLeft(node.getIndex());
        double top = majorCategoryLabelTop + AxisChartLayout.sum(majorCategoryLabelHeights, level);
        double right =  left +
                (node.getSpan() * majorCategoryWidth) +
                ((node.getSpan() - 1) * PIECHART_MARGIN);
        double bottom = top + majorCategoryLabelHeights[level];
        return new Rect(left, top, right, bottom);
    }


    public double getLengendTextWidth(){
        return Math.min(legendTextWidth, LEGEND_MAXIMUM_WIDTH);
    }

    @Override
    public String[] getMajorTextLines(PivotTable.Node leafNode){
        return majorCategoryTextLinesMap.get(leafNode);
    }

    public static double getMajorCategoryWidth(){
        return PIECHART_RADIUS * 2;
    }
}
