/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.pivot.viewModel.chart;

import org.activityinfo.analysis.pivot.viewModel.PercentFormatter;
import org.activityinfo.analysis.pivot.viewModel.PivotTable;
import org.activityinfo.analysis.pivot.viewModel.ReportStyle;
import org.activityinfo.i18n.shared.TextDirection;
import org.activityinfo.ui.vdom.shared.html.H;
import org.activityinfo.ui.vdom.shared.html.SvgTag;
import org.activityinfo.ui.vdom.shared.tree.*;

import java.text.DecimalFormat;
import java.util.*;

public class PieChartRenderer{
    private final HashMap<PivotTable.Node, HashMap<PivotTable.Node, double[]>> majorMinorTotals = new HashMap<>();
    private static final PercentFormatter PERCENT_FORMATTER = PercentFormatter.get();
    private final static String[] TEXTCOLORS = new ReportStyle("Greys").getColorScheme(7);

    protected final PieChartLayout layout;
    protected final ChartFontContext fontContext;

    //https://stackoverflow.com/questions/5737975/circle-drawing-with-svgs-arc-path
    final double CIRCLE_THRESHOLD = 0.0001;
    final double LABEL_THRESHOLD = 0.03; //labels for slices equal to or greater than this percentage will be rendered


    public PieChartRenderer(PieChartLayout layout, ChartFontContext fontContext) {
        this.layout = layout;
        this.fontContext = fontContext;
        computeTotalsAndPercentages();
    }

    public VNode render(PieChartLayout layout) {

        VAttrMap svgProps = GraphChartRenderer.getChartProperties(layout);
        return new VElement(SvgTag.SVG, svgProps,
                GraphChartRenderer.background(layout),
                renderPieChart(),
                GraphChartRenderer.majorCategoryLabels(layout.getPivotTable(),fontContext, layout,
                        layout.getPlotArea().getBottom(), PieChartLayout.PIECHART_MARGIN),
                renderLegend()
        );
    }

    private double[] polarToCartesian(double centerX, double centerY, double radius, double degrees){
        double radians = Math.toRadians(degrees -90);
        double[] coord = {centerX + radius * Math.cos(radians), centerY + radius * Math.sin(radians)};
        return coord;
    }

    private VNode renderArc(double x, double y, double radius, double startAngle, double endAngle, String color) {
        return renderArc(x, y, radius, startAngle, endAngle, color, color);
    }

    private VNode renderArc(double x, double y, double radius, double startAngle, double endAngle, String color, String strokeColor){
        /*
        https://stackoverflow.com/questions/5736398/how-to-calculate-the-svg-path-for-an-arc-of-a-circle
         */
        double[] start = polarToCartesian(x, y, radius, endAngle);
        double[] end = polarToCartesian(x, y, radius, startAngle);
        SvgPathBuilder path = new SvgPathBuilder();
        path.moveTo(start[0], start[1]);
        int largeArcFlag = endAngle - startAngle <= 180 ? 0 : 1;
        path.arcTo(radius, largeArcFlag, end[0], end[1]);
        path.lineTo(x, y);
        VAttrMap pathProps = VAttr.create();
        pathProps.set("stroke", strokeColor);
        pathProps.set("stroke-width", "1");
        pathProps.set("fill", color);
        pathProps.set("d", path.toString());
        return new VElement(SvgTag.PATH, pathProps);
    }

    private VNode renderLabel(double x, double y, double radius, double startAngle, double endAngle, double percent, int minorIdx){
        radius = radius/2;
        endAngle = (startAngle + endAngle)/2;
        double[] end = polarToCartesian(x, y, radius, endAngle);
        return renderLabel(end[0], end[1], percent, minorIdx);
    }

    private VNode renderLabel(double x, double y, double ratio, int minorIdx){
        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.MINOR_CATEGORY_LABELS);
        VAttrMap labelProps = VAttr.create();
        String label = PERCENT_FORMATTER.format(ratio);
        double widthDelta = font.measureWidth(label)/2;
        labelProps.set("x", x - widthDelta);
        labelProps.set("y", y);
        labelProps.set("fill", getTextColor(minorIdx));

        labelProps.set("font", font.getSvgFontString());


        return new VElement(SvgTag.TEXT, labelProps, new VText(label));
    }

    private VNode renderPieChart(){
        List<PivotTable.Node> major = AxisChartLayout.getMajorLeafCategories(layout.getPivotTable());
        List<PivotTable.Node> minor = AxisChartLayout.getMinorLeafCategories(layout.getPivotTable());
        List<VNode> pieCharts = new ArrayList<>();
        for (int i = 0; i < major.size(); i++) {
            pieCharts.add(renderPieChart(minor, major.get(i), i));
            //pieCharts.add(renderLabelsArcDebug(minor, major.get(i), i));
            pieCharts.add(renderLabels(minor, major.get(i), i));
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), pieCharts);
    }

    private VNode renderPieChart(List<PivotTable.Node> minor, PivotTable.Node majorCategory, int majorIdx) {

        Rect rect = layout.getPieRect(majorIdx);
        List<VNode> slices = new ArrayList<>();
        double startingDegree = 0;
        for(int i = 0; i < minor.size();i++){
            String color = GraphChartRenderer.graphElementColor(i);
            double[] ratio = majorMinorTotals.get(majorCategory).get(minor.get(i));
            if(ratio == null){
                continue;
            }else if(Math.abs(ratio[1] - 1) < CIRCLE_THRESHOLD){ //svg arcs do not properly render circles, see link next to CIRCLE_THRESHOLD declaration
                slices.add(renderCircle(rect.getCenterX(), rect.getCenterY(), PieChartLayout.PIECHART_RADIUS, color));
            }else{
                double end = startingDegree + ratio[1] * 360; //360 degree circle
                slices.add(renderArc(rect.getCenterX(), rect.getCenterY(), PieChartLayout.PIECHART_RADIUS, startingDegree, end, color));
                startingDegree = end;
            }
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), slices);
    }

    private VNode renderLabels(List<PivotTable.Node> minor, PivotTable.Node majorCategory, int majorIdx) {

        Rect rect = layout.getPieRect(majorIdx);
        List<VNode> slices = new ArrayList<>();
        double startingDegree = 0;
        for(int i = 0; i < minor.size();i++){
            double[] ratio = majorMinorTotals.get(majorCategory).get(minor.get(i));
            if(ratio == null){
                continue;
            }else if(Math.abs(ratio[1] - 1) < CIRCLE_THRESHOLD){ //svg arcs do not properly render circles, see link next to CIRCLE_THRESHOLD declaration
                slices.add(renderLabel(rect.getCenterX(), rect.getCenterY(), ratio[1], i));
            }else{
                double end = startingDegree + ratio[1] * 360; //360 degree circle
                if(ratio[1] >= LABEL_THRESHOLD) {
                    slices.add(renderLabel(rect.getCenterX(), rect.getCenterY(), PieChartLayout.PIECHART_RADIUS, startingDegree, end, ratio[1], i));
                }
                startingDegree = end;
            }
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), slices);
    }

    private VNode renderLabelsArcDebug(List<PivotTable.Node> minor, PivotTable.Node majorCategory, int majorIdx) {

        Rect rect = layout.getPieRect(majorIdx);
        List<VNode> slices = new ArrayList<>();
        double startingDegree = 0;
        for(int i = 0; i < minor.size();i++){
            double[] ratio = majorMinorTotals.get(majorCategory).get(minor.get(i));
            if(ratio == null){//defensive programming @ null when Double.isFinite(point) call false when building map
                continue;
            }else if(Math.abs(ratio[1] - 1) > CIRCLE_THRESHOLD){ //svg arcs do not properly render circles, see link next to CIRCLE_THRESHOLD declaration
                double end = startingDegree + ratio[1] * 360; //360 degree circle
                slices.add(renderArc(rect.getCenterX(), rect.getCenterY(), PieChartLayout.PIECHART_RADIUS/2, startingDegree, (startingDegree +end)/2,"transparent", "black"));
                startingDegree = end;
            }
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), slices);
    }


    protected VNode renderLegend() {

        if(!layout.hasLegend()){
            return H.t("");
        }
        final int LEFTMARGIN = 14;
        final int LABELMARGIN = 9;
        final int ICONWIDTH = 16;
        final double TEXTWIDTH = layout.getLengendTextWidth();
        final double x;
        if(TextDirection.currentLocale().equals(TextDirection.LTR)){
            x = LEFTMARGIN + layout.getPlotArea().getRight();
        }else {
            x = layout.getPlotArea().getLeft();
        }

        double y = layout.getPlotArea().getTop();

        List<PivotTable.Node> major = AxisChartLayout.getMajorLeafCategories(layout.getPivotTable());
        List<PivotTable.Node> minor = AxisChartLayout.getMinorLeafCategories(layout.getPivotTable());
        ChartFontMetrics font = fontContext.getFontMetrics(ChartFontStyle.MINOR_CATEGORY_LABELS);
        ArrayList<VNode> legend = new ArrayList<>();
        TextLineOperator textOperator = new TextLineOperator(font, TEXTWIDTH);
        HashSet<PivotTable.Node> visited = new HashSet<>();
        for(int j = 0; j < minor.size(); j++) {
            for(int i = 0; i< major.size(); i++){
                double[] point = majorMinorTotals.get(major.get(i)).get(minor.get(j));
                if(point != null  && !visited.contains(minor.get(j))) {
                    visited.add(minor.get(j));
                    String color = GraphChartRenderer.graphElementColor(j);
                    //render icon
                    VNode icon = renderSquare(x, y, ICONWIDTH, color);
                    legend.add(icon);

                    //render text
                    double delta;
                    if(TextDirection.currentLocale().equals(TextDirection.LTR)){
                        delta = ICONWIDTH + LABELMARGIN;
                    }else{
                        delta = -LABELMARGIN;
                    }
                    VAttrMap labelProps = VAttr.create();
                    labelProps.set("x", x + delta);
                    labelProps.set("y", y + font.getAscent());
                    labelProps.set("font", font.getSvgFontString());

                    String label = minor.get(j).flattenLabel();
                    legend.add(new VElement(SvgTag.TEXT, labelProps, new VText(textOperator.truncate(label))));
                    y += AxisChartLayout.MINOR_CATEGORY_LABEL_LINE_HEIGHT;
                }
            }
        }
        return new VElement(SvgTag.GROUP, VAttr.create(), legend);
    }

    /*

    Get totals fpr each major category

     */
    private void computeTotalsAndPercentages(){
        List<PivotTable.Node> major = AxisChartLayout.getMajorLeafCategories(layout.getPivotTable());
        List<PivotTable.Node> minor = AxisChartLayout.getMinorLeafCategories(layout.getPivotTable());

        for (int i = 0; i < major.size(); i++) {
            majorMinorTotals.put(major.get(i), new HashMap<>());
            for (int j = 0; j < minor.size(); j++) {
                double point = major.get(i).getCellValue(minor.get(j));
                if(Double.isFinite(point) ) {
                    majorMinorTotals.get(major.get(i)).put(minor.get(j), new double[2]);
                    majorMinorTotals.get(major.get(i)).get(minor.get(j))[0] = point;
                }
            }
        }
        for(HashMap<PivotTable.Node, double[]> majorMinor : majorMinorTotals.values()){
            populatePercentage(majorMinor.values());
        }
    }

    private void populatePercentage(Collection<double[]> values){
        double total = 0;
        for(double[] num : values){
            total += num[0];
        }

        for(double[] num : values){
            num[1] = num[0] / total;
        }
    }

    private VElement debugLabel(String title, HashMap<PivotTable.Node, double[]> nodes, double[] coords, double ascent) {
        List<VNode> spans = new ArrayList<>();
        VAttrMap textLineProps = VAttr.create();
        textLineProps.set("x", coords[0]);
        textLineProps.set("y", coords[1]);
        spans.add(new VElement(SvgTag.TEXT, textLineProps, new VText(title)));
        double dy = coords[1] + ascent;
        for(Map.Entry<PivotTable.Node, double[]> nodeLine : nodes.entrySet()){
            textLineProps = VAttr.create();
            textLineProps.set("x", coords[0]);
            textLineProps.set("y", dy);
            //textLineProps.set("text-anchor", "middle");
            DecimalFormat formatter = new DecimalFormat("#,##0.00");
            double[] value = nodeLine.getValue();
            String line = nodeLine.getKey().flattenLabel() + " Total: " + formatter.format(value[0]) + " Percentage %: " + formatter.format(value[1]);
            spans.add(new VElement(SvgTag.TEXT, textLineProps, new VText(line)));
            dy += ascent;
        }
        coords[1] = dy + ascent;

        VAttrMap labelProps = VAttr.create();
        return new VElement(SvgTag.GROUP, labelProps, spans.stream());
    }

    private VNode debugSamplePie(){
        double[] test = {0.4, 0.3, 0.1, 0.2};
        List<VNode> rectangles = new ArrayList<>();
        double start = 0;
        //AxisChartLayout.getMajorLeafCategories(layout.getPivotTable());
        Rect rect = layout.getPieRect(0);
        for(int i = 0; i < test.length; i++){
            String color = GraphChartRenderer.graphElementColor(i);
            double end = start + test[i] * 360;
            rectangles.add(renderArc(rect.getCenterX(), rect.getCenterY(), PieChartLayout.PIECHART_RADIUS, start, end, color));
            start = end;
        }
        rectangles.add(GraphChartRenderer.debugRect(rect));
        return new VElement(SvgTag.GROUP, VAttr.create(), rectangles);
    }

    private VNode renderCircle(double x, double y, double radius, String color) {
        VAttrMap circleProps = VAttr.create();
        circleProps.set("cx", x);
        circleProps.set("cy", y);
        circleProps.set("r", radius);
        circleProps.set("fill", color);
        return new VElement(SvgTag.CIRCLE, circleProps);
    }

    private VNode renderSquare(double x, double y, double length, String color) {
        VAttrMap rectProps = VAttr.create();
        rectProps.set("x", x);
        rectProps.set("y", y);
        rectProps.set("width", length);
        rectProps.set("height", length);
        rectProps.set("fill", color);

        return new VElement(SvgTag.RECT, rectProps);
    }

    private String getTextColor(int minorIdx){
        String color;
        if( minorIdx == 3 || minorIdx == 19 || minorIdx == 26){ //dark color on yellow slices
            color = TEXTCOLORS[3];
        }else{
            color = TEXTCOLORS[0];
        }
        return "#" + color;
    }
}
