/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.table;

import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.type.Cardinality;
import org.activityinfo.model.type.FieldType;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.subform.SubFormReferenceType;

public class ColumnFormatFactory {


    private String columnId;
    private ResolvedNode formula;

    public ColumnFormatFactory(String columnId, ResolvedNode formula) {
        this.columnId = columnId;
        this.formula = formula;
    }

    public static ColumnFormat create(String id, ResolvedNode formula) {

        // All valid formulas have a FormulaType, but if the formula is simply
        // referencing a single field, we may want to handle it specially

        if(formula.getFieldType().isPresent()) {
            FieldType fieldType = formula.getFieldType().get();
            if (fieldType instanceof SubFormReferenceType) {
                return new SubFormFormat(id, formula, ((SubFormReferenceType) fieldType).getClassId());
            }
        }

        if(formula.getResolvedType().isPresent()) {
            FormulaType formulaType = formula.getResolvedType().get();
            return createForType(id, formula, formulaType);
        }

        return new ErrorFormat();
    }

    private static ColumnFormat createForFormulaType(String id, ResolvedNode formula) {
        return formula
            .getResolvedType()
            .map(t -> createForType(id, formula, t))
            .orElse(new ErrorFormat());
    }

    private static ColumnFormat createForType(String id, ResolvedNode formula, FormulaType t) {
        switch (t) {
            case STRING:
                if (formula.getFieldType().isPresent()) {
                    if (formula.getFieldType().get() instanceof EnumType) {
                        EnumType enumType = (EnumType) formula.getFieldType().get();
                        Cardinality cardinality = enumType.getCardinality();
                        if (cardinality.equals(Cardinality.MULTIPLE)) {
                            return new MultiEnumFormat(id, formula, enumType);
                        } else {
                            return new SingleEnumFormat(id, formula, enumType);
                        }
                    }
                }
                return new TextFormat(id, formula);
            case BOOLEAN:
                return new BooleanFormat(id, formula);
            case LOCAL_DATE:
                return new DateFormat(id, formula);
            case INSTANT:
                return new InstantFormat(id, formula);
            case NUMBER:
                return new NumberFormat(id, formula);
            case GEOMETRY:
                return new GeoPointFormat(id, formula);
            case OPAQUE:
                return new ErrorFormat();

        }
        throw new IllegalArgumentException("type: "+  t);
    }

}
