/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.table;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.model.analysis.*;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formTree.LookupKey;
import org.activityinfo.model.formTree.LookupKeySet;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.query.SortDir;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.time.PeriodType;

import java.util.*;


/**
 * Constructs a reasonable table from a form, including all simple fields and key fields for
 * reference fields.
 */
public class DefaultTableAnalysisModelBuilder {

    private final FormTree formTree;
    private final List<TableColumn> columns = new ArrayList<>();

    private boolean includeRecordIds = false;

    private boolean includeLastEditTime = false;

    public DefaultTableAnalysisModelBuilder(FormTree formTree) {
        this.formTree = formTree;
    }

    public DefaultTableAnalysisModelBuilder includeRecordIds(boolean include) {
        this.includeRecordIds = include;
        return this;
    }

    public DefaultTableAnalysisModelBuilder includeLastEditTime(boolean include) {
        this.includeLastEditTime = include;
        return this;
    }

    public TableAnalysisModel build() {
        return new TableAnalysisModelBuilder()
            .formId(formTree.getRootFormId())
            .columns(buildColumnList())
            .ordering(buildSortOrder())
            .build();
    }

    /**
     * If this form has a date-like field, order by reverse date.
     */
    private Collection<SortOrder> buildSortOrder() {
        return buildSortOrder(formTree);
    }

    /**
     * If this form has a date-like field, order by reverse date.
     * @param formTree
     */
    public static Collection<SortOrder> buildSortOrder(FormTree formTree) {
        for (FormTree.Node rootField : formTree.getRootFields()) {
            if(rootField.getType() instanceof PeriodType) {
                return Collections.singleton(new SortOrder(SymbolNode.deparse(rootField.getFieldId()), SortDir.DESC));
            }
        }

        // otherwise do not sort
        return Collections.emptyList();
    }

    public List<TableColumn> buildColumnList() {
        if(includeRecordIds) {
            columns.add(new TableColumn("id", Optional.of(I18N.CONSTANTS.recordId()), 100, ColumnModel.RECORD_ID_SYMBOL));
        }

        if(includeLastEditTime){
            columns.add(new TableColumn("lastEditTime", Optional.of(I18N.CONSTANTS.lastEditTime()), 100, ColumnModel.LAST_EDIT_TIME_SYMBOL));
        }
        addDefaultColumns(formTree);
        return columns;
    }

    private void addDefaultColumns(FormTree formTree) {
        for (FormTree.Node node : formTree.getRootFields()) {
            if(node.getField().isTableVisible()) {
                if (SimpleTypeSupervisor.isSimpleType(node.getType())) {
                    columns.add(defaultColumnModel(node));

                } else if (node.getType() instanceof ReferenceType) {
                    addKeyColumns(node);

                } else if (node.isVisibleSubForm()) {
                    columns.add(defaultColumnModel(node));
                }
            }
        }
    }

    private TableColumn defaultColumnModel(FormTree.Node node) {
        return defaultColumnModel(node.getPath().toExpr());
    }

    private TableColumn defaultColumnModel(FormulaNode formulaNode) {
        String formulaString = formulaNode.deparse();

        // We need stable ids for our default columns, otherwise
        // the views will get confused and refresh unnecessarily
        String id = formulaString.replace('.', 'd');

        return new TableColumnBuilder()
            .id(id)
            .formula(formulaString)
            .build();
    }

    private void addKeyColumns(FormTree.Node node) {

        LookupKeySet lookupKeySet = new LookupKeySet(formTree, node.getField());
        Map<LookupKey, FormulaNode> formulas = lookupKeySet.getKeyFormulas(node.getFieldId());

        // If there are natural keys, but we have been asked to include record ids,
        // then explicitly include the parent record id for convenience.
        if(includeRecordIds && lookupKeySet.hasNaturalKeys()) {
            columns.add(new TableColumnBuilder()
                .id(node.getFieldId().asString() + "_id")
                .label(node.getField().getLabel() + " " + I18N.CONSTANTS.recordId())
                .formula(new SymbolNode(node.getFieldId()).deparse())
                .build());
        }

        for (LookupKey lookupKey : lookupKeySet.getLookupKeys()) {

            columns.add(new TableColumnBuilder()
                .id(node.getFieldId().asString() + "_k" + lookupKey.getKeyIndex())
                .formula(formulas.get(lookupKey).toString())
                .label(lookupKey.getKeyLabel())
                .build());
        }
    }
}
