/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.table;

import org.activityinfo.analysis.filter.FilterViewModel;
import org.activityinfo.model.analysis.SortOrder;
import org.activityinfo.model.analysis.TableAnalysisModel;
import org.activityinfo.model.analysis.TableColumn;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.FormulaParser;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.QueryModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.observable.Observable;
import org.activityinfo.store.query.shared.FormSource;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * The effective description of a table
 */
public class EffectiveTableModel {

    private FormTree formTree;
    private TableAnalysisModel tableModel;
    private List<EffectiveTableColumn> columns;
    private List<EffectiveSortOrder> ordering;
    private FilterViewModel filter;
    private final Set<String> filteredColumns;

    public EffectiveTableModel(FormTree formTree, TableAnalysisModel tableModel) {
        this.formTree = formTree;
        this.tableModel = tableModel;
        this.columns = new ArrayList<>();
        this.ordering = new ArrayList<>();

        for (SortOrder sortOrder : ordering(formTree, tableModel)) {
            FormulaParser.tryParse(sortOrder.getFormula()).ifPresent(formula -> {
                ordering.add(new EffectiveSortOrder(formula, sortOrder.isAscending()));
            });
        }

        for (TableColumn tableColumn : resolveColumns(tableModel)) {
            columns.add(new EffectiveTableColumn(formTree, tableColumn));
        }

        TableColumnMatcher columnMatcher = new TableColumnMatcher(columns);
        this.filter = FilterViewModel.parse(formTree, columnMatcher, tableModel.getFilter());
        this.filteredColumns = columnMatcher.findFilteredColumns(filter);
    }

    private Collection<SortOrder> ordering(FormTree formTree, TableAnalysisModel tableModel) {
        if (tableModel.getOrdering().isEmpty()) {
            return DefaultTableAnalysisModelBuilder.buildSortOrder(formTree);
        } else {
            return tableModel.getOrdering();
        }
    }

    public List<TableColumn> resolveColumns(TableAnalysisModel tableModel) {
        if(tableModel.getColumns().isEmpty() && formTree.getRootState() == FormTree.State.VALID) {
            return new DefaultTableAnalysisModelBuilder(formTree).buildColumnList();
        } else {
            return tableModel.getColumns();
        }
    }

    public FilterViewModel getFilter() {
        return filter;
    }

    public TableAnalysisModel getModel() {
        return tableModel;
    }

    public String getTitle() {
        return formTree.getRootFormClass().getLabel();
    }


    public ResourceId getFormId() {
        return formTree.getRootFormId();
    }

    public QueryModel buildQuery() {
        return buildQuery(columns);
    }

    private QueryModel buildQuery(List<EffectiveTableColumn> columns) {
        QueryModel queryModel = new QueryModel(formTree.getRootFormId());
        if(tableModel.getFilter().isPresent()) {
            queryModel.setFilter(tableModel.getFilter().get());
        }
        for (EffectiveTableColumn column : columns) {
            queryModel.addColumns(column.getQueryModel());
        }
        return queryModel;
    }

    public FormTree getFormTree() {
        return formTree;
    }

    public ResourceId getDatabaseId() {
        return formTree.getRootFormClass().getDatabaseId();
    }

    public List<EffectiveTableColumn> getColumns() {
        return columns;
    }

    public Observable<ColumnSet> queryColumns(FormSource formSource) {
        return formSource.query(buildQuery(columns));
    }

    public String getFormLabel() {
        if(formTree.getRootState() == FormTree.State.VALID) {
            return formTree.getRootFormClass().getLabel();
        } else {
            return "";
        }
    }

    /**
     * Determines whether this table is sorted by the given formula
     */
    public Sorting getSorting(ResolvedNode formula) {
        for (EffectiveSortOrder sortOrder : ordering) {
            if(sortOrder.getFormula().equals(formula.toFormula())) {
                return sortOrder.isAscending() ? Sorting.ASCENDING : Sorting.DESCENDING;
            }
        }
        return Sorting.NONE;
    }

    public boolean isFiltered(EffectiveTableColumn column) {
        return filteredColumns.contains(column.getId());
    }

    public List<TableColumn> getColumnModels() {
        return columns.stream().map(EffectiveTableColumn::getModel).collect(Collectors.toList());
    }

    public List<EffectiveSortOrder> getOrdering() {
        return ordering;
    }
}
