/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.analysis.table;

import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumType;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Column format for an enum that can take multiple values.
 */
public class MultiEnumFormat implements ColumnFormat {

    private final String columnId;
    private final ResolvedNode formula;
    private final EnumType enumType;

    private final int numItems;
    private final String[] labels;

    public MultiEnumFormat(String columnId, ResolvedNode formula, EnumType enumType) {
        this.columnId = columnId;
        this.formula = formula;
        this.enumType = enumType;
        this.numItems = enumType.getItems().size();
        this.labels = new String[numItems];
        for (int i = 0; i < numItems; i++) {
            labels[i] = enumType.getItems().get(i).getLabel();
        }
    }

    public EnumType getEnumType() {
        return enumType;
    }

    public Renderer createRenderer(ColumnSet columnSet) {
        return new Renderer(columnSet);
    }

    @Override
    public List<ColumnModel> getColumnModels() {
        List<ColumnModel> columns = new ArrayList<>();
        for (EnumItem enumItem : enumType.getItems()) {

            ColumnModel model = new ColumnModel();
            model.setId(getItemId(enumItem));
            model.setFormula(new CompoundExpr(formula.toFormula(), new SymbolNode(enumItem.getId())));

            columns.add(model);
        }

        return columns;
    }

    private String getItemId(EnumItem enumItem) {
        return columnId + ":" + enumItem.getId();
    }

    @Override
    public <T> T accept(EffectiveTableColumn columnModel, TableColumnVisitor<T> visitor) {
        return visitor.visitMultiEnumColumn(columnModel, this);
    }

    @Override
    public int getExportedColumnCount() {
        return 1;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        MultiEnumFormat that = (MultiEnumFormat) o;
        return columnId.equals(that.columnId) && enumType.equals(that.enumType);
    }

    @Override
    public int hashCode() {
        return Objects.hash(columnId, enumType);
    }

    public class Renderer {

        private ColumnView[] views;

        private Renderer(ColumnSet columnSet) {
            views = new ColumnView[numItems];
            for (int i = 0; i < numItems; i++) {
                EnumItem item = enumType.getItems().get(i);
                views[i] = columnSet.getColumnView(getItemId(item));
            }
        }

        public String render(int rowIndex) {
            StringBuilder sb = new StringBuilder();
            boolean needsSep = false;
            for (int i = 0; i < numItems; i++) {
                ColumnView view = views[i];
                if(view != null && view.getBoolean(rowIndex) == ColumnView.TRUE) {
                    if(needsSep) {
                        sb.append(", ");
                    }
                    sb.append(labels[i]);
                    needsSep = true;
                }
            }

            if(sb.length() > 0) {
                return sb.toString();
            } else {
                return null;
            }
        }
    }
}
