/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.i18n.shared;

@SuppressWarnings("checkstyle:AvoidEscapedUnicodeCharacters")
public class Numerals {

    static final String EASTERN_ARABIC_RE = ".*[\u0660-\u0669].*";
    static final char EASTERN_ARABIC_ZERO = '\u0660';
    static final char EASTERN_ARABIC_NINE = '\u0669';
    static final char EASTERN_ARABIC_THOUSANDS_SEP = '\u066C';
    static final char EASTERN_ARABIC_DECIMAL_SEP = '\u066b';

    static final char SPACE = ' ';
    static final char THIN_SPACE = '\u2009';

    /**
     * In some locales, the apostrophe is used as a thousands seperator...
     * https://en.wikipedia.org/wiki/Decimal_separator#Examples_of_use
     */
    static final char APOSTROPHE = '\'';

    /**
     * resembles an apostrophe
     */
    static final char DECIMAL_SEPARATOR_KEY_SYMBOL = '\u2396';

    static final char MIDDLE_DOT = '·';

    static final char DOT_ABOVE = '\u02D9';

    /**
     * Preprocess a string so that it can be parsed by {@link Double#parseDouble(String)}.
     *
     * Convert any Eastern Arabic (Indic) Numerals found in the {@code input} string
     * to Western Arabic Numerals (0-9).
     *
     * <p>Eastern arabic thousands seperators, spaces, and the localized thousands seperator are ignored and
     * dropped from the output.</p>
     *
     * <p>The local decimal seperator is replaced with a '.'.</p>
     *
     * @param input the input string
     * @param localDecimalSeperator the character used as a decimal separator in this locale
     * @param localThousandsSeparator the character used as a thousands separator in this locale
     */
    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    public static String delocalize(String input, char localDecimalSeperator, char localThousandsSeparator) {

        char extraThousandsSeparator = 0;

        // If commas are involved, check to see
        // if they are obviously reversed

        if(localDecimalSeperator == ',' || localThousandsSeparator == ',') {
            int lastComma = input.lastIndexOf(',');
            int lastDot = input.lastIndexOf('.');

            if(lastComma != -1 && lastDot != -1) {
                if(lastDot > lastComma) {
                    localDecimalSeperator = '.';
                    localThousandsSeparator = ',';
                } else {
                    localDecimalSeperator = ',';
                    localThousandsSeparator = '.';
                }
            }
        } else {
            // If this locale doesn't use a comma for either the thousands seperator OR
            // the decimal seperator, then accept it as an additional thousands seperator

            extraThousandsSeparator = ',';
        }

        StringBuilder output = new StringBuilder(input.length());
        for (int i = 0; i < input.length(); i++) {
            char c = input.charAt(i);
            if(c >= EASTERN_ARABIC_ZERO && c <= EASTERN_ARABIC_NINE) {
                output.append((char) (c - EASTERN_ARABIC_ZERO + '0'));

            } else if(c == EASTERN_ARABIC_DECIMAL_SEP ||
                      c == localDecimalSeperator ||
                      c == MIDDLE_DOT ||
                      c == DECIMAL_SEPARATOR_KEY_SYMBOL) {

                output.append('.');

            } else if(c == EASTERN_ARABIC_THOUSANDS_SEP ||
                      c == localThousandsSeparator ||
                      c == SPACE ||
                      c == THIN_SPACE ||
                      c == extraThousandsSeparator ||
                      c == DOT_ABOVE) {

                // skip, not needed for parsing

            } else {
                output.append(c);
            }
        }
        return output.toString();
    }
}
