/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.io.csv;

import java.util.ArrayList;
import java.util.List;

/**
 * Parses delimited text files into rows and columns
 */
public class RowParser {

    public static final char QUOTE_CHAR = '"';

    public interface EventListener {

        void field(String string);

        boolean endOfRecord();

        default void endOfFile() {
        }
    }

    private String text;
    private int length;
    private int currentPos = 0;
    private char delimiter;
    private String delimiterString;
    private int rowIndex;
    private int maxRowCount = Integer.MAX_VALUE;

    public RowParser(String text, char delimiter) {
        this.text = text;
        this.length = text.length();
        this.delimiter = delimiter;
        this.delimiterString = "" + delimiter;
    }

    public List<ParsedRow> parseRows() {
        return parseRows(Integer.MAX_VALUE);
    }

    public List<ParsedRow> parseRows(int maxRowCount) {


        List<ParsedRow> rows = new ArrayList<>();
        parse(new EventListener() {

            List<String> cells = new ArrayList<>();

            @Override
            public void field(String string) {
                cells.add(string);
            }

            @Override
            public boolean endOfRecord() {
                rows.add(new ParsedRow(cells));
                cells = new ArrayList<>();
                return rows.size() < maxRowCount;
            }

            @Override
            public void endOfFile() {
            }
        });

        return rows;
    }

    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    public void parse(EventListener eventListener) {

        int textLength = text.length();
        int currentPos = skipBlankLines(this.currentPos);

        while(currentPos < textLength) {

            // Start of column
            char c = text.charAt(currentPos);
            boolean lastColumn = false;

            if(c == QUOTE_CHAR) {

                currentPos++;
                int columnStart = currentPos;
                int columnEnd;

                boolean quotedQuotes = false;

                while(true) {
                    int nextQuote = text.indexOf(QUOTE_CHAR, currentPos);
                    if(nextQuote == -1) {
                        // unterminated quoted column
                        columnEnd = textLength;
                        break;
                    }
                    String followingQuote = text.substring(nextQuote + 1, Math.min(textLength, nextQuote + 3));
                    if(followingQuote.startsWith(delimiterString)) {
                        columnEnd = nextQuote;
                        currentPos = nextQuote + 2;
                        break;
                    }
                    if(followingQuote.isEmpty()) {
                        columnEnd = nextQuote;
                        currentPos = textLength;
                        lastColumn = true;
                        break;
                    }
                    if(followingQuote.startsWith("\n")) {
                        columnEnd = nextQuote;
                        currentPos = columnEnd + 2;
                        lastColumn = true;
                        break;
                    }
                    if(followingQuote.startsWith("\r\n")) {
                        columnEnd = nextQuote;
                        lastColumn = true;
                        currentPos = columnEnd + 3;
                        break;
                    }
                    quotedQuotes = true;
                    currentPos = nextQuote + 1;
                }

                String cell = text.substring(columnStart, columnEnd);
                if(quotedQuotes) {
                    cell = cell.replaceAll("\"\"", "\"");
                }
                eventListener.field(cell);


            } else {

                // Find end of unquoted field
                int columnStart = currentPos;
                int columnEnd = text.indexOf(delimiter, columnStart);
                int lineEnd = text.indexOf("\n", columnStart);

                if(columnEnd < 0 || (lineEnd > 0 && lineEnd < columnEnd)) {
                    columnEnd = lineEnd;
                    lastColumn = true;
                }
                if(columnEnd == -1) {
                    columnEnd = textLength;
                    lastColumn = true;
                }

                // Advance the cursor to the start of the next field
                currentPos = columnEnd + 1;

                // Backup if there is an \r at the end of this column
                if(lastColumn) {
                    if(text.charAt(columnEnd - 1) == '\r') {
                        columnEnd --;
                    }
                }

                // Have we reached the end of the file?
                if(currentPos == textLength) {
                    lastColumn = true;
                }

                eventListener.field(text.substring(columnStart, columnEnd));
            }

            if(lastColumn) {
                boolean continueParsing = eventListener.endOfRecord();
                if(!continueParsing) {
                    this.currentPos = currentPos;
                    return;
                }
                currentPos = skipBlankLines(currentPos);
            }
        }

        this.currentPos = textLength;
        eventListener.endOfFile();
    }

    private int skipBlankLines(int currentPos) {
        // a blank line is composed only of delimeters followed by newlines.
        int lastLineStart = currentPos;
        while(currentPos < text.length()) {
            char c = text.charAt(currentPos);
            if(c == '\n' || c == '\r') {
                lastLineStart = currentPos + 1;
            } else if(c != delimiter && c != ' ') {
                break;
            }
            currentPos++;
        }
        return lastLineStart;
    }


}
