/*
 * ActivityInfo
 * Copyright (C) 2009-2013 UNICEF
 * Copyright (C) 2014-2021 BeDataDriven Groep B.V.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.activityinfo.io.match.server.coord;

import net.lightoze.gwt.i18n.server.LocaleProxy;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Locale;

/**
 * Provides number formatting & parsing. Extracted from the class to allow
 * for testing.
 *
 * <p>N.B. there is an alternative version of this class in src/main/resource/.../emul/... that
 * is used when compiling to JavaScript.</p>
 */
public final class CoordinateNumbers {

    private final NumberFormat decimalFormat;
    private final NumberFormat shortDecimalFormat;
    private final NumberFormat intFormat;

    public static CoordinateNumbers get() {
        return new CoordinateNumbers(LocaleProxy.getLocale());
    }

    public CoordinateNumbers(Locale locale) {
        decimalFormat = NumberFormat.getInstance(locale);
        decimalFormat.setMinimumFractionDigits(6);
        decimalFormat.setMaximumFractionDigits(6);
        decimalFormat.setMinimumIntegerDigits(0);

        shortDecimalFormat = NumberFormat.getInstance(locale);
        shortDecimalFormat.setMinimumFractionDigits(2);
        shortDecimalFormat.setMaximumFractionDigits(2);
        shortDecimalFormat.setMinimumIntegerDigits(1);

        intFormat = NumberFormat.getIntegerInstance(locale);
    }

    /**
     * @return the decimal separator for the current locale. In American English, this would be "."
     */
    public char getDecimalSeparator() {
        return ((DecimalFormat) decimalFormat).getDecimalFormatSymbols().getDecimalSeparator();
    }


    /**
     * Return true if the given character {@code c} is a digit in the current locale
     */
    public boolean isDigit(char c) {
        return Character.isDigit(c);
    }


    /*
     * Formats a coordinate as Degree-decimal, with exactly six decimal places and always with a
     * sign prefix. Equivalent the formatting pattern "+0.000000;-0.000000"
     */
    public String formatDDd(double value) {
        if (value > 0) {
            return "+" + decimalFormat.format(value);
        } else {
            return decimalFormat.format(value);
        }
    }

    public String formatShortFraction(double value) {
        return shortDecimalFormat.format(value);
    }

    public String formatInt(double value) {
        return intFormat.format(value);
    }
}
