/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.io.match.shared.date;

import org.activityinfo.model.type.time.LocalDate;

import javax.annotation.Nonnull;
import java.util.Arrays;

/**
 * Extracts local dates from strings using a series of heuristics
 */
public class LatinDateParser {

    private static final String SEPARATORS = "-/\\, \t.";
    
    private static final int NUM_COMPONENTS = 3;

    private final int[] components = new int[NUM_COMPONENTS];

    /**
     * The pivot year is used to determine whether a two-digit year refers to 1900 or 2000. For example,
     * if the pivot year is set to 50, the default value, then:
     * <ul>
     *     <li>'19 Jan 99' will be parsed as 1999</li>
     *     <li>'15 May 01' will be parsed as 2015</li>
     *     <li>'30 May 51' will be parsed as 1951</li>
     *     <li>'30 May 49' will be parsed as 2049</li>
     * </ul>
     */
    private static final int PIVOT_YEAR = 50;


    @Nonnull
    public LocalDate parse(@Nonnull String string) {
        // basically we expect three components, in SOME format
        // that are separated by something normal

        int charIndex = 0;
        int component = 0;

        // if we stumble across the month name,
        // make sure we note this.
        int monthIndex = GregorianMonths.NOT_FOUND;

        while(component < NUM_COMPONENTS) {

            if(charIndex >= string.length()) {
                throw new IllegalArgumentException("Not enough components in '" + string + "', found: " +
                        Arrays.toString(components));
            }

            int start = charIndex;
            char c = string.charAt(start);

            if(Character.isDigit(c)) {

                // read all the digits in
                do {
                    charIndex++;
                }
                while(isDigit(string, charIndex));

                // parse as number
                components[component] = Integer.parseInt(string.substring(start, charIndex));

                // move on the next component
                component++;

            } else {

                // read until we hit a separator or digit
                do {
                    charIndex++;
                } while(isPartOfWord(string, charIndex));

                int monthNameLength = charIndex-start;
                if(monthNameLength > GregorianMonths.MIN_MONTH_NAME_LENGTH) {
                    int month = GregorianMonths.tryParseName(string.substring(start, charIndex));
                    if(month != GregorianMonths.NOT_FOUND) {
                        components[component] = month;
                        monthIndex = component;
                        component++;
                    }
                }
            }

            // advance through any separator chars
            while(isSeparator(string, charIndex)) {
                charIndex++;
            }
        }

        if(monthIndex != -1) {
            return parseUsingKnownMonthPosition(components, monthIndex);

        } else {
            // try to find the obvious year
            int yearIndex = findYearIndex(components, -1);
            if(yearIndex == -1) {
                // if we can't find a 4-digit year, we can only assume that it comes at
                // the end in some completely ambiguous form like 3/4/12
                yearIndex = 2;
            }

            return parseUsingKnownYearPosition(string, components, yearIndex);
        }
    }

    private LocalDate parseUsingKnownYearPosition(String string, int[] components, int yearIndex) {
        if(yearIndex == 0) {
            // usually YYYY-MM-dd
            return toDate(components, yearIndex, 1, 2);
        } else if(yearIndex == 1) {
            // date in the middle?? 31-2000-12 ??  i don't think so...
            throw new IllegalArgumentException(string);
        } else {
            // the classic ambiguous 5/3/2007
            return toDate(components, yearIndex, 1, 0);
        }
    }

    private boolean monthAndDayMatch(int[] components, int monthIndex, int dayIndex) {
        int month = components[monthIndex];
        int day = components[dayIndex];
        if(month > 12 || month == 0) {
            return false;
        }
        if(day > GregorianMonths.getMaxDaysInMonth(month) || day == 0) {
            return false;
        }
        return true;
    }

    private LocalDate parseUsingKnownMonthPosition(int components[], int monthIndex) {
        int yearIndex = findYearIndex(components, monthIndex);
        if(yearIndex != -1) {
            int dayIndex = remainingIndex(monthIndex, yearIndex);
            return new LocalDate(components[yearIndex], components[monthIndex], components[dayIndex]);
        } else {
            // best guess
            if(monthIndex == 1) {
                // usually 31st May 12
                return toDate(components, 2, monthIndex, 0);
            } else {
                // who knows...
                return toDate(components, 2, 0, 1);
            }
        }
    }

    private LocalDate toDate(int[] components, int yearIndex, int monthIndex, int dayIndex) {
        int year = components[yearIndex];
        if((year >= 100 && year < 1000) || (year >= 10000)) {
            throw new IllegalArgumentException();
        }
        if(!monthAndDayMatch(components, monthIndex, dayIndex)) {
            throw new IllegalArgumentException();
        }
        if(isTwoDigits(year)) {
            year += inferCentury(year);
        }
        return new LocalDate(year, components[monthIndex], components[dayIndex]);
    }

    private int inferCentury(int year) {
        if(year < PIVOT_YEAR) {
            return 2000;
        } else {
            return 1900;
        }
    }

    private boolean isTwoDigits(int year) {
        return (year >= 0 && year < 100);
    }

    private int remainingIndex(int monthIndex, int yearIndex) {
        for(int i=0;i!=NUM_COMPONENTS;++i) {
            if(i != monthIndex && i != yearIndex) {
                return i;
            }
        }
        throw new IllegalStateException();
    }

    private int findYearIndex(int components[], int monthIndex) {
        for(int i=0;i!=NUM_COMPONENTS;++i) {
            if(i != monthIndex) {
                if(components[i] >= 1000 && components[i] < 10000) {
                    return i;
                }
            }
        }
        return GregorianMonths.NOT_FOUND;
    }

    private boolean isSeparator(String string, int charIndex) {
        if(charIndex < string.length()) {
            return SEPARATORS.indexOf(string.charAt(charIndex)) != -1;
        } else {
            return false;
        }
    }

    private boolean isDigit(String string, int charIndex) {
        if(charIndex < string.length()) {
            return Character.isDigit(string.charAt(charIndex));
        } else {
            return false;
        }
    }

    private boolean isPartOfWord(String string, int charIndex) {
        if(charIndex < string.length()) {
            return Character.isLetter(string.charAt(charIndex));
        } else {
            return false;
        }
    }
}

