/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.io.match.shared.names;

import java.text.Normalizer;

/**
 * Normalizes unicode strings for (our) purposes of comparison, discarding diacritical marks and decomposing
 * ligatures.
 *
 */
@SuppressWarnings("checkstyle:avoidescapedunicodecharacters")
public final class NameNormalizer {

    private NameNormalizer() { }

    /**
     * Regular expression used to remove diacritcal marks from an NFD-normalized string.
     */
    private static final String DIACRITIC_PATTERN = "[\\u0300-\\u036f\\u064B-\\u065F]";

    /**
     * Normalizes a string for the purposes of matching.
     *
     * <ul>
     *     <li>If {@code input} is a latin letter precomposed with a diacritical mark, for example "é",
     *     {@code normalizeCharacter} will return a upper-case version of the letter in the sequence</li>
     *     <li>If {@code input} is a precomposed latin ligature, for example "Æ",  {@code normalizeCharacter} will
     *     return the lower case versions of their constituent letters</li>
     *     <li>Otherwise the {@code normalizeCharacter} will return {@code input.toUpperCase()}</li>
     * </ul>
     *
     */
    public static String normalizeString(String input) {
        // First transform the string to uppercase, and decompose any pre-composed chacters. This will yield
        // á => A \u0301
        // ñ => N \u0303
        String decomposed = Normalizer.normalize(input.toUpperCase(), Normalizer.Form.NFD);

        // Now remove the diacritical marks like accent ague for latin scripts or hamza's in Arabic.
        String withoutDiacriticals = decomposed.replaceAll(NameNormalizer.DIACRITIC_PATTERN, "");

        // Now renormalize to eliminate things like positional forms of arabic characters
        return Normalizer.normalize(withoutDiacriticals, Normalizer.Form.NFKC);
    }
}
