/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.json;

import jsinterop.base.Js;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.*;
import java.util.function.Function;

public class JsonArrays {

    public static <T> T enumFromJson(JsonValue string, Function<String, T> valueOf, T defaultValue) {
        String s = string.asString();
        if(s == null) {
            return defaultValue;
        } else {
            return valueOf.apply(s.toUpperCase());
        }
    }

    public static <T> @Nullable JsonValue optionalToJson(Optional<T> value, Function<T, JsonValue> serializer) {
        if(value.isPresent()) {
            T t = Js.uncheckedCast(value.get());
            JsonValue jv = Js.uncheckedCast(serializer.apply(t));
            return jv;
        } else {
            return null;
        }
    }

    public static <T> List<T> listFromJson(JsonValue array, FromJsonFunction<T> deserializer) {
        List<T> list = new ArrayList<>();
        if(array.isJsonArray()) {
            for (int i = 0; i < array.length(); i++) {
                list.add(deserializer.fromJson(array.get(i)));
            }
        }
        return list;
    }

    public static <K, V> Map<K, V> mapFromJson(JsonValue object, Function<String, K> keyDeserializer, FromJsonFunction<V> valueDeserializer) {
        Map<K, V> map = new HashMap();
        if(object.isJsonObject()) {
            String[] keys = object.keys();
            for (int i = 0; i < keys.length; i++) {
                map.put(keyDeserializer.apply(keys[i]), valueDeserializer.fromJson(object.get(keys[i])));
            }
        }
        return map;
    }

    public static <T> Optional<T> optionalFromJson(JsonValue jsonValue, FromJsonFunction<T> deserializer) {
        if(jsonValue == null || jsonValue.isJsonNull()) {
            return Optional.empty();
        } else {
            return Optional.<T>ofNullable(deserializer.fromJson(jsonValue));
        }
    }

    public static <T> Set<T> setFromJson(JsonValue array, FromJsonFunction<T> deserializer) {
        Set<T> set = new HashSet<>();
        if(array.isJsonArray()) {
            for (int i = 0; i < array.length(); i++) {
                set.add(deserializer.fromJson(array.get(i)));
            }
        }
        return set;
    }

    public static <T> JsonValue toJsonArray(Iterable<T> elements, ToJsonFunction<T> serializer) {
        JsonValue array = Json.createArray();
        for (T element : elements) {
            JsonValue jsonElement = serializer.toJson(element);
            array.add(jsonElement);
        }
        return array;
    }

    public static JsonValue toJsonArrayFromEnums(Iterable<? extends Enum<?>> elements) {
        JsonValue array = Json.createArray();
        for (Enum<?> element : elements) {
            array.add(Json.create(element.name()));
        }
        return array;
    }

    public static JsonValue intToJson(Integer integer) {
        return Json.create(integer);
    }

    public static JsonValue boolToJson(Boolean bool) {
        return Json.create(bool);
    }

    public static JsonValue doubleToJson(Double d) {
        return Json.create(d);
    }

    public static JsonValue toJsonString(Enum<?> value) {
        return Json.create(value.name());
    }

    public static JsonValue toJsonArray(Iterable<? extends JsonSerializable> objects) {
        JsonValue array = Json.createArray();
        for (JsonSerializable object : objects) {
            array.add(object.toJson());
        }
        return array;
    }
}
