/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.json;

import jsinterop.annotations.JsType;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

@GwtIncompatible
public class JsonSerdeFactory {

    /**
     * Tries to create a {@link JsonSerde} for a given Java type.
     *
     * <p>The following types are supported:</p>
     * <ul>
     *     <li>The {@link JsonValue} class.</li>
     *     <li>Classes annotated with {@link JsType}</li>
     *     <li>Classes annotated with {@link AutoJson}</li>
     *     <li>Classes implementing {@link JsonSerializable} with a static {@code fromJson} method</li>
     *     <li>A parameterized {@link List} of any of the above.</li>
     * </ul>
     */
    public static Optional<JsonSerde> tryCreate(Type type) {

        if (type instanceof ParameterizedType) {
            ParameterizedType parameterizedType = (ParameterizedType) type;
            if(parameterizedType.getRawType().equals(List.class)) {
                Type elementType = parameterizedType.getActualTypeArguments()[0];
                if(elementType instanceof Class) {
                    return tryCreateFromClass(((Class<?>) elementType)).map(es -> new ListStrategy(es));
                }
            }
        }

        if(type instanceof Class) {
            return tryCreateFromClass((Class<?>) type);
        }

        return Optional.empty();
    }

    private static Optional<JsonSerde> tryCreateFromClass(Class<?> type) {
        Class<?> typeClass = type;
        if(typeClass.isAnnotationPresent(AutoJson.class)) {
            return Optional.of(fromAutoJsonClass(typeClass));

        } else if(typeClass.isAnnotationPresent(JsType.class)){
            return Optional.of(new JsTypeStrategy(type));

        } else if(JsonSerializable.class.isAssignableFrom(typeClass)) {
            return Optional.of(fromJsonSerializable(typeClass));

        } else if(JsonValue.class.isAssignableFrom(typeClass)) {
            return Optional.of(new Noop());
        }
        return Optional.empty();
    }

    private static JsonSerde fromAutoJsonClass(Class<?> type) {

        try {
            Class jsonClass = Class.forName(type.getName() + "Json");

            Method reader = jsonClass.getMethod("fromJson", JsonValue.class);
            Method writer = jsonClass.getMethod("toJson", type);

            return new MethodHandleStrategy(reader, writer);

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static JsonSerde fromJsonSerializable(Class<?> typeClass) {

        try {
            Method reader = typeClass.getMethod("fromJson", JsonValue.class);
            Method writer = typeClass.getMethod("toJson");

            return new MethodHandleStrategy(reader, writer);

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private static class Noop implements JsonSerde {

        @Override
        public Object fromJson(JsonValue jsonValue) {
            return jsonValue;
        }

        @Override
        public JsonValue toJson(Object value) {
            return (JsonValue) value;
        }
    }

    private static class JsTypeStrategy implements JsonSerde {

        private Class<?> jsTypeClass;

        private JsTypeStrategy(Class<?> type) {
            this.jsTypeClass = type;
        }

        @Override
        @Nullable
        public Object fromJson(JsonValue jsonValue) {
            try {
                return Json.fromJson(jsTypeClass, jsonValue);
            } catch (JsonMappingException e) {
                throw new RuntimeException(e);
            }
        }

        @Override
        public JsonValue toJson(Object value) {
            return Json.toJson(value);
        }
    }

    private static class MethodHandleStrategy implements JsonSerde {

        private final MethodHandle reader;
        private final MethodHandle writer;

        private MethodHandleStrategy(Method reader, Method writer) {
            MethodHandles.Lookup lookup = MethodHandles.publicLookup();
            try {
                this.reader = lookup.unreflect(reader);
                this.writer = lookup.unreflect(writer);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            }

        }

        @Override
        public Object fromJson(JsonValue jsonValue) {
            try {
                return reader.invoke(jsonValue);
            } catch (Throwable caught) {
                throw new RuntimeException(caught);
            }
        }

        @Override
        public JsonValue toJson(Object value) {
            try {
                return (JsonValue) writer.invoke(value);
            } catch (Throwable caught) {
                throw new RuntimeException(caught);
            }
        }
    }

    private static class ListStrategy implements JsonSerde {

        private final JsonSerde elementStrategy;

        private ListStrategy(JsonSerde elementStrategy) {
            this.elementStrategy = elementStrategy;
        }

        @Override
        public Object fromJson(JsonValue jsonValue) {
            List list = new ArrayList<>();
            for (int i = 0; i < jsonValue.length(); i++) {
                list.add(elementStrategy.fromJson(jsonValue.get(i)));
            }
            return list;
        }

        @Override
        public JsonValue toJson(Object value) {
            JsonValue array = Json.createArray();
            List list = (List) value;
            for (int i = 0; i < list.size(); i++) {
                array.add(elementStrategy.toJson(list.get(i)));
            }
            return array;
        }
    }

}
