/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.json.impl;

import org.activityinfo.json.*;
import org.checkerframework.checker.nullness.qual.NonNull;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.*;

/**
 * Server-side implementation of JsonObject.
 */
public class JreJsonObject extends JreJsonValue implements JsonValue {

    private static final long serialVersionUID = 1L;

    private static List<String> stringifyOrder(String[] keys) {
        List<String> toReturn = new ArrayList<String>();
        List<String> nonNumeric = new ArrayList<String>();
        for (String key : keys) {
            if (key.matches("\\d+")) {
                toReturn.add(key);
            } else {
                nonNumeric.add(key);
            }
        }
        Collections.sort(toReturn);
        toReturn.addAll(nonNumeric);
        return toReturn;
    }

    private transient JsonFactory factory;
    private transient Map<@NonNull String, JsonValue> map = new LinkedHashMap<@NonNull String, JsonValue>();

    public JreJsonObject(JsonFactory factory) {
        this.factory = factory;
    }

    @Override
    public boolean asBoolean() {
        return true;
    }

    @Override
    public double asNumber() {
        return Double.NaN;
    }

    @Override
    public String asString() {
        return "[object Object]";
    }

    public JsonValue get(String key) {
        JsonValue value = map.get(key);
        if(value == null) {
            return JreJsonNull.NULL_INSTANCE;
        }
        return value;
    }

    public boolean getBoolean(String key) {
        return get(key).asBoolean();
    }

    public double getNumber(String key) {
        return get(key).asNumber();
    }


    public @Nullable String getString(String key) {
        return get(key).asString();
    }

    @Override
    public String getRequiredString(String key) throws JsonException {
        String s = get(key).asString();
        if(s == null) {
            throw new JsonException(key);
        }
        return s;
    }

    public JsonType getType() {
        return JsonType.OBJECT;
    }

    @Override
    public boolean hasKey(String key) {
        return map.containsKey(key);
    }

    @Override
    @SuppressWarnings("nullness")
    public String[] keys() {
        return map.keySet().toArray(new String[map.size()]);
    }

    public void put(String key, JsonValue value) {
        if (value == null) {
            value = factory.createNull();
        }
        map.put(key, value);
    }

    public void put(String key, @Nullable String value) {
        put(key, factory.createFromNullable(value));
    }

    public void put(String key, double value) {
        put(key, factory.create(value));
    }

    public void put(String key, boolean bool) {
        put(key, factory.create(bool));
    }

    @Override
    public void remove(String key) {
        map.remove(key);
    }

    @Override
    public void add(String key, JsonValue value) {
        put(key, value);
    }

    public void set(String key, JsonValue value) {
        put(key, value);
    }

    public String toJson() {
        return JsonUtil.stringify(this);
    }

    public String toString() {
        throw new UnsupportedOperationException("SHOULD NOT BE CALLED!!");
    }

    @Override
    @SuppressWarnings("keyfor")
    public Iterable<Map.Entry<String, JsonValue>> entrySet() {
        return map.entrySet();
    }

    @Override
    public void traverse(JsonVisitor visitor, JsonContext ctx) {
        if (visitor.visitObject(this, ctx)) {
            JsonObjectContext objCtx = new JsonObjectContext(this);
            for (String key : stringifyOrder(keys())) {
                if (visitor.visitKey(key, objCtx)) {
                    visitor.accept(get(key), objCtx);
                    objCtx.setFirst(false);
                }
            }
        }
        visitor.endObjectVisit(this, ctx);
    }

    @Override
    public boolean isJsonObject() {
        return true;
    }


    @com.google.gwt.core.shared.GwtIncompatible
    private void readObject(ObjectInputStream stream)
            throws IOException, ClassNotFoundException {
        JreJsonObject instance = parseJson(stream);
        this.factory = Json.instance();
        this.map = instance.map;
    }

    @com.google.gwt.core.shared.GwtIncompatible
    private void writeObject(ObjectOutputStream stream) throws IOException {
        stream.writeObject(toJson());
    }
}
