/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.json.impl;

import com.google.gwt.core.client.JavaScriptObject;
import com.google.gwt.core.client.JsArrayString;
import jsinterop.base.Js;
import org.activityinfo.json.JsonException;
import org.activityinfo.json.JsonType;
import org.activityinfo.json.JsonValue;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.HashMap;
import java.util.Map;

/**
 * JSO backed implementation of JsonValue.
 */
public final class JsJsonValue extends JavaScriptObject implements JsonValue {

    static native JsonValue box(JsonValue value) /*-{
        // box for DevMode, not ProdMode
        return @com.google.gwt.core.client.GWT::isScript()() || value == null ? value : Object(value);
    }-*/;

    static native JsonValue debox(JsonValue value) /*-{
        // we don't debox (currently), because ProdMode is now unboxed, and DevMode should stay boxed
        return value;
    }-*/;

    static native String getJsType(Object obj) /*-{
        return typeof obj;
    }-*/;

    static native boolean isArray(Object obj) /*-{
        // ensure that array detection works cross-frame
        return Object.prototype.toString.apply(@org.activityinfo.json.impl.JsJsonValue::debox(Lorg/activityinfo/json/JsonValue;)(obj)) === '[object Array]';
    }-*/;

    private static native boolean isNull(JsJsonValue jsJsonValue) /*-{
        // TODO(cromwellian): if this moves to GWT, we may have to support more leniency
        return jsJsonValue === null;
    }-*/;

    protected JsJsonValue() {
    }

    @Override
    final public native boolean asBoolean() /*-{
        return !!this;
    }-*/;

    @Override
    final public native double asNumber() /*-{
        if (this == null) {
            return 0;
        }
        return @com.google.gwt.core.client.GWT::isScript()() ?
            +@org.activityinfo.json.impl.JsJsonValue::debox(Lorg/activityinfo/json/JsonValue;)(this) :
            (+@org.activityinfo.json.impl.JsJsonValue::debox(Lorg/activityinfo/json/JsonValue;)(this)).valueOf();
    }-*/;

    @Override
    // avoid casts, as compiler will throw CCE trying to cast a raw JS String to an interface
    final public native String asString() /*-{
        return this == null ? null : "" + this;
    }-*/;

    final public JsonType getType() {
        if (isJsonNull()) {
            return JsonType.NULL;
        }
        String jsType = getJsType(this);
        if ("string".equals(jsType)) {
            return JsonType.STRING;
        } else if ("number".equals(jsType)) {
            return JsonType.NUMBER;
        } else if ("boolean".equals(jsType)) {
            return JsonType.BOOLEAN;
        } else if ("object".equals(jsType)) {
            return isArray(this) ? JsonType.ARRAY : JsonType.OBJECT;
        }
        throw new UnsupportedOperationException("Unknown Json Type");
    }


    final public native String toJson() /*-{
        // skip hashCode field
        return $wnd.JSON.stringify(this, function (keyName, value) {
            if (keyName == "$H") {
                return undefined; // skip hashCode property
            }
            return value;
        }, 0);
    }-*/;


    @Override
    public final int asInt() {
        return (int) asNumber();
    }

    @Override
    public final long asLong() {
        return Long.parseLong(asString());
    }

    @Override
    public final native boolean isString() /*-{
        return (typeof this) === "string";
    }-*/;

    @Override
    public final native boolean isJsonNull() /*-{
        return this == null;
    }-*/;

    @Override
    public final boolean isJsonArray() {
        return isArray(this);
    }


    @Override
    public final boolean isJsonPrimitive() {
        switch (getType()) {
            case STRING:
            case NUMBER:
            case BOOLEAN:
                return true;
        }
        return false;
    }

    @Override
    public final boolean isBoolean() {
        return getType() == JsonType.BOOLEAN;
    }

    @Override
    public final boolean isJsonObject() {
        return getType() == JsonType.OBJECT;
    }

    @Override
    public final boolean isNumber() {
        return getType() == JsonType.NUMBER;
    }

    @Override
    @SuppressWarnings("keyfor")
    public Iterable<Map.Entry<String, JsonValue>> entrySet() {
        Map<String, JsonValue> map = new HashMap<>();
        for (String key : keys()) {
            map.put(key, get(key));
        }
        return map.entrySet();
    }

    @Override
    public boolean getBoolean(String key) {
        return get(key).asBoolean();
    }


    public final native JsonValue get(int index) /*-{
        return this[index];
    }-*/;


    public final native JsonValue get(String key) /*-{
        return this[key];
    }-*/;


    @Override
    public Iterable<JsonValue> values() {
        return new JsonArrayIterable(this);
    }

    public boolean getBoolean(int index) {
        return get(index).asBoolean();
    }

    public double getNumber(int index) {
        return get(index).asNumber();
    }

    @Override
    public double getNumber(String key) {
        return get(key).asNumber();
    }

    @Override
    public @Nullable String getString(String key) {
        return get(key).asString();
    }

    @Override
    public native void put(String key, JsonValue value) /*-{
        this[key] = value;
    }-*/;


    @Override
    public void put(String key, @Nullable String value) {
        put(key, (JsonValue)JsJsonFactory.uncheckedCast(value));
    }

    @Override
    public void put(String key, double value) {
        put(key, JsJsonFactory.createNumber(value));
    }

    @Override
    public void put(String key, boolean bool) {
        put(key, JsJsonFactory.createBoolean(bool));
    }

    @Override
    public native boolean hasKey(String key) /*-{
        return key in this;
    }-*/;

    @Override
    public native void remove(String key) /*-{
        delete this[key];
    }-*/;

    @Override
    public void add(String key, JsonValue value) {
        put(key, value);
    }

    public @Nullable String getString(int index) {
        return get(index).asString();
    }

    @Override
    public String getRequiredString(String key) throws JsonException {
        String s = get(key).asString();
        if(s == null) {
            throw new JsonException(key);
        }
        return s;
    }

    public String[] keys() {
        JsArrayString str = keys0();
        return reinterpretCast(str);
    }

    public native JsArrayString keys0() /*-{
        var keys = [];
        for (var key in this) {
            if (Object.prototype.hasOwnProperty.call(this, key) && key != '$H') {
                keys.push(key);
            }
        }
        return keys;
    }-*/;


    private native String[] reinterpretCast(JsArrayString arrayString) /*-{
        return arrayString;
    }-*/;

    @Override
    public native int length() /*-{
        return this.length;
    }-*/;

    public native void set(int index, JsonValue value) /*-{
        this[index] = value;
    }-*/;


    public native void add(JsonValue value) /*-{
        this.push(value);
    }-*/;

    public void set(int index, String string) {
        set(index, Js.<JsonValue>uncheckedCast(string));
    }

    public void set(int index, double number) {
        set(index, JsJsonFactory.createNumber(number));
    }

    public void set(int index, boolean bool) {
        set(index, JsJsonFactory.createBoolean(bool));
    }

}
