/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.account;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.time.LocalDate;

import java.util.Date;
import java.util.Optional;

/**
 * The status of the user's billing account
 */
public class AccountStatus implements JsonSerializable {

    public static final String V41_FEATURE_FLAG = "v41";

    public static final String MOBILE_DATA_COLLECTION_FLAG = "mdc";

    public static final int FREE_TRIAL_LIMIT = 10;

    public static final int DAYS_PER_WEEK = 7;

    private final UserAccount userAccount;
    private final Optional<BillingAccount> billingAccount;

    public AccountStatus(UserAccount userAccount, Optional<BillingAccount> billingAccount) {
        this.userAccount = userAccount;
        this.billingAccount = billingAccount;
    }

    public boolean isTrial() {
        return billingAccount.map(BillingAccount::isTrial).orElse(false);
    }

    public ResourceId getUserAccountId() {
        return userAccount.getId();
    }

    public String getUserAccountEmail() {
        return userAccount.getEmail();
    }

    public String getUserAccountName() {
        return userAccount.getName();
    }

    public int getExpirationTime() {
        return billingAccount.map(b -> (int)(b.getExpirationTime() / 1000)).orElse(0);
    }

    public int getDatabaseCount() {
        return billingAccount.map(b -> b.getDatabaseCount()).orElse(0);
    }

    public long getBillingAccountId() {
        return billingAccount.map(b -> b.getId()).orElse(0L);
    }

    public Integer getExpectedPaymentDate() {
        return billingAccount.map(b -> (int)(b.getExpectedPaymentTime() / 1000)).orElse(0);
    }

    public UserAccount getUserAccount() {
        return userAccount;
    }

    public Optional<BillingAccount> getBillingAccount() {
        return billingAccount;
    }

    public int hoursUntilExpiration(Date now) {
        int secondsUntil = secondsUntilExpiration(now);
        int hours = Math.floorDiv(secondsUntil, 3600);
        return hours;
    }

    private int secondsUntilExpiration(Date now) {
        int secondsNow = (int) (now.getTime() / 1000);
        return getExpirationTime() - secondsNow;
    }


    public int daysUntilExpiration(Date now) {
        int hours = hoursUntilExpiration(now);
        return Math.floorDiv(hours, 24);
    }

    private int secondsUntilExpectedPayment(Date now) {
        int secondsNow = (int) (now.getTime() / 1000);
        return getExpectedPaymentDate() - secondsNow;
    }

    public int hoursUntilExpectedPayment(Date now) {
        int secondsUntil = secondsUntilExpectedPayment(now);
        int hours = Math.floorDiv(secondsUntil, 3600);
        return hours;
    }

    public int daysUntilExpectedPayment(Date now) {
        int hours = hoursUntilExpectedPayment(now);
        return Math.floorDiv(hours, 24);
    }

    /**
     * @return human readable string
     */
    public String expiringIn(Date now) {
        int hours = hoursUntilExpiration(now);
        if (hours <= 48) {
            return I18N.MESSAGES.inHours(hours);
        } else {
            int days = Math.floorDiv(hours, 24);
            if (days < 21) {
                return I18N.MESSAGES.inDays(days);
            } else {
                int weeks = Math.floorDiv(days, 7);
                return I18N.MESSAGES.inWeeks(weeks);
            }
        }
    }

    /**
     * @return human readable string
     */
    public String paymentExpectedIn(Date now) {
        int hours = hoursUntilExpectedPayment(now);
        if (hours < 1) {
            return "overdue";
        } else if (hours <= 48) {
            return "due in " + hours + " hours";
        } else {
            int days = Math.floorDiv(hours, 24);
            if (days < 10) {
                return "due in " + days + " days";
            } else {
                int weeks = Math.floorDiv(days, 7);
                return "due in " + weeks + " weeks";
            }
        }
    }

    /**
     * True if the user should be warned about expiration
     */
    public boolean shouldWarn(Date now) {
        if(getDatabaseCount() == 0) {
            return false;
        }
        int daysLeft = daysUntilExpiration(now);
        return getBillingAccount()
            .map(billing -> {
                if(billing.isTrial()) {
                    return daysLeft <= (2 * DAYS_PER_WEEK);
                } else if(!billing.isAutomaticCollection()) {
                    return daysLeft <= (8 * DAYS_PER_WEEK);
                } else {
                    return false;
                }
            })
            .orElse(false);
    }

    /**
     * True if the user should be nudged about expected payment
     */
    public boolean shouldNudgeForPayment(Date now) {
        if (!isPaymentExpected()) {
            return false;
        }
        if(getDatabaseCount() == 0 || isTrial()) {
            return false;
        }
        return daysUntilExpectedPayment(now) <= (2 * DAYS_PER_WEEK);
    }

    /**
     * @return the date after which the user should be again warned
     */
    public LocalDate snoozeDate(Date now) {
        int daysLeft = daysUntilExpiration(now);
        LocalDate today = new LocalDate(now);

        if(isTrial()) {
            if(daysLeft < 7) {
                return today.plusDays(1);
            } else {
                return today.plusDays(4);
            }
        } else {
            LocalDate expirationDay = new LocalDate(new Date(getExpirationTime() * 1000L));
            int weeksLeft = Math.floorDiv(daysLeft, 7);
            if (weeksLeft <= 1) {
                return today.plusDays(1);
            } else if (weeksLeft <= 2) {
                return expirationDay.plusDays(-1 * DAYS_PER_WEEK);
            } else if (weeksLeft <= 3) {
                return expirationDay.plusDays(-2 * DAYS_PER_WEEK);
            } else if (weeksLeft <= 4) {
                return expirationDay.plusDays(-3 * DAYS_PER_WEEK);
            } else {
                return expirationDay.plusDays(-4 * DAYS_PER_WEEK);
            }
        }
    }

    /**
     * @return the date after which the user should be nudged again for payment
     */
    public LocalDate paymentSnoozeDate(Date now) {
        int daysLeft = daysUntilExpectedPayment(now);
        LocalDate today = new LocalDate(now);

        LocalDate expectedPaymentDay = new LocalDate(new Date(getExpectedPaymentDate() * 1000L));
        if (daysLeft <= 3) {
            // Show each day for last three days prior to payment
            return today.plusDays(1);
        } else if (daysLeft <= 7) {
            // Show again on third day before payment
            return expectedPaymentDay.plusDays(-3);
        } else {
            // Show again next week
            return expectedPaymentDay.plusDays(-1 * DAYS_PER_WEEK);
        }
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = Json.createObject();

        object.put("user", UserAccountJson.toJson(userAccount));

        billingAccount.ifPresent(billing -> {
            object.put("billingAccount", BillingAccountJson.toJson(billing));
        });

        // Expected by older versions of the application
        object.put("userAccountId", userAccount.getId().asString());
        object.put("userAccountEmail", userAccount.getEmail());
        object.put("userAccountName", userAccount.getName());
        object.put("userLocale", userAccount.getLocale());
        return object;
    }

    public static AccountStatus fromJson(JsonValue object) {

        UserAccount user;
        if(object.hasKey("user")) {
            user = UserAccountJson.fromJson(object.get("user"));
        } else {
            user = new UserAccountBuilder()
                .id(ResourceId.valueOf(object.getString("userAccountId")))
                .name(object.getString("userAccountName"))
                .email(object.getString("userAccountEmail"))
                .build();
        }
        Optional<BillingAccount> billingAccount;
        if(object.hasKey("billingAccount")) {
            billingAccount = Optional.of(BillingAccountJson.fromJson(object.get("billingAccount")));
        } else {
            billingAccount = Optional.empty();
        }
        return new AccountStatus(user, billingAccount);
    }

    public boolean isExpired() {
        return secondsUntilExpiration(new Date()) < 0;
    }

    public boolean isSuspended() {
        return billingAccount.map(b -> b.getStatus() == BillingAccountStatus.SUSPENDED).orElse(false);
    }

    public boolean isPaidAccount() {
        return billingAccount.map(b -> b.getStatus() == BillingAccountStatus.SUSPENDED).orElse(false);
    }

    public boolean isAddDatabaseAllowed() {
        return billingAccount
            .map(b -> b.getStatus() != BillingAccountStatus.SUSPENDED)
            .orElse(false);
    }

    public boolean isPaymentExpected() {
        return billingAccount.map(b -> b.isPaymentExpected()).orElse(false);
    }

    public boolean isPaymentOverdue() {
        return billingAccount.map(b -> {
            if(b.isPaymentExpected()) {
                return secondsUntilExpectedPayment(new Date()) < 0;
            } else {
                return false;
            }
        }).orElse(false);
    }

    /**
     *
     * True if the {@code userCount} field is being updated.
     */
    public boolean isStaleCounts() {
        return billingAccount.map(b -> b.isStaleCounts()).orElse(false);
    }

    public boolean isFeatureEnabled(String flag) {
        return false;
    }

    public boolean hasBillingAccount() {
        return billingAccount.isPresent();
    }

}
