/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.account;

import java.util.Optional;

/*
 * <p>Validates the format of an email address.</p>
 *
 * <p>Note that this validator is more restrictive that the formal specifications for email address format.
 * Specifically:</p>
 *
 * <ul>
 * <li>Addresses with quoted portions are rejected as invalid.</li>
 * <li>Addresses that use literal IP address domains are rejected.</li>
 * <li>Addresses that use non-ascii characters are rejected.</li>
 * </ul>
 *
 * <p>Leading and following whitespace is trimmed.</p>
 */
public class EmailAddressFormValidator {

    private static final int MAXIMUM_LOCAL_PART_LENGTH_OCTETS = 64;

    /**
     * Tries to validate a string as an email address.
     *
     * <p>The result is at least well-formed; it may still not be a valid domain or working
     * email address</p>
     */
    public static Optional<EmailAddress> validate(String text) {
        if(text == null) {
            return Optional.empty();
        }
        text = text.trim();
        if(text.isEmpty()) {
            return Optional.empty();
        }

        int domainSeparator = text.lastIndexOf('@');
        if(domainSeparator == -1) {
            return Optional.empty();
        }

        String localPart = text.substring(0, domainSeparator);
        String domain = text.substring(domainSeparator + 1);

        if(!isValidLocalPart(localPart) || !isValidDomain(domain)) {
            return Optional.empty();
        }

        return Optional.of(new EmailAddress(localPart, domain));
    }


    private static boolean isValidLocalPart(String localPart) {

        if (localPart.isEmpty() || localPart.length() > MAXIMUM_LOCAL_PART_LENGTH_OCTETS) {
            return false;
        }

        // If unquoted, it may use any of these ASCII characters:
        // uppercase and lowercase Latin letters A to Z and a to z;
        // digits 0 to 9;
        // printable characters !#$%&'*+-/=?^_`{|}~;
        // dot ., provided that it is not the first or last character and provided also that it does
        // not appear consecutively (e.g., John..Doe@example.com is not allowed.

        // Email address can technically include quotes in the local part, which
        // can then include spaces and other characters.

        char lastChar = 0;
        for(int i = 0; i < localPart.length(); ++i) {
            char c = localPart.charAt(i);
            if(c == '.') {
                // Cannot start or end with a dot
                if (i == 0 || i + 1 == localPart.length()) {
                    return false;
                }
                // Two consecutive dots is not allowed
                if(lastChar == '.') {
                    return false;
                }
            }
            if(!isValidLocalPartChar(c)) {
                return false;
            }
            lastChar = c;
        }

        return true;
    }


    private static boolean isValidDomain(String domain) {

        if (domain.isEmpty()) {
            return false;
        }

        // The domain name part of an email address has to conform to strict guidelines:
        // it must match the requirements for a hostname, a list of dot-separated DNS labels.

        String[] labels = domain.split("\\.");
        for (int i = 0; i < labels.length; i++) {
            if(!isValidDnsLabel(labels[i])) {
                return false;
            }
        }

        return true;
    }

    private static boolean isValidDnsLabel(String label) {

        // each label being limited to a length of 63 characters and consisting of:

        // * uppercase and lowercase Latin letters A to Z and a to z;
        // * digits 0 to 9, provided that top-level domain names are not all-numeric;
        // * hyphen -, provided that it is not the first or last character.

        if(label.isEmpty() || label.length() > 63) {
            return false;
        }
        if(label.startsWith("-") || label.endsWith("-")) {
            return false;
        }

        for(int i = 0; i < label.length(); ++i) {
            if(!isValidDnsLabelChar(label.charAt(i))) {
                return false;
            }
        }
        return true;
    }

    private static boolean isValidDnsLabelChar(char c) {
        return isAlphaNumeric(c) || c == '-';
    }

    private static boolean isValidLocalPartChar(char c) {
        if (isAlphaNumeric(c)) {
            return true;
        }
        //  !#$%&'*+-/=?^_`{|}~;
        switch (c) {
            case '.':
            case '!':
            case '#':
            case '$':
            case '%':
            case '&':
            case '\'':
            case '*':
            case '+':
            case '-':
            case '/':
            case '=':
            case '?':
            case '^':
            case '_':
            case '`':
            case '{':
            case '|':
            case '}':
            case '~':
                return true;
            default:
                return false;
        }
    }


    private static boolean isAlphaNumeric(char c) {
        if(c >= 'A' && c <= 'Z') {
            return true;
        }
        if(c >= 'a' && c <= 'z') {
            return true;
        }
        if(c >= '0' && c <= '9') {
            return true;
        }
        return false;
    }

}
