/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.account;

import com.google.common.base.Strings;
import org.activityinfo.i18n.shared.ApplicationLocale;
import org.activityinfo.json.AutoJson;
import org.activityinfo.json.GwtIncompatible;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.resource.ResourceId;

import java.util.Locale;
import java.util.Optional;

@AutoJson
@AutoBuilder
public class UserAccount {
    private final ResourceId id;
    private final String email;
    private final String name;
    private final String locale;
    private final Optional<BillingRoleAssignment> billingRole;
    private final PlatformRole platformRole;
    private final ActivationStatus activationStatus;
    private final DeliveryStatus deliveryStatus;
    private final String idp;

    @Builder
    public UserAccount(ResourceId id, String email, String name, String locale, Optional<BillingRoleAssignment> billingRole, PlatformRole platformRole, ActivationStatus activationStatus, DeliveryStatus deliveryStatus, String idp) {
        this.id = id;
        this.email = email;
        this.name = name;
        this.locale = locale;
        this.platformRole = platformRole;
        this.billingRole = billingRole;
        this.activationStatus = activationStatus;
        this.idp = Strings.nullToEmpty(idp);
        this.deliveryStatus = deliveryStatus;
    }

    public UserAccount(ResourceId id, String email, String name) {
        this(id, email, name, "en", Optional.empty(), PlatformRole.NONE, ActivationStatus.ACTIVE, DeliveryStatus.UNKNOWN, "");
    }

    public UserAccount(ResourceId id, String email, String name, ApplicationLocale locale) {
        this(id, email, name, locale.getCode(), Optional.empty(), PlatformRole.NONE, ActivationStatus.ACTIVE, DeliveryStatus.UNKNOWN, "");
    }

    public ResourceId getId() {
        return id;
    }

    public String getEmail() {
        return email;
    }

    public EmailAddress getEmailAddress() {
        return EmailAddressFormValidator.validate(email).orElseThrow(() -> new IllegalStateException(email));
    }

    public String getName() {
        return name;
    }

    public String getLocale() {
        return locale;
    }

    public ApplicationLocale getApplicationLocale() {
        return ApplicationLocale.validateLocale(locale).orElse(ApplicationLocale.EN);
    }

    public ActivationStatus getActivationStatus() {
        return activationStatus;
    }

    public DeliveryStatus getDeliveryStatus() {
        return deliveryStatus;
    }

    public Optional<BillingRoleAssignment> getBillingRole() {
        return billingRole;
    }

    public PlatformRole getPlatformRole() {
        if(platformRole == null) {
            return PlatformRole.NONE;
        }
        return platformRole;
    }

    /**
     * @return this user's Identity Provider.
     */
    public String getIdp() {
        return idp;
    }

    @GwtIncompatible
    public Locale getLocaleObject() {
        return Locale.forLanguageTag(locale);
    }

}
