/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.analysis;

import org.activityinfo.json.AutoJson;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.resource.ResourceId;

import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Consumer;

/**
 * The user's table model
 */
@AutoJson
@AutoBuilder
public class TableAnalysisModel implements AnalysisModel  {

    public static final String TYPE = "table";

    /**
     * The id of the form from which the table should be created.
     */
    private final ResourceId formId;

    /**
     * A list of columns to include in the table. If the list is empty, a list of
     * columns will be computed from the form's fields.
     */
    private final List<TableColumn> columns;

    /**
     * List of columns to use in ordering the form's records.
     */
    private final List<SortOrder> ordering;

    /**
     * A boolean-valued formula which filters which records are included.
     */
    private final Optional<String> filter;

    public TableAnalysisModel(ResourceId formId, List<TableColumn> columns, List<SortOrder> ordering, Optional<String> filter) {
        this.formId = formId;
        this.columns = columns;
        this.ordering = ordering;
        this.filter = filter;
    }

    public ResourceId getFormId() {
        return formId;
    }

    public List<TableColumn> getColumns() {
        return columns;
    }

    public List<SortOrder> getOrdering() {
        return ordering;
    }

    /**
     * A boolean-valued formula that determines which rows to include in the
     * results.
     */
    public Optional<String> getFilter() {
        return filter;
    }

    @Override
    public String getTypeId() {
        return TYPE;
    }

    @Override
    public boolean isEmpty() {
        return false;
    }

    public TableAnalysisModel withColumns(List<TableColumn> columns) {
        return new TableAnalysisModelBuilder(this)
                .columns(columns)
                .build();
    }

    public TableAnalysisModel mutate(Consumer<TableAnalysisModelBuilder> mutator) {
        TableAnalysisModelBuilder builder = new TableAnalysisModelBuilder(this);
        mutator.accept(builder);
        return builder.build();
    }

    @Override
    public JsonValue toJson() {
        return TableAnalysisModelJson.toJson(this);
    }

    @Override
    public <T> T accept(AnalysisModelVisitor<T> visitor) {
        return visitor.visitTable(this);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof TableAnalysisModel)) {
            return false;
        }
        TableAnalysisModel that = (TableAnalysisModel) o;
        return formId.equals(that.formId) &&
                columns.equals(that.columns) &&
                ordering.equals(that.ordering) &&
                filter.equals(that.filter);
    }

    @Override
    public int hashCode() {
        return Objects.hash(formId, columns, ordering, filter);
    }
}
