/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.analysis.map;

import org.activityinfo.json.AutoJson;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.analysis.AnalysisModel;
import org.activityinfo.model.analysis.AnalysisModelVisitor;
import org.activityinfo.model.analysis.pivot.ImmutableLists;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.geo.GeoPoint;

import java.util.List;
import java.util.Set;

@AutoJson
@AutoBuilder
public class MapModel implements AnalysisModel {

    public static final String TYPE = "map";

    private final String baseMap;
    private final List<Layer> layers;

    /**
     * Zoom level 0-22, and -1 means that the map should be fitted to the data.
     */
    private final double zoomLevel;

    private final GeoPoint center;


    public MapModel(String baseMap, List<Layer> layers, double zoomLevel, GeoPoint center) {
        this.baseMap = baseMap;
        this.layers = layers;
        this.zoomLevel = zoomLevel;
        this.center = center;
    }

    public MapModel withLayer(Layer layer) {
        return new MapModelBuilder(this)
                .layers(ImmutableLists.update(getLayers(), layer, l -> l.getId()))
                .build();
    }

    public MapModel withoutLayer(String layerId) {
        return new MapModelBuilder(this)
                .layers(ImmutableLists.remove(getLayers(), layerId, l -> l.getId()))
                .build();
    }

    public MapModel withBaseMap(String id) {
        return new MapModelBuilder(this)
                .baseMap(id)
                .build();
    }

    public MapModel withZoom(double updatedZoomLevel) {
        return new MapModel(baseMap, layers, updatedZoomLevel, center);
    }

    public MapModel withCenter(GeoPoint updatedCenter) {
        return new MapModel(baseMap, layers, zoomLevel, updatedCenter);
    }

    @Override
    public String getTypeId() {
        return TYPE;
    }

    @Override
    public JsonValue toJson() {
        return MapModelJson.toJson(this);
    }

    @Override
    public boolean isEmpty() {
        return layers.isEmpty();
    }

    @Override
    public <T> T accept(AnalysisModelVisitor<T> visitor) {
        return visitor.visitMap(this);
    }

    public String getBaseMap() {
        return baseMap;
    }

    public List<Layer> getLayers() {
        return layers;
    }

    public double getZoomLevel() {
        return zoomLevel;
    }

    public GeoPoint getCenter() {
        return center;
    }

}
