/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.analysis.pivot;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.json.AutoJson;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;

/**
 * A Dimension has a number of discrete categories into which measures are disaggregated.
 *
 * <p>Importantly, a dimension must break up quantitative data from multiple data sources into common
 * categories. </p>
 */
@AutoJson
@AutoBuilder
public class DimensionModel {

    public static final String STATISTIC_ID = "statistic";

    public static final String MEASURE_ID = "measure";

    private final String id;
    private final String label;
    private final List<DimensionMapping> mappings;
    private final Axis axis;
    private final boolean totals;
    private final boolean missingIncluded;
    private final boolean percentage;
    private final Optional<DateLevel> dateLevel;
    private final Optional<String> totalLabel;
    private final Optional<String> missingLabel;
    private final List<CategoryMapping> categoryMappings;
    private final Optional<ResourceId> mapTemplateId;

    public DimensionModel(String id, String label,
                          List<DimensionMapping> mappings,
                          Axis axis,
                          boolean totals,
                          boolean missingIncluded,
                          boolean percentage,
                          Optional<DateLevel> dateLevel,
                          Optional<String> totalLabel,
                          Optional<String> missingLabel,
                          List<CategoryMapping> categoryMappings,
                          Optional<ResourceId> mapTemplateId) {
        this.id = id;
        this.label = label;
        this.mappings = mappings;
        this.axis = axis;
        this.totals = totals;
        this.missingIncluded = missingIncluded;
        this.percentage = percentage;
        this.dateLevel = dateLevel;
        this.totalLabel = totalLabel;
        this.missingLabel = missingLabel;
        this.categoryMappings = categoryMappings;
        this.mapTemplateId = mapTemplateId;
    }

    public static DimensionModelBuilder builder() {
        return new DimensionModelBuilder();
    }

    public String getId() {
        return id;
    }

    public String getLabel() {
        return label;
    }

    public List<DimensionMapping> getMappings() {
        return mappings;
    }

    public Optional<ResourceId> getMapTemplateId() {
        return mapTemplateId;
    }

    public Axis getAxis() {
        return axis;
    }

    public boolean isTotals() {
        return totals;
    }

    public boolean isMissingIncluded() {
        return missingIncluded;
    }

    public Optional<DateLevel> getDateLevel() {
        return dateLevel;
    }

    public Optional<String> getTotalLabel() {
        return totalLabel;
    }

    public Optional<String> getMissingLabel() {
        return missingLabel;
    }

    public String getEffectiveMissingLabel() {
        return missingLabel.orElse(I18N.CONSTANTS.emptyDimensionCategory());
    }

    public List<CategoryMapping> getCategoryMappings() {
        return categoryMappings;
    }

    public boolean isPercentage() {
        return percentage;
    }

    public DimensionModelBuilder with() {
        return new DimensionModelBuilder(this);
    }

    public DimensionModel withCategoriesHidden(Set<String> categories) {
        List<CategoryMapping> newMappings = new ArrayList<>();
        Set<String> toHide = new HashSet<>(categories);

        for (CategoryMapping mapping : categoryMappings) {
            if(categories.contains(mapping.getLabel())) {
                newMappings.add(new CategoryMapping(mapping.getLabel(), mapping.getSource(), false));
                toHide.remove(mapping.getLabel());
            } else {
                newMappings.add(mapping);
            }
        }

        for (String category : toHide) {
            newMappings.add(new CategoryMapping(category, false));
        }

        return with().categoryMappings(newMappings).build();
    }

    public DimensionModel withCategoryVisible(String category, boolean visible) {

        List<CategoryMapping> newMappings = new ArrayList<>();

        boolean hasMapping = false;
        for (CategoryMapping mapping : categoryMappings) {
            if(mapping.getLabel().equals(category)) {
                newMappings.add(new CategoryMapping(mapping.getLabel(), mapping.getSource(), visible));
                hasMapping = true;
            } else {
                newMappings.add(mapping);
            }
        }

        if(!hasMapping) {
            newMappings.add(new CategoryMapping(category, Collections.singleton(category), visible));
        }

        return with().categoryMappings(newMappings).build();
    }


    /**
     * Checks whether a (mapped) category is visible.
     */
    public boolean isCategoryVisible(String mappedCategory) {
        for (CategoryMapping categoryMapping : categoryMappings) {
            if(!categoryMapping.isVisible() && categoryMapping.getLabel().equals(mappedCategory)) {
                return false;
            }
        }
        return true;
    }

    public boolean hasAnyHiddenCategories() {
        for (CategoryMapping categoryMapping : categoryMappings) {
            if(!categoryMapping.isVisible()) {
                return true;
            }
        }
        return false;
    }

    @Override
    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof DimensionModel)) {
            return false;
        }
        DimensionModel that = (DimensionModel) o;
        return totals == that.totals &&
            missingIncluded == that.missingIncluded &&
            percentage == that.percentage &&
            id.equals(that.id) &&
            label.equals(that.label) &&
            mappings.equals(that.mappings) &&
            axis == that.axis &&
            dateLevel.equals(that.dateLevel) &&
            totalLabel.equals(that.totalLabel) &&
            missingLabel.equals(that.missingLabel) &&
            categoryMappings.equals(that.categoryMappings);
    }

    @Override
    public int hashCode() {
        return Objects.hash(id, label);
    }

    @Override
    public String toString() {
        return "DimensionModel{" +
                "id='" + id + '\'' +
                ", label='" + label + '\'' +
                ", mappings=" + mappings +
                ", axis=" + axis +
                ", totals=" + totals +
                ", missingIncluded=" + missingIncluded +
                ", percentage=" + percentage +
                ", dateLevel=" + dateLevel +
                ", totalLabel=" + totalLabel +
                ", missingLabel=" + missingLabel +
                '}';
    }

}
