/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.analysis.pivot;

import org.activityinfo.json.AutoJson;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.analysis.AnalysisModel;
import org.activityinfo.model.analysis.AnalysisModelVisitor;
import org.activityinfo.model.annotation.AutoBuilder;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Defines a pivot model from one or more data sources.
 */
@AutoJson
@AutoBuilder
public class PivotModel implements AnalysisModel {

    public static final String TYPE = "pivot";

    private final VisualizationType visualization;

    private final List<MeasureModel> measures;
    private final List<DimensionModel> dimensions;

    PivotModel(VisualizationType visualization,
               List<MeasureModel> measures,
               List<DimensionModel> dimensions) {
        this.visualization = visualization;
        this.measures = measures;
        this.dimensions = dimensions;
    }

    public VisualizationType getVisualization() {
        return visualization;
    }

    public List<MeasureModel> getMeasures() {
        return measures;
    }

    public List<DimensionModel> getDimensions() {
        return dimensions;
    }

    public boolean hasAnyHiddenCategories() {
        return dimensions.stream().anyMatch(d -> d.hasAnyHiddenCategories());
    }

    /**
     * Updates the model to update the given measure with the same id.
     */
    public PivotModel withMeasure(MeasureModel measureModel) {
        return new PivotModelBuilder(this)
                .measures(ImmutableLists.update(getMeasures(), measureModel, m -> m.getId()))
                .build();
    }

    public PivotModel withMeasure(MeasureModel measureModel, int insertIndex) {

        List<MeasureModel> newList = new ArrayList<>();
        int currentIndex = 0;
        for (MeasureModel existingMeasure : getMeasures()) {
            if(currentIndex == insertIndex) {
                newList.add(measureModel);
                currentIndex++;
            }
            if(!existingMeasure.getId().equals(measureModel.getId())) {
                newList.add(existingMeasure);
                currentIndex++;
            }
        }

        if(insertIndex >= currentIndex) {
            newList.add(measureModel);
        }

        return new PivotModelBuilder(this)
                .measures(newList)
                .build();
    }

    public PivotModel withDimension(DimensionModel dimensionModel) {
        return new PivotModelBuilder(this)
                .dimensions(ImmutableLists.update(getDimensions(), dimensionModel, m -> m.getId()))
                .build();
    }

    public PivotModel withDimension(DimensionModel dimensionModel, int insertIndex) {

        List<DimensionModel> newList = new ArrayList<>();
        int currentIndex = 0;
        for (DimensionModel existingDimension : getDimensions()) {
            if(currentIndex == insertIndex) {
                newList.add(dimensionModel);
                currentIndex++;
            }
            if(!existingDimension.getId().equals(dimensionModel.getId())) {
                newList.add(existingDimension);
                if(existingDimension.getAxis() == dimensionModel.getAxis()) {
                    currentIndex++;
                }
            }
        }

        if(insertIndex >= currentIndex) {
            newList.add(dimensionModel);
        }

        return new PivotModelBuilder(this)
                .dimensions(newList)
                .build();
    }

    public PivotModel withoutTotals(boolean removeTotals) {
        if(removeTotals) {
            List<DimensionModel> newList = new ArrayList<>();
            for (DimensionModel existingDimension : getDimensions()) {
                if (existingDimension.isTotals()) {
                    newList.add(existingDimension.with().totals(false).build());
                }else{
                    newList.add(existingDimension);
                }
            }
            return new PivotModelBuilder(this)
                    .dimensions(newList)
                    .build();
        }else{
            return this;
        }
    }

    public PivotModel withoutMeasure(String measureId) {
        return new PivotModelBuilder(this)
                .measures(ImmutableLists.remove(getMeasures(), measureId, m -> m.getId()))
                .build();
    }

    public PivotModel withoutDimension(String dimensionId) {
        return new PivotModelBuilder(this)
                .dimensions(ImmutableLists.remove(getDimensions(), dimensionId, d -> d.getId()))
                .build();
    }

    public static PivotModelBuilder builder() {
        return new PivotModelBuilder();
    }

    /**
     *
     * @return true if any of the measures defined have multiple statistics.
     */
    public boolean isMeasureDefinedWithMultipleStatistics() {
        for (MeasureModel measure : getMeasures()) {
            if(measure.getStatistics().size() > 1) {
                return true;
            }
        }
        for (DimensionModel dimensionModel : getDimensions()) {
            if(dimensionModel.isPercentage()) {
                return true;
            }
        }
        return false;
    }

    public boolean isChart(){
        return visualization == VisualizationType.BARCHART
                || visualization == VisualizationType.LINECHART
                || visualization == VisualizationType.PIECHART;
    }

    @Override
    public String getTypeId() {
        return TYPE;
    }

    @Override
    public JsonValue toJson() {
        return PivotModelJson.toJson(this);
    }

    @Override
    public boolean isEmpty() {
        return getMeasures().isEmpty() && getDimensions().isEmpty();
    }

    @Override
    public <T> T accept(AnalysisModelVisitor<T> visitor) {
        return visitor.visitPivotModel(this);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof PivotModel)) {
            return false;
        }
        PivotModel that = (PivotModel) o;
        return visualization == that.visualization &&
                measures.equals(that.measures) &&
                dimensions.equals(that.dimensions);
    }

    @Override
    public int hashCode() {
        return Objects.hash(visualization, measures, dimensions);
    }

    @Override
    public String toString() {
        return "PivotModel{" +
                "visualization=" + visualization +
                ", measures=" + measures +
                ", dimensions=" + dimensions +
                '}';
    }

}
