/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.audit;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.model.database.*;
import org.activityinfo.model.resource.ResourceId;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;

public class DatabaseTreeAuditEventFactory {

    public static final String ID_PREFIX = "dbtree:";

    public static final Iterable<AuditResourceType> RESOURCE_TYPES = Arrays.asList(
        AuditResourceType.DATABASE,
        AuditResourceType.FOLDER,
        AuditResourceType.LOCK,
        AuditResourceType.FORM,
        AuditResourceType.ROLE);


    public static String composeDescription(DatabaseTree previousTree, UpdateDatabaseDelta delta) {
        List<String> changes = new ArrayList<>();

        appendResourceChanges(previousTree, delta.getResources(), changes);
        appendRoleChanges(delta.getRoles(), changes);
        appendLocks(previousTree, delta.getLocks(), changes);

        if(changes.isEmpty()) {
            return "Updated database";
        } else {
            return changes.stream().collect(Collectors.joining(", "));
        }
    }

    private static void appendResourceChanges(DatabaseTree previousTree, ListDelta<Resource> delta, List<String> changes) {
        for (Resource resource : delta.getAdded().values()) {
            switch (resource.getType()) {
                case FORM:
                case FOLDER:
                    changes.add(I18N.MESSAGES.addedResource(resourceType(resource), resource.getLabel()));
                    break;
                case SUB_FORM:
                    String parentLabel = previousTree.getResource(resource.getParentId())
                        .map(r -> r.getLabel())
                        .orElse("");
                    changes.add(I18N.MESSAGES.addedSubform(resource.getLabel(), parentLabel));
                    break;
            }
        }
        for (ListDelta.Update<Resource> update : delta.getUpdated().values()) {
            describeRenaming(update, Resource::getLabel, DatabaseTreeAuditEventFactory::resourceType, changes);
            describeVisibilityUpdate(changes, update);
            describeMoving(previousTree, update, changes);
        }

        for (Resource resource : delta.getDeleted().values()) {
            changes.add(I18N.MESSAGES.deletedResource(
                resourceType(resource),
                resource.getLabel()));
        }
    }

    private static void appendLocks(DatabaseTree previousTree, ListDelta<RecordLock> delta, List<String> changes) {
        for (RecordLock lock : delta.getAdded().values()) {
            changes.add(I18N.MESSAGES.addedResource(I18N.CONSTANTS.resourceTypeLock(), lock.getLabel()));
        }

        for (ListDelta.Update<RecordLock> update : delta.getUpdated().values()) {
            describeRenaming(update, RecordLock::getLabel, lock -> I18N.CONSTANTS.resourceTypeLock(), changes);
            describeLockChange(previousTree, update, changes);
        }

        for (RecordLock lock : delta.getDeleted().values()) {
            changes.add(I18N.MESSAGES.deletedResource(I18N.CONSTANTS.resourceTypeLock(), lock.getLabel()));
        }
    }

    private static void appendRoleChanges(ListDelta<Role> delta, List<String> changes) {
        for (Role role : delta.getAdded().values()) {
            changes.add(I18N.MESSAGES.addedResource(I18N.CONSTANTS.resourceTypeRole(), role.getLabel()));
        }

        for (Role role : delta.getDeleted().values()) {
            changes.add(I18N.MESSAGES.deletedResource(I18N.CONSTANTS.resourceTypeRole(), role.getLabel()));
        }

        for (ListDelta.Update<Role> update : delta.getUpdated().values()) {
            if(!describeRenaming(update, Role::getLabel, r -> I18N.CONSTANTS.resourceTypeRole(), changes)) {
                changes.add(I18N.MESSAGES.updatedRole(update.getOld().getLabel()));
            }
        }
    }

    private static void describeMoving(DatabaseTree tree, ListDelta.Update<Resource> update, List<String> changes) {
        if(!update.getOld().getParentId().equals(update.getUpdated().getParentId())) {
            Resource resource = update.getOld();
            String oldFolder = folderLabel(tree, update.getOld());
            String updatedFolder =  folderLabel(tree, update.getUpdated());
            changes.add(I18N.MESSAGES.movedResource(
                resourceType(resource),
                resource.getLabel(),
                oldFolder,
                updatedFolder));
        }
    }

    private static String folderLabel(DatabaseTree tree, Resource resource) {
        if(resource.getParentId().equals(tree.getDatabaseId())) {
            return tree.getLabel();
        } else {
            return tree.getResource(resource.getParentId()).map(Resource::getLabel).orElse("(unknown)");
        }
    }

    private static void describeLockChange(DatabaseTree previousTree, ListDelta.Update<RecordLock> update, List<String> changes) {
        if (update.getOld().getDateRange().equals(update.getUpdated().getDateRange())) {
            ResourceId resourceId = update.getOld().getResourceId();
            changes.add(previousTree.getResource(resourceId)
                .map(resource -> I18N.MESSAGES.updatedLockDateRangeOnResource(update.getOld().getLabel(),
                    resourceType(resource),
                    resource.getLabel()))
                .orElseGet(() ->  I18N.MESSAGES.updatedLockDateRange(update.getOld().getLabel())));
        }
    }

    private static void describeVisibilityUpdate(List<String> changes, ListDelta.Update<Resource> update) {
        if(!update.getOld().getVisibility().equals(update.getUpdated().getVisibility())) {
            changes.add(I18N.MESSAGES.updatedVisibility(resourceType(update.getOld()),
                update.getOld().getLabel(),
                visibility(update.getOld()),
                visibility(update.getUpdated())));
        }
    }

    private static  <T> boolean describeRenaming(ListDelta.Update<T> update, Function<T, String> getLabel, Function<T, String> getType, List<String> changes) {
        String oldLabel = getLabel.apply(update.getOld());
        String updatedLabel = getLabel.apply(update.getUpdated());
        if(!oldLabel.equals(updatedLabel)) {
            changes.add(I18N.MESSAGES.renamedResource(getType.apply(update.getOld()),
                oldLabel,
                updatedLabel));
            return true;
        } else {
            return false;
        }
    }

    private static String visibility(Resource resource) {
        switch (resource.getVisibility()) {
            case PRIVATE:
                return I18N.CONSTANTS.resourceVisibilityPrivate();
            case PUBLIC:
                return I18N.CONSTANTS.resourceVisibilityPublic();
            case REFERENCE:
                return I18N.CONSTANTS.resourceVisibilityReference();
            default:
                return resource.getVisibility().name();
        }
    }

    private static String resourceType(Resource resource) {
        switch (resource.getType()) {
            case FORM:
                return I18N.CONSTANTS.resourceTypeForm();
            case FOLDER:
                return I18N.CONSTANTS.resourceTypeFolder();
            case SUB_FORM:
                return I18N.CONSTANTS.resourceTypeSubform();
            default:
                return resource.getType().name();
        }
    }
}
