/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.form.SubFormKind;
import org.activityinfo.model.form.TypedFormRecord;
import org.activityinfo.model.formula.*;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.*;
import org.activityinfo.model.type.attachment.AttachmentType;
import org.activityinfo.model.type.barcode.BarcodeType;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.expr.CalculatedFieldType;
import org.activityinfo.model.type.geo.GeoAreaType;
import org.activityinfo.model.type.geo.GeoPointType;
import org.activityinfo.model.type.number.QuantityType;
import org.activityinfo.model.type.primitive.TextType;
import org.activityinfo.model.type.subform.SubFormReferenceType;
import org.activityinfo.model.type.time.*;

import java.util.Collection;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Duplicates a database and or forms.
 *
 * <p>This class can be used to create a new copy of a database or form, with any globally unique ids
 * mapped to new ids, depending on the context.</p>
 */
public class DatabaseDuplicator {
    private final Function<ResourceId, ResourceId> ids;

    private boolean preserveReferenceFields = false;

    /**
     *
     * @param ids a function which maps resource ids to new resource ids.
     */
    public DatabaseDuplicator(Function<ResourceId, ResourceId> ids) {
        this.ids = ids;
    }

    private ResourceId newId(ResourceId id) {

        if(VisibleDatabaseTree.isUserForm(id)) {
            ResourceId databaseId = VisibleDatabaseTree.getDatabaseIdFromUserFormId(id);
            ResourceId newDatabaseId = newId(databaseId);
            return VisibleDatabaseTree.getUserFormIdFromDatabaseId(newDatabaseId);
        }

        return ids.apply(id);
    }


    private RecordRef newId(RecordRef ref) {
        return new RecordRef(ids.apply(ref.getFormId()), ref.getRecordId());
    }

    /**
     * Do not change reference fields.
     */
    public void preserveReferenceFields(boolean preserve) {
        this.preserveReferenceFields = preserve;

    }

    public DatabaseTree duplicate(VisibleDatabaseTree tree) {
        DatabaseTreeBuilder newTree = new DatabaseTreeBuilder();
        newTree.databaseId(newId(tree.getDatabaseId()));
        newTree.label(tree.getLabel());
        newTree.ownerId(tree.getOwnerId());
        newTree.version(1L);

        for (Resource resource : tree.getResources()) {
            newTree.addResource(duplicate(resource));
        }

        return newTree.build();
    }

    public Resource duplicate(Resource resource) {
        return new Resource.Builder(resource)
            .setId(newId(resource.getId()))
            .setParentId(newId(resource.getParentId()))
            .build();
    }

    public FormClass duplicate(FormClass old) {

        if(VisibleDatabaseTree.isUserForm(old.getId())) {
            throw new IllegalArgumentException("Cannot duplicate user form");
        }

        FormClass newForm = new FormClass(newId(old.getId()));
        newForm.setSchemaVersion(1L);
        newForm.setLabel(old.getLabel());
        newForm.setDatabaseId(newId(old.getDatabaseId()));
        if(old.isSubForm()) {
            newForm.setSubFormKind(SubFormKind.REPEATING);
            newForm.setParentFormId(newId(old.getParentFormId().get()));
        }
        newForm.setDescription(old.getDescription());

        for (FormField field : old.getFields()) {
            newForm.addElement(duplicate(field));
        }
        return newForm;
    }

    private FormField duplicate(FormField field) {
        // Form field ids do not need to be changed
        FormField newField = new FormField(field);
        newField.setType(duplicateType(field.getType()));
        return newField;
    }

    private FieldType duplicateType(FieldType type) {
        return type.accept(new FieldTypeVisitor<FieldType>() {
            @Override
            public FieldType visitAttachment(AttachmentType attachmentType) {
                return attachmentType;
            }

            @Override
            public FieldType visitCalculated(CalculatedFieldType calculatedFieldType) {
                return calculatedFieldType;
            }

            @Override
            public FieldType visitReference(ReferenceType referenceType) {
                if(preserveReferenceFields) {
                    return referenceType;
                } else {
                    return new ReferenceType(
                        referenceType.getCardinality(),
                        referenceType.getRange()
                            .stream()
                            .map(formId -> newId(formId))
                            .collect(Collectors.toSet()));
                }
            }

            @Override
            public FieldType visitNarrative(NarrativeType type) {
                return type;
            }

            @Override
            public FieldType visitQuantity(QuantityType type) {
                return type;
            }

            @Override
            public FieldType visitGeoPoint(GeoPointType type) {
                return type;
            }

            @Override
            public FieldType visitGeoArea(GeoAreaType type) {
                return type;
            }

            @Override
            public FieldType visitEnum(EnumType type) {
                return type;
            }

            @Override
            public FieldType visitBarcode(BarcodeType type) {
                return type;
            }

            @Override
            public FieldType visitSubForm(SubFormReferenceType type) {
                return new SubFormReferenceType(newId(type.getClassId()));
            }

            @Override
            public FieldType visitLocalDate(LocalDateType type) {
                return type;
            }

            @Override
            public FieldType visitMonth(MonthType monthType) {
                return monthType;
            }

            @Override
            public FieldType visitYear(YearType yearType) {
                return yearType;
            }

            @Override
            public FieldType visitFortnight(FortnightType fortnightType) {
                return fortnightType;
            }

            @Override
            public FieldType visitWeek(EpiWeekType epiWeekType) {
                return epiWeekType;
            }

            @Override
            public FieldType visitLocalDateInterval(LocalDateIntervalType localDateIntervalType) {
                return localDateIntervalType;
            }

            @Override
            public FieldType visitText(TextType textType) {
                return textType;
            }

            @Override
            public FieldType visitSerialNumber(SerialNumberType serialNumberType) {
                return serialNumberType;
            }

            @Override
            public FieldType visitSectionHeader(SectionHeaderType sectionHeaderType) {
                return sectionHeaderType;
            }
        });
    }

    public TypedFormRecord duplicate(TypedFormRecord old) {
        TypedFormRecord newRecord = new TypedFormRecord(newId(old.getRef()));
        old.getParentRecordRef().ifPresent(parentRef -> {
            newRecord.setParentRecordRef(newId(parentRef));
        });
        old.getFieldValueMap().forEach((fieldId, value) -> {
            newRecord.set(fieldId, duplicate(value));
        });
        return newRecord;
    }

    private FieldValue duplicate(FieldValue value) {
        if(value instanceof SerialNumber) {
            return null;
        }
        if(value instanceof ReferenceValue) {
            ReferenceValue referenceValue = (ReferenceValue) value;
            return new ReferenceValue(referenceValue.getReferences()
                .stream()
                .map(this::newId)
                .collect(Collectors.toSet()));
        } else {
            return value;
        }
    }

    public Role duplicate(Role role, Collection<Resource> originalResources) {

        Set<ResourceId> formIds = originalResources
            .stream()
            .filter(r -> r.isFormOrSubform())
            .map(r -> r.getId())
            .collect(Collectors.toSet());

        return new RoleBuilder()
            .id(role.getId())
            .label(role.getLabel())
            .parameters(role.getParameters()
                .stream()
                .map(p -> duplicateParameter(p))
                .collect(Collectors.toSet()))
            .permissions(role.getPermissions()
                .stream()
                .map(p -> duplicatePermission(p, formIds))
                .collect(Collectors.toList()))
            .filters(role.getFilters()
                .stream()
                .map(f -> duplicateRoleFilter(f, formIds))
                .collect(Collectors.toList()))
            .build();
    }

    private RoleParameter duplicateParameter(RoleParameter parameter) {
        return new RoleParameter(parameter.getParameterId(), parameter.getLabel(), newId(parameter.getRange()));
    }

    private RoleFilter duplicateRoleFilter(RoleFilter filter, Set<ResourceId> referencedFormIds) {
        return new RoleFilter(filter.getId(), filter.getLabel(),
            duplicateFilterFormula(filter.getFilter(), referencedFormIds));
    }

    private FilteredPermission duplicatePermission(FilteredPermission permission, Set<ResourceId> referencedFormIds) {
        return new FilteredPermission(
            permission.getOperation(),
            permission.getFilter().map(f -> duplicateFilterFormula(f, referencedFormIds)),
            permission.getSecurityCategories());
    }

    private String duplicateFilterFormula(String formulaString, Set<ResourceId> referencedFormIds) {
        FormulaNode ast = FormulaParser.parse(formulaString);
        FormulaNode rewritten = ast.accept(new FormulaVisitor<FormulaNode>() {
            @Override
            public FormulaNode visitConstant(ConstantNode node) {
                return node;
            }

            @Override
            public FormulaNode visitSymbol(SymbolNode symbolNode) {
                final ResourceId id = ResourceId.valueOf(symbolNode.getName());
                if(VisibleDatabaseTree.isUserForm(id) || referencedFormIds.contains(id)) {
                    return new SymbolNode(newId(id));
                } else {
                    return symbolNode;
                }
            }

            @Override
            public FormulaNode visitGroup(GroupNode expr) {
                return new GroupNode(expr.accept(this));
            }

            @Override
            public FormulaNode visitCompoundExpr(CompoundExpr compoundExpr) {
                return new CompoundExpr(
                    compoundExpr.getValue().accept(this),
                    compoundExpr.getField());
            }

            @Override
            public FormulaNode visitFunctionCall(FunctionCallNode call) {
                return new FunctionCallNode(
                    call.getFunctionName(),
                    call.getArguments()
                        .stream().map(a -> a.accept(this))
                        .collect(Collectors.toList()));
            }

            @Override
            public FormulaNode visitError(ErrorFormulaNode errorFormulaNode) {
                return errorFormulaNode;
            }
        });
        return rewritten.deparse();
    }

}
