/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import org.activityinfo.json.AutoJson;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.promise.Maybe;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * <p>Hierarchical resource tree, lock set, and descriptive metadata of a Database.
 * Provided by a {@code DatabaseMetaProvider}.</p>
 *
 * <p>A {@link DatabaseTree} defines the <b>shared metadata of a Database</b> in the system. However a User will never
 * access a {@link DatabaseTree} directly as visibility of the data is dependent on the the rights of the User. </p>
 *
 * <p>Therefore, a {@link DatabaseTree} and {@link DatabaseUserGrantSet} are provided as input to a {@code DatabaseProvider} to
 * produce a {@link VisibleDatabaseTree}. A {@link VisibleDatabaseTree} defines the metadata of a Database <b>visible to the
 * requesting User.</b></p>
 */
@AutoJson
@AutoBuilder
public class DatabaseTree {

    private final ResourceId databaseId;
    private final ResourceId ownerId;
    private final long billingAccountId;
    private final String ownerName;
    private final String ownerEmail;
    private final long version;

    private final String label;
    private final String description;

    private final boolean publishedTemplate;
    private final boolean suspended;

    private String storage;

    private List<Role> roles = new ArrayList<>();

    private List<SecurityCategory> securityCategories = new ArrayList<>();

    // Resources and Locks present on this database
    private final Map<ResourceId, Resource> resources = new HashMap<>();
    private final Multimap<ResourceId, RecordLock> locks = HashMultimap.create();

    public DatabaseTree(ResourceId databaseId,
                        ResourceId ownerId,
                        long billingAccountId, String ownerName, String ownerEmail, long version,
                        String label,
                        String description,
                        boolean publishedTemplate,
                        boolean suspended,
                        String storage,
                        List<Resource> resources,
                        List<RecordLock> locks,
                        List<Role> roles,
                        List<SecurityCategory> securityCategories) {
        this.databaseId = databaseId;
        this.ownerId = ownerId;
        this.billingAccountId = billingAccountId;
        this.ownerName = ownerName;
        this.ownerEmail = ownerEmail;
        this.version = version;
        this.label = label;
        this.description = description;
        this.publishedTemplate = publishedTemplate;
        this.suspended = suspended;
        this.storage = storage;
        this.roles = roles;
        this.securityCategories = securityCategories;

        for (Resource resource : resources) {
            this.resources.put(resource.getId(), resource);
        }
        for (RecordLock lock : locks) {
            this.locks.put(lock.getResourceId(), lock);
        }
    }


    public @Nonnull ResourceId getDatabaseId() {
        return databaseId;
    }

    public ResourceId getOwnerId() {
        return ownerId;
    }

    public long getBillingAccountId() {
        return billingAccountId;
    }

    public long getVersion() {
        return version;
    }

    public @Nonnull String getLabel() {
        return label;
    }

    public @Nullable String getDescription() {
        return description;
    }

    public boolean isPublishedTemplate() {
        return publishedTemplate;
    }

    public boolean isSuspended() {
        return suspended;
    }

    public String getStorage() {
        return storage;
    }

    public Map<ResourceId, Resource> getResourceMap() {
        return resources;
    }

    public Optional<Resource> getResource(ResourceId id) {
        return Optional.ofNullable(resources.get(id));
    }

    /**
     * Returns true if the given {@code resourceId} is the database id or a resource that is currently
     * part of this database tree.
     */
    public boolean isResourcePresent(ResourceId resourceId) {
        return resourceId.equals(databaseId) || resources.containsKey(resourceId);
    }

    public Collection<Resource> getResources() {
        return resources.values();
    }

    public Collection<RecordLock> getLocks() {
        return locks.values();
    }

    public Multimap<ResourceId, RecordLock> getLockMap() {
        return locks;
    }

    public List<Role> getRoles() {
        return roles;
    }

    public ResourceId getUserFormId() {
        return VisibleDatabaseTree.getUserFormIdFromDatabaseId(getDatabaseId());
    }

    public Set<ResourceId> findAllChildren(Collection<ResourceId> resources) {
        Set<ResourceId> set = new HashSet<>();
        for (ResourceId id : resources) {
            if(set.add(id)) {
                collectChildren(id, set);
            }
        }
        return set;
    }

    private void collectChildren(ResourceId parentId, Set<ResourceId> set) {
        for (Resource resource : resources.values()) {
            if (resource.getParentId().equals(parentId)) {
                if(set.add(resource.getId())) {
                    collectChildren(resource.getId(), set);
                }
            }
        }
    }

    public Set<Resource> getDescendants(ResourceId resourceId) {
        Set<Resource> children = new HashSet<>();
        Multimap<ResourceId, Resource> resourceParentIdMap = getResourceParentIdMap();
        Set<Resource> firstLevelChildren = new HashSet<>();
        firstLevelChildren.addAll(resourceParentIdMap.get(resourceId));
        children.addAll(firstLevelChildren);
        while(true) {
            firstLevelChildren = collectNextLevelChildren(resourceParentIdMap, firstLevelChildren);
            if (firstLevelChildren.isEmpty()) {
                break;
            }
            children.addAll(firstLevelChildren);
        }
        return children;
    }


    private Set<Resource> collectNextLevelChildren(Multimap<ResourceId, Resource> resourceParentIdMap, Set<Resource> firstLevelChildren) {
        Set<Resource> nextLevelChildren = new HashSet<>();
        for (Resource resource : firstLevelChildren) {
            nextLevelChildren.addAll(resourceParentIdMap.get(resource.getId()));
        }
        return nextLevelChildren;
    }


    private Multimap<ResourceId, Resource> getResourceParentIdMap() {
        HashMultimap<ResourceId, Resource> resourceMap = HashMultimap.create();
        for (Resource resource : resources.values()) {
            resourceMap.put(resource.getParentId(), resource);
        }
        return resourceMap;
    }


    public String getOwnerName() {
        return ownerName;
    }

    public String getOwnerEmail() {
        return ownerEmail;
    }

    public VisibleDatabaseTree ownerVisible() {
        return visible(Optional.of(ownerId), Optional.empty()).get();
    }

    public Maybe<VisibleDatabaseTree> visible(DatabaseUserGrantSet grant) {
        return visible(Optional.of(grant.getUserId()), Optional.of(grant));
    }
    public Maybe<VisibleDatabaseTree> visible(ResourceId userId, Optional<DatabaseUserGrantSet> grant) {
        return visible(Optional.of(userId), grant);
    }

    public Maybe<VisibleDatabaseTree> visible(Optional<ResourceId> userId, Optional<DatabaseUserGrantSet> user) {
        boolean owner = userId.map(uId -> this.ownerId.equals(uId)).orElse(false);
        RoleAssignment roleAssignment = user.map(u -> u.getRole()).orElse(RoleAssignment.EMPTY);
        EffectiveRole role = user.map(gs -> EffectiveRole.compute(userId, roles, gs.getRole())).orElse(EffectiveRole.EMPTY);

        Map<ResourceId, DatabaseUserGrant> grants = user.map(DatabaseUserGrantSet::getGrantMap).orElse(Collections.emptyMap());

        // Is this database visible to the user at all?
        if(!isDatabaseVisible(owner, role, grants, resources)) {
            return Maybe.forbidden();
        }

        // Find the list of resources that are visible to the given user
        Collection<Resource> visibleResources = findVisibleResources(owner, role, grants);
        Set<ResourceId> visibleSet = visibleResources.stream().map(r -> r.getId()).collect(Collectors.toSet());

        // Retain only locks which apply to visible resources
        List<RecordLock> locks = this.locks.values()
            .stream()
            .filter(lock -> lock.getResourceId().equals(databaseId) ||
                            visibleSet.contains(lock.getResourceId()))
            .collect(Collectors.toList());

        return Maybe.of(new VisibleDatabaseTree(
            databaseId,
            userId,
            visibleVersion(user),
            label,
            description,
            new UserRef(ownerId, ownerName, ownerEmail),
            suspended,
            storage,
            roleAssignment,
            roles,
            securityCategories,
            visibleResources,
            locks,
            grants.values(),
            billingAccountId,
            publishedTemplate));
    }

    private String visibleVersion(Optional<DatabaseUserGrantSet> user) {
        return Long.toString(this.version) + user.map(u -> "#" + u.getVersion()).orElse("");
    }

    private boolean isDatabaseVisible(boolean owner, EffectiveRole role, Map<ResourceId, DatabaseUserGrant> grants, Map<ResourceId, Resource> resources) {
        if(this.publishedTemplate) {
            return true;
        }
        if(owner) {
            return true;
        }
        if(!role.isEmpty()) {
            return true;
        }

        // Are there any public resources?
        if(resources.values().stream().anyMatch(r -> r.getVisibility() == Resource.Visibility.PUBLIC)) {
            return true;
        }

        // Does the user have any explicit grants?
        return grants
            .values()
            .stream()
            .anyMatch(g -> g.hasOperation(Operation.VIEW));
    }

    private Collection<Resource> findVisibleResources(boolean owner, EffectiveRole role, Map<ResourceId, DatabaseUserGrant> grants) {

        // Is the whole tree visible?
        if(this.publishedTemplate || owner) {
            return resources.values();
        }

        Set<Resource> visibleSet = new HashSet<>();

        // Add all resources (and their children) that are visible
        // by virtue of the user's role
        for (ResourceId resourceId : role.getResources()) {
            addResourceAndChildren(resourceId, grants.keySet(), visibleSet);
        }

        // Now add additional resources that have explicit grants
        addGrantedResources(grants, visibleSet);

        // Add public and reference resources
        addPublicResources(role, grants, visibleSet);

        // Expand the visible set to include any parents of included resources.
        // For example, if we grant access to a form inside a folder, the user needs to be
        // able to see the folder, otherwise they can't navigate to the form.
        for (Resource resource : new ArrayList<>(visibleSet)) {
            addParents(resource, visibleSet);
        }

        return visibleSet;
    }

    private void addGrantedResources(Map<ResourceId, DatabaseUserGrant> grants, Set<Resource> visibleSet) {
        for (DatabaseUserGrant grant : grants.values()) {
            if(grant.hasOperation(Operation.VIEW)) {
                addResourceAndChildren(grant.getResourceId(), grants.keySet(), visibleSet);
            }
        }
    }

    private void addPublicResources(EffectiveRole role, Map<ResourceId, DatabaseUserGrant> grants, Set<Resource> visibleSet) {

        boolean databaseUser = !role.isEmpty() || !grants.isEmpty();

        for (Resource resource : resources.values()) {
            switch (resource.getVisibility()) {
                case PUBLIC:
                case REFERENCE:
                    visibleSet.add(resource);
                    break;

                case DATABASE_USERS:
                    if(databaseUser) {
                        visibleSet.add(resource);
                    }
                    break;
            }
        }
    }


    private void addResourceAndChildren(ResourceId parentId, Set<ResourceId> grantKeySet, Set<Resource> visibleSet) {
        Resource parent = resources.get(parentId);
        if(parent != null) {
            visibleSet.add(parent);
        }
        for (Resource child : resources.values()) {
            if(child.getParentId().equals(parentId)) {
                // If there is an explicit grant associated with this child, then
                // it will not inherit visibility
                if (!grantKeySet.contains(child.getId())) {
                    addResourceAndChildren(child.getId(), grantKeySet, visibleSet);
                }
            }
        }
    }


    private void addParents(Resource resource, Set<Resource> visibleSet) {
        Resource parent = resources.get(resource.getParentId());
        if(parent != null) {
            boolean added = visibleSet.add(parent);
            if(added) {
                addParents(parent, visibleSet);
            }
        }
    }

    public Optional<Role> getRole(ResourceId roleId) {
        return roles.stream().filter(r -> r.getId().equals(roleId)).findFirst();
    }

    public boolean hasResource(ResourceId resourceId) {
        return resources.containsKey(resourceId);
    }

    public List<SecurityCategory> getSecurityCategories() {
        return securityCategories;
    }
}
