/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.form.FormRecord;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;

/**
 * <p>
 *     A {@link DatabaseUserGrant} defines the set of allowed {@link Operation}s (and their optional {@link FormRecord} filters)
 *     on a specific {@link Resource}. A {@link DatabaseUserGrant} is inherited by any child {@link Resource}s. Any
 *     {@link Operation}s not defined on a {@link DatabaseUserGrant} are <strong>not allowed</strong> to be performed by the
 *     User.
 * </p>
 * <p>
 *     As well as a binary permission (can perform/cannot perform), the Database Owner may set a filter which should be
 *     applied when performing an {@link Operation}. These filters allow the Database Owner to limit the data the User
 *     is allowed to view or operate on at a Record level. Filters are composed of {@link FormulaNode}s.
 * </p>
 * <p>
 *      For example, a User may be limited to viewing data which is associated with a specific Partner field. In this
 *      instance, the User's {@link DatabaseUserGrant} will include an {@link Operation#VIEW} entry with a <br><br>
 *          {@code P0000001234 == "p0000004321"}
 *      <br><br> filter, specifying that they are only allowed to view records which are associated with the
 *      {@code FormRecord p00000043211} in the <i>referenced</i> {@code Form P0000001234}.
 * </p>
 * <p>
 *     A collection of {@link DatabaseUserGrant}s compose a {@link DatabaseUserGrantSet}. A {@link DatabaseUserGrantSet} gives the
 *     full set of allowed {@link Operation}s for a specific User, and their associated filters, across all of the
 *     {@link Resource}s within a Database.
 * </p>
 */
public class DatabaseUserGrant implements JsonSerializable, HasPermissions {

    private ResourceId resourceId;
    private Map<Operation, FilteredPermission> filteredPermissions = new HashMap<>();

    private DatabaseUserGrant() {
    }

    public DatabaseUserGrant(ResourceId resourceId) {
        this.resourceId = resourceId;
    }


    /**
     * @return The id of the {@link Resource} to which this grant applies. The {@code Resource} may be a
     * {@link ResourceType#FORM}, {@link ResourceType#SUB_FORM}, {@link ResourceType#FOLDER}, or {@link ResourceType#DATABASE}
     */
    public ResourceId getResourceId() {
        return resourceId;
    }

    /**
     * @return The set of {@link Operation}s the user is allowed to perform within the {@link Resource}
     * this grant applies to.
     */
    public Set<Operation> getOperations() {
        return filteredPermissions.keySet();
    }

    public boolean hasOperation(Operation operation) {
        return filteredPermissions.containsKey(operation);
    }

    /**
     * @return The (optional) filter to be applied when performing the specified operation on the
     * {@link Resource} this grant applies to.
     */
    public Optional<String> getFilter(Operation operation) {
        if(filteredPermissions.containsKey(operation)) {
            return filteredPermissions.get(operation).getFilter();
        } else {
            return Optional.empty();
        }
    }

    public Set<ResourceId> getSecurityCategories(Operation operation) {
        if(filteredPermissions.containsKey(operation)) {
            return filteredPermissions.get(operation).getSecurityCategories();
        } else {
            return Collections.emptySet();
        }
    }

    @Override
    public JsonValue toJson() {
        JsonValue opArray = Json.createArray();
        for (FilteredPermission permission : filteredPermissions.values()) {
            JsonValue opObject = Json.createObject();
            opObject.add("operation", Json.create(permission.getOperation().name()));
            opObject.add("filter", permission.getFilter().isPresent() ?
                    Json.create(permission.getFilter().get()) : Json.createNull());

            JsonValue scArray = Json.createArray();
            for (ResourceId sc : permission.getSecurityCategories()) {
                scArray.add(Json.create(sc.asString()));
            }
            opObject.add("securityCategories", scArray);

            opArray.add(opObject);
        }

        JsonValue object = Json.createObject();
        object.put("resourceId", resourceId.asString());
        object.put("operations", opArray);
        return object;
    }

    public static DatabaseUserGrant fromJson(JsonValue object) {
        DatabaseUserGrant grant = new DatabaseUserGrant();

        // Must accommodate legacy grant models (which specified only folder grants)
        JsonValue resourceId = object.hasKey("folderId") ? object.get("folderId") : object.get("resourceId");
        grant.resourceId = ResourceId.valueOf(resourceId.asString());

        if (object.hasKey("operations") && object.get("operations").isJsonArray()) {
            JsonValue opArray = object.get("operations");
            for (int i=0; i<opArray.length(); i++) {
                JsonValue opObject = opArray.get(i);
                Operation operation;
                try {
                    // Ignore permissions that we don't know about yet...
                    operation = Operation.valueOf(opObject.getString("operation").replace("CREATE_", "ADD_"));
                } catch (IllegalArgumentException e) {
                    continue;
                }
                Optional<String> filter = Optional.ofNullable(opObject.get("filter").asString());

                Set<ResourceId> securityCategories = new HashSet<>();
                if (opObject.hasKey("securityCategories") && opObject.get("securityCategories").isJsonArray()) {
                    JsonValue scArray = opObject.get("securityCategories");
                    for (int j = 0; j < scArray.length(); j++) {
                        String sc = scArray.getString(j);
                        if (sc != null) {
                            securityCategories.add(ResourceId.valueOf(sc));
                        }
                    }
                }
                grant.filteredPermissions.put(operation, new FilteredPermission(operation, filter, securityCategories));
            }
        }

        if(object.get("admin").asBoolean()) {
            grant.filteredPermissions.put(Operation.ADD_RESOURCE,
                    new FilteredPermission(Operation.ADD_RESOURCE, Optional.empty(), Collections.emptySet()));
            grant.filteredPermissions.put(Operation.EDIT_RESOURCE,
                    new FilteredPermission(Operation.EDIT_RESOURCE, Optional.empty(), Collections.emptySet()));
            grant.filteredPermissions.put(Operation.DELETE_RESOURCE,
                    new FilteredPermission(Operation.DELETE_RESOURCE, Optional.empty(), Collections.emptySet()));
            grant.filteredPermissions.put(Operation.LOCK_RECORDS,
                    new FilteredPermission(Operation.LOCK_RECORDS, Optional.empty(), Collections.emptySet()));
            grant.filteredPermissions.put(Operation.MANAGE_REFERENCE_DATA,
                    new FilteredPermission(Operation.MANAGE_REFERENCE_DATA, Optional.empty(), Collections.emptySet()));
            grant.filteredPermissions.put(Operation.MANAGE_COLLECTION_LINKS,
                    new FilteredPermission(Operation.MANAGE_COLLECTION_LINKS, Optional.empty(), Collections.emptySet()));
        }

        return grant;
    }

    public Optional<FilteredPermission> getPermission(Operation operation) {
        if(filteredPermissions.containsKey(operation)) {
            return Optional.of(filteredPermissions.get(operation));
        } else {
            return Optional.empty();
        }
    }

    public boolean hasPermission(Operation operation) {
        return filteredPermissions.containsKey(operation);
    }

    @Override
    public Collection<FilteredPermission> getPermissions() {
        return filteredPermissions.values();
    }

    public static class Builder {

        private DatabaseUserGrant model = new DatabaseUserGrant();

        public Builder() {
        }

        public Builder(DatabaseUserGrant toCopy) {
            model.resourceId = toCopy.getResourceId();
            model.filteredPermissions = new HashMap<>(toCopy.filteredPermissions);
        }

        /**
         * Sets the {@link Resource} to which this grant applies. This can be the id of a
         * {@link ResourceType#FORM}, {@link ResourceType#SUB_FORM}, {@link ResourceType#FOLDER} or {@link ResourceType#DATABASE}.
         */
        public Builder setResourceId(ResourceId resourceId) {
            model.resourceId = resourceId;
            return this;
        }

        /**
         * Adds a permitted {@link Operation} to this grant.
         */
        public Builder addOperation(Operation operation) {
            model.filteredPermissions.put(operation, new FilteredPermission(operation));
            return this;
        }

        public Builder addOperation(Operation operation, Optional<String> filter) {
            model.filteredPermissions.put(operation, new FilteredPermission(operation, filter));
            return this;
        }

        public Builder addOperation(Operation operation, Optional<String> filter, Set<ResourceId> securityCategories) {
            model.filteredPermissions.put(operation, new FilteredPermission(operation, filter, securityCategories));
            return this;
        }

        public Builder addOperation(Operation operation, boolean grant) {
            if(grant) {
                addOperation(operation);
            } else {
                removeOperation(operation);
            }
            return this;
        }

        public Builder addOperations(Set<Operation> operations) {
            for (Operation operation : operations) {
                model.filteredPermissions.put(operation, new FilteredPermission(operation));
            }
            return this;
        }

        /**
         * Adds a permitted {@link Operation} and filter to this grant.
         */
        public Builder addOperation(Operation operation, String recordFilter) {
            model.filteredPermissions.put(operation, new FilteredPermission(operation, recordFilter));
            return this;
        }

        public Builder removeOperation(Operation operation) {
            if(operation == Operation.VIEW) {
                model.filteredPermissions.clear();
            } else {
                model.filteredPermissions.remove(operation);
            }
            return this;
        }

        /**
         * Adds a filter to the specified operation on this grant. This filter is applied to the records
         * within the {@link Resource} this grant applies to when performing the specified operation.
         */
        public Builder addFilter(Operation operation, String recordFilter) {
            model.filteredPermissions.put(operation, new FilteredPermission(operation, recordFilter));
            return this;
        }

        public Builder addFilter(Operation operation, Optional<String> recordFilter) {
            model.filteredPermissions.put(operation, new FilteredPermission(operation, recordFilter));
            return this;
        }

        public Builder setAdmin(boolean admin) {
            if(admin && model.filteredPermissions.containsKey(Operation.VIEW)) {
                addOperation(Operation.VIEW);
            }
            return this;
        }

        public Builder setOperation(Operation operation, boolean granted) {
            if(granted) {
                addOperation(operation);
            } else {
                removeOperation(operation);
            }
            return this;
        }

        public DatabaseUserGrant build() {
            return model;
        }

    }
}
