/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.json.AutoJson;
import org.activityinfo.model.account.ActivationStatus;
import org.activityinfo.model.account.DeliveryStatus;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.time.LocalDate;

import java.util.*;
import java.util.stream.Collectors;

/**
 * <p>Collection of metadata for a User's permissions on a Database. Provided by a {@code DatabaseGrantProvider}.</p>
 *
 * <p>{@code DatabaseGrant} defines the <b>assigned permissions of a User on a Database</b> in the system. It defines
 * the Resources visible to the User and the permitted operations on those Resources.</p>
 *
 * <p>A {@link DatabaseTree} and {@code DatabaseGrant} are provided as input to a {@code DatabaseProvider} to
 * produce a {@link VisibleDatabaseTree}. A {@link VisibleDatabaseTree} defines the metadata of a Database <b>visible to the
 * requesting User.</b></p>
 */
@AutoJson
@AutoBuilder
public class DatabaseUserGrantSet {

    private final ResourceId databaseId;
    private final ResourceId userId;
    private final long version;

    private final String email;
    private final String name;
    private final double inviteTime;
    private final ActivationStatus activationStatus;
    private final DeliveryStatus deliveryStatus;
    private final double lastLoginTime;
    private final RoleAssignment role;
    private final Map<ResourceId, DatabaseUserGrant> grants = new HashMap<>();

    @Builder
    public DatabaseUserGrantSet(ResourceId databaseId,
                                ResourceId userId,
                                long version,
                                String email,
                                String name,
                                double inviteTime,
                                ActivationStatus activationStatus,
                                DeliveryStatus deliveryStatus,
                                double lastLoginTime,
                                RoleAssignment role,
                                List<DatabaseUserGrant> grants) {
        this.databaseId = databaseId;
        this.userId = userId;
        this.version = version;
        this.email = email;
        this.name = name;
        this.inviteTime = inviteTime;
        this.activationStatus = activationStatus;
        this.deliveryStatus = deliveryStatus;
        this.role = role;
        this.lastLoginTime = lastLoginTime;

        for (DatabaseUserGrant grant : grants) {
            this.grants.put(grant.getResourceId(), grant);
        }
    }

    public ResourceId getDatabaseId() {
        return databaseId;
    }

    // The set of grants assigned to this User.

    public ResourceId getUserId() {
        return userId;
    }

    public long getVersion() {
        return version;
    }

    public String getEmail() {
        return email;
    }

    public String getName() {
        return name;
    }

    public double getInviteTime() {
        return inviteTime;
    }

    public double getLastLoginTime() {
        return lastLoginTime;
    }

    public ActivationStatus getActivationStatus() {
        if(activationStatus == null) {
            return ActivationStatus.ACTIVE;
        } else {
            return activationStatus;
        }
    }

    public DeliveryStatus getDeliveryStatus() {
        if(deliveryStatus == null) {
            return DeliveryStatus.UNKNOWN;
        }
        return deliveryStatus;
    }

    public DatabaseUser toDatabaseUser() {
        return new DatabaseUser(databaseId, userId, name, email, role, version,
            new LocalDate(new Date((long)inviteTime)).toString(),
            getDeliveryStatus(),
            activationStatus != ActivationStatus.PENDING);
    }

    public Collection<DatabaseUserGrant> getGrants() {
        return grants.values();
    }

    public Map<ResourceId, DatabaseUserGrant> getGrantMap() {
        return grants;
    }

    public Optional<DatabaseUserGrant> getGrantIfPresent(ResourceId resourceId) {
        return Optional.ofNullable(grants.get(resourceId));
    }

    public RoleAssignment getRole() {
        return role;
    }

    public List<RoleAssignment> getRoles() {
        if(role.isEmpty()) {
            return Collections.emptyList();
        } else {
            return Collections.singletonList(role);
        }
    }

    public DatabaseUserGrantSet withoutDeletedResources(DatabaseTree tree) {
        return new DatabaseUserGrantSetBuilder(this)
            .role(role.withoutDeletedResources(tree))
            .grants(this.grants.values()
                .stream()
                .filter(g -> tree.isResourcePresent(g.getResourceId()))
                .collect(Collectors.toList()))
            .build();
    }

}
