/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.model.formula.*;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.ReferenceValue;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

public class EffectiveRole {

    private static final Logger LOGGER = Logger.getLogger(EffectiveRole.class.getName());

    public static final EffectiveRole EMPTY = new EffectiveRole();
    public static final SymbolNode USER_SYMBOL = new SymbolNode("@user");

    private final ResourceId userId;
    private Role role;
    private Map<ResourceId, FieldValue> parameters;
    private Set<ResourceId> resources;
    private List<FilteredPermission> permissions;

    private EffectiveRole() {
        this.userId = ResourceId.valueOf("0");
        this.role = Role.EMPTY;
        this.parameters = Collections.emptyMap();
        this.resources = Collections.emptySet();
        this.permissions = Collections.emptyList();
    }

    public EffectiveRole(ResourceId userId, Role role, RoleAssignment assignment) {
        this.userId = userId;
        this.role = role;
        this.resources = assignment.getRoleResources();
        this.parameters = new HashMap<>();
        for (RoleParameter parameter : role.getParameters()) {
            FieldValue value = assignment.getParameterValue(parameter);
            if(value != null) {
                parameters.put(parameter.getParameterId(), value);
            }
        }

        // Compute the effective permissions
        permissions = new ArrayList<>();
        for (FilteredPermission permission : role.getPermissions()) {
            permissions.add(new FilteredPermission(permission.getOperation(), permission.getFilter().map(this::substitute), permission.getSecurityCategories()));
        }
    }


    private Optional<String> substitute(Optional<String> filter) {
        return filter.map(this::substitute);
    }

    private String substitute(String formula) {
        if(!formula.contains("@user")) {
            return formula;
        }
        FormulaNode node;
        try {
            node = FormulaParser.parse(formula);
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, "Failed to parse permission formula", e);
            return "FALSE";
        }
        return node
            .transform(this::substituteRoleParameters)
            .transform(this::subsituteUserId)
            .deparse();
    }

    private FormulaNode substituteRoleParameters(FormulaNode formulaNode) {
        if(formulaNode instanceof CompoundExpr) {
            CompoundExpr expr = (CompoundExpr) formulaNode;
            if(expr.getValue().equals(USER_SYMBOL)) {
                FieldValue fieldValue = parameters.get(expr.getField().asResourceId());
                if(fieldValue instanceof ReferenceValue) {
                    ReferenceValue value = (ReferenceValue) fieldValue;
                    return new ConstantNode(value.getReferences().iterator().next().getRecordId().asString());
                }
            }
        }
        return formulaNode;
    }

    private FormulaNode subsituteUserId(FormulaNode formulaNode) {
        if(formulaNode.equals(USER_SYMBOL)) {
            return new ConstantNode(userId.toString());
        } else {
            return formulaNode;
        }
    }


    public Optional<ReferenceValue> getParameterAsReferenceValue(ResourceId parameterId) {
        FieldValue value = parameters.get(parameterId);
        if(value instanceof ReferenceValue) {
            return Optional.of((ReferenceValue) value);
        } else {
            return Optional.empty();
        }
    }

    public Role getRole() {
        return role;
    }

    public Map<ResourceId, FieldValue> getParameters() {
        return parameters;
    }

    public Set<ResourceId> getResources() {
        return resources;
    }

    public boolean isEmpty() {
        return resources.isEmpty() && permissions.isEmpty() && parameters.isEmpty();
    }

    public static EffectiveRole compute(VisibleDatabaseTree tree, RoleAssignment assignment) {
        return compute(tree.getUserId(), assignment, tree.getRoles());
    }

    public static EffectiveRole compute(Optional<ResourceId> userId, RoleAssignment assignment, List<Role> roles) {
        return compute(userId, roles, assignment);
    }

    public static EffectiveRole compute(Optional<ResourceId> userId, List<Role> roles, RoleAssignment assignment) {
        return roles.stream()
            .filter(r -> r.getId().equals(assignment.getRoleId()))
            .findAny()
            .flatMap(role -> userId.map(uId -> new EffectiveRole(uId, role, assignment)))
            .orElse(EffectiveRole.EMPTY);
    }

    /**
     * Returns the equivalent {@link DatabaseUserGrant} for this role, with all role parameters
     * resolved to the user's parameter values.
     *
     * <p>For example, if this role includes a filter <pre>partner == @user.partner</pre> for the VIEW operation, then
     * this method will substitute the user's actual partner (e.g. "NRC") for the expression <pre>@user.partner</pre></p>
     */
    public DatabaseUserGrant getResolvedGrant(ResourceId resourceId) {
        DatabaseUserGrant.Builder grant = new DatabaseUserGrant.Builder();
        grant.setResourceId(resourceId);
        for (FilteredPermission permission : permissions) {
            grant.addOperation(permission.getOperation(), permission.getFilter(), permission.getSecurityCategories());
        }
        return grant.build();
    }

    /**
     * Returns the equivalent {@link DatabaseUserGrant} for this role.
     *
     * <p>For example, if this role includes a filter <pre>partner == @user.partner</pre> for the VIEW operation, then
     * this grant will retain the symbolic <pre>@user.partner</pre> expression. That way, if the user's partner is
     * changed, their resolved grant will change automatically.
     */
    public DatabaseUserGrant getGrant(ResourceId resourceId) {
        DatabaseUserGrant.Builder grant = new DatabaseUserGrant.Builder();
        grant.setResourceId(resourceId);
        for (FilteredPermission permission : role.getPermissions()) {
            grant.addOperation(permission.getOperation(), permission.getFilter(), permission.getSecurityCategories());
        }
        return grant.build();
    }

    public DatabaseUserGrant resolve(DatabaseUserGrant grant) {
        DatabaseUserGrant.Builder resolved = new DatabaseUserGrant.Builder();
        resolved.setResourceId(grant.getResourceId());

        for (Operation operation : grant.getOperations()) {
            resolved.addOperation(operation, substitute(grant.getFilter(operation)), grant.getSecurityCategories(operation));
        }
        return resolved.build();
    }

    public ResourceId getRoleId() {
        return role.getId();
    }

}
