/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;
import java.util.stream.Collectors;

public interface HasPermissions {

    Optional<FilteredPermission> getPermission(Operation operation);

    boolean hasPermission(Operation operation);

    default boolean hasUnconditionalPermission(Operation operation) {
        return getPermission(operation).map(p -> !p.isFiltered()).orElse(false);
    }

    Collection<FilteredPermission> getPermissions();

    public static HasPermissions VIEW_ONLY = viewOnly(Optional.empty());

    public static HasPermissions viewOnly(FormulaNode formulaNode) {
        return viewOnly(Optional.of(formulaNode.deparse()));
    }

    public static HasPermissions viewOnly(Optional<String> filter) {

        FilteredPermission view = new FilteredPermission(Operation.VIEW, filter);

        return new HasPermissions() {
            @Override
            public Optional<FilteredPermission> getPermission(Operation operation) {
                if(operation == Operation.VIEW) {
                    return Optional.of(view);
                } else {
                    return Optional.empty();
                }
            }

            @Override
            public boolean hasPermission(Operation operation) {
                return operation == Operation.VIEW;
            }

            @Override
            public Collection<FilteredPermission> getPermissions() {
                return Collections.singleton(view);
            }
        };
    }

    public static HasPermissions NONE = new HasPermissions() {
        @Override
        public Optional<FilteredPermission> getPermission(Operation operation) {
            return Optional.empty();
        }

        @Override
        public boolean hasPermission(Operation operation) {
            return false;
        }

        @Override
        public Collection<FilteredPermission> getPermissions() {
            return Collections.emptySet();
        }
    };

    public static HasPermissions all(VisibleDatabaseTree tree) {
        final Set<ResourceId> securityCategories = tree.getSecurityCategories()
            .stream()
            .map(c -> c.getId())
            .collect(Collectors.toSet());

        return new HasPermissions() {
            @Override
            public Optional<FilteredPermission> getPermission(Operation operation) {
                if (operation == Operation.EDIT_RECORD) {
                    return Optional.of(new FilteredPermission(operation, securityCategories));
                } else {
                    return Optional.of(new FilteredPermission(operation));
                }
            }

            @Override
            public boolean hasPermission(Operation operation) {
                return true;
            }

            @Override
            public Collection<FilteredPermission> getPermissions() {
                return Arrays.stream(Operation.values())
                    .map(o -> {
                        if (o == Operation.EDIT_RECORD) {
                            return new FilteredPermission(o, securityCategories);
                        } else {
                            return new FilteredPermission(o);
                        }
                    })
                    .collect(Collectors.toList());
            }
        };
    }

}
