/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import com.google.common.collect.Iterables;
import com.google.common.collect.Streams;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

import static java.util.function.Function.identity;

public class ListDelta<T> {

    public static class Update<T> {
        private final T old;
        private final T updated;

        public Update(T old, T updated) {
            this.old = old;
            this.updated = updated;
        }

        public T getOld() {
            return old;
        }

        public T getUpdated() {
            return updated;
        }
    }

    private final Map<ResourceId, T> added;
    private final Map<ResourceId, Update<T>> updated;
    private final Map<ResourceId, T> deleted;
    private final Set<ResourceId> missingDeleted;
    private final List<T> result;

    public ListDelta(Function<T, ResourceId> id,
                     Collection<T> existing,
                     List<T> updates,
                     Collection<ResourceId> deletions) {

        Map<ResourceId, T> existingMap = existing
            .stream()
            .collect(Collectors.toMap(id, identity()));


        added = new HashMap<>();
        updated = new HashMap<>();
        for (T update : updates) {
            ResourceId key = id.apply(update);
            T old = existingMap.get(key);
            if (old == null) {
                added.put(key, update);
            } else {
                updated.put(key, new Update(old, update));
            }
        }

        deleted = new HashMap<>();
        missingDeleted = new HashSet<>();
        for (ResourceId key : deletions) {
            T old = existingMap.get(key);
            if(old == null) {
                missingDeleted.add(key);
            } else {
                deleted.put(key, old);
            }
        }

        // Combine retained, updated, and added
        result = Streams
            .concat(
                existing.stream()
                    .filter(t -> !updated.containsKey(id.apply(t)))
                    .filter(t -> !deleted.containsKey(id.apply(t))),
                added.values().stream(),
                updated.values().stream().map(u -> u.getUpdated()))
            .collect(Collectors.toList());
    }

    public Map<ResourceId, T> getAdded() {
        return added;
    }

    public Map<ResourceId, Update<T>> getUpdated() {
        return updated;
    }

    public Map<ResourceId, T> getDeleted() {
        return deleted;
    }

    public Set<ResourceId> getMissingDeleted() {
        return missingDeleted;
    }

    public boolean isDeleted(ResourceId id) {
        return deleted.containsKey(id);
    }

    public boolean isEmpty() {
        return added.isEmpty() && deleted.isEmpty() && updated.isEmpty() && missingDeleted.isEmpty();
    }

    public List<T> getAddedOrUpdated() {
        List<T> list = new ArrayList<>();
        list.addAll(added.values());
        updated.values().forEach(u -> list.add(u.getUpdated()));
        return list;
    }

    public Iterable<ResourceId> getUpdatedOrDeletedIds() {
        return Iterables.concat(updated.keySet(), deleted.keySet());
    }

    public List<T> getAddedUpdatedOrDeleted() {
        List<T> list = new ArrayList<>();
        list.addAll(added.values());
        updated.values().forEach(u -> list.add(u.getUpdated()));
        deleted.values().forEach(u -> list.add(u));
        return list;
    }

    public List<T> getResult() {
        return result;
    }
}
