/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.json.JsonArrays;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.TypedFormRecord;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.time.LocalDateInterval;
import org.activityinfo.model.type.time.PeriodType;
import org.activityinfo.model.type.time.PeriodValue;

import java.util.*;
import java.util.stream.Collectors;

public class RecordLockSet implements Iterable<RecordLock>, JsonSerializable {

    public static final RecordLockSet EMPTY = new RecordLockSet(Collections.emptyList());

    private List<RecordLock> locks;

    public RecordLockSet(List<RecordLock> locks) {
        this.locks = locks;
    }

    @Override
    public Iterator<RecordLock> iterator() {
        return locks.iterator();
    }

    public boolean isEmpty() {
        return locks.isEmpty();
    }

    @Override
    public JsonValue toJson() {
        return JsonArrays.toJsonArray(locks, RecordLockJson::toJson);
    }

    public static RecordLockSet fromJson(JsonValue array) {
        return new RecordLockSet(JsonArrays.listFromJson(array, RecordLockJson::fromJson));
    }

    public boolean isLocked(PeriodValue period) {
        return isLocked(period.asInterval());
    }

    public boolean isLocked(LocalDateInterval period) {
        for (RecordLock lock : locks) {
            if(lock.getDateRange().overlaps(period)) {
                return true;
            }
        }
        return false;
    }

    public boolean isLocked(FieldValue value) {
        if(value instanceof PeriodValue) {
            return isLocked(((PeriodValue) value));
        } else {
            return false;
        }
    }

    public boolean isLocked(FormTree formTree, TypedFormRecord record) {
        for (ResourceId fieldId :  findLockedFields(formTree.getRootFormClass())) {
            if (isLocked(record.get(fieldId))) {
                return true;
            }
        }
        return false;
    }

    public List<RecordLock> getLocks() {
        return locks;
    }

    /**
     * Finds the set of fields ids to which this {@code RecordLockSet} affects.
     */
    public Set<ResourceId> findLockedFields(FormClass formClass) {
        if(isEmpty()) {
            return Collections.emptySet();
        } else {
            return formClass.getFields()
                .stream()
                .filter(f -> f.isKey() && f.getType() instanceof PeriodType)
                .map(f -> f.getId())
                .collect(Collectors.toSet());
        }
    }

    public Map<ResourceId, List<RecordLock>> getResourceLockMap() {
        Map<ResourceId, List<RecordLock>> resourceLockMap = new HashMap<>();
        for (RecordLock lock : locks) {
            ResourceId resourceId = lock.getResourceId();
            if (resourceLockMap.containsKey(resourceId)) {
                List<RecordLock> lockList = resourceLockMap.get(resourceId);
                lockList.add(lock);
                resourceLockMap.put(resourceId, lockList);
            } else {
                List<RecordLock> lockList = new ArrayList<>();
                lockList.add(lock);
                resourceLockMap.put(resourceId, lockList);
            }
        }
        return resourceLockMap;
    }

}
