/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.resource.ResourceId;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

public class Resource implements JsonSerializable {



    public enum Visibility {
        /**
         * Public.
         * Visible to *all* Users on system, regardless of assigned permissions.
         */
        PUBLIC,

        /**
         * Visible to any user who has been add
         */
        @Deprecated
        DATABASE_USERS,

        /**
         * Visible only to users given explicit permissions.
         */
        PRIVATE,

        /**
         * Visible only if another Resource on the Database is visible to the current User. Used to provide Forms
         * with data to be referenced, but not requiring an explicit grant to be generated for the User.
         */
        REFERENCE
    }

    public static class Node {

        private Node parentNode;
        private Resource resource;
        private List<Node> childNodes = new ArrayList<>(0);

        public Node() {
        }

        public Node(@Nonnull Resource resource) {
            this.resource = resource;
        }

        public Node(@Nonnull Node parentNode, @Nonnull Resource resource) {
            this.parentNode = parentNode;
            this.resource = resource;
        }

        public boolean isRoot() {
            return parentNode == null;
        }

        public boolean isLeaf() {
            return childNodes.isEmpty();
        }

        public void addChildNode(Node childNode) {
            childNodes.add(childNode);
        }

        public void addChildNodes(List<Node> children) {
            childNodes.addAll(children);
        }

        public @Nullable Node getParentNode() {
            return parentNode;
        }

        public void setParentNode(Node parentNode) {
            this.parentNode = parentNode;
        }

        public @Nonnull Resource getResource() {
            return resource;
        }

        public List<Node> getChildNodes() {
            return childNodes;
        }
    }

    /**
     * The resource's id. Form and subform resources must have globally-unique ids (across all of ActivityInfo)
     * but folder ids are only guaranteed to be unique within a single database.
     */
    private ResourceId id;

    /**
     * The id of this resource's parent, for example, a folder id. If this is a top-level folder or form, the
     * the parentId will be the id of the database itself.
     */
    private ResourceId parentId;

    /**
     * The resource type.
     */
    private ResourceType type;

    /**
     * The resource's human-readable label.
     */
    private String label;

    /**
     * The resource's visibility.
     */
    private Visibility visibility = Visibility.PRIVATE;

    private Resource() {}

    public Resource(ResourceId id, ResourceType type, ResourceId parentId, String label) {
        this.id = id;
        this.type = type;
        this.parentId = parentId;
        this.label = label;
    }

    @org.activityinfo.model.annotation.Builder
    public Resource(ResourceId id, ResourceType type, ResourceId parentId, String label, Visibility visibility) {
        this(id, type, parentId, label);
        this.visibility = visibility;
    }

    public ResourceId getId() {
        return id;
    }

    public ResourceId getParentId() {
        return parentId;
    }

    public String getLabel() {
        if(label == null) {
            return "UNLABELED";
        }
        return label;
    }

    public ResourceType getType() {
        return type;
    }

    public Visibility getVisibility() {
        return visibility;
    }

    public boolean isPublic() {
        return visibility == Visibility.PUBLIC;
    }

    public boolean isPrivate() {
        return visibility == Visibility.PRIVATE;
    }

    public boolean isReference() {
        return visibility == Visibility.REFERENCE;
    }

    public boolean isFormOrSubform() {
        return ResourceType.FORM.equals(type) || ResourceType.SUB_FORM.equals(type);
    }


    public Resource withLabel(String newLabel) {
        return new Resource(id, type, parentId, newLabel, this.visibility);
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = Json.createObject();
        object.put("id", id.asString());
        object.put("parentId", parentId.asString());
        object.put("label", label);
        object.put("type", type.name());
        object.put("visibility", visibility.name());
        return object;
    }

    public static Resource fromJson(JsonValue object) {
        Resource resource = new Resource();
        resource.id = ResourceId.valueOf(object.getString("id"));
        resource.parentId = ResourceId.valueOf(object.getString("parentId"));
        resource.label = object.getString("label");
        resource.type = ResourceType.valueOf(object.getString("type"));
        resource.visibility = Visibility.valueOf(object.getString("visibility"));
        return resource;
    }

    public static class Builder {
        private Resource resource = new Resource();

        public Builder() {
        }

        public Builder(Resource toCopy) {
            resource.id = toCopy.id;
            resource.type = toCopy.type;
            resource.label = toCopy.label;
            resource.visibility = toCopy.visibility;
            resource.parentId = toCopy.parentId;
        }

        public Builder setId(ResourceId id) {
            resource.id = id;
            return this;
        }

        public Builder setLabel(@Nonnull String label) {
            resource.label = label;
            return this;
        }

        public Builder setParentId(ResourceId id) {
            resource.parentId = id;
            return this;
        }

        public Builder setType(ResourceType type) {
            resource.type = type;
            return this;
        }

        public Builder setVisibility(Visibility visibility) {
            resource.visibility = visibility;
            return this;
        }

        public Builder setVisibleToPublic() {
            return setVisibility(Visibility.PUBLIC);
        }

        public Builder setPrivate() {
            return setVisibility(Visibility.PRIVATE);
        }

        public Builder setVisibleAsReference() {
            return setVisibility(Visibility.REFERENCE);
        }

        public Resource build() {
            assert resource.id != null : "id is missing";
            assert resource.label != null : "label is missing";
            assert resource.type != null : "type is missing";
            return resource;
        }
    }

    @Override
    public int hashCode() {
        return Objects.hashCode(getId().asString());
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (!(obj instanceof Resource)) {
            return false;
        }
        return toString().equals(obj.toString());
    }

    @Override
    public String toString() {
        return toJson().toJson();
    }

}
