/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.json.AutoJson;
import org.activityinfo.model.annotation.AutoBuilder;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * Describes a named, potentially parametrized set of permissions that can be granted to users.
 */
@AutoJson
@AutoBuilder
public class Role implements HasPermissions {

    public static final Role EMPTY = new Role();

    /**
     * This role's id
     */
    private final ResourceId id;

    /**
     * This role's human-readable label
     */
    private final String label;

    /**
     * Permissions granted this role.
     */
    private final List<FilteredPermission> permissions;

    /**
     * Parameters defined for this role. Parameters can be referenced in filtering formulas.
     */
    private final List<RoleParameter> parameters;

    /**
     * Pre-defined filters. Role filters allow other users to choose filters for permissions
     * without having to write formulas themselves.
     */
    private final List<RoleFilter> filters;
    private final List<DatabaseUserGrant> grants;
    private final int version;
    private final boolean grantBased;

    private Role() {
        this.id = ResourceId.valueOf("empty");
        this.label = "";
        this.permissions = Collections.emptyList();
        this.parameters = Collections.emptyList();
        this.filters = Collections.emptyList();
        this.grants = Collections.emptyList();
        this.version = 0;
        this.grantBased = true;
    }

    @Builder
    public Role(ResourceId id, String label, List<FilteredPermission> permissions, List<RoleParameter> parameters, List<RoleFilter> filters, List<DatabaseUserGrant> grants, int version, boolean grantBased) {
        this.id = id;
        this.label = label;
        this.permissions = permissions;
        this.parameters = parameters;
        this.filters = filters;
        this.grants = grants;
        this.version = version;
        this.grantBased = grantBased;
    }

    public ResourceId getId() {
        return id;
    }

    public String getLabel() {
        return label;
    }

    public List<FilteredPermission> getPermissions() {
        return permissions;
    }

    public List<RoleParameter> getParameters() {
        return parameters;
    }

    public List<RoleFilter> getFilters() {
        return filters;
    }

    public List<DatabaseUserGrant> getGrants() {
        return grants;
    }

    /**
     * @return
     */
    public int getVersion() {
        return version;
    }

    public boolean isGrantBased() {
        return grantBased;
    }

    public Optional<FilteredPermission> getPermission(Operation operation) {
        return permissions
            .stream()
            .filter(p -> p.getOperation() == operation)
            .findAny();
    }

    public boolean hasPermission(Operation operation) {
        return permissions
            .stream()
            .anyMatch(p -> p.getOperation() == operation);
    }

    public Optional<RoleFilter> matchFilter(String filter) {
        for (RoleFilter roleFilter : filters) {
            if(roleFilter.getFilter().equals(filter)) {
                return Optional.of(roleFilter);
            }
        }
        return Optional.empty();
    }

    public Role withoutPermission(Operation operation) {
        return new RoleBuilder(this)
            .permissions(permissions
                .stream()
                .filter(p -> p.getOperation() != operation)
                .collect(Collectors.toList()))
            .build();
    }
}
