/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import org.activityinfo.json.Json;
import org.activityinfo.json.JsonArrays;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * Describes a user's assignment to a role within a database.
 */
public class RoleAssignment implements JsonSerializable {

    public static final RoleAssignment EMPTY = new RoleAssignment(ResourceId.valueOf("default"), Json.createObject(), Collections.emptySet());

    /**
     * The id of the role assigned to this user.
     */
    private ResourceId roleId;

    /**
     * The values of the parameters defined for this role.
     */
    private JsonValue roleParameters;

    /**
     * The resources (database, folder, forms, or subforms) to which the role's permissions
     * are to apply.
     */
    private Set<ResourceId> roleResources;

    @Builder
    public RoleAssignment(ResourceId roleId, JsonValue roleParameters, Set<ResourceId> roleResources) {
        this.roleId = roleId;
        this.roleParameters = roleParameters;
        this.roleResources = roleResources;
    }

    public RoleAssignment(ResourceId roleId, ResourceId resourceId) {
        this.roleId = roleId;
        this.roleParameters = Json.createObject();
        this.roleResources = Collections.singleton(resourceId);
    }

    /**
     * The id of the role to which this user has been assigned.
     */
    public ResourceId getRoleId() {
        return roleId;
    }

    /**
     * A json object containing the values of the role parameter values.
     */
    public JsonValue getRoleParameters() {
        return roleParameters;
    }

    /**
     * The resources to which the user
     * @return
     */
    public Set<ResourceId> getRoleResources() {
        return roleResources;
    }

    public FieldValue getParameterValue(RoleParameter parameter) {
        JsonValue value = roleParameters.get(parameter.getParameterId().asString());
        if(value.isJsonNull()) {
            return null;
        } else {
            return parameter.parseJsonValue(value);
        }
    }

    public boolean isEmpty() {
        return roleId.equals(EMPTY.getRoleId());
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = Json.createObject();
        object.put("id", roleId.asString());
        object.put("parameters", roleParameters);
        object.put("resources", JsonArrays.toJsonArray(roleResources, ResourceId::toJson));
        return object;
    }

    public static RoleAssignment fromJson(JsonValue value) {
        if(value.isJsonObject()) {
            return new RoleAssignment(
                ResourceId.valueOf(value.getString("id")),
                value.get("parameters"),
                JsonArrays.setFromJson(value.get("resources"), ResourceId::fromJson));
        } else {
            return RoleAssignment.EMPTY;
        }
    }

    public RoleAssignment withoutDeletedResources(DatabaseTree tree) {
        Set<ResourceId> newSet = new HashSet<>();
        for (ResourceId roleResourceId : roleResources) {
            if (tree.isResourcePresent(roleResourceId)) {
                newSet.add(roleResourceId);
            }
        }
        return new RoleAssignment(roleId, roleParameters, newSet);
    }
}
