/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import com.google.common.collect.Sets;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

public class UpdateDatabaseDelta {
    private final DatabaseTree originalTree;
    private final Optional<String> updatedDatabaseLabel;
    private final ListDelta<Resource> resources;
    private final ListDelta<Role> roles;
    private final ListDelta<RecordLock> locks;
    private final Map<ResourceId, Resource> cascadingDeletes;
    private final List<RecordLock> orphanedLocks;
    private final Optional<Boolean> publishTemplate;

    public UpdateDatabaseDelta(DatabaseTree originalTree,
                               Optional<String> updatedDatabaseLabel,
                               ListDelta<Resource> resources,
                               ListDelta<Role> roles,
                               ListDelta<RecordLock> locks,
                               Map<ResourceId, Resource> cascadingDeletes,
                               List<RecordLock> orphanedLocks,
                               Optional<Boolean> publishTemplate) {
        this.originalTree = originalTree;
        this.updatedDatabaseLabel = updatedDatabaseLabel;
        this.resources = resources;
        this.roles = roles;
        this.locks = locks;
        this.cascadingDeletes = cascadingDeletes;
        this.orphanedLocks = orphanedLocks;
        this.publishTemplate = publishTemplate;
    }

    public DatabaseTree getOriginalTree() {
        return originalTree;
    }

    public Optional<String> getUpdatedDatabaseLabel() {
        return updatedDatabaseLabel;
    }

    public ListDelta<Resource> getResources() {
        return resources;
    }

    public ListDelta<Role> getRoles() {
        return roles;
    }

    public ListDelta<RecordLock> getLocks() {
        return locks;
    }

    public Map<ResourceId, Resource> getCascadingDeletes() {
        return cascadingDeletes;
    }

    public List<RecordLock> getOrphanedLocks() {
        return orphanedLocks;
    }

    public boolean isResourceDeleted(ResourceId resourceId) {
        return
            resources.getDeleted().containsKey(resourceId) ||
            cascadingDeletes.containsKey(resourceId);
    }

    public Optional<Boolean> getPublishTemplate() {
        return publishTemplate;
    }

    public List<Resource> getAllDeletedResources() {
        List<Resource> all = new ArrayList<>();
        all.addAll(resources.getDeleted().values());
        all.addAll(cascadingDeletes.values());
        return all;
    }

    /**
     * Gets the updated resource list, after applying all additions, updates, deletions,
     * and cascading deletes of child resources.
     */
    public List<Resource> getUpdatedResourceList() {
        return resources
            .getResult()
            .stream()
            .filter(r -> !cascadingDeletes.containsKey(r.getId()))
            .collect(Collectors.toList());
    }

    /**
     * Gets the updated list of record locks, after applying all additions, updates, deletions,
     * and removing locks of deleted resources.
     */
    public List<RecordLock> getUpdatedLockList() {
        return locks
            .getResult()
            .stream()
            .filter(lock -> !orphanedLocks.contains(lock))
            .collect(Collectors.toList());
    }

    public List<Role> getUpdatedRoleList() {
        return roles.getResult();
    }

    public static UpdateDatabaseDelta build(DatabaseTree tree, UpdateDatabaseRequest request) {

        Optional<String> updatedDatabaseLabel = request.getResourceUpdates()
            .stream()
            .filter(r -> r.getId().equals(tree.getDatabaseId()))
            .findAny()
            .flatMap(update -> {
                if (!update.getLabel().equals(tree.getLabel())) {
                    return Optional.of(update.getLabel());
                } else {
                    return Optional.empty();
                }
            });

        ListDelta<Resource> resourceDiff = new ListDelta<>(Resource::getId,
            tree.getResources(),
            request.getResourceUpdates()
                .stream()
                .filter(r -> !r.getId().equals(tree.getDatabaseId()))
                .collect(Collectors.toList()),
            request.getResourceDeletions());

        ListDelta<Role> roleDiff = new ListDelta<>(Role::getId,
            tree.getRoles(),
            request.getRoleUpdates(),
            request.getRoleDeletions());

        ListDelta<RecordLock> lockDiff = new ListDelta<>(RecordLock::getId,
            tree.getLocks(),
            request.getLockUpdates(),
            request.getLockDeletions());

        Map<ResourceId, Resource> cascadingDeletes = findCascadingDeletes(tree, resourceDiff.getDeleted().keySet());

        Set<ResourceId> deletedResources = Sets.union(
            resourceDiff.getDeleted().keySet(),
            cascadingDeletes.keySet());

        Set<ResourceId> retainedResources = Sets.difference(
            tree.getResources().stream().map(Resource::getId).collect(Collectors.toSet()),
            deletedResources);

        List<RecordLock> orphanedLocks = tree.getLocks()
            .stream()
            .filter(lock -> !lock.getResourceId().equals(tree.getDatabaseId()))
            .filter(lock -> !lockDiff.isDeleted(lock.getId()))
            .filter(lock -> !retainedResources.contains(lock.getResourceId()))
            .collect(Collectors.toList());

        return new UpdateDatabaseDelta(tree, updatedDatabaseLabel, resourceDiff, roleDiff,
                lockDiff, cascadingDeletes, orphanedLocks, request.getPublishedTemplate());
    }


    static Map<ResourceId, Resource> findCascadingDeletes(DatabaseTree tree, Set<ResourceId> deleted) {

        Set<ResourceId> deletedParents = new HashSet<>(deleted);
        List<Resource> cascading = new ArrayList<>();

        boolean deleting;
        do {
            deleting = false;
            for (Resource resource : tree.getResources()) {
                if(deletedParents.contains(resource.getParentId())) {
                    if(deletedParents.add(resource.getId())) {
                        deleting = true;
                        cascading.add(resource);
                    }
                }
            }
        } while(deleting);

        return cascading
            .stream()
            .collect(Collectors.toMap(Resource::getId, Function.identity()));
    }


}
