/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.database;

import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import org.activityinfo.json.AutoJson;
import org.activityinfo.json.JsonIgnore;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.annotation.Undocumented;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;
import java.util.function.Consumer;

/**
 * <p>Collection of metadata for a specific User's view of a Database. Provided by a {@code DatabaseProvider}.</p>
 *
 * <p>{@link VisibleDatabaseTree} defines the <b>visible metadata and permissions of a User in a Database</b> in the system.
 * It defines the metadata of the Database visible to the User, the Resources visible to the User,
 * the permitted operations on those Resources, and any Locks on the Database.</p>
 *
 * <p>A {@link DatabaseTree} and {@link DatabaseUserGrantSet} are provided as input to a {@code DatabaseProvider} to produce
 * a {@link VisibleDatabaseTree}.</p>
 */
@AutoJson
public class VisibleDatabaseTree  {

    public static final String VERSION_SEP = "#";

    public static final String LEGACY_MYSQL_STORAGE = "m";
    public static final String USER_FORM_ID_SUFFIX = "@users";

    /**
     * This database's id
     */
    private final ResourceId databaseId;

    /**
     * The id of the requesting user. The contents of the tree depends on
     * the permissions of the requesting user.
     */
    private final Optional<ResourceId> userId;

    /**
     * The monotonically-increasing version of the database tree. The version number is
     * incremented whenever a change is made that affects the database tree.
     */
    private final String version;

    /**
     * This database's human-readable label.
     */
    private final String label;
    private final String description;

    private final UserRef ownerRef;

    private final long billingAccountId;

    /**
     * The requesting user's assigned role in this database.
     */
    private final RoleAssignment role;

    @JsonIgnore
    private final EffectiveRole effectiveRole;


    /**
     * True if this database is suspended for billing reasons
     */
    private final boolean suspended;

    /**
     * Opaque string that provides information about where and how this database is stored. Used
     * to support data migration.
     */
    @Undocumented
    private final String storage;

    private final boolean publishedTemplate;


    // Set of visible Resources and Grants assigned to this User, and Locks on the database
    @JsonIgnore
    private final Map<ResourceId, Resource.Node> resourceNodeMap = new HashMap<>();


    /**
     * The set of resources (folders, forms, and subforms) that belong to this database.
     */
    private final Map<ResourceId, Resource> resources = new HashMap<>();

    /**
     * The permission grants that have been made to the requesting user for this database.
     */
    private final Map<ResourceId, DatabaseUserGrant> grants = new HashMap<>();

    /**
     * The record locks that have been defined on this database.
     */
    private final Multimap<ResourceId, RecordLock> locks = HashMultimap.create();

    /**
     * The roles that have been defined for this database.
     */
    private final List<Role> roles;

    /**
     * The security categories that have been defined for this database.
     */
    private final List<SecurityCategory> securityCategories;

    @JsonIgnore
    private final Resource userForm;

    @Builder
    public VisibleDatabaseTree(ResourceId databaseId,
                               Optional<ResourceId> userId,
                               String version,
                               String label,
                               String description,
                               UserRef ownerRef,
                               boolean suspended,
                               String storage,
                               RoleAssignment role,
                               List<Role> roles,
                               List<SecurityCategory> securityCategories,
                               Collection<Resource> resources,
                               List<RecordLock> locks,
                               Collection<DatabaseUserGrant> grants,
                               long billingAccountId,
                               boolean publishedTemplate) {
        this.databaseId = databaseId;
        this.userId = userId;
        this.version = version;
        this.label = label;
        this.description = description;
        this.suspended = suspended;
        this.storage = storage;
        this.role = role;
        this.roles = roles;
        this.securityCategories = securityCategories;
        this.ownerRef = ownerRef;
        this.billingAccountId = billingAccountId;
        this.publishedTemplate = publishedTemplate;

        for (Resource resource : resources) {
            this.resources.put(resource.getId(), resource);
        }
        for (RecordLock lock : locks) {
            this.locks.put(lock.getResourceId(), lock);
        }

        // Add grants dicted by role
        this.effectiveRole = EffectiveRole.compute(userId, roles, role);
        for (ResourceId roleResource : effectiveRole.getResources()) {
            this.grants.put(roleResource, effectiveRole.getResolvedGrant(roleResource));
        }

        for (DatabaseUserGrant grant : grants) {
            this.grants.put(grant.getResourceId(), effectiveRole.resolve(grant));
        }

        userForm = new Resource.Builder()
            .setId(getUserFormId())
            .setParentId(databaseId)
            .setLabel("Users")
            .setType(ResourceType.FORM)
            .setVisibility(Resource.Visibility.REFERENCE)
            .build();
    }

    public long getBillingAccountId() {
        return billingAccountId;
    }

    public static boolean isUserForm(ResourceId formId) {
        return formId.asString().endsWith(USER_FORM_ID_SUFFIX);
    }

    public static ResourceId getDatabaseIdFromUserFormId(ResourceId formId) {
        String idString = formId.asString();
        if(idString.endsWith(USER_FORM_ID_SUFFIX)) {
            return ResourceId.valueOf(idString.substring(0, idString.length() - USER_FORM_ID_SUFFIX.length()));
        } else {
            return formId;
        }
    }

    public static ResourceId getUserFormIdFromDatabaseId(ResourceId databaseId) {
        return ResourceId.valueOf(databaseId.asString() + USER_FORM_ID_SUFFIX);
    }

    public ResourceId getDatabaseId() {
        return databaseId;
    }

    public Optional<ResourceId> getUserId() {
        return userId;
    }

    public String getVersion() {
        return version;
    }

    public long getUserVersion() {
        return isOwner()
            ? Long.valueOf(version)
            : Long.valueOf(version.substring(version.indexOf(VERSION_SEP)+1, version.length()));
    }

    public String getLabel() {
        return label;
    }

    public String getDescription() {
        return description;
    }

    public UserRef getOwnerRef() {
        return ownerRef;
    }

    public boolean isSuspended() {
        return suspended;
    }

    public boolean isPublishedTemplate() {
        return publishedTemplate;
    }

    public ResourceId getUserFormId() {
        return ResourceId.valueOf(databaseId.asString() + "@users");
    }

    public String getStorage() {
        return storage;
    }

    public Collection<Resource> getResources() {
        return resources.values();
    }

    public boolean hasResource(ResourceId resourceId) {
        return resourceId.equals(userForm.getId()) ||
            resources.containsKey(resourceId);
    }

    public Optional<Resource> getResource(ResourceId resourceId) {
        if(resourceId.equals(userForm.getId())) {
            return Optional.of(userForm);
        }
        return Optional.ofNullable(resources.get(resourceId));
    }

    public Map<ResourceId, Resource.Node> getResourceNodeMap() {
        return resourceNodeMap;
    }

    public Collection<DatabaseUserGrant> getGrants() {
        return grants.values();
    }

    public List<Role> getRoles() {
        return roles;
    }

    public boolean hasGrant(ResourceId resourceId) {
        return grants.containsKey(resourceId);
    }

    public Optional<DatabaseUserGrant> getGrant(ResourceId resourceId) {
        return Optional.ofNullable(grants.get(resourceId));
    }

    public Collection<RecordLock> getLocks() {
        return locks.values();
    }

    public RecordLockSet getEffectiveLocks(ResourceId resourceId) {
        List<RecordLock> effective = new ArrayList<>();
        do {
            effective.addAll(this.locks.get(resourceId));
            Resource resource = resources.get(resourceId);
            if(resource == null) {
                break;
            }
            Preconditions.checkState(!resource.getId().equals(resource.getParentId()));
            resourceId = resource.getParentId();
        } while(true);

        return new RecordLockSet(effective);
    }

    public void visitDepthFirst(Consumer<Resource> visitor) {
        for (Resource.Node node : getResourceNodeMap().values()) {
            if (node.isRoot()) {
                visitDepthFirstRecursive(node.getResource(), visitor);
            }
        }
    }

    private void visitDepthFirstRecursive(Resource parent, Consumer<Resource> visitor) {
        visitor.accept(parent);
        for (Resource.Node node : getResourceNodeMap().values()) {
            if (!node.isRoot() && node.getResource().getParentId().equals(parent.getId())) {
                visitDepthFirstRecursive(node.getResource(), visitor);
            }
        }
    }

    /**
     * Return a list starting with the given {@code resourceId}, and climbing to the database
     * root.
     *
     * @param resourceId
     */
    public List<ResourceId> getAncestors(ResourceId resourceId) {
        List<ResourceId> ancestors = new ArrayList<>();
        ancestors.add(resourceId);
        while (true) {
            Optional<Resource> resource = getResource(resourceId);
            if (!resource.isPresent()) {
                break;
            }
            resourceId = resource.get().getParentId();
            ancestors.add(resourceId);
        }
        return ancestors;
    }

    public Set<Resource> getDescendants(ResourceId resourceId) {
        Set<Resource> children = new HashSet<>();
        Multimap<ResourceId, Resource> resourceParentIdMap = getResourceParentIdMap();
        Set<Resource> firstLevelChildren = new HashSet<>();
        firstLevelChildren.addAll(resourceParentIdMap.get(resourceId));
        children.addAll(firstLevelChildren);
        while(true) {
            firstLevelChildren = collectNextLevelChildren(resourceParentIdMap, firstLevelChildren);
            if (firstLevelChildren.isEmpty()) {
                break;
            }
            children.addAll(firstLevelChildren);
        }
        return children;
    }

    public Set<Resource> getSelfAndDescendants(ResourceId resourceId) {
        Set<Resource> set = getDescendants(resourceId);
        getResource(resourceId).ifPresent(set::add);
        return set;
    }

    private Set<Resource> collectNextLevelChildren(Multimap<ResourceId, Resource> resourceParentIdMap, Set<Resource> firstLevelChildren) {
        Set<Resource> nextLevelChildren = new HashSet<>();
        for (Resource resource : firstLevelChildren) {
            nextLevelChildren.addAll(resourceParentIdMap.get(resource.getId()));
        }
        return nextLevelChildren;
    }

    private Multimap<ResourceId, Resource> getResourceParentIdMap() {
        HashMultimap<ResourceId, Resource> resourceMap = HashMultimap.create();
        for (Resource resource : resources.values()) {
            resourceMap.put(resource.getParentId(), resource);
        }
        return resourceMap;
    }

    public ResourceId getOwnerId() {
        return ownerRef.getId();
    }

    public String getOwnerName() {
        return ownerRef.getName();
    }

    public String getOwnerEmail() {
        return ownerRef.getEmail();
    }

    public boolean isLegacyStorage() {
        return LEGACY_MYSQL_STORAGE.equals(getStorage());
    }

    public RoleAssignment getRole() {
        return role;
    }

    @Override
    public int hashCode() {
        return Objects.hash(databaseId, userId, version);
    }

    @Override
    public boolean equals(Object obj) {
        if(this == obj) {
            return true;
        }
        if(!(obj instanceof VisibleDatabaseTree)) {
            return false;
        }
        VisibleDatabaseTree that = (VisibleDatabaseTree) obj;
        return this.databaseId.equals(that.databaseId) &&
               this.userId.equals(that.userId) &&
               this.version.equals(that.version);
    }

    @Override
    public String toString() {
        return "VisibleDatabaseTree{id=" + databaseId + "}";
    }

    public JsonValue toJson() {
        return VisibleDatabaseTreeJson.toJson(this);
    }

    public List<SecurityCategory> getSecurityCategories() {
        return securityCategories;
    }

    public boolean isOwner() {
        return userId.map(uId -> uId.equals(ownerRef.getId())).orElse(false);
    }
}
