/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.form;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import org.activityinfo.json.*;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.subform.SubFormReferenceType;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.activityinfo.json.Json.createObject;

/**
 * The FormClass defines structure and semantics for {@code Resource}s.
 * <p/>
 * {@code Resources} which fulfill the contract described by a {@code FormClass}
 * are called {@code FormInstances}.
 */
public class FormClass implements JsonSerializable {

    public static final String CLASSIC_STORAGE = "classic";

    public static final ResourceId PARENT_FIELD_ID = ResourceId.valueOf("@parent");


    @Nonnull
    private ResourceId id;

    private ResourceId databaseId;

    private String label;
    private String description;
    private final List<FormField> elements;

    private ResourceId parentFormId = null;
    private SubFormKind subFormKind = null;

    private long schemaVersion;

    private String storage;

    public FormClass(ResourceId id) {
        Preconditions.checkNotNull(id);
        this.id = id;
        this.schemaVersion = 1L;
        this.elements = new ArrayList<>();
    }

    public FormClass(FormClass toCopy) {
        this.id = toCopy.id;
        this.databaseId = toCopy.databaseId;
        this.label = toCopy.label;
        this.description = toCopy.description;
        this.parentFormId = toCopy.parentFormId;
        this.subFormKind = toCopy.subFormKind;
        this.schemaVersion = toCopy.schemaVersion;
        this.elements = new ArrayList<>(toCopy.elements);
        this.storage = toCopy.storage;
    }

    public Set<String> findReferencedFormIds() {
        return elements
            .stream()
            .flatMap(f -> {
                if (f.getType() instanceof ReferenceType) {
                    return ((ReferenceType) f.getType()).getRange().stream();
                } else {
                    return Stream.empty();
                }
            })
            .map(id -> id.asString())
            .collect(Collectors.toSet());
    }

    public ResourceId getDatabaseId() {
        return databaseId;
    }

    public FormClass setDatabaseId(ResourceId databaseId) {
        this.databaseId = databaseId;
        return this;
    }

    public void removeField(ResourceId formElementId) {
        elements.removeIf(e -> e.getId().equals(formElementId));
    }

    public ResourceId getId() {
        return id;
    }

    public void setId(ResourceId id) {
        this.id = id;
    }

    public String getLabel() {
        return label;
    }

    public FormClass setLabel(String label) {
        this.label = label;
        return this;
    }

    /**
     * @return the version of this form schema
     */
    public long getSchemaVersion() {
        return schemaVersion;
    }

    public FormClass setSchemaVersion(long schemaVersion) {
        this.schemaVersion = schemaVersion;
        return this;
    }

    public String getDescription() {
        return description;
    }

    public FormClass setDescription(String description) {
        this.description = description;
        return this;
    }

    public List<FormField> getElements() {
        return elements;
    }

    public List<FormField> getFields() {
        return elements;
    }

    public FormField getField(ResourceId fieldId) {
        if (fieldId.equals(PARENT_FIELD_ID)) {
            return getParentField().get();
        }
        Optional<FormField> field = getFieldIfPresent(fieldId);
        if(field.isPresent()) {
            return field.get();
        } else {
            throw new IllegalArgumentException("No such field: " + fieldId);
        }
    }

    public Optional<FormField> getFieldIfPresent(ResourceId fieldId) {
        for (FormField field : getFields()) {
            if (field.getId().equals(fieldId)) {
                return Optional.of(field);
            }
        }
        return Optional.empty();
    }


    public FormClass addElement(FormField element) {
        elements.add(element);
        return this;
    }

    public FormField addField() {
        return addField(ResourceId.generateId());
    }


    public FormField addField(ResourceId fieldId) {
        FormField field = new FormField(fieldId);
        elements.add(field);
        return field;
    }

    public FormClass insertElement(int index, FormField element) {
        elements.add(index, element);
        return this;
    }


    public FormClass insertElements(int index, Collection<FormField> elements) {
        this.elements.addAll(index, elements);
        return this;
    }


    public void forEachRelatedForm(Consumer<ResourceId> relatedFormConsumer) {
        if(parentFormId != null) {
            relatedFormConsumer.accept(parentFormId);
        }
        for (FormField field : elements) {
            if(field.getType() instanceof ReferenceType) {
                ReferenceType type = (ReferenceType) field.getType();
                for (ResourceId formId : type.getRange()) {
                    relatedFormConsumer.accept(formId);
                }
            } else if(field.getType() instanceof SubFormReferenceType) {
                SubFormReferenceType type = (SubFormReferenceType) field.getType();
                relatedFormConsumer.accept(type.getClassId());
            }
        }
    }

    public java.util.Optional<ResourceId> getParentFormId() {
        return java.util.Optional.ofNullable(parentFormId);
    }

    /**
     * Returns a {@code FormField} of type {@code ReferenceType} that will refer to the master submission
     * of a subform.
     * 
     * <p>If this {@code FormClass} is a sub form of another form, then each of its {@code FormSubmissions} will
     * have a {@code @parent} field value that refers back to the master form instance. There is no corresponding
     * {@code FormField} instance in this {@code FormClass}'s field list, because it is not a user-editable field.</p>
     */
    public java.util.Optional<FormField> getParentField() {
        return getParentFormId().map(parentFormId -> {
            FormField formField = new FormField(PARENT_FIELD_ID);
            formField.setLabel("Parent");
            formField.setType(ReferenceType.single(parentFormId));
            formField.setRequired(true);
            return formField;
        });
    }

    public FormClass setParentFormId(ResourceId parentFormId) {
        this.parentFormId = parentFormId;
        return this;
    }
    
    public boolean isSubForm() {
        return parentFormId != null;
    }

    public SubFormKind getSubFormKind() {
        if(!isSubForm()) {
            throw new IllegalStateException("Not a sub form");
        }
        if(subFormKind == null) {
            return SubFormKind.REPEATING;
        }
        return subFormKind;
    }

    public String getStorage() {
        return storage;
    }

    public void setStorage(String storage) {
        this.storage = storage;
    }

    public FormClass setSubFormKind(SubFormKind subFormKind) {
        this.subFormKind = subFormKind;
        return this;
    }

    @Override
    public String toString() {
        return "<FormClass: " + getLabel() + ">";
    }

    @Deprecated
    public JsonValue toJsonObject() {
        return toJson();
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = createObject();
        object.put("id", id.asString());
        object.put("schemaVersion", schemaVersion);
        
        if(databaseId != null) {
            object.put("databaseId", databaseId.asString());
        }
        object.put("label", label);

        if(storage != null) {
            object.put("storage", storage);
        }

        if(!Strings.isNullOrEmpty(description)) {
            object.put("description", description);
        }
        
        if(subFormKind != null) {
            object.put("parentFormId", parentFormId.asString());
            object.put("subFormKind", subFormKind.name().toLowerCase());
        }
        object.put("elements", JsonArrays.toJsonArray(elements, FormField::toJson));
        return object;
    }

    public FormClass copy() {
        return fromJson(toJson());
    }

    static JsonValue toJsonArray(Iterable<FormField> elements) {
        JsonValue elementsArray = Json.createArray();
        for (FormField element : elements) {
            elementsArray.add(element.toJson());
        }
        return elementsArray;
    }

    public String toJsonString() {
        return toJson().toJson();
    }

    public static FormClass fromJson(JsonValue object) {
        // Deal with previous encoding

        ResourceId id;
        if(object.hasKey("@id")) {
            id = ResourceId.valueOf(object.getString("@id"));
        } else {
            id = ResourceId.valueOf(object.getString("id"));
        }
        
        FormClass formClass = new FormClass(id);

        if(object.hasKey("schemaVersion")) {
            formClass.setSchemaVersion(object.get("schemaVersion").asLong());
        }

        if(object.hasKey("databaseId")) {
            formClass.setDatabaseId(ResourceId.valueOf(object.get("databaseId").asString()));
        }
        
        if(object.hasKey("_class_label")) {
            formClass.setLabel(JsonParsing.toNullableString(object.get("_class_label")));
        } else {
            formClass.setLabel(JsonParsing.toNullableString(object.get("label")));
        }

        if(object.hasKey("subFormKind")) {
            formClass.setSubFormKind(SubFormKind.valueOf(object.get("subFormKind").asString().toUpperCase()));
            formClass.setParentFormId(ResourceId.valueOf(object.get("parentFormId").asString()));
        }
        if(formClass.getParentFormId().isPresent() && formClass.subFormKind == null) {
            formClass.setSubFormKind(SubFormKind.REPEATING);
        }

        formClass.setStorage(object.getString("storage"));

        if(object.hasKey("elements")) {
            JsonValue elements = object.get("elements");
            if(elements.isJsonArray()) {
                JsonValue elementsArray = elements;
                formClass.elements.addAll(fromJsonArray(elementsArray));
            }
        }
        return formClass;
    }

    static List<FormField> fromJsonArray(JsonValue elementsArray) {
        List<FormField> elements = new ArrayList<>();
        for (int i = 0; i < elementsArray.length(); i++) {
            JsonValue elementObject = elementsArray.get(i);
            elements.add(FormField.fromJson(elementObject));
        }
        return elements;
    }

    public static FormClass fromJson(String json) {
        return fromJson(new JsonParser().parse(json));
    }

}
