/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.form;

import com.google.common.base.Strings;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.*;
import org.activityinfo.model.type.expr.CalculatedFieldType;

import javax.annotation.Nonnull;
import java.util.Objects;
import java.util.Optional;

import static com.google.common.base.Preconditions.checkNotNull;
import static org.activityinfo.json.Json.createObject;

/**
 * The smallest logical unit of data entry.
 */
public final class FormField implements JsonSerializable {

    private final ResourceId id;
    private String code;
    private String label;
    private String description;
    private String relevanceCondition;
    private String validationCondition;
    private FieldType type;
    private boolean dataEntryVisible = true;
    private boolean tableVisible = true;
    private boolean required;
    private boolean key;
    private Optional<ResourceId> securityCategoryId = Optional.empty();

    public FormField(ResourceId id) {
        checkNotNull(id);
        this.id = id;
    }

    public FormField(FormField toCopy) {
        this.id = toCopy.getId();
        this.code = toCopy.getCode();
        this.label = toCopy.getLabel();
        this.description = toCopy.getDescription();
        this.relevanceCondition = toCopy.getRelevanceCondition();
        this.validationCondition = toCopy.getValidationCondition();
        this.type = toCopy.getType();
        this.dataEntryVisible = toCopy.isDataEntryVisible();
        this.tableVisible = toCopy.isTableVisible();
        this.required = toCopy.isRequired();
        this.key = toCopy.isKey();
        this.securityCategoryId = toCopy.securityCategoryId;
    }

    public ResourceId getId() {
        return id;
    }

    public String getName() {
        return id.asString();
    }

    /**
     * @return user-assigned code for this field that can be
     * used in expressions.
     */
    public String getCode() {
        return code;
    }


    public FormField setCode(String code) {
        this.code = Strings.emptyToNull(code);
        return this;
    }

    public boolean hasCode() {
        return code != null;
    }

    @Nonnull
    public String getLabel() {
        return label;
    }

    public FormField setLabel(String label) {
        assert label != null;
        this.label = label;
        return this;
    }

    public boolean hasRelevanceCondition() {
        return !Strings.isNullOrEmpty(relevanceCondition);
    }

    public boolean hasValidationCondition() {
        return !Strings.isNullOrEmpty(validationCondition);
    }

    public String getRelevanceCondition() {
        return relevanceCondition;
    }

    public String getValidationCondition() {
        return validationCondition;
    }

    public FormField setRelevanceCondition(String relevanceCondition) {
        this.relevanceCondition = relevanceCondition;
        return this;
    }

    public FormField setValidationCondition(String validationCondition) {
        this.validationCondition = validationCondition;
        return this;
    }

    /**
     * @return an extended description of this field, presented to be
     * presented to the user during data entry
     */
    public String getDescription() {
        return description;
    }

    public FormField setDescription(String description) {
        this.description = description;
        return this;
    }

    public FieldType getType() {
        assert type != null : "type is missing for " + id;
        return type;
    }

    public FormField setType(FieldType type) {
        this.type = type;
        return this;
    }

    public boolean isKey() {
        return key;
    }

    public FormField setKey(boolean key) {
        this.key = key;
        return this;
    }

    /**
     *
     * @return true if this field requires a response before submitting the form
     */
    public boolean isRequired() {
        return required;
    }

    public FormField setRequired(boolean required) {
        this.required = required;
        return this;
    }

    public boolean isDataEntryVisible() {
        return dataEntryVisible;
    }

    public boolean isTableVisible() {
        return tableVisible;
    }

    public void setDataEntryVisible(boolean dataEntryVisible) {
        this.dataEntryVisible = dataEntryVisible;
    }

    public void setTableVisible(boolean tableVisible) {
        this.tableVisible = tableVisible;
    }

    public Optional<ResourceId> getSecurityCategoryId() {
        return securityCategoryId;
    }

    public void setSecurityCategoryId(Optional<ResourceId> securityCategoryId) {
        this.securityCategoryId = securityCategoryId;
    }

    public boolean isSecurityCategory(ResourceId securityCategoryId) {
        if (!this.securityCategoryId.isPresent()) {
            return false;
        } else {
            return this.securityCategoryId.get().equals(securityCategoryId);
        }
    }

    @Override
    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod", "OverlyComplexBooleanExpression"})
    public boolean equals(Object o) {
        if (this == o) return true;
        if (!(o instanceof FormField)) return false;
        FormField field = (FormField) o;
        return dataEntryVisible == field.dataEntryVisible &&
                tableVisible == field.tableVisible &&
                required == field.required &&
                key == field.key &&
                Objects.equals(id, field.id) &&
                Objects.equals(code, field.code) &&
                Objects.equals(label, field.label) &&
                Objects.equals(description, field.description) &&
                Objects.equals(relevanceCondition, field.relevanceCondition) &&
                Objects.equals(validationCondition, field.validationCondition) &&
                Objects.equals(type, field.type) &&
                Objects.equals(securityCategoryId, field.securityCategoryId);
    }

    @Override
    public int hashCode() {
        return Objects.hash(id, code, label, description, relevanceCondition, validationCondition,
                type, tableVisible, dataEntryVisible, required, key, securityCategoryId);
    }

    @Override
    public String toString() {
        return "FormField{" +
                "id=" + id +
                ", label=" + label +
                ", type=" + type.getTypeClass().getId() +
                '}';
    }

    public static class Builder {
        private FormField field;

        public Builder(ResourceId fieldId) {
            this.field = new FormField(fieldId);
        }

        public Builder(FormField toCopy) {
            this.field = new FormField(toCopy.getId());
            this.field.code = toCopy.code;
            this.field.label = toCopy.label;
            this.field.description = toCopy.description;
            this.field.key = toCopy.key;
            this.field.required = toCopy.required;
            this.field.type = toCopy.type;
            this.field.dataEntryVisible = toCopy.dataEntryVisible;
            this.field.tableVisible = toCopy.tableVisible;
            this.field.relevanceCondition = toCopy.relevanceCondition;
            this.field.validationCondition = toCopy.validationCondition;
            this.field.securityCategoryId = toCopy.securityCategoryId;
        }

        public Builder setCode(String code) {
            this.field.code = Strings.emptyToNull(code);
            return this;
        }

        public Builder setLabel(String label) {
            this.field.label = label;
            return this;
        }

        public Builder setDescription(String description) {
            this.field.description = description;
            return this;
        }

        public Builder setKey(boolean key) {
            this.field.key = key;
            return this;
        }

        public Builder setDataEntryVisible(boolean visible) {
            this.field.dataEntryVisible = visible;
            if(!visible) {
                this.field.required = false;
            }
            return this;
        }

        public Builder setTableVisible(boolean visible) {
            this.field.tableVisible = visible;
            return this;
        }

        public Builder setRequired(boolean required) {
            this.field.required = required;
            if(required) {
                this.field.dataEntryVisible = true;
            }
            return this;
        }

        public Builder setRelevanceCondition(String relevanceCondition) {
            this.field.relevanceCondition = relevanceCondition;
            return this;
        }

        public Builder setValidationCondition(String validationCondition) {
            this.field.validationCondition = validationCondition;
            return this;
        }

        public Builder setType(FieldType type) {
            this.field.type = type;
            return this;
        }

        public Builder setSecurityCategoryId(Optional<ResourceId> securityCategoryId) {
            this.field.securityCategoryId = securityCategoryId;
            return this;
        }

        public FormField build() {
            return this.field;
        }
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = createObject();
        object.put("id", id.asString());
        object.put("code", code);
        object.put("label", label);
        object.put("description", description);
        object.put("relevanceCondition", relevanceCondition);
        object.put("validationCondition", validationCondition);

        if(!dataEntryVisible) {
            object.put("dataEntryVisible", dataEntryVisible);
        }
        if(!tableVisible) {
            object.put("tableVisible", tableVisible);
        }
        object.put("required", required);

        object.put("type", type.getTypeClass().getId());

        if(key) {
            object.put("key", true);
        }

        securityCategoryId.ifPresent(resourceId -> {
            object.put("securityCategoryId", resourceId.asString());
        });

        if(type instanceof ParametrizedFieldType) {
            object.put("typeParameters", ((ParametrizedFieldType) type).getParametersAsJson());
        }
        
        return object;
    }


    public static FormField fromJson(JsonValue jsonObject) {
        FormField field = new FormField(ResourceId.valueOf(jsonObject.get("id").asString()));
        field.setLabel(Strings.nullToEmpty(jsonObject.getString("label")));
        field.setCode(jsonObject.getString("code"));
        field.setDescription(jsonObject.getString("description"));

        if(jsonObject.hasKey("relevanceCondition")) {
            field.setRelevanceCondition(jsonObject.getString("relevanceCondition"));
        } else if(jsonObject.hasKey("relevanceConditionExpression")) {
            field.setRelevanceCondition(jsonObject.getString("relevanceConditionExpression"));
        }

        if(jsonObject.hasKey("validationCondition")) {
            field.setValidationCondition(jsonObject.getString("validationCondition"));
        }

        if(jsonObject.hasKey("required")) {
            field.setRequired(jsonObject.getBoolean("required"));
        }

        if(jsonObject.hasKey("key")) {
            field.setKey(jsonObject.getBoolean("key"));
        }

        if (jsonObject.hasKey("securityCategoryId")) {
            Optional<ResourceId> securityCategoryId = Optional.of(
                    ResourceId.valueOf(jsonObject.getString("securityCategoryId")));
            field.setSecurityCategoryId(securityCategoryId);
        }

        readFieldType(jsonObject, field);
        readVisibility(jsonObject, field);

        return field;
    }

    private static void readVisibility(JsonValue jsonObject, FormField field) {
        if(jsonObject.hasKey("visible")) {
            boolean legacyVisible = jsonObject.getBoolean("visible");
            if (!legacyVisible) {
                field.setDataEntryVisible(false);
                field.setTableVisible(false);
            } else if (field.getType() instanceof CalculatedFieldType) {
                field.setDataEntryVisible(false);
            }
        }
        if(jsonObject.hasKey("dataEntryVisible")) {
            field.setDataEntryVisible(jsonObject.getBoolean("dataEntryVisible"));
        }
        if(jsonObject.hasKey("tableVisible")) {
            field.setTableVisible(jsonObject.getBoolean("tableVisible"));
        }
    }

    private static void readFieldType(JsonValue jsonObject, FormField field) {
        String type;
        JsonValue typeParameters;
        JsonValue typeElement = jsonObject.get("type");

        if(typeElement.isJsonPrimitive()) {
            type = typeElement.asString();
            typeParameters = jsonObject.get("typeParameters");
        } else {
            JsonValue typeObject = typeElement;
            type = typeObject.get("typeClass").asString();
            typeParameters = typeObject.get("parameters");
        }

        FieldTypeClass typeClass = TypeRegistry.get().getTypeClass(type);
        if(typeClass instanceof ParametrizedFieldTypeClass && !typeParameters.isJsonNull()) {
            field.setType(((ParametrizedFieldTypeClass) typeClass).deserializeType(typeParameters));
        } else {
            field.setType(typeClass.createType());
        }
    }


}
