/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.form;

import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.permission.FormPermissions;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.promise.Maybe;

import java.util.Collections;
import java.util.List;
import java.util.Optional;

import static org.activityinfo.json.Json.createObject;

/**
 * Provides user-specific metadata for a given form, including permissions
 * and versioning.
 */
public class FormMetadata implements JsonSerializable {

    ResourceId id;


    /**
     * The version of the Schema. The version number is incremented
     * whenever a change is made to the FormClass.
     */
    long schemaVersion;

    /**
     * True if this form has been deleted.
     */
    boolean deleted = false;

    /**
     * The permissions for the current user.
     */
    FormPermissions permissions;

    FormClass schema;

    boolean visible = true;

    private boolean suspended;

    public static FormMetadata notFound(ResourceId formId) {
        FormMetadata metadata = new FormMetadata();
        metadata.id = formId;
        metadata.visible = false;
        metadata.deleted = true;
        metadata.permissions = FormPermissions.none();
        return metadata;
    }

    public static FormMetadata forbidden(ResourceId formId) {
        FormMetadata metadata = new FormMetadata();
        metadata.id = formId;
        metadata.visible = false;
        metadata.permissions = FormPermissions.none();
        return metadata;
    }

    public static FormMetadata of(FormClass schema, FormPermissions permissions) {
        FormMetadata metadata = new FormMetadata();
        metadata.id = schema.getId();
        metadata.visible = true;
        metadata.permissions = permissions;
        metadata.schema = schema;
        metadata.schemaVersion = schema.getSchemaVersion();
        return metadata;
    }

    public ResourceId getId() {
        return id;
    }

    public long getSchemaVersion() {
        return schemaVersion;
    }

    public FormPermissions getPermissions() {
        return permissions;
    }

    public boolean isVisible() {
        return visible;
    }

    public boolean isDeleted() {
        return deleted;
    }

    public boolean isAccessible() {
        return visible && !deleted;
    }

    public boolean isSuspended() {
        return suspended;
    }

    public FormClass getSchema() {
        assert visible : "form is not visible to user";
        assert !deleted : "form has been deleted.";
        return schema;
    }

    public Optional<ResourceId> getDatabaseId() {
        if(schema != null) {
            return Optional.of(schema.getDatabaseId());
        } else {
            return Optional.empty();
        }
    }

    public Maybe<FormClass> getSchemaIfVisible() {
        if(isVisible()) {
            return Maybe.of(getSchema());
        } else if(!isAccessible()) {
            return Maybe.forbidden();
        } else {
            return Maybe.notFound();
        }
    }

    public List<FormField> getFields() {
        if(isVisible()) {
            return getSchema().getFields();
        } else {
            return Collections.emptyList();
        }
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = createObject();
        object.put("id", id.asString());
        if(!visible) {
            object.put("visible", false);
        }
        if(deleted) {
            object.put("deleted", true);
        }
        if(schema != null) {
            object.put("schema", schema.toJson());
        }
        if(visible) {
            object.put("schemaVersion", schemaVersion);
            object.put("permissions", permissions.toJson());
        }
        if(suspended) {
            object.put("suspended", suspended);
        }
        return object;
    }

    public static FormMetadata fromJson(JsonValue object) {
        FormMetadata metadata = new FormMetadata();
        metadata.id = ResourceId.valueOf(object.get("id").asString());

        if(object.hasKey("schemaVersion")) {
            metadata.schemaVersion = object.get("schemaVersion").asLong();
        }
        if(object.hasKey("schema")) {
            metadata.schema = FormClass.fromJson(object.get("schema"));
        }
        if(object.hasKey("permissions")) {
            metadata.permissions = FormPermissions.fromJson(object.get("permissions"));
        }
        if(object.hasKey("visible")) {
            metadata.visible = object.get("visible").asBoolean();
        }
        if(object.hasKey("deleted")) {
            metadata.deleted = object.get("deleted").asBoolean();
        }
        if(object.hasKey("suspended")) {
            metadata.suspended = object.getBoolean("suspended");
        }
        return metadata;
    }


    public static class Builder {
        private FormMetadata meta = new FormMetadata();

        public Builder setId(ResourceId id) {
            meta.id = id;
            return this;
        }

        public Builder setSchema(FormClass schema) {
            meta.schema = schema;
            meta.schemaVersion = schema.getSchemaVersion();
            return this;
        }

        public Builder setPermissions(FormPermissions permissions) {
            meta.permissions = permissions;
            return this;
        }

        public Builder setVersion(long version) {
            return this;
        }

        public Builder setSuspended(boolean suspended) {
            meta.suspended = suspended;
            return this;
        }

        public FormMetadata build() {
            return meta;
        }

    }

}
