/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.form;

import org.activityinfo.json.JsonParser;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.RecordRef;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static org.activityinfo.json.Json.createObject;

public class FormRecord implements JsonSerializable {
    
    private String recordId;
    private String formId;
    private String parentRecordId;
    private double lastEditTime;
    private JsonValue fields;

    private FormRecord() {
    }

    public FormRecord(RecordRef ref, String parentRecordId, double lastEditTime, JsonValue fields) {
        this.formId = ref.getFormId().asString();
        this.recordId = ref.getRecordId().asString();
        this.parentRecordId = parentRecordId;
        this.lastEditTime = lastEditTime;
        this.fields = fields;
    }

    public String getRecordId() {
        return recordId;
    }

    public String getFormId() {
        return formId;
    }

    public RecordRef getRef() {
        return new RecordRef(ResourceId.valueOf(formId), ResourceId.valueOf(recordId));
    }

    public String getParentRecordId() {
        return parentRecordId;
    }

    public JsonValue getFields() {
        return fields;
    }

    public double getLastEditTime() {
        return lastEditTime;
    }

    public static FormRecord fromJson(String json) {
        JsonParser parser = new org.activityinfo.json.JsonParser();
        return fromJson(parser.parse(json));
    }
    
    public static FormRecord fromJson(JsonValue element) {
        JsonValue jsonObject = element;

        FormRecord formRecord = new FormRecord();
        formRecord.recordId = jsonObject.get("recordId").asString();
        formRecord.formId = jsonObject.get("formId").asString();
        formRecord.lastEditTime = jsonObject.getNumber("lastEditTime"); //is there some sort of schema/doc to be edited

        if(jsonObject.hasKey("parentRecordId")) {
            formRecord.parentRecordId = jsonObject.get("parentRecordId").asString();
        }

        formRecord.fields = jsonObject.get("fields");

        return formRecord;
    }

    @Override
    public JsonValue toJson() {
        
        assert recordId != null;
        assert formId != null;

        JsonValue jsonObject = createObject();
        jsonObject.put("recordId", recordId);
        jsonObject.put("formId", formId);
        jsonObject.put("lastEditTime", lastEditTime);
        if(parentRecordId != null) {
            jsonObject.put("parentRecordId", parentRecordId);
        }
        jsonObject.put("fields", fields);
        return jsonObject;
    }
    
    public static List<FormRecord> fromJsonArray(JsonValue array) {
        List<FormRecord> list = new ArrayList<>();
        for (JsonValue jsonElement : array.values()) {
            list.add(fromJson(jsonElement));
        }
        return list;
    }

    public static FormRecord fromInstance(TypedFormRecord typedRecord) {
        FormRecord record = new FormRecord();
        record.parentRecordId = typedRecord.getParentRecordId().map(ResourceId::asString).orElse(null);
        record.recordId = typedRecord.getId().asString();
        record.formId = typedRecord.getFormId().asString();
        record.lastEditTime = typedRecord.getLastEditTime();
        record.fields = createObject();

        for (Map.Entry<ResourceId, FieldValue> entry : typedRecord.getFieldValueMap().entrySet()) {
            String field = entry.getKey().asString();
            if(!field.equals("classId")) {
                if(entry.getValue() != null) {
                    record.fields.put(field, entry.getValue().toJson());
                }
            }
        }
        return record;
    }

    public static Builder builder() {
        return new Builder();
    }


    public static class Builder {
        
        private FormRecord record;

        public Builder() {
            record = new FormRecord();
            record.fields = createObject();
        }

        public Builder setRecordId(ResourceId id) {
            this.record.recordId = id.asString();
            return this;
        }
        
        public Builder setFormId(ResourceId id) {
            this.record.formId = id.asString();
            return this;
        }

        public Builder setLastEditTime(double lastEditTime) {
            this.record.lastEditTime = lastEditTime;
            return this;
        }
        
        public Builder setParentRecordId(ResourceId id) {
            this.record.parentRecordId = id.asString();
            return this;
        }

        public Builder setFieldValue(ResourceId fieldId, String value) {
            if(value == null) {
                this.record.fields.remove(fieldId.asString());
            } else {
                this.record.fields.put(fieldId.asString(), value);
            }
            return this;
        }
        
        public Builder setFieldValue(ResourceId fieldId, FieldValue value) {
            if(value == null) {
                this.record.fields.remove(fieldId.asString());
            } else {
                this.record.fields.put(fieldId.asString(), value.toJson());
            }
            return this;
        }
        
        public FormRecord build() {
            assert record.recordId != null;
            assert record.formId != null;
            
            return record;
        }

    }
    
}
