/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.form;

import com.google.common.base.Preconditions;
import com.google.gwt.i18n.client.DateTimeFormat;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.resource.RecordUpdate;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldType;
import org.activityinfo.model.type.FieldTypeClass;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.model.type.attachment.Attachment;
import org.activityinfo.model.type.attachment.AttachmentRef;
import org.activityinfo.model.type.attachment.AttachmentValue;
import org.activityinfo.model.type.number.Quantity;
import org.activityinfo.model.type.primitive.HasStringValue;
import org.activityinfo.model.type.primitive.TextValue;
import org.activityinfo.model.type.time.LocalDate;

import javax.annotation.Nonnull;
import java.util.*;

/**
 *
 * Wrapper for a {@code Record} or {@code Resource} that exposes its properties
 * as {@code FieldValue}s
 *
 * @author yuriyz on 1/29/14.
 */
public class TypedFormRecord {

    private ResourceId id;
    private ResourceId formId;
    private Optional<RecordRef> parentRecordRef = Optional.empty();
    private double lastEditTime;

    private Map<ResourceId, FieldValue> fieldMap;

    /**
     * Constructs a new TypedFormRecord. To obtain an id for a new instance
     * use
     *
     * @param id the id of the instance.
     * @param formId the id of this form's class
     */
    public TypedFormRecord(@Nonnull ResourceId id, @Nonnull ResourceId formId) {
        Preconditions.checkNotNull(id);
        Preconditions.checkNotNull(formId);
        this.id = id;
        this.formId = formId;
        this.fieldMap = new HashMap<>();
        this.lastEditTime = Double.NaN;
    }

    public TypedFormRecord(RecordRef recordRef) {
        this(recordRef.getRecordId(), recordRef.getFormId());
    }

    public static TypedFormRecord toTypedFormRecord(FormClass formClass, FormRecord record) {
        TypedFormRecord typedRecord = new TypedFormRecord(ResourceId.valueOf(record.getRecordId()), formClass.getId());
        typedRecord.setLastEditTime(record.getLastEditTime());
        if (record.getParentRecordId() != null && formClass.getParentFormId().isPresent()) {
            ResourceId parentRecordId = ResourceId.valueOf(record.getParentRecordId());
            RecordRef parentRef = new RecordRef(formClass.getParentFormId().get(), parentRecordId);
            typedRecord.setParentRecordRef(parentRef);
        }

        JsonValue fieldsObject = record.getFields();
        for (FormField field : formClass.getFields()) {
            if(fieldsObject.hasKey(field.getName())) {
                JsonValue fieldValue = fieldsObject.get(field.getName());
                if (!fieldValue.isJsonNull()) {
                    typedRecord.set(field.getId(), field.getType().parseJsonValue(fieldValue));
                }
            }
        }
        return typedRecord;
    }

    public ResourceId getId() {
        return id;
    }

    public double getLastEditTime() {
        return lastEditTime;
    }

    public RecordRef getRef() {
        return new RecordRef(getFormId(), getId());
    }

    public TypedFormRecord setId(ResourceId id) {
        this.id = id;
        return this;
    }

    public TypedFormRecord setFormId(ResourceId formId) {
        this.formId = formId;
        return this;
    }

    public TypedFormRecord setLastEditTime(double lastEditTime) {
        this.lastEditTime = lastEditTime;
        return this;
    }

    public ResourceId getFormId() {
        return formId;
    }

    public void setParentRecordId(Optional<ResourceId> parentFormId, Optional<ResourceId> parentRecordId) {
        if(parentFormId.isPresent() && parentRecordId.isPresent()) {
            this.parentRecordRef = Optional.of(new RecordRef(parentFormId.get(), parentRecordId.get()));
        } else {
            this.parentRecordRef = Optional.empty();
        }
    }

    public Optional<ResourceId> getParentRecordId() {
        return parentRecordRef.map(ref -> ref.getRecordId());
    }

    public Optional<RecordRef> getParentRecordRef() {
        return parentRecordRef;
    }

    public void setParentRecordRef(Optional<RecordRef> parentRecordRef) {
        this.parentRecordRef = parentRecordRef;
    }

    public TypedFormRecord setParentRecordRef(RecordRef parentRef) {
        this.parentRecordRef = Optional.of(parentRef);
        return this;
    }

    public Map<ResourceId, FieldValue> getFieldMap() {
        return fieldMap;
    }

    public void setFieldMap(Map<ResourceId, FieldValue> fieldMap) {
        this.fieldMap = fieldMap;
    }

    public Map<ResourceId, FieldValue> getFieldValueMap() {
        return new HashMap<>(fieldMap);
    }

    public void removeAll(Set<ResourceId> fieldIds) {
        for (ResourceId fieldId : fieldIds) {
            fieldMap.remove(fieldId);
        }
    }

    public void setAll(Map<ResourceId, FieldValue> valueMap) {
        fieldMap.putAll(valueMap);
    }

    public TypedFormRecord set(@Nonnull ResourceId fieldId, String value) {
        if(value == null) {
            fieldMap.remove(fieldId);
        } else {
            fieldMap.put(fieldId, TextValue.valueOf(value));
        }
        return this;
    }

    public Optional<Attachment> findAttachment(ResourceId fieldId, String blobId) {
        FieldValue value = get(fieldId);
        if(!(value instanceof AttachmentValue)) {
            return Optional.empty();
        }

        AttachmentValue attachmentValue = (AttachmentValue) value;
        return attachmentValue.getValues()
                .stream()
                .filter(a -> a.getBlobId().equals(blobId))
                .findFirst();
    }


    public TypedFormRecord set(String fieldId, String value) {
        return set(ResourceId.valueOf(fieldId), value);
    }

    public TypedFormRecord set(String fieldId, double value) {
        return set(ResourceId.valueOf(fieldId), new Quantity(value));
    }

    public TypedFormRecord set(@Nonnull ResourceId fieldId, double value) {
        return set(fieldId, new Quantity(value));
    }

    public TypedFormRecord set(@Nonnull ResourceId fieldId, FieldValue fieldValue) {
        fieldMap.put(fieldId, fieldValue);
        return this;
    }

    public FieldValue get(ResourceId fieldId, FieldType fieldType) {
        FieldValue value = fieldMap.get(fieldId);
        if(value != null && value.getTypeClass() == fieldType.getTypeClass()) {
            return value;
        } else {
            return null;
        }
    }

    public FieldValue get(ResourceId fieldId) {
        return fieldMap.get(fieldId);
    }


    public Optional<FieldValue> getIfPresent(ResourceId fieldId) {
        return Optional.ofNullable(fieldMap.get(fieldId));
    }

    /**
     * Returns the value of {@code fieldId} if the value is present and of
     * the specified {@code typeClass}, or {@code null} otherwise.
     */
    public FieldValue get(ResourceId fieldId, FieldTypeClass typeClass) {
        FieldValue value = get(fieldId);
        if(value != null && value.getTypeClass() == typeClass) {
            return value;
        } else {
            return null;
        }
    }

    public String getString(ResourceId fieldId) {
        FieldValue value = fieldMap.get(fieldId);
        if(value instanceof HasStringValue) {
            return ((HasStringValue) value).asString();
        }
        return null;
    }

    public LocalDate getDate(ResourceId fieldId) {
        FieldValue value = get(fieldId);
        if (value instanceof LocalDate) {
            return (LocalDate) value;
        }
        return null;
    }


    public Double getDouble(ResourceId fieldId) {
        FieldValue value = get(fieldId);
        if(value instanceof Quantity) {
            return ((Quantity) value).getValue();
        }
        return null;
    }

    public TypedFormRecord copy() {
        final TypedFormRecord copy = new TypedFormRecord(getId(), getFormId());
        copy.parentRecordRef = this.parentRecordRef;
        copy.fieldMap.putAll(fieldMap);
        copy.lastEditTime = this.lastEditTime;
        return copy;
    }

    public int size() {
        return fieldMap.size();
    }

    @Override
    public String toString() {
        String dateStr =                 "lastEditDate=" + DateTimeFormat.
                getFormat(DateTimeFormat.
                        PredefinedFormat.DATE_TIME_FULL).format(new Date((long)lastEditTime));
        return "TypedFormRecord{" +
                "id=" + id +
                "lastEditTime" + dateStr +
                ", classId=" + formId +
                ", parentRecordId=" + parentRecordRef +
                ", fieldMap=" + fieldMap +
                '}';
    }

    public boolean isEmpty() {
        return fieldMap.isEmpty();
    }

    public boolean isEmpty(String... fieldsToIgnore) {
        Map<ResourceId, FieldValue> copy = new HashMap<>(fieldMap);
        for (String field : fieldsToIgnore) {
            copy.remove(ResourceId.valueOf(field));
        }
        return copy.isEmpty();
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        TypedFormRecord instance = (TypedFormRecord) o;

        return !(id != null ? !id.equals(instance.id) : instance.id != null);

    }

    @Override
    public int hashCode() {
        return id != null ? id.hashCode() : 0;
    }

    public RecordUpdate toRecordUpdate() {
        RecordUpdate update = new RecordUpdate();
        update.setFormId(getFormId());
        update.setRecordId(getRef().getRecordId());
        update.setParentRecordId(getParentRecordId().map(ResourceId::asString).orElse(null));

        for (Map.Entry<ResourceId, FieldValue> entry : fieldMap.entrySet()) {
            update.setFieldValue(entry.getKey(), entry.getValue());
        }
        return update;
    }


    public FormRecord toFormRecord() {
        FormRecord.Builder builder = new FormRecord.Builder();
        builder.setFormId(formId);
        builder.setRecordId(getRef().getRecordId());
        builder.setLastEditTime(getLastEditTime());
        getParentRecordId().ifPresent(parent -> builder.setParentRecordId(parent));

        for (Map.Entry<ResourceId, FieldValue> entry : this.fieldMap.entrySet()) {
            builder.setFieldValue(entry.getKey(), entry.getValue());
        }
        return builder.build();
    }
}
