/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formTree;

import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.form.FormMetadata;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldType;
import org.activityinfo.model.type.FieldTypeClass;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.expr.CalculatedFieldType;
import org.activityinfo.model.type.subform.SubFormReferenceType;
import org.activityinfo.promise.Maybe;

import java.util.*;

/**
 * Contains a tree of fields based on references to other {@code FormClasses}
 */
public class FormTree implements FormClassProvider, FormMetadataProvider, JsonSerializable {

    private ResourceId rootFormId;



    public enum State {
        VALID,
        DELETED,
        FORBIDDEN,
        SUSPENDED
    }

    public class Node {

        private Node parent;
        private FormField field;

        private FieldPath path;
        private FormMetadata form;
        private List<Node> children = Lists.newArrayList();

        private int depth;

        public boolean isRoot() {
            return parent == null;
        }

        public boolean isReference() {
            return field.getType() instanceof ReferenceType ||
                    field.getType() instanceof SubFormReferenceType;
        }

        public boolean isParentReference() {
            if(!form.getSchema().isSubForm()) {
                return false;
            }
            if(!(field.getType() instanceof ReferenceType)) {
                return false;
            }
            ReferenceType type = (ReferenceType) getType();
            if(type.getRange().size() != 1) {
                return false;
            }
            ResourceId rangeFormId = Iterables.getOnlyElement(type.getRange());
            ResourceId parentFormId = form.getSchema().getParentFormId().get();

            return rangeFormId.equals(parentFormId);
        }

        public Node addChild(FormMetadata declaringClass, FormField field) {
            Node childNode = new Node();
            childNode.parent = this;
            childNode.field = field;
            childNode.path = new FieldPath(this.path, field.getId());
            childNode.form = declaringClass;
            children.add(childNode);
            nodeMap.put(childNode.path, childNode);

            if (childNode.parent != null) {
                childNode.depth = childNode.parent.depth + 1;
            }
            return childNode;
        }


        /**
         *
         * @return the fields that are defined on the classes in this Field's range.
         */
        public List<Node> getChildren() {
            return children;
        }

        /**
         * @return the fields that are defined one of the classes in this field's range
         */
        public Iterable<Node> getChildren(ResourceId formClassId) {
            List<Node> matching = Lists.newArrayList();
            for (Node child : children) {
                if(child.getDefiningFormClass().getId().equals(formClassId)) {
                    matching.add(child);
                }
            }
            return matching;
        }

        public FieldPath getPath() {
            return path;
        }

        public FormField getField() {
            return field;
        }

        /**
         *
         * @return the form class which has defined this form
         */
        public FormClass getDefiningFormClass() {
            return form.getSchema();
        }

        public FormMetadata getForm() {
            return form;
        }

        public ResourceId getFieldId() {
            return field.getId();
        }

        /**
         *
         * @return for Reference fields, the range of this field
         */

        public Collection<ResourceId> getOnlyReferenceRange() {
            if (field.getType() instanceof ReferenceType) {
                return ((ReferenceType) field.getType()).getRange();
            } else {
                return Collections.emptySet();
            }
        }

        public Collection<ResourceId> getOnlyKeyReferenceRange() {
            if (field.getType() instanceof ReferenceType && field.isKey()) {
                return ((ReferenceType) field.getType()).getRange();
            } else {
                return Collections.emptySet();
            }
        }

        public Collection<ResourceId> getRange() {
            if(field.getType() instanceof ReferenceType) {
                return ((ReferenceType) field.getType()).getRange();

            } else if(field.getType() instanceof SubFormReferenceType) {
                SubFormReferenceType subFormType = (SubFormReferenceType) field.getType();
                ResourceId subFormClassId = subFormType.getClassId();
                return Collections.singleton(subFormClassId);

            } else {
                return Collections.emptySet();
            }
        }

        public FieldType getType() {
            return field.getType();
        }

        public FieldTypeClass getTypeClass() {
            return field.getType().getTypeClass();
        }

        public Node getParent() {
            return parent;
        }

        /**
         *
         * @return a readable path for this node for debugging
         */
        public String debugPath() {
            StringBuilder path = new StringBuilder();
            path.append(toString(this.getField().getLabel(), this.getDefiningFormClass()));
            Node parent = this.parent;
            while(parent != null) {
                path.insert(0, toString(parent.getField().getLabel(), parent.getDefiningFormClass()) + ".");
                parent = parent.parent;
            }
            return path.toString();
        }

        @Override
        public String toString() {
            return toString(field.getLabel(), this.getDefiningFormClass()) + ":" + field.getType();
        }

        public String fieldLabel() {
            return toLabel(field.getLabel());
        }

        public String formFieldLabel() {
            return getDefiningFormClass() != null
                    ? toLabel(getDefiningFormClass().getLabel()) + "." + toLabel(field.getLabel())
                    : toLabel(field.getLabel());
        }

        private String toLabel(String label) {
            return label.toLowerCase().trim().replace(" ","_");
        }

        private String toString(String label, FormClass definingFormClass) {
            String field = "[";
            if(definingFormClass != null && definingFormClass.getLabel() != null)  {
                field += definingFormClass.getLabel() + ":";
            }
            field += label;
            field += "]";

            return field;
        }

        public boolean isLeaf() {
            return children.isEmpty();
        }


        public FormClass getRootFormClass() {
            if(isRoot()) {
                return getDefiningFormClass();
            } else {
                return getParent().getRootFormClass();
            }
        }

        public boolean isCalculated() {
            return getType() instanceof CalculatedFieldType;
        }

        public boolean isSubForm() {
            return getType() instanceof SubFormReferenceType;
        }

        public boolean isVisibleSubForm() {
            if(!isSubForm()) {
                return false;
            }
            SubFormReferenceType type = (SubFormReferenceType) field.getType();
            FormMetadata subForm = getFormMetadata(type.getClassId());
            return subForm.isVisible();
        }
    }

    private State rootState = State.VALID;
    private List<Node> rootFields = Lists.newArrayList();
    private Map<FieldPath, Node> nodeMap = Maps.newHashMap();
    private Map<ResourceId, FormMetadata> formMap = new HashMap<>();

    public FormTree(ResourceId rootFormId) {
        this.rootFormId = rootFormId;
    }

    public ResourceId getRootFormId() {
        return rootFormId;
    }

    public State getRootState() {
        return rootState;
    }

    public void setRootState(State rootState) {
        this.rootState = rootState;
    }

    public Node addRootField(FormMetadata form, FormField field) {
        Node node = new Node();
        node.form = form;
        node.field = field;
        node.path = new FieldPath(field.getId());
        rootFields.add(node);
        nodeMap.put(node.path, node);
        return node;
    }

    public void addFormMetadata(FormMetadata form) {
        assert form != null;
        if (!formMap.containsKey(form.getId())) {
            formMap.put(form.getId(), form);
        }
    }

    public List<Node> getRootFields() {
        return rootFields;
    }


    public Maybe<FormTree> toMaybe() {
        switch (getRootState()) {
            case VALID:
                return Maybe.of(this);
            case FORBIDDEN:
                return Maybe.forbidden();
            default:
            case DELETED:
                return Maybe.deleted();
        }
    }


    public FormClass getRootFormClass() {
        return formMap.get(rootFormId).getSchema();
    }

    public Iterator<FormClass> getSelfAndParentIterator() {
        return new Iterator<FormClass>() {

            private FormClass next = getRootFormClass();

            @Override
            public boolean hasNext() {
                return next != null;
            }

            @Override
            public FormClass next() {
                FormClass toReturn = next;
                next = next.getParentFormId()
                        .flatMap(parentId -> getFormClassIfPresent(parentId))
                        .orElse(null);
                return toReturn;
            }
        };
    }

    public FormMetadata getRootMetadata() {
        return formMap.get(rootFormId);
    }

    public FormClass getFormClass(ResourceId formClassId) {
        Optional<FormClass> formClass = getFormClassIfPresent(formClassId);
        if(!formClass.isPresent()) {
            throw new IllegalArgumentException("No such FormClass: " + formClassId);
        }
        return formClass.get();
    }


    public FormMetadata getFormMetadata(ResourceId formId) {
        FormMetadata metadata = formMap.get(formId);
        if(metadata == null) {
            return FormMetadata.notFound(formId);
        }
        return metadata;
    }

    public Collection<FormMetadata> getForms() {
        return formMap.values();
    }

    public Optional<FormClass> getFormClassIfPresent(ResourceId formId) {
        FormMetadata formMetadata = getFormMetadata(formId);
        if(formMetadata == null || !formMetadata.isVisible()) {
            return Optional.empty();
        }
        return Optional.of(formMetadata.getSchema());
    }

    public Node getNodeByPath(FieldPath path) {
        Node node = nodeMap.get(path);
        if (node == null) {
            throw new IllegalArgumentException();
        }
        return node;
    }


    public Node getRootField(ResourceId fieldId) {
        return nodeMap.get(new FieldPath(fieldId));
    }

    private void findLeaves(List<Node> leaves, Iterable<Node> children) {
        for(Node child : children) {
            if(child.isLeaf()) {
                leaves.add(child);
            } else {
                findLeaves(leaves, child.getChildren());
            }
        }
    }

    public List<Node> getLeaves() {
        List<Node> leaves = Lists.newArrayList();
        findLeaves(leaves, rootFields);
        return leaves;
    }

    /**
     *
     * @param formId
     * @return FormTree for any form/subform of this formTree
     */

    public FormTree tree(ResourceId formId) {
        if (formId.equals(this.rootFormId)) {
            return this;
        }
        FormTreeBuilder treeBuilder = new FormTreeBuilder((FormMetadataProvider)this);
        return treeBuilder.queryTree(formId);
    }

    @Override
    public String toString() {
        StringBuilder s = new StringBuilder();
        for(Node node : getLeaves()) {
            s.append(node.debugPath()).append("\n");
        }
        return s.toString();
    }

    public static boolean isUnchanged(FormTree a, FormTree b) {
        if(!a.getRootFormId().equals(b.getRootFormId())) {
            return false;
        }
        if(a.getForms().size() != b.getForms().size()) {
            return false;
        }
        for (FormMetadata fa : a.getForms()) {
            FormMetadata fb = b.getFormMetadata(fa.getId());
            if(fa.isVisible() != fb.isVisible()) {
                return false;
            }
            if(fa.isVisible() && fa.getSchemaVersion() != fb.getSchemaVersion()) {
                return false;
            }
        }
        return true;
    }


    @Override
    public JsonValue toJson() {
        return JsonFormTreeBuilder.toJson(this);
    }

    public static FormTree fromJson(JsonValue jsonValue) {
        return JsonFormTreeBuilder.fromJson(jsonValue);
    }
}
