/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formTree;

import com.google.common.collect.Lists;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.form.FormMetadata;
import org.activityinfo.model.permission.FormPermissions;
import org.activityinfo.model.resource.ResourceId;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.logging.Logger;

/**
 * Constructs a tree of related Forms from a given root FormClass.
 */
public class FormTreeBuilder {

    private static final Logger LOGGER = Logger.getLogger(FormTreeBuilder.class.getName());

    private final FormMetadataProvider metadataProvider;

    public FormTreeBuilder(final FormClassProvider schemaProvider) {
        metadataProvider = new FormMetadataProvider() {
            @Override
            public FormMetadata getFormMetadata(ResourceId formId) {
                FormClass formClass;
                try {
                    formClass = schemaProvider.getFormClass(formId);
                } catch (Exception e) {
                    return FormMetadata.notFound(formId);
                }
                if(formClass == null) {
                    return FormMetadata.notFound(formId);
                }
                return FormMetadata.of(formClass, FormPermissions.readonly());
            }
        };
    }

    public FormTreeBuilder(FormMetadataProvider metadataProvider) {
        this.metadataProvider = metadataProvider;
    }

    public FormTree queryTree(ResourceId rootFormId) {

        FormTree tree = new FormTree(rootFormId);
        FormMetadata root = metadataProvider.getFormMetadata(rootFormId);

        if(root.isDeleted()) {
            tree.setRootState(FormTree.State.DELETED);
            return tree;
        }


        if(!root.isVisible()) {
            tree.setRootState(FormTree.State.FORBIDDEN);
            return tree;
        }

        if(root.isSuspended()) {
            tree.setRootState(FormTree.State.SUSPENDED);
        }

        FormClass rootSchema = root.getSchema();

        List<ResourceId> stack = Lists.newArrayList(rootFormId);

        fetchParent(tree, null, root, stack);

        tree.addFormMetadata(root);

        // Add fields defined by this FormClass
        for(FormField field : rootSchema.getFields()) {
            FormTree.Node node = tree.addRootField(root, field);
            if(node.isReference()) {
                fetchChildren(stack, tree, node, node.getRange());
            }
        }
        return tree;
    }

    private void fetchParent(FormTree tree, FormTree.Node parent, FormMetadata form, List<ResourceId> stack) {
        if(!form.isVisible()) {
            return;
        }
        Optional<FormField> parentField = form.getSchema().getParentField();
        if(parentField.isPresent()) {
            if(!stack.contains(parentField.get().getId())) {
                tree.addFormMetadata(form);
                FormTree.Node node;
                if(parent == null) {
                    node = tree.addRootField(form, parentField.get());
                } else {
                    node = parent.addChild(form, parentField.get());
                }
                fetchChildren(stack, tree, node,
                        form.getSchema().getParentFormId()
                                .map(id -> Collections.singleton(id))
                                .orElse(Collections.emptySet()));
            }
        }
    }

    /**
     * Now that we have the actual FormClass model that corresponds to this node's
     * formClassId, add it's children.
     *
     */
    private void fetchChildren(List<ResourceId> stack, FormTree tree, FormTree.Node parent, Iterable<ResourceId> childFormIds)  {
        for(ResourceId childFormId : childFormIds) {
            if(!stack.contains(childFormId)) {
                addChildren(stack, tree, parent, metadataProvider.getFormMetadata(childFormId));
            }
        }
    }

    private void addChildren(List<ResourceId> stack, FormTree tree, FormTree.Node parent, FormMetadata childForm) {
        stack.add(parent.getDefiningFormClass().getId());
        try {
            tree.addFormMetadata(childForm);

            if (childForm.isVisible()) {
                fetchParent(tree, parent, childForm, stack);
            }

            for (FormField field : childForm.getFields()) {
                FormTree.Node childNode = parent.addChild(childForm, field);
                if (childNode.isReference()) {
                    fetchChildren(stack, tree, childNode, childNode.getRange());
                }
            }
        } finally {
            stack.remove(stack.size() - 1);
        }
    }

}
