/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formTree;

import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.form.TypedFormRecord;
import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.SerialNumber;
import org.activityinfo.model.type.SerialNumberType;
import org.activityinfo.model.type.primitive.HasStringValue;
import org.activityinfo.model.type.primitive.TextType;

import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * A level within a hierarchy of lookup choices.
 */
public class LookupKey {

    private int keyIndex;

    List<ParentKey> parentLevels = new ArrayList<>();

    private ResourceId formId;

    private String formLabel;

    /**
     * The name of the field that serves as a key, or {@code null} if this is
     * the raw unique record ID.
     */
    private final String fieldLabel;

    /**
     * This key's field id.
     */
    private SymbolNode fieldId;

    private final FormulaNode formula;

    boolean multiRange;

    /**
     * True if fuzzy matching is appropriate for this key.
     */
    boolean fuzzyMatchable;

    final FormulaType type;

    private List<LookupKey> childLevels = new ArrayList<>();

    /**
     * This key's distinctive label. This is set by LookupKeySet after all
     * LookupKeys are constructed.
     */
    String keyLabel;

    /**
     * A key cannot know itself whether it is leaf key - keys both with and without children can be valid candidates
     * for a leaf key depending on what level we are selecting from. This is set by {@link LookupKeySet} if it has
     * determined the key is a leaf key.
     */
    boolean leaf = false;



    /**
     * Defines a parent of a key, including the field path for navigating to the parent in the form tree
     */
    final static class ParentKey {

        private LookupKey key;
        private SymbolNode fieldId;

        public ParentKey(LookupKey key) {
            this.key = key;
        }

        public ParentKey(LookupKey key, String fieldId) {
            this(key);
            this.fieldId = fieldId == null ? null : new SymbolNode(fieldId);
        }

        public LookupKey getKey() {
            return key;
        }

        public SymbolNode getFieldId() {
            return fieldId;
        }
    }

    private LookupKey(int keyIndex,
                      String parentFieldId,
                      LookupKey parentLevel,
                      FormClass formClass,
                      SymbolNode keyFormula,
                      String fieldLabel,
                      FormulaNode formula,
                      FormulaType type) {
        this.keyIndex = keyIndex;
        if (parentLevel != null) {
            parentLevels.add(new ParentKey(parentLevel, parentFieldId));
            parentLevel.childLevels.add(this);
        }
        this.formId = formClass.getId();
        this.formLabel = formClass.getLabel();
        this.fieldLabel = fieldLabel;
        this.fieldId = keyFormula;
        this.formula = formula;
        this.type = type;
    }

    /**
     * Constructs a new LookupKey for the given parent, form, and key field.
     */
    LookupKey(int keyIndex,
              String parentFieldId,
              LookupKey parentLevel,
              FormClass formClass,
              FormField formField,
              FormulaNode formula) {

        this(keyIndex, parentFieldId, parentLevel, formClass,
            new SymbolNode(formField.getId()),
            formField.getLabel(),
            formula,
            formField.getType().getFormulaType());
        fuzzyMatchable = isFuzzyMatchable(formField);
    }

    LookupKey(int keyIndex, List<ParentKey> parentKeys, FormClass formClass, FormField formField, FormulaNode formula) {
        this(keyIndex, null, null, formClass, new SymbolNode(formField.getId()),
            formField.getLabel(),
            formula,
            formField.getType().getFormulaType());
        this.fuzzyMatchable = isFuzzyMatchable(formField);
        for (ParentKey parentKey : parentKeys) {
            parentLevels.add(parentKey);
            parentKey.getKey().childLevels.add(this);
        }
    }

    /**
     * Constructs a new LookupKey for the given form and key field.
     */
    LookupKey(int keyIndex, FormClass formClass, FormField field, FormulaNode formula) {
        this(keyIndex, null, null, formClass, field, formula);
    }

    private static boolean isFuzzyMatchable(FormField field) {
        if (field.getType() instanceof TextType) {
            TextType textType = (TextType) field.getType();
            return !textType.hasInputMask();
        }
        return false;
    }

    public boolean isFuzzyMatchable() {
        return fuzzyMatchable;
    }

    public FormulaType getType() {
        return type;
    }

    /**
     * Constructs a new LookupKey for the given form using the raw record id.
     */
    LookupKey(int keyIndex, FormClass formClass, FormulaNode formula) {
        this(keyIndex, null, null, formClass, new SymbolNode(ColumnModel.RECORD_ID_SYMBOL), "ID",
            formula, FormulaType.STRING);
    }

    /**
     * @return this key's one-based index
     */
    public int getKeyIndex() {
        return keyIndex;
    }

    public boolean isRoot() {
        return parentLevels.isEmpty();
    }

    public boolean isLeaf() {
        return leaf;
    }

    public void setLeaf(boolean leaf) {
        this.leaf = leaf;
    }

    public String getFormLabel() {
        return formLabel;
    }

    public String getFieldLabel() {
        return fieldLabel;
    }

    public ResourceId getFormId() {
        return formId;
    }

    public SymbolNode getKeyField() {
        return fieldId;
    }

    public List<LookupKey> getDirectParents() {
        assert !parentLevels.isEmpty();
        return parentLevels.stream()
                .map(ParentKey::getKey)
                .collect(Collectors.toList());
    }

    public Set<LookupKey> getParentLevels() {
        Set<LookupKey> parents = new HashSet<>();
        collectParents(parents);
        return parents;
    }

    private void collectParents(Set<LookupKey> parents) {
        if(this.isRoot()) {
            return;
        }
        for (LookupKey parent : getDirectParents()) {
            parents.add(parent);
            parent.collectParents(parents);
        }
    }

    public List<LookupKey> getChildLevels() {
        return childLevels;
    }

    @Override
    public String toString() {
        return "[" + formLabel + "." + fieldLabel + ": " +  formId + "." + fieldId + "]";
    }

    private void collectKeys(@Nullable FormulaNode baseField, Map<LookupKey, FormulaNode> keys) {

        if(multiRange) {
            collectKeysMultiRange(baseField, keys);
            return;
        }

        // If this key has already been added, then we do not want to overwrite
        if (!keys.containsKey(this)) {
            keys.put(this, prefix(baseField, formula));
        }

        if(!isRoot()) {
            parentLevels.forEach(parentLevel -> {
                parentLevel.getKey().collectKeys(baseField, keys);
            });
        }
    }

    /**
     * This provides special case handling for fields that reference more than one form. Currently, this isn't
     * possible to construct using the user interface, but ActivityInfo classic Location Types were migrated to forms
     * that use this feature. Properly handling keys for such forms is quite complex, but the routine below works
     * in all of the cases that exist in the system.
     *
     * <p>It either needs to be properly generalized, (see branch keys-features), or we need to sunset the ability
     * to reference more than one form.</p>
     *
     */
    private void collectKeysMultiRange(@Nullable FormulaNode baseField, Map<LookupKey, FormulaNode> keys) {
        // If this key has already been added, then we do not want to overwrite
        if (!keys.containsKey(this)) {
            keys.put(this, join(new SymbolNode(formId), fieldId));
        }

        if(!isRoot()) {
            parentLevels.forEach(parentLevel -> {
                parentLevel.getKey().collectKeysMultiRange(join(baseField, parentLevel.getFieldId()), keys);
            });
        }
    }

    private FormulaNode prefix(FormulaNode baseField, FormulaNode formula) {
        if(baseField == null) {
            return formula;
        } else if (formula.equals(new SymbolNode(ColumnModel.RECORD_ID_SYMBOL))) {
            return baseField;
        } else if(formula instanceof SymbolNode) {
            return new CompoundExpr(baseField, ((SymbolNode) formula));
        } else if(formula instanceof CompoundExpr) {
            CompoundExpr split = ((CompoundExpr) formula);
            return new CompoundExpr(prefix(baseField, split.getValue()), split.getField());
        } else {
            // All formulas constructed by LookupKeySet should include only Symbols and compound nodes.
            return formula;
        }
    }

    private FormulaNode join(@Nullable FormulaNode base, @Nullable SymbolNode field) {
        if(base == null && field == null) {
            return null;
        } else if (base == null) {
            return field;
        } else if (field == null) {
            return base;
        } else {
            return new CompoundExpr(base, field);
        }
    }

    public Map<LookupKey, FormulaNode> getKeyFormulas() {
        return getKeyFormulas(null);
    }

    public Map<LookupKey, FormulaNode> getKeyFormulas(@Nullable FormulaNode baseField) {
        Map<LookupKey, FormulaNode> keys = new HashMap<>();
        collectKeys(baseField, keys);
        return keys;
    }

    public String label(TypedFormRecord record) {
        if(fieldId.getName().equals(ColumnModel.RECORD_ID_SYMBOL)) {
            return record.getId().asString();
        }

        FieldValue fieldValue = record.get(ResourceId.valueOf(fieldId.getName()));
        if(fieldValue == null) {
            return record.getId().asString();
        }

        if(fieldValue instanceof SerialNumber) {
            SerialNumberType type = new SerialNumberType();
            return type.format(((SerialNumber) fieldValue));

        } else if(fieldValue instanceof HasStringValue) {
            return ((HasStringValue) fieldValue).asString();

        } else {
            return "";
        }
    }

    public String getKeyLabel() {
        return keyLabel;
    }

    void addParentLevel(ParentKey parent) {
        this.parentLevels.add(parent);
        parent.getKey().childLevels.add(this);
    }

    void addParentLevels(List<ParentKey> parents) {
        parents.forEach(this::addParentLevel);
    }

    public String getKeyId() {
        return formId + "." + getKeyField();
    }
}
