/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formTree;

import com.google.common.base.Strings;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Iterables;
import com.google.common.collect.Multimap;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FormulaParser;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.SerialNumberType;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.primitive.TextType;
import org.activityinfo.model.type.subform.SubFormReferenceType;
import org.activityinfo.model.type.time.PeriodType;

import javax.annotation.Nullable;
import java.util.*;

/**
 * Set of keys that can be used to lookup a value for a reference field.
 *
 * <p>Fields of type {@link ReferenceType} refer to Records in other Forms. In order to be select
 * a value, users need a way to "look up" the ID of a Record. We could of course ask users to the enter
 * the auto-generated RecordId, but that would be difficult.</p>
 *
 * <p>One intuitive way to look up Records is to use the fields that form designers have identified as "keys".</p>
 *
 * <p>In the simplest case, a Form will have a single text key field. Such a reference field will have a single
 * {@link LookupKey}. The user can select one value from the {@code LookupKey} and that will be sufficient to
 * uniquely identify the record.</p>
 *
 * <p>In other cases, a Form might have two text keys, such as "Last Name" and "First Name". Neither key is
 * sufficient to identify a record uniquely, but together they can be used to look up a record.</p>
 */
public class LookupKeySet {


    public enum LabelContext {
        /**
         * These keys are to be used in a context where they must be distinguished from other fields. Therefore,
         * the key lables must include some reference to the field.
         */
        FIELD_LABEL_REQUIRED,

        /**
         * These keys are to be used in a context where the field is already clear to the user. The key labels
         * should not reference again the field label.
         */
        FIELD_LABEL_IMPLIED
    }


    private List<LookupKey> lookupKeys = new ArrayList<>();

    /**
     * Map from a referenced form to it's leaf lookup key.
     *
     * For example, if you have a reference field that includes [PROVINCE, TERRITORY, SECTOR]
     * in its reference range, that means that _either_ a PROVINCE can be selected, _or_ a TERRITORY _or_ a SECTOR
     * can be selected.
     *
     * If those three forms have the following keys:
     *
     * <pre>
     *     PROVINCE:  k0[PROVINCE.name]
     *     TERRITORY: k0[PROVINCE.name], k1[TERRITORY.name]
     *     SECTOR:    k0[PROVINCE.name], k1[TERRITORY.name], k2[SECTOR.name]
     * </pre>
     *
     * Then the "leaf key" of the PROVINCE form is k0, while the leaf key of TERRITORY is k1, and
     * the leaf key of SECTOR is k2.
     *
     */
    private Map<ResourceId, LookupKey> leafKeyMap = new HashMap<>();

    /**
     * Maps each visited form to its leaf key.
     *
     * <p>For reference fields with multiple forms included in the range, the keys may overlap. We want
     * to
     *
     */
    private Map<ResourceId, LookupKey> formMap = new HashMap<>();

    private Multimap<ResourceId, ResourceId> parentMap = HashMultimap.create();

    private FormTree formTree;
    private String referenceFieldLabel;

    /**
     * True if there are user-defined keys (natural keys) defined.
     */
    private boolean naturalKeysFound;

    private final LabelContext labelContext;
    private final boolean includeParents;

    private int multiRangeStack = 0;

    /**
     *
     * @param formTree the form tree that includes this field
     * @param field a {@code FormField} with type {@link SubFormReferenceType} or {@link ReferenceType}
     */
    public LookupKeySet(FormTree formTree, FormField field) {
        this(formTree, field, LabelContext.FIELD_LABEL_REQUIRED, true);
    }

    /**
     * @param formTree the form tree that includes this field
     * @param field a {@code FormField} with type {@link SubFormReferenceType} or {@link ReferenceType}
     * @param labelContext
     * @param includeParents
     */
    public LookupKeySet(FormTree formTree, FormField field, LabelContext labelContext, boolean includeParents) {
        this.formTree = formTree;
        this.referenceFieldLabel = field.getLabel();
        this.labelContext = labelContext;
        this.includeParents = includeParents;

        if(field.getType() instanceof ReferenceType) {
            ReferenceType referenceType = (ReferenceType) field.getType();
            if(referenceType.getRange().size() > 1) {
                multiRangeStack++;
            }
            for (ResourceId referenceFormId : referenceType.getRange()) {
                if (!formTree.getFormMetadata(referenceFormId).isAccessible()) {
                    continue;
                }
                leafKeyMap.put(referenceFormId, addKeys(formTree.getFormClass(referenceFormId)));
                leafKeyMap.get(referenceFormId).setLeaf(true);
            }
        } else if(field.getType() instanceof SubFormReferenceType) {
            SubFormReferenceType subFormType = (SubFormReferenceType) field.getType();
            leafKeyMap.put(subFormType.getClassId(), addKeys(formTree.getFormClass(subFormType.getClassId())));
            leafKeyMap.get(subFormType.getClassId()).setLeaf(true);
        }

        // Now that we know how many keys there are and from which forms,
        // we can compose labels
        for (LookupKey lookupKey : lookupKeys) {
            composeLabel(lookupKey);
        }
    }


    public LookupKeySet(FormTree formTree) {
        this.formTree = formTree;
        this.referenceFieldLabel = "";
        this.labelContext = LabelContext.FIELD_LABEL_IMPLIED;
        this.includeParents = true;

        addKeys(formTree.getRootFormClass());

        // Now that we know how many keys there are and from which forms,
        // we can compose labels
        for (LookupKey lookupKey : lookupKeys) {
            composeLabel(lookupKey);
        }
    }

    private LookupKey addKeys(FormClass formClass) {
        return addKeys(Optional.empty(), Collections.emptyList(), formClass);
    }

    /**
     * Adds keys to the {@code lookupKeys} list, using a single serial number field if it is available, otherwise,
     * looking for "natural" keys.
     *
     * <p>If a reference field is used as a key, we recurse to find the keys defined in the referenced form.</p>
     *
     */
    private LookupKey addKeys(Optional<FormulaNode> prefix, List<LookupKey.ParentKey> parentKeys, FormClass formClass) {

        ResourceId formId = formClass.getId();

        // In the case of multiple, overlapping heirarchies, keys for this form
        // may already be defined.

        if(formMap.containsKey(formId)) {
            return formMap.get(formId);
        }

        // if serial number is present, we use that exclusively.

        Optional<FormField> serialNumberField = findSerialNumberField(formClass);
        if(serialNumberField.isPresent()) {
            LookupKey lookupKey = serialNumberLevel(prefix, parentKeys, formClass, serialNumberField.get());
            lookupKeys.add(lookupKey);
            formMap.put(formId, lookupKey);
            naturalKeysFound = true;
            return lookupKey;

        } else {
            return addNaturalKeys(prefix, parentKeys, formClass, formId);
        }
    }

    private LookupKey addNaturalKeys(Optional<FormulaNode> prefix, List<LookupKey.ParentKey> parentKeys, FormClass formClass, ResourceId formId) {
        LookupKey lookupKey = null;

        List<FormField> fields = formClass.getFields();

        if(formClass.isSubForm() && includeParents) {
            parentKeys = addReferenceParent(prefix, formClass.getParentField().get(),
                formClass.getParentFormId().get(), parentKeys);
        }

        for (FormField field : fields) {
            if (isReferenceKey(field)) {
                parentKeys = addReferenceParent(prefix, field, formId, parentKeys);
            }
        }

        for (FormField field : fields) {
            if(isTextLikeKey(field)) {
                lookupKey = addTextParent(prefix, parentKeys, formClass, field);
                parentKeys = Collections.singletonList(new LookupKey.ParentKey(lookupKey, null));
            }
        }

        // If this form's only key is a reference to another form, then
        // it has a 1-to-1 relatonship with that form, and we don't need our own key
        if(lookupKey == null && parentKeys.size() == 1) {
            return parentKeys.get(0).getKey();
        }

        // If there are really no other key fields, then use the autogenerated id as a key
        if (lookupKey == null) {
            lookupKey = idLevel(prefix, formClass);
            if (!parentKeys.isEmpty()) {
                lookupKey.addParentLevels(parentKeys);
            }
            lookupKeys.add(lookupKey);
        }

        formMap.put(formId, lookupKey);

        return lookupKey;
    }

    private boolean isReferenceKey(FormField field) {
        return field.isKey() && field.getType() instanceof ReferenceType;
    }

    private LookupKey addTextParent(Optional<FormulaNode> prefix, List<LookupKey.ParentKey> parentKeys, FormClass formClass, FormField formField) {
        LookupKey lookupKey = textKeyLevel(prefix, parentKeys, formClass, formField);
        lookupKey.multiRange = (multiRangeStack > 0);
        lookupKeys.add(lookupKey);
        naturalKeysFound = true;
        return lookupKey;
    }


    private LookupKey addLookupFormula(FormClass form, FormField field, String formula) {
        LookupKey lookupKey =  new LookupKey(nextKeyIndex(), Collections.emptyList(), form, field,
            FormulaParser.tryParse(formula).orElse(new SymbolNode("_id")));
        lookupKeys.add(lookupKey);
        naturalKeysFound = true;
        return lookupKey;
    }

    private List<LookupKey.ParentKey> addReferenceParent(Optional<FormulaNode> prefix, FormField referenceKey, ResourceId formId, List<LookupKey.ParentKey> parentKeys) {
        ReferenceType referenceType = (ReferenceType) referenceKey.getType();
        Optional<FormulaNode> newPrefix = Optional.of(join(prefix, referenceKey.getId()));

        if(referenceType.getRange().size() > 1) {
            multiRangeStack++;
        }


        List<LookupKey.ParentKey> newParents = new ArrayList<>();
        for(ResourceId referencedFormId : referenceType.getRange()) {
            Optional<FormClass> optionalFormClass = formTree.getFormClassIfPresent(referencedFormId);
            if(!optionalFormClass.isPresent()){
                continue;
            }
            FormClass referencedFormClass = optionalFormClass.get();
            parentMap.put(formId, referencedFormId);
            LookupKey parent = addKeys(newPrefix, parentKeys, referencedFormClass);
            newParents.add(new LookupKey.ParentKey(parent, referenceKey.getId().asString()));
        }

        if(referenceType.getRange().size() > 1) {
            multiRangeStack--;
        }

        return newParents;
    }

    private boolean isTextLikeKey(FormField formField) {
        return formField.isKey() &&
                (formField.getType() instanceof TextType ||
                 formField.getType() instanceof EnumType ||
                 formField.getType() instanceof PeriodType);
    }


    private int nextKeyIndex() {
        return lookupKeys.size() + 1;
    }


    private LookupKey serialNumberLevel(Optional<FormulaNode> prefix, List<LookupKey.ParentKey> parentKeys, FormClass form, FormField field) {
        return new LookupKey(nextKeyIndex(), parentKeys, form, field, join(prefix, field.getId()));
    }

    private LookupKey textKeyLevel(Optional<FormulaNode> prefix, List<LookupKey.ParentKey> parentKeys, FormClass form, FormField formField) {
        return new LookupKey(nextKeyIndex(), parentKeys, form, formField, join(prefix, formField.getId()));
    }

    private LookupKey idLevel(Optional<FormulaNode> prefix, FormClass formSchema) {
        return new LookupKey(nextKeyIndex(), formSchema, join(prefix, new SymbolNode(ColumnModel.RECORD_ID_SYMBOL)));
    }

    private FormulaNode join(Optional<FormulaNode> prefix, ResourceId fieldId) {
        return join(prefix, new SymbolNode(fieldId));
    }

    private FormulaNode join(Optional<FormulaNode> prefix, SymbolNode symbolNode) {
        return prefix.map(p -> (FormulaNode)new CompoundExpr(p, symbolNode)).orElse(symbolNode);
    }

    private Optional<FormField> findSerialNumberField(FormClass formClass) {
        for (FormField formField : formClass.getFields()) {
            if(formField.getType() instanceof SerialNumberType) {
                return Optional.of(formField);
            }
        }
        return Optional.empty();
    }


    private void composeLabel(LookupKey lookupKey) {
        StringBuilder label = new StringBuilder();

        if (formMap.size() > 1) {
            // If there are keys that live on multiple forms,
            // we need to distinguish them by their form name
            label.append(lookupKey.getFormLabel().trim());

        } else if(labelContext == LabelContext.FIELD_LABEL_REQUIRED) {
            // If there is only one form referenced,
            // we always use the referencing field's label
            label.append(Strings.nullToEmpty(referenceFieldLabel));
        }

        if(lookupKey.getFieldLabel() == null) {
            if(label.length() > 0) {
                label.append(" ");
            }
            label.append("ID");
        } else {
            if(!label.toString().equals(lookupKey.getFieldLabel())) {
                if(label.length() > 0) {
                    label.append(" ");
                }
                label.append(lookupKey.getFieldLabel().trim());
            }
        }

        lookupKey.keyLabel = label.toString();
    }

    public List<LookupKey> getLookupKeys() {
        return lookupKeys;
    }

    /**
     * @return a List of LookupKeys with no children.
     */
    public Collection<LookupKey> getLeafKeys() {
        return leafKeyMap.values();
    }


    public LookupKey getKey(int i) {
        return lookupKeys.get(i);
    }

    public LookupKey getLeafKey(ResourceId referencedFormId) {
        LookupKey lookupKey = leafKeyMap.get(referencedFormId);
        assert lookupKey != null : "No leaf key for " + referencedFormId;
        return lookupKey;
    }

    public Set<ResourceId> getAncestorForms(ResourceId formId) {
        Set<ResourceId> set = new HashSet<>();
        collectAncestors(Collections.singletonList(formId), set);
        return set;
    }

    private void collectAncestors(Collection<ResourceId> formIds, Set<ResourceId> set) {
        for (ResourceId formId : formIds) {
            Collection<ResourceId> parents = parentMap.get(formId);
            set.addAll(parents);
            collectAncestors(parents, set);
        }
    }

    public Map<LookupKey, FormulaNode> getKeyFormulas(@Nullable FormulaNode baseField) {

        if(leafKeyMap.size() == 1) {
            return Iterables.getOnlyElement(getLeafKeys()).getKeyFormulas(baseField);

        } else {
            // For the case where a field can refer to one of several different forms,
            // query the keys using the syntax [formId].[keyFieldId]

            Map<LookupKey, FormulaNode> keyMap = new HashMap<>();
            for (LookupKey lookupKey : lookupKeys) {
                keyMap.put(lookupKey, new CompoundExpr(new SymbolNode(lookupKey.getFormId()), lookupKey.getKeyField()));
            }

            return keyMap;
        }
    }

    public Map<LookupKey, FormulaNode> getKeyFormulas(ResourceId fieldId) {
        return getKeyFormulas(new SymbolNode(fieldId));
    }

    public boolean hasLeafKey(ResourceId referencedFormId) {
        return leafKeyMap.containsKey(referencedFormId);
    }

    public int size() {
        return lookupKeys.size();
    }

    /**
     * Returns true if there any fields actually marked as key fields defined by the database owner, like
     * a serial number or a text key field.
     */
    public boolean hasNaturalKeys() {
        return naturalKeysFound;
    }
}