/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.formula.resolved.ResolvedNodeVisitor;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;
import org.activityinfo.model.type.number.Quantity;
import org.activityinfo.model.type.primitive.TextValue;
import org.activityinfo.model.type.time.LocalDate;

import javax.annotation.Nonnull;
import java.util.Collections;
import java.util.Optional;

public class ConstantNode extends FormulaNode implements ResolvedNode {

    public static final ConstantNode BLANK_STRING = new ConstantNode(FormulaType.STRING, ColumnViews.BLANK_STRING);

    public static final ConstantNode TRUE = new ConstantNode(FormulaType.BOOLEAN, ColumnViews.TRUE);

    public static final ConstantNode FALSE = new ConstantNode(FormulaType.BOOLEAN, ColumnViews.FALSE);


    private final ColumnView value;
    private final FormulaType type;

    public ConstantNode(FormulaType type, ColumnView value) {
        this.type = type;
        this.value = value;
    }

    public ConstantNode(double value) {
        this(FormulaType.NUMBER, ColumnViews.of(value));
    }

    public ConstantNode(boolean value) {
        this(FormulaType.BOOLEAN, ColumnViews.of(value));
    }

    public ConstantNode(String value) {
        this(FormulaType.STRING, ColumnViews.of(value));
    }

    public ConstantNode(Quantity value) {
        this(FormulaType.NUMBER, ColumnViews.of(value.getValue()));
    }

    public ConstantNode(TextValue value) {
        this(FormulaType.STRING, ColumnViews.of(value.asString()));
    }

    public ConstantNode(LocalDate localDate) {
        this(FormulaType.LOCAL_DATE, ColumnViews.of(localDate));
    }

    public ConstantNode(boolean value, SourceRange source) {
        this(value);
        this.sourceRange = source;
    }

    public ConstantNode(double value, SourceRange sourceRange) {
        this(value);
        this.sourceRange = sourceRange;
    }

    public ConstantNode(String string, SourceRange sourceRange) {
        this(string);
        this.sourceRange = sourceRange;
    }

    @Nonnull
    public ColumnView getValue() {
        return value;
    }


    public String getStringValue() {
        return value.getString(0);
    }

    public double getDoubleValue() {
        return value.getDouble(0);
    }

    @Override
    public String deparse() {

        if(type == FormulaType.LOCAL_DATE) {
            LocalDate date = value.getLocalDate(0);
            if(date == null) {
                // Blank date literal??
                return "\"\"";
            } else {
                return "DATE(" + date.getYear() + ", " + date.getMonthOfYear() + ", " + date.getDayOfMonth() + ")";
            }
        }

        switch (value.getType()) {
            case NUMBER:
                if(value.isBlank(0)) {
                    return "NaN";
                } else {
                    return deparse(value.getDouble(0));
                }
            case STRING:
                if(value.isBlank(0)) {
                    return "\"\"";
                } else {
                    // TODO: Escaping
                    return "\"" + value.getString(0) + "\"";
                }
            case BOOLEAN:
                if(value.getBoolean(0) == ColumnView.TRUE) {
                    return "TRUE";
                } else {
                    return "FALSE";
                }
        }
        throw new UnsupportedOperationException("constant: " + value);
    }

    private String deparse(double doubleValue) {
        if(Math.floor(doubleValue) == doubleValue && !Double.isInfinite(doubleValue)) {
            return Integer.toString((int)doubleValue);
        } else {
            return "" + doubleValue;
        }
    }

    @Override
    public <T> T accept(FormulaVisitor<T> visitor) {
        return visitor.visitConstant(this);
    }


    @Override
    public <T> T accept(ResolvedNodeVisitor<T> visitor) {
        return visitor.visitConstant(this);
    }


    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        ConstantNode that = (ConstantNode) o;

        if (!value.equals(that.value)) {
            return false;
        }

        return true;
    }

    @Override
    public int hashCode() {
        return value.hashCode();
    }


    @Override
    public String toString() {
        return deparse();
    }

    @Override
    public boolean isValid() {
        return true;
    }

    public FormulaType getType() {
        return type;
    }

    @Override
    public Optional<FormulaType> getResolvedType() {
        return Optional.of(type);
    }

    @Override
    public FormulaNode toFormula() {
        return this;
    }

    @Override
    public Iterable<ResolvedNode> getChildren() {
        return Collections.emptyList();
    }

}
