/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import com.google.common.base.Strings;
import org.activityinfo.model.formula.resolved.*;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;

import java.util.Map;

public class FormulaBeautifier implements ResolvedNodeVisitor<FormulaNode>, ReferenceVisitor<FormulaNode> {

    private ResourceId formId;
    private Map<ResourceId, String> aliasMap;

    public FormulaBeautifier(ResourceId formId, Map<ResourceId, String> aliasMap) {
        this.formId = formId;
        this.aliasMap = aliasMap;
    }

    @Override
    public FormulaNode visitRecordId(RecordIdNode recordIdNode) {
        return new SymbolNode(ColumnModel.RECORD_ID_SYMBOL);
    }

    @Override
    public FormulaNode visitFormId(FormIdNode formIdNode) {
        return new SymbolNode(ColumnModel.FORM_ID_SYMBOL);
    }

    @Override
    public FormulaNode visitLastEditTime(LastEditTimeNode node) {
        return new SymbolNode(ColumnModel.LAST_EDIT_TIME_SYMBOL);
    }

    @Override
    public FormulaNode visitField(FieldReferenceNode field) {
        if (field.isParentField()) {
            return new SymbolNode("@parent");
        }
        if (field.getReferencedForms().contains(formId)) {
            if (Strings.isNullOrEmpty(field.getField().getCode())) {
                if (aliasMap.isEmpty()) {
                    return getLabelOrId(field.getField().getId(), field.getField().getLabel());
                } else {
                    return new SymbolNode(aliasMap.get(field.getField().getId()));
                }
            }
        }
        if (Strings.isNullOrEmpty(field.getField().getCode())) {
            return getLabelOrId(field.getField().getId(), field.getField().getLabel());
        } else {
            return new SymbolNode(field.getField().getCode());
        }
    }

    public static SymbolNode getLabelOrId(ResourceId id, String label) {
        // If the label is short and only includes basic latin
        // characters, numbers, and spaces, then use the label
        if(label.length() < 20 && label.matches("^[A-Za-z][A-Za-z0-9_ ]*$")) {
            return new SymbolNode(label);
        }
        // Otherwise we have to use the ID
        return new SymbolNode(id);
    }

    @Override
    public FormulaNode visitCalculatedField(CalculatedFieldNode field) {
        return visitField(field);
    }

    @Override
    public FormulaNode visitEnumItem(EnumItemReference enumItemReference) {
        return new CompoundExpr(enumItemReference.getField().accept(this), enumItemReference.getEnumItem().getLabel());
    }

    @Override
    public FormulaNode visitJoin(JoinQueryNode node) {
        FormulaNode left = node.getLeft().accept(this);
        ReferenceNode rightNode = node.getRight();
        while (rightNode instanceof JoinQueryNode) {
            FormulaNode tempLeft = ((JoinQueryNode) rightNode).getLeft().accept(this);
            left = new CompoundExpr(left, (SymbolNode) tempLeft);
            rightNode = ((JoinQueryNode) rightNode).getRight();
        }
        if (rightNode instanceof FieldReferenceNode) {
            FormulaNode right = rightNode.accept(this);
            if (right instanceof SymbolNode) {
                return new CompoundExpr(left, ((SymbolNode) right));
            }
        }
        if (rightNode instanceof EnumItemReference) {
            FormulaNode right = rightNode.accept(this);
            if (right instanceof CompoundExpr) {
                FormulaNode value = ((CompoundExpr) right).getValue();
                if (value instanceof SymbolNode) {
                    left = new CompoundExpr(left, (SymbolNode) value);
                    return new CompoundExpr(left, ((CompoundExpr) right).getField());
                }
            }

        }
        throw new UnrewrittableException();
    }

    @Override
    public FormulaNode visitUnion(UnionNode node) {
        throw new UnrewrittableException();
    }

    @Override
    public FormulaNode visitAggregation(AggregateNode node) {
        throw new UnrewrittableException();
    }

    @Override
    public FormulaNode visitConstant(ConstantNode node) {
        return node;
    }

    @Override
    public FormulaNode visitInvalidSyntax(InvalidSyntaxNode node) {
        return node.toFormula();
    }

    @Override
    public FormulaNode visitReferenceNode(ResolvedReference node) {
        try {
            return node.getRoot().accept(this);
        } catch (UnrewrittableException e) {
            return node.toFormula();
        }
    }

    @Override
    public FormulaNode visitInvalidReferenceNode(InvalidReferenceNode node) {
        return node.toFormula();
    }

    @Override
    public FormulaNode visitCallNode(ResolvedCallNode node) {
        return FormulaUglifier.transformCall(node, this);
    }

    @Override
    public FormulaNode visitInvalidCall(InvalidCallNode invalidCallNode) {
        return new FunctionCallNode(invalidCallNode.getFunctionName(),
                invalidCallNode.getChildren()
                        .stream()
                        .map(argument -> argument.accept(this))
                        .toArray(FormulaNode[]::new));
    }

    @Override
    public FormulaNode visitInvalidArgumentNode(InvalidArgumentNode invalidArgumentNode) {
        return invalidArgumentNode.toFormula();
    }

    private class UnrewrittableException extends RuntimeException {

    }
}
