/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import com.google.common.base.Charsets;
import com.google.common.escape.Escaper;
import com.google.common.escape.Escapers;
import com.google.common.io.Files;
import com.google.gwt.core.shared.GwtIncompatible;
import org.activityinfo.model.formula.resolved.*;

import java.io.File;
import java.io.IOException;

@GwtIncompatible
public class FormulaGrapher implements ResolvedNodeVisitor<Void>, ReferenceVisitor<Void> {

    private static final Void NOTHING = null;

    private StringBuilder dot;
    private final Escaper labelEscaper;

    public FormulaGrapher(StringBuilder dot) {
        labelEscaper = Escapers.builder()
            .addEscape('"', "\\\"")
            .build();
        this.dot = dot;
    }

    public static String toGraph(ResolvedNode node) {
        StringBuilder s = new StringBuilder();
        s.append("digraph G {\n");
        node.accept(new FormulaGrapher(s));
        s.append("}\n");
        return s.toString();
    }

    public static void dumpAndGraph(ResolvedNode node) throws IOException, InterruptedException {
        File dotFile = File.createTempFile("formula", ".dot");
        Files.asCharSink(dotFile, Charsets.UTF_8).write(toGraph(node));

        File pngFile = File.createTempFile("formula", ".png");

        System.err.println("Wrote graph to " + dotFile.getAbsolutePath());

        // Convert to png
        int result = new ProcessBuilder("dot", "-Tpng", dotFile.getAbsolutePath(), "-o", pngFile.getAbsolutePath())
            .inheritIO()
            .start()
            .waitFor();

        if(result != 0) {
            throw new RuntimeException("Could not render graph");
        }

        // open
        new ProcessBuilder("xdg-open", pngFile.getAbsolutePath())
            .start()
            .waitFor();
    }

    private String nodeId(Object node) {
        return "N" + Integer.toHexString(System.identityHashCode(node));
    }

    private void describeNode(Object node, FormulaNode label) {
        describeNode(node, label.deparse());
    }

    private void edge(Object from, Object to) {
        dot.append(String.format("%s -> %s;\n", nodeId(from), nodeId(to)));
    }

    private void describeNode(Object node, String label) {
        dot.append(String.format("%s [label=\"%s\"];\n", nodeId(node), labelEscaper.escape(label)));
    }

    @Override
    public Void visitConstant(ConstantNode node) {
        describeNode(node, node.toFormula());
        return NOTHING;
    }


    @Override
    public Void visitInvalidSyntax(InvalidSyntaxNode node) {
        describeNode(node, node.toFormula());
        return NOTHING;
    }

    @Override
    public Void visitReferenceNode(ResolvedReference node) {
        describeNode(node, node.toFormula());
        edge(node, node.getRoot());
        node.getRoot().accept(this);
        return NOTHING;
    }

    @Override
    public Void visitInvalidReferenceNode(InvalidReferenceNode node) {
        return NOTHING;
    }

    @Override
    public Void visitCallNode(ResolvedCallNode node) {
        describeNode(node, node.getFunction().getHelpLabel());
        for (int i = 0; i < node.getArgumentCount(); i++) {
            edge(node, node.getArgument(i));
            node.getArgument(i).accept(this);
        }
        return NOTHING;
    }


    @Override
    public Void visitInvalidCall(InvalidCallNode invalidCallNode) {
        describeNode(invalidCallNode, invalidCallNode.toFormula());
        for (ResolvedNode child : invalidCallNode.getChildren()) {
            edge(invalidCallNode, child);
            child.accept(this);
        }
        return NOTHING;
    }

    @Override
    public Void visitInvalidArgumentNode(InvalidArgumentNode invalidArgumentNode) {
        return NOTHING;
    }

    @Override
    public Void visitRecordId(RecordIdNode recordIdNode) {
        describeNode(recordIdNode, "recordId");
        return NOTHING;
    }

    @Override
    public Void visitFormId(FormIdNode formIdNode) {
        describeNode(formIdNode, "formId");
        return NOTHING;
    }

    @Override
    public Void visitLastEditTime(LastEditTimeNode node) {
        describeNode(node, "lastEditTime");
        return NOTHING;
    }

    @Override
    public Void visitField(FieldReferenceNode field) {
        describeNode(field, field.getField().getLabel());
        return NOTHING;
    }

    @Override
    public Void visitCalculatedField(CalculatedFieldNode field) {
        describeNode(field, field.getField().getLabel());
        edge(field, field.getResolved());
        field.getResolved().accept(this);
        return NOTHING;
    }

    @Override
    public Void visitEnumItem(EnumItemReference node) {
        describeNode(node, "enum");
        describeNode(node.getEnumItem(), node.getEnumItem().getLabel());
        edge(node, node.getField());
        edge(node, node.getEnumItem());
        node.getField().accept(this);
        return NOTHING;
    }

    @Override
    public Void visitJoin(JoinQueryNode node) {
        describeNode(node, "join");
        edge(node, node.getLeft());
        edge(node, node.getRight());
        node.getLeft().accept(this);
        node.getRight().accept(this);
        return NOTHING;
    }

    @Override
    public Void visitUnion(UnionNode node) {
        describeNode(node, "union");
        for (ReferenceNode child : node.getChildren()) {
            edge(child, node);
        }
        for (ReferenceNode child : node.getChildren()) {
            child.accept(this);
        }
        return NOTHING;
    }

    @Override
    public Void visitAggregation(AggregateNode node) {
        describeNode(node, node.getStatFunction().getId());

        edge(node.getSubFormField(), node);
        edge(node.getValueField(), node);

        node.getSubFormField().accept(this);
        node.getValueField().accept(this);

        return NOTHING;
    }
}
