/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import com.google.common.base.Strings;
import org.activityinfo.model.formula.diagnostic.FormulaException;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.formula.functions.FormulaFunction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

import static java.util.Collections.singletonList;

/**
 * Recursive descent parser for the ActivityInfo formula language.
 *
 */
public class FormulaParser {

    private FilteredFormulaLexer lexer;

    public FormulaParser(FormulaLexer lexer) {
        this.lexer = new FilteredFormulaLexer(lexer);
    }

    /**
     * For operators that use infix or prefix notation, return their precedence order, with higher numbers
     * having greater precedence when parsing.
     */
    public static int operatorPrecedence(FormulaFunction function) {
        switch (function.getId()) {
            case "!":
                return 7;
            case "*":
            case "/":
                return 6;
            case "+":
            case "-":
                return 5;
            case ">":
            case "<":
            case ">=":
            case "<=":
                return 4;
            case "==":
            case "!=":
                return 3;
            case "&&":
                return 2;
            case "||":
                return 1;
            default:
                return 0;
        }
    }

    /**
     * Returns true if the given function uses infix notation. For example, "1 + 2" rather than "+(1, 2)"
     */
    public static boolean isInfix(String functionName) {
        return functionName.matches("[><=!+\\-&|*/]+");
    }

    /**
     * Consume and return the next token if it is of type {@code tokenType}, or
     * throw an {@link FormulaSyntaxException} if there are no remaining tokens or
     * the next token is of the wrong type.
     * @param tokenType
     * @return
     */
    private Token expect(TokenType tokenType) {
        if(lexer.hasNext()) {
            Token token = lexer.next();
            if (token.getType() == tokenType) {
                return token;
            }
        }
        throw new FormulaSyntaxException("Expected " + tokenType);
    }

    private FormulaNode disjunction() {
        FormulaNode left = conjunction();
        while(lexer.hasNext() && lexer.peek().isOrOperator()) {
            Token op = lexer.next();
            FormulaNode right = conjunction();

            left = binaryInfixCall(op, left, right);
        }
        return left;
    }

    private FormulaNode conjunction() {
        FormulaNode left = equality();
        while(lexer.hasNext() && lexer.peek().isAndOperator()) {
            Token op = lexer.next();
            FormulaNode right = equality();

            left = binaryInfixCall(op, left, right);
        }
        return left;
    }

    private FormulaNode equality() {
        FormulaNode left = relational();
        while(lexer.hasNext() && lexer.peek().isEqualityOperator()) {
            Token op = lexer.next();
            FormulaNode right = relational();

            left = binaryInfixCall(op, left, right);
        }
        return left;
    }

    private FormulaNode relational() {
        FormulaNode left = term();
        while(lexer.hasNext() && lexer.peek().isRelationalOperator()) {
            Token op = lexer.next();
            FormulaNode right = term();

            left = binaryInfixCall(op, left, right);
        }
        return left;
    }

    private FormulaNode term() {

        // <term> ::= <factor> | <term> + <factor> | <term> - <factor>

        FormulaNode left = factor();
        while (lexer.hasNext() &&  lexer.peek().isAdditiveOperator()) {
            Token op = lexer.next();
            FormulaNode right = factor();

            left = binaryInfixCall(op, left, right);
        }
        return left;
    }


    public FormulaNode factor() {

        // <factor> ::= <unary> | <factor> * <unary> | <factor> / <unary>

        FormulaNode left = unary();
        while(lexer.hasNext() && lexer.peek().isMultiplicativeOperator()) {
            Token op = lexer.next();
            FormulaNode right = unary();

            left = binaryInfixCall(op, left, right);
        }
        return left;
    }

    private FormulaNode unary() {
        // <unary> ::=  + <unary> | - <unary> | <unary2>

        if(!lexer.hasNext()) {
            throw unexpectedEnd();
        }
        Token token = lexer.peek();
        if(token.getType() == TokenType.OPERATOR) {
            if (token.getString().equals("-") ||
                token.getString().equals("+")) {

                Token opToken = lexer.next();
                FormulaNode operand = unary();
                SourceRange sourceRange = new SourceRange(opToken.getStart(), operand.getSourceRange().getEnd());
                return new FunctionCallNode(opToken.getString(), singletonList(operand), sourceRange);
            }
        }

        return unary2();
    }

    private FormulaSyntaxException unexpectedEnd() {
        Token lastToken = lexer.last();
        SourceRange sourceRange;
        if(lastToken == null) {
            sourceRange = new SourceRange(new SourcePos(0, 0), new SourcePos(0, 1));
        } else if(lastToken.getType() == TokenType.WHITESPACE) {
            sourceRange = new SourceRange(lastToken);
        } else {
            int lastLine = lastToken.getEnd().getLine();
            int lastColumn = lastToken.getEnd().getColumn();

            sourceRange = new SourceRange(
                    new SourcePos(lastLine, lastColumn - 1),
                    new SourcePos(lastLine, lastColumn));
        }
        return new FormulaSyntaxException(sourceRange, "Unexpected end of formula");
    }

    private FormulaNode unary2() {
        if(!lexer.hasNext()) {
            throw unexpectedEnd();
        }
        Token token = lexer.peek();
        if(token.getType() == TokenType.OPERATOR) {
            if(token.getString().equals("!")) {
                Token opToken = lexer.next();
                FormulaNode operand = primary();
                SourceRange sourceRange = new SourceRange(opToken.getStart(), operand.getSourceRange().getEnd());
                return new FunctionCallNode(opToken.getString(), singletonList(operand), sourceRange);
            }
        }
        return primary();
    }

    private FormulaNode primary() {
        switch (lexer.peek().getType()) {
            case SYMBOL:
                return symbolOrCall();
            case NUMBER:
                return number();
            case BOOLEAN_LITERAL:
                return booleanLiteral();
            case STRING_LITERAL:
                return stringLiteral();
            case PAREN_START:
                Token openToken = lexer.next();
                FormulaNode e = expression();
                Token closeToken = expect(TokenType.PAREN_END);
                return new GroupNode(e, new SourceRange(openToken, closeToken));

            default:
                throw new FormulaSyntaxException(new SourceRange(lexer.peek()),
                        "Expected a symbol, a number, a string, or '('");
        }
    }

    private FormulaNode stringLiteral() {
        Token token = lexer.next();
        if (token.getType() != TokenType.STRING_LITERAL) {
            throw new FormulaSyntaxException("Expected token type is " + TokenType.STRING_LITERAL + " but was " + token.getType());
        }
        return new ConstantNode(token.getString(), new SourceRange(token));
    }

    private FormulaNode booleanLiteral() {
        Token token = lexer.next();
        if (token.getType() != TokenType.BOOLEAN_LITERAL) {
            throw new FormulaSyntaxException("Expected token type is " + TokenType.BOOLEAN_LITERAL + " but was " + token.getType());
        }
        boolean value = token.getString().toLowerCase().equals("true");
        SourceRange source = new SourceRange(token);
        return new ConstantNode(value, source);
    }


    private FunctionCallNode binaryInfixCall(Token op, FormulaNode left, FormulaNode right) {
        SourceRange source = new SourceRange(left.getSourceRange(), right.getSourceRange());
        return new FunctionCallNode(op.getString(), Arrays.asList(left, right), source);
    }

    private FormulaNode symbolOrCall() {
        Token symbolToken = lexer.next();
        if (symbolToken.getType() != TokenType.SYMBOL) {
            throw new FormulaSyntaxException("Expected token type is " + TokenType.SYMBOL + " but was " + symbolToken.getType());
        }

        if(lexer.hasNext() && lexer.peek().getType() == TokenType.PAREN_START) {
            return call(symbolToken);
        } else {
            return compound(symbol(symbolToken));
        }
    }

    private FormulaNode call(Token functionToken) {
        expect(TokenType.PAREN_START);

        List<FormulaNode> arguments = new ArrayList<>();

        while(true) {
            if (!lexer.hasNext()) {
                throw unexpectedEnd();
            }
            TokenType nextToken = lexer.peek().getType();
            if (nextToken == TokenType.COMMA) {
                // Consume comma and parse next argument
                lexer.next();
                continue;
            }
            if (nextToken == TokenType.PAREN_END) {
                // consume paren and complete argument list
                Token closingParen = lexer.next();
                return new FunctionCallNode(functionToken, arguments, new SourceRange(functionToken, closingParen));
            }

            // Otherwise parse the next argument
            arguments.add(expression());
        }
    }

    private FormulaNode compound(SymbolNode symbol) {
        FormulaNode result = symbol;

        while(lexer.hasNext() && lexer.peek().isDot()) {
            lexer.next();
            SymbolNode field = symbol();
            result = new CompoundExpr(result, field, new SourceRange(result.getSourceRange(), field.getSourceRange()));
        }
        return result;
    }

    private SymbolNode symbol() {
        Token token = lexer.next();
        return symbol(token);
    }

    private SymbolNode symbol(Token token) {
        if (token.getType() != TokenType.SYMBOL) {
            throw new FormulaSyntaxException("Expected token type is " + TokenType.SYMBOL + " but was " + token.getType());
        }
        return new SymbolNode(token);
    }

    private FormulaNode number() {
        Token token = lexer.next();
        double value;
        try {
            value = Double.parseDouble(token.getString());
        } catch (NumberFormatException e) {
            throw new FormulaSyntaxException(new SourceRange(token), "Invalid number '" + token.getString() + "': " +
                    e.getMessage());
        }
        return new ConstantNode(value, new SourceRange(token));
    }

    public FormulaNode expression() {
        // Start with the operator with the highest precedence
        return disjunction();
    }

    public FormulaNode parse() {
        FormulaNode expr = expression();
        if(lexer.hasNext()) {
            Token extraToken = lexer.next();
            throw new FormulaSyntaxException(new SourceRange(extraToken), "Missing an operator like + - / *, etc.");
        }
        return expr;
    }

    public static Optional<FormulaNode> tryParse(String formula) {
        try {
            return Optional.of(parse(formula));

        } catch (FormulaException e) {
            return Optional.empty();
        }
    }

    public static FormulaNode parse(String expression) {
        if(Strings.isNullOrEmpty(expression)) {
            return new ConstantNode((String)null);
        }
        FormulaLexer lexer = new FormulaLexer(expression);
        FormulaParser parser = new FormulaParser(lexer);
        return parser.parse();
    }
}
