/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import org.activityinfo.model.formula.functions.*;
import org.activityinfo.model.formula.resolved.*;
import org.activityinfo.model.query.ColumnModel;

import java.util.ArrayList;
import java.util.List;

import static org.activityinfo.model.formula.FormulaParser.operatorPrecedence;

/**
 * Rewrites formulas to use ids instead of codes or labels where possible.
 */
public class FormulaUglifier implements ResolvedNodeVisitor<FormulaNode>, ReferenceVisitor<FormulaNode> {

    public static final FormulaUglifier INSTANCE = new FormulaUglifier();

    private FormulaUglifier() {
    }

    @Override
    public FormulaNode visitConstant(ConstantNode node) {
        return node;
    }

    @Override
    public FormulaNode visitInvalidSyntax(InvalidSyntaxNode node) {
        return node.toFormula();
    }

    @Override
    public FormulaNode visitReferenceNode(ResolvedReference node) {
        try {
            return node.getRoot().accept(this);
        } catch (UnrewrittableException e) {
            return node.toFormula();
        }
    }

    @Override
    public FormulaNode visitInvalidReferenceNode(InvalidReferenceNode node) {
        return node.toFormula();
    }

    @Override
    public FormulaNode visitCallNode(ResolvedCallNode node) {
        return transformCall(node, this);
    }

    public static FormulaNode transformCall(ResolvedCallNode node, ResolvedNodeVisitor<FormulaNode> visitor) {
        List<FormulaNode> arguments = new ArrayList<>();
        for (int i = 0; i < node.getArgumentCount(); i++) {
            ResolvedNode argument = node.getArgument(i);
            FormulaNode uglified = argument.accept(visitor);
            if(isGroupingRequired(node, i, argument)) {
                arguments.add(new GroupNode(uglified));
            } else {
                arguments.add(uglified);
            }
        }
        return new FunctionCallNode(node.getFunction(), arguments.toArray(new FormulaNode[0]));
    }

    private static boolean isGroupingRequired(ResolvedCallNode callNode, int argumentIndex, ResolvedNode argument) {
        FormulaFunction f = callNode.getFunction();
        if(!(f.isInfix() && argument instanceof ResolvedCallNode)) {
            return false;
        }

        ResolvedCallNode innerCall = (ResolvedCallNode) argument;
        FormulaFunction g = innerCall.getFunction();

        // Grouping only required for two adjacent infix operators
        if(!g.isInfix()) {
            return false;
        }
        // If the argument has greater precedence, correct order is implied
        if(operatorPrecedence(g) > operatorPrecedence(f)) {
            return false;
        }

        // If it's the same function, and the function is commutative, no grouping necessary
        if(f.equals(g) && f.isCommutative()) {
            return false;
        }

        // Special handling for 3 + 3 - 2 and 3 * 3 / 2
        if (isNotNeededForReasonsThatAreObviousButICannotQuiteArticulate(argumentIndex, f, g)) {
            return false;
        }

        return true;
    }

    private static boolean isNotNeededForReasonsThatAreObviousButICannotQuiteArticulate(int argumentIndex, FormulaFunction f, FormulaFunction g) {
        if(argumentIndex == 0) {
            if(f == SubtractFunction.INSTANCE && g == AddFunction.INSTANCE) {
                return true;
            }
            if(f == DivideFunction.INSTANCE && g == MultiplyFunction.INSTANCE) {
                return true;
            }
        }
        return false;
    }

    @Override
    public FormulaNode visitInvalidCall(InvalidCallNode invalidCallNode) {
        return invalidCallNode.toFormula();
    }

    @Override
    public FormulaNode visitInvalidArgumentNode(InvalidArgumentNode invalidArgumentNode) {
        throw new UnrewrittableException();
    }

    @Override
    public FormulaNode visitRecordId(RecordIdNode recordIdNode) {
        return new SymbolNode(ColumnModel.RECORD_ID_SYMBOL);
    }

    @Override
    public FormulaNode visitFormId(FormIdNode formIdNode) {
        return new SymbolNode(ColumnModel.FORM_ID_SYMBOL);
    }

    @Override
    public FormulaNode visitLastEditTime(LastEditTimeNode node) {
        return new SymbolNode(ColumnModel.LAST_EDIT_TIME_SYMBOL);
    }

    @Override
    public FormulaNode visitField(FieldReferenceNode field) {
        return new SymbolNode(field.getField().getId());
    }

    @Override
    public FormulaNode visitCalculatedField(CalculatedFieldNode field) {
        return new SymbolNode(field.getField().getId());
    }

    @Override
    public FormulaNode visitEnumItem(EnumItemReference enumItemReference) {
        return new CompoundExpr(enumItemReference.getField().accept(this), enumItemReference.getEnumItem().getId());
    }

    @Override
    public FormulaNode visitJoin(JoinQueryNode node) {
        ReferenceNode rightNode = node.getRight();
        FormulaNode tempLeft = node.getLeft().accept(this);
        while (rightNode instanceof JoinQueryNode) {
            FormulaNode left = ((JoinQueryNode) rightNode).getLeft().accept(this);
            tempLeft = new CompoundExpr(tempLeft, (SymbolNode) left);
            rightNode = ((JoinQueryNode) rightNode).getRight();
        }
        FormulaNode tempRight = rightNode.accept(this);
        if (tempRight instanceof CompoundExpr) {
            FormulaNode value = ((CompoundExpr) tempRight).getValue();
            FormulaNode field = ((CompoundExpr) tempRight).getField();
            tempLeft = new CompoundExpr(tempLeft, (SymbolNode) value);
            return new CompoundExpr(tempLeft, (SymbolNode) field);
        }
        FormulaNode right = rightNode.accept(this);
        if (right instanceof SymbolNode) {
            return new CompoundExpr(tempLeft, (SymbolNode) right);
        } else {
            return new ErrorFormulaNode("unhandled reference node: " + rightNode);
        }
    }

    @Override
    public FormulaNode visitUnion(UnionNode node) {
        throw new UnrewrittableException();
    }

    @Override
    public FormulaNode visitAggregation(AggregateNode node) {
        throw new UnrewrittableException();
    }

    private class UnrewrittableException extends RuntimeException {

    }
}
