/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import org.activityinfo.model.formula.functions.*;
import org.activityinfo.model.formula.functions.date.DateFunction;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.time.LocalDate;

import java.util.*;

/**
 * Functions for manipulating formulas
 */
public class Formulas {
    
    public static SymbolNode symbol(String name) {
        return new SymbolNode(name);
    }
    
    public static SymbolNode symbol(ResourceId resourceId) {
        return new SymbolNode(resourceId);
    }
    
    public static FunctionCallNode equals(FormulaNode x, FormulaNode y) {
        return new FunctionCallNode(EqualFunction.INSTANCE, x, y);
    }
    
    public static FunctionCallNode equals(ResourceId x, ResourceId y) {
        return new FunctionCallNode(EqualFunction.INSTANCE, symbol(x), symbol(y));
    }
    
    public static FormulaNode allTrue(Iterable<FormulaNode> nodes) {
        return binaryTree(AndFunction.INSTANCE, nodes);
    }

    public static FormulaNode allTrue(FormulaNode... nodes) {
        return allTrue(Arrays.asList(nodes));
    }
    
    public static FormulaNode anyTrue(Collection<FormulaNode> nodes) {
        if(nodes.isEmpty()) {
            return new ConstantNode(false);
        } else {
            return binaryTree(OrFunction.INSTANCE, nodes);
        }
    }
    
    private static FormulaNode binaryTree(FormulaFunction function, Iterable<FormulaNode> nodes) {
        Iterator<FormulaNode> it = nodes.iterator();
        FormulaNode expr = it.next();
        while(it.hasNext()) {
            expr = new FunctionCallNode(function, expr, it.next());
        }
        return expr;
    }

    public static FormulaNode idConstant(ResourceId id) {
        return new ConstantNode(id.asString());
    }
    
    public static FormulaNode idEqualTo(ResourceId id) {
        return equals(symbol(ColumnModel.RECORD_ID_SYMBOL), idConstant(id));
    }
    
    public static FormulaNode idEqualTo(Set<ResourceId> ids) {
        List<FormulaNode> conditions = new ArrayList<>();
        for (ResourceId id : ids) {
            conditions.add(idEqualTo(id));
        }
        return anyTrue(conditions);
    }

    public static FunctionCallNode call(FormulaFunction function, FormulaNode... argumentList) {
        return new FunctionCallNode(function, argumentList);
    }

    /**
     * Tries to decompose a tree of binary operations into a list of operands. (A && B && C) or
     * (A || B || C) or (A + B + C) => [A, B, C]
     */
    public static List<FormulaNode> findBinaryTree(FormulaNode rootNode, FormulaFunction operator) {
        List<FormulaNode> list = new ArrayList<>();
        findBinaryTree(rootNode, list, operator);

        return list;
    }

    private static void findBinaryTree(FormulaNode node, List<FormulaNode> list, FormulaFunction operator) {

        // Unwrap group expressions ((A))
        node = simplify(node);

        if(isBinaryOperation(node, operator)) {
            // If this expression is in the form A && B, then descend
            // recursively
            FunctionCallNode callNode = (FunctionCallNode) node;
            findBinaryTree(callNode.getArgument(0), list, operator);
            findBinaryTree(callNode.getArgument(1), list, operator);

        } else {
            // If not a conjunction, then add this node to the list
            list.add(node);
        }
    }

    public static FormulaNode simplify(FormulaNode node) {
        while(node instanceof GroupNode) {
            node = ((GroupNode) node).getExpr();
        }
        return node;
    }

    private static boolean isBinaryOperation(FormulaNode node, FormulaFunction operator) {
        if(node instanceof FunctionCallNode) {
            FunctionCallNode callNode = (FunctionCallNode) node;
            return callNode.getArgumentCount() == 2 &&
                   callNode.getFunctionName().equals(operator.getId());
        }
        return false;
    }

    public static FormulaNode date(LocalDate localDate) {
        return new FunctionCallNode(DateFunction.INSTANCE,
            new ConstantNode(localDate.getYear()),
            new ConstantNode(localDate.getMonthOfYear()),
            new ConstantNode(localDate.getDayOfMonth()));
    }

    public static Optional<LocalDate> localDateConstant(FormulaNode node) {
        if(node instanceof FunctionCallNode) {
            FunctionCallNode callNode = (FunctionCallNode) node;
            if(callNode.getFunctionName().equals(DateFunction.INSTANCE.getId()) && callNode.getArgumentCount() == 3) {
                int year = dateComponent(callNode.getArgument(0));
                int month = dateComponent(callNode.getArgument(1));
                int date = dateComponent(callNode.getArgument(2));

                if(year > 0 && month > 0 && date > 0) {
                    return Optional.of(new LocalDate(year, month, date));
                }
            }
        }
        return Optional.empty();
    }

    private static int dateComponent(FormulaNode node) {
        if(node instanceof ConstantNode) {
            ConstantNode constantNode = (ConstantNode) node;
            double doubleValue = constantNode.getDoubleValue();
            if (!Double.isNaN(doubleValue)) {
                return (int)doubleValue;
            }
        }
        return -1;
    }

    public static FormulaNode latitudeCentroid(FormulaNode geoArea) {
        return centroid(geoArea, BoundingBoxFunction.YMIN, BoundingBoxFunction.YMAX);
    }

    public static FormulaNode longitudeCentroid(FormulaNode geoArea) {
        return centroid(geoArea, BoundingBoxFunction.XMIN, BoundingBoxFunction.XMAX);
    }

    private static FunctionCallNode centroid(FormulaNode geoArea, BoundingBoxFunction min, BoundingBoxFunction max) {
        FunctionCallNode xmin = new FunctionCallNode(min, geoArea);
        FunctionCallNode xmax = new FunctionCallNode(max, geoArea);
        FunctionCallNode sum = new FunctionCallNode(AddFunction.INSTANCE, xmin, xmax);
        FunctionCallNode average = new FunctionCallNode(DivideFunction.INSTANCE, sum, new ConstantNode(2.0));
        return average;
    }

    public static FunctionCallNode parentFilter(String parentId) {
        return equals(new SymbolNode("@parent"), new ConstantNode(parentId));
    }

    public static FunctionCallNode grandParentFilter(String grandParentId) {
        return equals(new CompoundExpr(new SymbolNode("@parent"),
                new SymbolNode("@parent")), new ConstantNode(grandParentId));
    }
}
