/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import com.google.common.base.Joiner;
import org.activityinfo.model.formula.functions.FormulaFunction;
import org.activityinfo.model.formula.functions.FormulaFunctions;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.function.Function;

public class FunctionCallNode extends FormulaNode {

    private String function;

    private List<FormulaNode> arguments;


    public FunctionCallNode(String function, List<FormulaNode> arguments, SourceRange sourceRange) {
        super();
        this.function = function;
        this.arguments = arguments;
        this.sourceRange = sourceRange;
    }

    public FunctionCallNode(Token function, List<FormulaNode> arguments, SourceRange sourceRange) {
        this(function.getString(), arguments, sourceRange);
    }

    public FunctionCallNode(String function, List<FormulaNode> arguments) {
        this(function, arguments, null);
    }

    public FunctionCallNode(String function, FormulaNode... arguments) {
        this(function, Arrays.asList(arguments));
    }

    public FunctionCallNode(FormulaFunction function, FormulaNode... arguments) {
        this(function.getId(), Arrays.asList(arguments));
    }

    /**
     * To REMOVE
     */
    public FormulaFunction getFunction() {
        return FormulaFunctions.get(function).get();
    }

    public String getFunctionName() {
        return function;
    }

    public List<FormulaNode> getArguments() {
        return arguments;
    }

    public int getArgumentCount() {
        return arguments.size();
    }


    public FormulaNode getArgument(int i) {
        return arguments.get(i);
    }

    @Override
    public String toString() {
        return "(" + function + " " + Joiner.on(" ").join(arguments) + ")";
    }

    @Override
    public String deparse() {
        if(FormulaParser.isInfix(function)) {
            if(arguments.size() == 2) {
                return arguments.get(0).deparse() + " " + function + " " + arguments.get(1).deparse();
            } else if(arguments.size() == 1) {
                return function + arguments.get(0).deparse();
            }
        }
        StringBuilder expr = new StringBuilder();
        expr.append(function.toUpperCase());
        expr.append("(");

        boolean needsComma = false;
        for (FormulaNode arg : arguments) {
            if(needsComma) {
                expr.append(", ");
            }
            expr.append(arg.deparse());
            needsComma = true;
        }
        expr.append(")");

        return expr.toString();
    }


    @Override
    public <T> T accept(FormulaVisitor<T> visitor) {
        return visitor.visitFunctionCall(this);
    }

    @Override
    public FormulaNode transform(Function<FormulaNode, FormulaNode> function) {
        List<FormulaNode> transformedArgs = new ArrayList<>();
        for (FormulaNode argument : arguments) {
            transformedArgs.add(argument.transform(function));
        }
        return function.apply(new FunctionCallNode(this.function, transformedArgs));
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + (arguments.hashCode());
        result = prime * result + (function.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        FunctionCallNode other = (FunctionCallNode) obj;
        return other.function.equals(function) && other.arguments.equals(arguments);
    }


}
