/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;


import org.activityinfo.model.formula.functions.AndFunction;
import org.activityinfo.model.formula.functions.FormulaFunction;
import org.activityinfo.model.formula.functions.OrFunction;

import java.util.ArrayList;
import java.util.List;

/**
 * Prior to unifying the resolution of formulas between the query executor
 * and the relevance calculator, a number of differences between the implementations
 * arose.
 *
 * These included:
 *
 * <ul>
 *     <li>Enum fields evaluated to their ids in the relevance calculator, rather than their labels,
 *     as in the query engine.</li>
 *     <li>Enum literals could be expressed as symbols. This is introduces quite a bit of ambiguity in the evaluation
 *     of formulas, and so will be removed here. Fortunately, all enum ids started with 't'.</li>
 *     <li>The {@code containsAny} and {@code constainsAll} functions were a bit too magical: everywhere else an
 *     select field evaluates to a string, but only as the first argument to this function did they have a set
 *     value.</li>
 *
 * </ul>
 */
public class LegacyFormulaRewriter {

    public static FormulaNode rewrite(FormulaNode node) {
        if(node instanceof FunctionCallNode) {
            FunctionCallNode call = (FunctionCallNode) node;

            // Handle enum references as symbols
            // {Q1007820397}!={t1844566507}
            if(isEnumItem(call)) {
                return rewriteEnumItem(call);

            } else if(call.getArguments().size() > 1 &&
                    (call.getFunctionName().equals("containsAny") ||
                     call.getFunctionName().equals("containsAll"))) {

                return rewriteContains(call);

            } else {
                boolean changed = false;
                List<FormulaNode> rewrittenArguments = new ArrayList<>();
                for (FormulaNode argument : call.getArguments()) {
                    FormulaNode rewrittenArgument = rewrite(argument);
                    changed = changed || rewrittenArgument != argument;
                    rewrittenArguments.add(rewrittenArgument);
                }
                if(changed) {
                    return new FunctionCallNode(call.getFunctionName(), rewrittenArguments);
                }
            }
        } else if(node instanceof GroupNode) {
            GroupNode groupNode = (GroupNode) node;
            FormulaNode rewritten = rewrite(groupNode.getExpr());
            if(rewritten != groupNode.getExpr()) {
                return new GroupNode(rewritten);
            }
        }
        return node;
    }

    private static FormulaNode rewriteContains(FunctionCallNode call) {
        FormulaFunction joiner;
        if(call.getFunctionName().equals("containsAny")) {
            joiner = OrFunction.INSTANCE;
        } else {
            joiner = AndFunction.INSTANCE;
        }
        SymbolNode field = (SymbolNode) call.getArgument(0);
        FormulaNode result = new CompoundExpr(field, (SymbolNode) call.getArgument(1));

        for (int i = 2; i < call.getArgumentCount(); i++) {
            result = new FunctionCallNode(joiner, result, new CompoundExpr(field, (SymbolNode)call.getArgument(i)));
        }
        return result;
    }


    private static boolean isEnumItem(FunctionCallNode call) {
        if(call.getFunctionName().equals("==") || call.getFunctionName().equals("!=")) {
            if(call.getArgument(0) instanceof SymbolNode && call.getArgument(1) instanceof SymbolNode) {
                SymbolNode rhs = (SymbolNode) call.getArgument(1);
                if(rhs.getName().startsWith("t")) {
                    return true;
                }
            }
        }
        return false;
    }

    private static FormulaNode rewriteEnumItem(FunctionCallNode call) {
        FormulaNode itemRef = new CompoundExpr(call.getArgument(0), (SymbolNode) call.getArgument(1));
        if(call.getFunctionName().equals("!=")) {
            return new FunctionCallNode("!", itemRef);
        } else {
            return itemRef;
        }
    }

}
