/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula;

import org.activityinfo.model.formula.resolved.*;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;

import java.util.Map;

/**
 * Evaluates the current value of validation expression for reference fields, excepting a self field
 */
public class ValidationFormulaRewriter implements ResolvedNodeVisitor<FormulaNode>, ReferenceVisitor<FormulaNode> {

    private ResourceId selfFieldId;
    private Map<ResolvedNode, ConstantNode> fieldValueMap;

    /**
     * @param selfFieldId The ID of a self field
     * @param fieldValueMap The map of values for evaluated resolved nodes
     */

    public ValidationFormulaRewriter(ResourceId selfFieldId, Map<ResolvedNode, ConstantNode> fieldValueMap) {
        this.selfFieldId = selfFieldId;
        this.fieldValueMap = fieldValueMap;
    }

    @Override
    public FormulaNode visitRecordId(RecordIdNode recordIdNode) {
        return new SymbolNode(ColumnModel.RECORD_ID_SYMBOL);
    }

    @Override
    public FormulaNode visitFormId(FormIdNode formIdNode) {
        return new SymbolNode(ColumnModel.FORM_ID_SYMBOL);
    }

    @Override
    public FormulaNode visitLastEditTime(LastEditTimeNode node) {
        return new SymbolNode(ColumnModel.LAST_EDIT_TIME_SYMBOL);
    }

    @Override
    public FormulaNode visitField(FieldReferenceNode field) {
        /**
         * In the following case the reference node can be a FieldReferenceNode
         * Reference_Field == Record_Id
         * That is why we need to check if the reference is a self field or not
         */
        if (field.getField().getId().equals(selfFieldId)) {
            return new SymbolNode(ColumnModel.RECORD_ID_SYMBOL);
        } else {
            ResolvedReference resolvedReference = new ResolvedReference(new SymbolNode(field.getField().getId()), field);
            if (fieldValueMap.containsKey(resolvedReference)) {
                return fieldValueMap.get(resolvedReference);
            } else {
                return new SymbolNode(field.getField().getId());
            }
        }
    }

    @Override
    public FormulaNode visitCalculatedField(CalculatedFieldNode field) {
        return visitField(field);
    }

    @Override
    public FormulaNode visitEnumItem(EnumItemReference enumItemReference) {
        ResolvedReference resolvedReference = new ResolvedReference(enumItemReference.toFormula(), enumItemReference);
        if (fieldValueMap.containsKey(resolvedReference)) {
            return fieldValueMap.get(resolvedReference);
        } else {
            return enumItemReference.toFormula();
        }
    }

    @Override
    public FormulaNode visitJoin(JoinQueryNode node) {
        /**
         * A reference field can be expressed by a JoinQueryNode => we need to check if the reference field is
         * a self field or not.
         * For a self field the formula node of the right part of the join should be returned.
         */
        if (node.getLeft().getField().getId().equals(selfFieldId)) {
            ReferenceNode right = node.getRight();
            return right.accept(new ReferenceVisitor<FormulaNode>() {
                @Override
                public FormulaNode visitRecordId(RecordIdNode recordIdNode) {
                    return new SymbolNode(ColumnModel.RECORD_ID_SYMBOL);
                }

                @Override
                public FormulaNode visitFormId(FormIdNode formIdNode) {
                    return new SymbolNode(ColumnModel.FORM_ID_SYMBOL);
                }

                @Override
                public FormulaNode visitLastEditTime(LastEditTimeNode node) {
                    return new SymbolNode(ColumnModel.LAST_EDIT_TIME_SYMBOL);
                }

                @Override
                public FormulaNode visitField(FieldReferenceNode field) {
                    return new SymbolNode(field.getField().getId());
                }

                @Override
                public FormulaNode visitCalculatedField(CalculatedFieldNode field) {
                    return visitField(field);
                }

                @Override
                public FormulaNode visitEnumItem(EnumItemReference enumItemReference) {
                    return new CompoundExpr(new SymbolNode(enumItemReference.getField().getField().getId()),
                            enumItemReference.getEnumItem().getId());
                }

                @Override
                public FormulaNode visitJoin(JoinQueryNode node) {
                    FormulaNode formulaNode = getFormulaNode(node);
                    return formulaNode;
                }

                @Override
                public FormulaNode visitUnion(UnionNode node) {
                    return new ErrorFormulaNode("union node: " + node);
                }

                @Override
                public FormulaNode visitAggregation(AggregateNode node) {
                    //TODO
                    return new ErrorFormulaNode("aggregate node: " + node);
                }
            });
        } else {
            /**
             * If a reference field from validation expression is not a self field, the evaluated current value of this
             * field is returned
             */
            FormulaNode formulaNode = getFormulaNode(node);
            ResolvedNode resolvedFormula = new ResolvedReference(formulaNode, node);
            if (fieldValueMap.containsKey(resolvedFormula)) {
                return fieldValueMap.get(resolvedFormula);
            } else {
                return formulaNode;
            }
        }
    }

    // TODO add toFormula() method to ReferenceNode
    private FormulaNode getFormulaNode(JoinQueryNode join) {
        ReferenceNode rightNode = join.getRight();
        FormulaNode left = new SymbolNode(join.getLeft().getField().getId());
        while (rightNode instanceof JoinQueryNode) {
            ResourceId tempLeft = ((JoinQueryNode) rightNode).getLeft().getField().getId();
            left = new CompoundExpr(left, tempLeft);
            rightNode = ((JoinQueryNode) rightNode).getRight();
        }
        if (rightNode instanceof FieldReferenceNode) {
            return new CompoundExpr(left, ((FieldReferenceNode) rightNode).getField().getId());
        } else if (rightNode instanceof EnumItemReference) {
            EnumItemReference enumRef = (EnumItemReference) rightNode;
            left = new CompoundExpr(left, enumRef.getField().getField().getId());
            return new CompoundExpr(left, enumRef.getEnumItem().getId());
        } else if (rightNode instanceof RecordIdNode) {
            return left;
        } else {
            return new ErrorFormulaNode("unhandled right join query node: " + rightNode);
        }
    }

    @Override
    public FormulaNode visitUnion(UnionNode node) {
        return new ErrorFormulaNode("union node: " + node);
    }

    @Override
    public FormulaNode visitAggregation(AggregateNode node) {
        //TODO
        return new ErrorFormulaNode("aggregation node: " + node);
    }

    @Override
    public FormulaNode visitConstant(ConstantNode node) {
        return node;
    }

    @Override
    public FormulaNode visitInvalidSyntax(InvalidSyntaxNode node) {
        return new ErrorFormulaNode("union node: " + node);
    }

    @Override
    public FormulaNode visitReferenceNode(ResolvedReference node) {
        if (node.getRoot() instanceof FieldReferenceNode) {
            return visitField((FieldReferenceNode) node.getRoot());
        } else if (node.getRoot() instanceof JoinQueryNode) {
            return visitJoin((JoinQueryNode) node.getRoot());
        } else {
            if (fieldValueMap.containsKey(node)) {
                return fieldValueMap.get(node);
            } else {
                return node.toFormula();
            }
        }
    }

    @Override
    public FormulaNode visitInvalidReferenceNode(InvalidReferenceNode node) {
        return new ErrorFormulaNode("invalid reference node: " + node);
    }

    @Override
    public FormulaNode visitCallNode(ResolvedCallNode node) {
        return FormulaUglifier.transformCall(node, this);
    }

    @Override
    public FormulaNode visitInvalidCall(InvalidCallNode invalidCallNode) {
        return new ErrorFormulaNode("invalid call node: " + invalidCallNode);
    }

    @Override
    public FormulaNode visitInvalidArgumentNode(InvalidArgumentNode invalidArgumentNode) {
        return new ErrorFormulaNode("invalid argument node: " + invalidArgumentNode);
    }
}
