/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.eval;

import org.activityinfo.model.formula.ConstantNode;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.functions.ColumnFunction;
import org.activityinfo.model.formula.resolved.*;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldType;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.model.type.ReferenceValue;
import org.activityinfo.model.type.enumerated.EnumValue;
import org.activityinfo.observable.Observable;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;

/**
 * Evaluates a formula by querying for individual records.
 *
 */
public final class RecordBasedEvaluator {

    private static final Observable<ColumnView> JUST_BLANK = Observable.just(ColumnViews.BLANK_STRING);

    private final ResourceId formId;
    private final FieldValueProvider fieldValueProvider;

    public RecordBasedEvaluator(ResourceId formId, FieldValueProvider fieldValueProvider) {
        this.formId = formId;
        this.fieldValueProvider = fieldValueProvider;
    }

    public Observable<ColumnView> evaluate(ResourceId recordId, ResolvedNode node) {
        return evaluate(new RecordRef(formId, recordId), node);
    }

    public Observable<ColumnView> evaluate(RecordRef ref, ResolvedNode node) {
        return node.accept(new ResolvedNodeVisitor<Observable<ColumnView>>() {
            @Override
            public Observable<ColumnView> visitConstant(ConstantNode node) {
                return Observable.just(node.getValue());
            }

            @Override
            public Observable<ColumnView> visitInvalidSyntax(InvalidSyntaxNode node) {
                return JUST_BLANK;
            }

            @Override
            public Observable<ColumnView> visitReferenceNode(ResolvedReference node) {
                return evaluateReference(ref, node.getRoot());
            }

            @Override
            public Observable<ColumnView> visitInvalidCall(InvalidCallNode invalidCallNode) {
                return JUST_BLANK;
            }

            @Override
            public Observable<ColumnView> visitInvalidArgumentNode(InvalidArgumentNode invalidArgumentNode) {
                return JUST_BLANK;
            }

            @Override
            public Observable<ColumnView> visitInvalidReferenceNode(InvalidReferenceNode node) {
                return JUST_BLANK;
            }

            @Override
            public Observable<ColumnView> visitCallNode(ResolvedCallNode node) {

                List<Observable<ColumnView>> inputs = new ArrayList<>();
                for (int i = 0; i < node.getArgumentCount(); i++) {
                    inputs.add(evaluate(ref, node.getArgument(i)));
                }

                return Observable.flatten(inputs).transform(evaluated -> {
                    return ((ColumnFunction) node.getFunction()).apply(1, evaluated);
                });
            }

        });
    }

    private Observable<ColumnView> evaluateReference(RecordRef ref, ReferenceNode node) {
        return node.accept(new ReferenceVisitor<Observable<ColumnView>>() {
            @Override
            public Observable<ColumnView> visitRecordId(RecordIdNode recordIdNode) {
                return Observable.just(ColumnViews.of(ref.getRecordId().asString()));
            }

            @Override
            public Observable<ColumnView> visitFormId(FormIdNode formIdNode) {
                return Observable.just(ColumnViews.of(ref.getFormId().asString()));
            }

            @Override
            public Observable<ColumnView> visitLastEditTime(LastEditTimeNode node) {
                return fieldValueProvider.getLastEditTime(ref).transform(time -> ColumnViews.of(time));
            }

            @Override
            public Observable<ColumnView> visitField(FieldReferenceNode field) {
                FieldType fieldType = field.getField().getType();

                return fieldValueProvider.getFieldValue(ref, field.getField().getId()).transform(optional -> {
                    return optional
                        .map(fieldValue -> fieldType.toFormulaValue(fieldValue))
                        .orElse(ColumnViews.blank(fieldType.getFormulaType()));
                });
            }

            @Override
            public Observable<ColumnView> visitCalculatedField(CalculatedFieldNode field) {
                return evaluate(ref, field.getResolved());
            }

            @Override
            public Observable<ColumnView> visitEnumItem(EnumItemReference enumItemRef) {
                return fieldValueProvider.getFieldValue(ref, enumItemRef.getField().getField().getId())
                    .transform(optional -> {
                        return optional.map(fieldValue -> {
                            if(fieldValue instanceof EnumValue) {
                                EnumValue enumValue = (EnumValue) fieldValue;
                                return ColumnViews.of(enumValue.getResourceIds().contains(enumItemRef.getEnumItem().getId()));
                            } else {
                                return ColumnViews.BLANK_BOOLEAN;
                            }
                        }).orElse(ColumnViews.BLANK_BOOLEAN);
                    });
            }

            @Override
            public Observable<ColumnView> visitJoin(JoinQueryNode node) {

                Observable<Optional<RecordRef>> referencedRecord;
                if(node.getLeft().getField().getId().asString().equals(ColumnModel.PARENT_SYMBOL)) {
                    referencedRecord = fieldValueProvider.getParentRef(ref);
                } else {
                    referencedRecord = fieldValueProvider.getFieldValue(ref, node.getLeft().getField().getId())
                        .transform(optional -> {
                            return optional.flatMap(fieldValue -> {
                                if (fieldValue instanceof ReferenceValue) {
                                    ReferenceValue referenceValue = (ReferenceValue) fieldValue;
                                    for (RecordRef reference : referenceValue.getReferences()) {
                                        if (reference.getFormId().equals(node.getRightFormId())) {
                                            return Optional.of(reference);
                                        }
                                    }
                                }
                                return Optional.<RecordRef>empty();
                            });
                        });
                }

                return referencedRecord.join(optional -> optional
                    .map(r -> evaluateReference(r, node.getRight()))
                    .orElse(node.getResolvedFormulaType().map(
                        type -> {
                            if (type == FormulaType.NUMBER) {
                                return Observable.just(ColumnViews.BLANK_NUMBER);
                            } else {
                                return JUST_BLANK;
                            }
                        }).orElse(JUST_BLANK))
                    );
            }

            @Override
            public Observable<ColumnView> visitUnion(UnionNode node) {
                Iterator<ReferenceNode> it = node.getChildren().iterator();
                Observable<ColumnView> result = evaluateReference(ref, it.next());
                while(it.hasNext()) {
                    ReferenceNode next = it.next();
                    result = result.join(v -> {
                        if(!v.isBlank(0)) {
                            return Observable.just(v);
                        } else {
                            return evaluateReference(ref, next);
                        }
                    });
                }
                return result;
            }

            @Override
            public Observable<ColumnView> visitAggregation(AggregateNode node) {
                // TODO to handle aggregate nodes in relevance/validation rules
                return Observable.just(ColumnViews.BLANK_NUMBER);
            }
        });
    }
}
