/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import com.google.common.annotations.VisibleForTesting;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.query.*;

import java.util.List;

/**
 * Takes the first non-null value of its arguments
 */
public class CoalesceFunction extends FormulaFunction implements ColumnFunction {

    public static final CoalesceFunction INSTANCE = new CoalesceFunction();

    private CoalesceFunction() {
    }

    @Override
    public String getId() {
        return "coalesce";
    }

    @Override
    public FunctionCategory getCategory() {
        return FunctionCategory.MISC;
    }

    @Override
    public int getReferenceTopicId() {
        return 889877;
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        if(argumentTypes.isEmpty()) {
            return FormulaType.NUMBER;
        }
        FormulaType resultType = argumentTypes.get(0);

        return resultType;
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        if(arguments.isEmpty()) {
            return new BlankColumnView(ColumnType.NUMBER, numRows);
        }
        ColumnView[] cols = arguments.toArray(new ColumnView[arguments.size()]);
        ColumnType columnType = cols[0].getType();

        switch(columnType) {
            case STRING:
                return combineString(cols);
            case NUMBER:
                return combineDouble(cols);
            default:
            case BOOLEAN:
                return combineBoolean(cols);
        }
    }


    private ColumnView combineString(ColumnView[] cols) {
        int numRows = cols[0].numRows();
        int numCols = cols.length;

        String[] values = new String[numRows];

        for(int i=0;i!=numRows;++i) {
            for(int j=0;j!=numCols;++j) {
                String value = cols[j].getString(i);
                if(value != null) {
                    values[i] = value;
                    break;
                }
            }
        }

        return new StringArrayColumnView(values);
    }

    @VisibleForTesting
    static ColumnView combineDouble(ColumnView[] cols) {
        int numRows = cols[0].numRows();
        int numCols = cols.length;

        double[] values = new double[numRows];

        for(int i=0;i!=numRows;++i) {
            values[i] = Double.NaN;
            for(int j=0;j!=numCols;++j) {
                double value = cols[j].getDouble(i);
                if(!Double.isNaN(value)) {
                    values[i] = value;
                    break;
                }
            }
        }
        return new DoubleArrayColumnView(values);
    }


    private ColumnView combineBoolean(ColumnView[] cols) {
        int numRows = cols[0].numRows();
        int numCols = cols.length;

        int[] values = new int[numRows];

        for(int i=0;i!=numRows;++i) {
            values[i] = ColumnView.BLANK;
            for(int j=0;j!=numCols;++j) {
                int value = cols[j].getBoolean(i);
                if(value != ColumnView.BLANK) {
                    values[i] = value;
                    break;
                }
            }
        }
        return new BooleanColumnView(values);
    }
}
