/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.query.BooleanColumnView;
import org.activityinfo.model.query.ColumnType;
import org.activityinfo.model.query.ColumnView;

import java.util.List;

public abstract class ComparisonOperator extends FormulaFunction implements ColumnFunction {

    private final String name;

    public ComparisonOperator(String name) {
        this.name = name;
    }

    @Override
    public final String getId() {
        return name;
    }

    @Override
    public final FunctionCategory getCategory() {
        return FunctionCategory.LOGICAL;
    }

    /**
     *
     * @return the {@code ComparisonOperator} that whose result is the opposite
     * of this Operator.
     */
    public abstract ComparisonOperator negate();

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        checkArity(arguments,2);

        ColumnView a = arguments.get(0);
        ColumnView b = arguments.get(1);

        ColumnType typeA = a.getType();
        ColumnType typeB = b.getType();

        if ((typeA == ColumnType.NUMBER || typeA == ColumnType.BOOLEAN) &&
            (typeB == ColumnType.NUMBER || typeB == ColumnType.BOOLEAN)) {
            return columnApplyNumber(a, b);

        } else if(typeA == ColumnType.STRING && typeB == ColumnType.STRING) {
            return columnApplyString(a, b);

        } else {
            throw new FormulaSyntaxException("Comparison between incompatible types: " + typeA + ", " + typeB);
        }
    }

    private ColumnView columnApplyNumber(ColumnView x, ColumnView y) {
        int result[] = new int[x.numRows()];
        for (int i = 0; i < result.length; i++) {
            double xi = x.getDouble(i);
            double yi = y.getDouble(i);
            if(Double.isNaN(xi) || Double.isNaN(yi)) {
                result[i] = ColumnView.BLANK;
            } else {
                result[i] = apply(xi, yi) ? 1 : 0;
            }
        }
        return new BooleanColumnView(result);
    }

    private ColumnView columnApplyString(ColumnView a, ColumnView b) {
        int result[] = new int[a.numRows()];
        for(int i=0; i < result.length; ++i) {
            String xi = a.getString(i);
            String yi = b.getString(i);
            if(xi == null || yi == null) {
                result[i] = ColumnView.BLANK;
            } else {
                result[i] = apply(xi, yi) ? 1 : 0;
            }
        }
        return new BooleanColumnView(result);
    }

    private boolean apply(String xi, String yi) {
        // We can use the comparison result (-1, 0, 1)
        // as in input into the numeric apply() function
        // for example:
        // xi == yi iff xi.compare(yi) == 0
        // xi <= yi iff xi.compare(yi) <= 0
        // etc
        int comparison = xi.compareTo(yi);

        return apply(comparison, 0);
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        checkArity(argumentTypes, 2);

        FormulaType argumentType1 = argumentTypes.get(0);
        FormulaType argumentType2 = argumentTypes.get(1);
        if (!argumentType1.equals(argumentType2)) {
            throw new FormulaSyntaxException("Comparison between incompatible types: " + argumentType1 + ", " + argumentType2);
        }

        return FormulaType.BOOLEAN;
    }

    protected abstract boolean apply(double x, double y);
}
