/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.FormulaParser;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.ArgumentException;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;

import java.util.List;
import java.util.Set;

public abstract class FormulaFunction {

    /**
     * Returns string representation of function (as it is used in expression, e.g. + - ||)
     *
     * @return string representation of function (as it is used in expression)
     */
    public abstract String getId();

    /**
     * Returns a string that can be used to further describe an operator, or "" if {@link #getId()} is sufficient.
     */
    public String getHelpLabel() {
        return "";
    }

    public boolean isCommutative() {
        return false;
    }

    /**
     * Returns the user interface category to which this function belongs
     */
    public abstract FunctionCategory getCategory();

    /**
     * Returns the id of the ScreenSteps topic in the formula reference manual.
     */
    public abstract int getReferenceTopicId();

    @Override
    public String toString() {
        return getId();
    }

    /**
     * @return the FieldTypeClass of the result given the argumentTypes as input
     * @param argumentTypes
     */
    public abstract FormulaType resolveResultType(List<FormulaType> argumentTypes);

    protected void checkArity(List<?> arguments, int expectedArity) {
        if(arguments.size() != expectedArity) {
            throw new FormulaSyntaxException("The " + getId() + "() function expects " + expectedArity + " argument(s).");
        }
    }

    protected void checkArgumentType(List<FormulaType> argumentTypes, int argumentIndex, FormulaType expectedType) {
        if(argumentTypes.get(argumentIndex) != expectedType) {
            throw new ArgumentException(argumentIndex, expectedType, argumentTypes.get(argumentIndex));
        }
    }

    protected void checkArgumentType(List<FormulaType> argumentTypes, int argumentIndex, Set<FormulaType> expectedTypes) {
        if(!expectedTypes.contains(argumentTypes.get(argumentIndex))) {
            throw new ArgumentException(argumentIndex, expectedTypes, argumentTypes.get(argumentIndex));
        }
    }

    public final boolean isInfix() {
        return FormulaParser.isInfix(getId());
    }

}
