/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.functions.date.*;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

public final class FormulaFunctions {


    /**
     * Avoids race conditions by using the <a href="https://en.wikipedia.org/wiki/Initialization-on-demand_holder_idiom">
     *     Initialization-on-demand holder idiom</a>
     */
    private static class MapHolder {
        private static final MapHolder INSTANCE = new MapHolder();
        
        private Map<String, FormulaFunction> lookupMap = new HashMap<>();

        private void register(FormulaFunction function) {
            lookupMap.put(function.getId().toLowerCase(), function);
        }
        
        public MapHolder() {
            register(AndFunction.INSTANCE);
            register(DivideFunction.INSTANCE);
            register(EqualFunction.INSTANCE);
            register(SubtractFunction.INSTANCE);
            register(MultiplyFunction.INSTANCE);
            register(PowerFunction.INSTANCE);
            register(NotEqualFunction.INSTANCE);
            register(NotFunction.INSTANCE);
            register(OrFunction.INSTANCE);
            register(AddFunction.INSTANCE);
            register(BooleanFunctions.GREATER);
            register(BooleanFunctions.GREATER_OR_EQUAL);
            register(BooleanFunctions.LESS);
            register(BooleanFunctions.LESS_OR_EQUAL);

            register(SumFunction.INSTANCE);
            register(AverageFunction.INSTANCE);
            register(MedianFunction.INSTANCE);
            register(MaxFunction.INSTANCE);
            register(MinFunction.INSTANCE);
            register(CountFunction.INSTANCE);
            register(CountDistinctFunction.INSTANCE);

            register(BoundingBoxFunction.XMIN);
            register(BoundingBoxFunction.YMIN);
            register(BoundingBoxFunction.XMAX);
            register(BoundingBoxFunction.YMAX);
            register(IfFunction.INSTANCE);

            register(DateFunction.INSTANCE);
            register(YearFunction.INSTANCE);
            register(MonthFunction.INSTANCE);
            register(QuarterFunction.INSTANCE);
            register(DayFunction.INSTANCE);
            register(YearFracFunction.INSTANCE);
            register(TodayFunction.INSTANCE);

            register(CeilingFunction.INSTANCE);
            register(FloorFunction.INSTANCE);

            register(SearchFunction.INSTANCE);
            register(IsNumberFunction.INSTANCE);
            register(IsBlankFunction.INSTANCE);

            register(AddDateFunction.INSTANCE);
            register(DaysFunction.INSTANCE);

            register(CoalesceFunction.INSTANCE);

            register(ConcatenateFunction.INSTANCE);
            register(LeftFunction.INSTANCE);
            register(RightFunction.INSTANCE);
            register(MidFunction.INSTANCE);
            register(RegexMatchFunction.INSTANCE);
            register(RegexExtractFunction.INSTANCE);
            register(RegexReplaceFunction.INSTANCE);

            register(TextFunction.INSTANCE);
        }
        
        public Optional<FormulaFunction> get(String name) {
            return Optional.ofNullable(lookupMap.get(name.toLowerCase()));
        }
    }
    
    private FormulaFunctions() {
    }

    public static Optional<FormulaFunction> get(String name) {
        return MapHolder.INSTANCE.get(name);
    }

    public static Stream<FormulaFunction> stream() {
        return MapHolder.INSTANCE.lookupMap.values().stream();
    }
}
