/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.ArgumentException;
import org.activityinfo.model.query.*;

import java.util.Arrays;
import java.util.List;


public class IfFunction extends FormulaFunction implements ColumnFunction {


    public static final IfFunction INSTANCE = new IfFunction();

    private IfFunction() {
    }

    @Override
    public String getId() {
        return "if";
    }

    @Override
    public FunctionCategory getCategory() {
        return FunctionCategory.LOGICAL;
    }

    @Override
    public int getReferenceTopicId() {
        return 878518;
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        checkArity(argumentTypes, 3);

        FormulaType conditionType = argumentTypes.get(0);
        if(conditionType != FormulaType.BOOLEAN) {
            throw new ArgumentException(0, FormulaType.BOOLEAN, argumentTypes.get(0));
        }

        FormulaType trueType = argumentTypes.get(1);
        FormulaType falseType = argumentTypes.get(2);
        if(trueType != falseType) {
            throw new ArgumentException(2, "Must have the same type as the TRUE argument.");
        }
        return trueType;
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        checkArity(arguments, 3);

        ColumnView condition = arguments.get(0);
        ColumnView x = arguments.get(1);
        ColumnView y = arguments.get(2);

        if(x.getType() == ColumnType.NUMBER) {
            double values[] = new double[x.numRows()];
            Arrays.fill(values, Double.NaN);

            for (int i = 0; i < values.length; i++) {
                switch (condition.getBoolean(i)) {
                    case ColumnView.TRUE:
                        values[i] = x.getDouble(i);
                        break;
                    default:
                        values[i] = y.getDouble(i);
                        break;
                }
            }
            return new DoubleArrayColumnView(values);

        } else if(x.getType() == ColumnType.STRING) {
            String[] values = new String[x.numRows()];
            for (int i = 0; i < values.length; i++) {
                switch (condition.getBoolean(i)) {
                    case ColumnView.TRUE:
                        values[i] = x.getString(i);
                        break;
                    default:
                        values[i] = y.getString(i);
                        break;
                }
            }
            return new StringArrayColumnView(values);

        } else {
            int values[] = new int[x.numRows()];
            Arrays.fill(values, ColumnView.BLANK);
            for (int i = 0; i < values.length; i++) {
                switch (condition.getBoolean(i)) {
                    case ColumnView.TRUE:
                        values[i] = x.getBoolean(i);
                        break;
                    default:
                        values[i] = y.getBoolean(i);
                        break;
                }
            }
            return new BooleanColumnView(values);
        }
    }
}