/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.ArgumentException;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.DoubleArrayColumnView;

import java.util.List;

public abstract class RealValuedFunction extends FormulaFunction implements ColumnFunction {

    private String name;

    protected RealValuedFunction(String name) {
        this.name = name;
    }

    @Override
    public final FunctionCategory getCategory() {
        return FunctionCategory.MATH;
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        //must be unary or binary function
        if(arguments.size() == 1) {
            return unaryFunctionColumnCall(numRows,arguments.get(0));
        } else if(arguments.size() == 2) {
            return binaryFunctionColumnCall(numRows,arguments);
        } else {
            throw new FormulaSyntaxException("The " + getLabel() + "() function expects exactly 1 or 2 argument(s).");
        }
    }

    private ColumnView unaryFunctionColumnCall(int numRows, ColumnView argument) {
        double[] result = new double[numRows];
        for (int i = 0; i < numRows; i++) {
            double d = argument.getDouble(i);
            if(Double.isNaN(d)) {
                result[i] = Double.NaN;
            } else {
                result[i] = apply(d);
            }
        }
        return new DoubleArrayColumnView(result);
    }

    private ColumnView binaryFunctionColumnCall(int numRows, List<ColumnView> arguments) {
        ColumnView x = arguments.get(0);
        ColumnView y = arguments.get(1);

        double result[] = new double[x.numRows()];
        for (int i = 0; i < result.length; i++) {
            double xd = x.getDouble(i);
            double yd = y.getDouble(i);
            if(Double.isNaN(xd) && Double.isNaN(yd)) {
                result[i] = Double.NaN;
            } else {
                if (Double.isNaN(xd)) {
                    xd = 0;
                }
                if( Double.isNaN(yd)) {
                    yd = 0;
                }
                result[i] = apply(xd, yd);
            }
        }
        return new DoubleArrayColumnView(result);
    }

    @Override
    public final String getId() {
        return name;
    }

    public final String getLabel() {
        return name;
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {

        for (int i = 0; i < argumentTypes.size(); i++) {
            if (argumentTypes.get(i) != FormulaType.NUMBER) {
                throw new ArgumentException(i, FormulaType.NUMBER, argumentTypes.get(i));
            }
        }

        return FormulaType.NUMBER;
    }

    protected abstract double apply(double a);

    protected abstract double apply(double a, double b);

}
