/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import com.google.common.base.Strings;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ConstantColumnView;
import org.activityinfo.model.query.DoubleArrayColumnView;

import java.util.List;

/**
 * Returns the location of a substring in a string. The search is NOT case-sensitive.
 *
 * <p>If no match is found, the result is NaN</p>
 *
 * <p>The return index as well as the search index is 1-based.</p>
 */
public class SearchFunction extends FormulaFunction implements ColumnFunction {

    public static final SearchFunction INSTANCE = new SearchFunction();

    private SearchFunction() {
    }

    @Override
    public String getId() {
        return "SEARCH";
    }

    @Override
    public FunctionCategory getCategory() {
        return FunctionCategory.TEXT;
    }

    @Override
    public int getReferenceTopicId() {
        return 889879;
    }

    private double apply(String substring, String string, double startPositionArgument) {

        if(!Double.isFinite(startPositionArgument) || startPositionArgument > Integer.MAX_VALUE) {
            return Double.NaN;
        }

        if(Strings.isNullOrEmpty(substring) || Strings.isNullOrEmpty(string)) {
            return Double.NaN;
        }

        int startIndex = (int)startPositionArgument - 1;

        int index = string.toLowerCase().indexOf(substring.toLowerCase(), startIndex);
        if(index == -1) {
            return Double.NaN;
        } else {
            return index + 1;
        }
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        checkArity(argumentTypes.size());
        checkArgumentType(argumentTypes, 0, FormulaType.STRING);
        checkArgumentType(argumentTypes, 1, FormulaType.STRING);
        if (argumentTypes.size() == 3) {
            checkArgumentType(argumentTypes, 2, FormulaType.NUMBER);
        }
        return FormulaType.NUMBER;
    }

    private void checkArity(int count) {
        if(count < 2 || count > 3) {
            throw new FormulaSyntaxException("SEARCH() expects two or three arguments.");
        }
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        ColumnView substring = arguments.get(0);
        ColumnView string = arguments.get(1);
        ColumnView startIndex;
        if(arguments.size() == 3) {
            startIndex = arguments.get(2);
        } else {
            startIndex = new ConstantColumnView(numRows, 1d);
        }

        double[] result = new double[numRows];
        for (int i = 0; i < numRows; i++) {
            result[i] = apply(
                substring.getString(i),
                string.getString(i),
                startIndex.getDouble(i));
        }

        return new DoubleArrayColumnView(result);
    }
}
