/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.DoubleArrayColumnView;

import java.util.List;

/**
 * Base class for functions which return a statistical summary of values
 */
public abstract class StatFunction extends FormulaFunction implements ColumnFunction {


    @Override
    public final FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        return FormulaType.NUMBER;
    }

    @Override
    public final FunctionCategory getCategory() {
        return FunctionCategory.STATISTICAL;
    }


    /**
     * Computes the statistic over the range [{@code start}, {@code end}) 
     * of values in the given array.
     *
     * <p>The implementation may choose to sort the section of the array [start, end) in place. </p>
     */
    public abstract double compute(double[] values, int start, int end);


    /**
     * @return the result of this function given an empty set of input values.
     */
    public double emptyValue() {
        return Double.NaN;
    }

    @Override
    public final ColumnView apply(int numRows, List<ColumnView> arguments) {

        // Apply the statistic to each row in the table,
        // over the columns

        int numCols = arguments.size();

        double[] result = new double[numRows];

        double[] argumentValues = new double[arguments.size()];

        for(int i=0;i<numRows;++i) {

            // Collect the value from each column in the row
            for(int j=0;j<numCols;++j) {
                argumentValues[j] = arguments.get(j).getDouble(i);
            }

            // Compute the statistic for the row
            result[i] = compute(argumentValues, 0, numCols);
        }

        return new DoubleArrayColumnView(result);
    }

}
