/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.ArgumentException;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.StringArrayColumnView;

import java.util.List;

public abstract class SubstringFunction extends FormulaFunction implements ColumnFunction {

    private static final int MAX_ARGS = 3;

    protected static final int TEXT_ARG = 0;
    protected static final int NUM_CHARS_ARG = 1;
    protected static final int START_POS_ARG = 2;

    private String name;
    private int numArgs;

    public SubstringFunction(String name, int numArgs) {
        assert numArgs <= MAX_ARGS;
        this.name = name;
        this.numArgs = numArgs;
    }

    @Override
    public String getId() {
        return name;
    }

    @Override
    public FunctionCategory getCategory() {
        return FunctionCategory.TEXT;
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        checkArity(argumentTypes, numArgs);
        ArgumentException.check(argumentTypes, 0, FormulaType.STRING);
        ArgumentException.check(argumentTypes, 1, FormulaType.NUMBER);
        if(numArgs == 3) {
            ArgumentException.check(argumentTypes, 2, FormulaType.NUMBER);
        }

        return FormulaType.STRING;
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        checkArity(arguments, numArgs);

        String[] substring = new String[numRows];
        for (int i=0; i<numRows; i++) {
            substring[i] = substring(text(i,arguments), startPosition(i,arguments), numChars(i,arguments));
        }

        return new StringArrayColumnView(substring);
    }

    private String substring(String text, double startPosition, double numChars) {
        if (text == null) {
            return null;
        }
        if (Double.isNaN(startPosition) || Double.isInfinite(startPosition) || startPosition > Integer.MAX_VALUE) {
            return null;
        }
        if (Double.isNaN(numChars) || Double.isInfinite(numChars) || numChars > Integer.MAX_VALUE) {
            return null;
        }
        return computeSubstring(text, startPosition, numChars);
    }

    private String text(int row, List<ColumnView> arguments) {
        return arguments.get(TEXT_ARG).getString(row);
    }

    private double numChars(int row, List<ColumnView> arguments) {
        return arguments.get(NUM_CHARS_ARG).getDouble(row);
    }

    protected abstract double startPosition(int row, List<ColumnView> arguments);

    protected abstract String computeSubstring(String text, double startPosition, double numChars);



}
