/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import com.google.common.base.Strings;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.query.BlankColumnView;
import org.activityinfo.model.query.ColumnType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.StringArrayColumnView;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.function.DoubleFunction;
import java.util.logging.Logger;
import java.util.stream.Collectors;

public class TextFunction extends FormulaFunction implements ColumnFunction {

    private static final Logger LOGGER = Logger.getLogger(TextFunction.class.getName());

    public static final TextFunction INSTANCE = new TextFunction();

    private TextFunction() {
    }

    @Override
    public String getId() {
        return "text";
    }

    @Override
    public FunctionCategory getCategory() {
        return FunctionCategory.TEXT;
    }

    @Override
    public int getReferenceTopicId() {
        return 0;
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        List<FormulaType> formulaTypes= arguments
                .stream()
                .map(a -> FormulaType.of(a.getType()))
                .collect(Collectors.toList());
        checkArity(formulaTypes);

        ColumnView argument = arguments.get(0);
        ColumnView formatPattern = findFormatPattern(numRows, arguments);

        ColumnType columnType = argument.getType();
        String[] strings = new String[numRows];
        apply(numRows, argument, formatPattern, columnType, strings);
        return new StringArrayColumnView(strings);
    }

    private ColumnView findFormatPattern(int numRows, List<ColumnView> arguments) {
        ColumnView formatPattern;
        if (arguments.size() == 2) {
            formatPattern = arguments.get(1);
        } else {
            formatPattern = new BlankColumnView(ColumnType.STRING, numRows);
        }
        return formatPattern;
    }

    private void apply(int numRows, ColumnView argument, ColumnView formatPattern, ColumnType columnType, String[] strings) {

        String lastPattern = "";
        DoubleFunction<String> formatter = null;

        for (int i = 0; i < numRows; i++) {
            if (!argument.isBlank(i)) {
                switch (columnType) {
                    case STRING:
                        strings[i] = argument.getString(i);
                        break;
                    case BOOLEAN:
                        strings[i] = argument.isTrue(i) ? "TRUE" : "FALSE";
                        break;
                    case NUMBER:
                        String formatString = formatPattern.getString(i);
                        if(formatString != lastPattern) {
                            formatter = createFormatter(formatString);
                            lastPattern = formatString;
                        }
                        strings[i] = formatter.apply(argument.getDouble(i));
                        break;
                    default:
                        LOGGER.severe("unknown column type for TEXT() function: " + columnType);
                        break;
                }
            }
        }
    }

    private DoubleFunction<String> createFormatter(String formatString) {
        if(Strings.isNullOrEmpty(formatString)) {
            return Double::toString;
        }
        return FormatterFactory.createFormatter(formatString);
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        checkArity(argumentTypes);
        Set<FormulaType> expectedTypes = createExpectedTypeSet();
        checkArgumentType(argumentTypes, 0, expectedTypes);
        if (argumentTypes.size() == 2) {
            checkArgumentType(argumentTypes, 1, FormulaType.STRING);
        }
        return FormulaType.STRING;
    }

    private void checkArity(List<FormulaType> argumentTypes) {
        if (argumentTypes.get(0) == FormulaType.NUMBER) {
            if (argumentTypes.size() < 1 || argumentTypes.size() > 2) {
                throw new FormulaSyntaxException("The TEXT() function applied to a number expects one or two arguments");
            }
        } else {
            checkArity(argumentTypes, 1);
        }
    }

    private Set<FormulaType> createExpectedTypeSet() {
        Set<FormulaType> expectedTypes = new HashSet<>();
        expectedTypes.add(FormulaType.LOCAL_DATE);
        expectedTypes.add(FormulaType.BOOLEAN);
        expectedTypes.add(FormulaType.NUMBER);
        expectedTypes.add(FormulaType.STRING);
        return expectedTypes;
    }

}
