/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.functions;

import com.google.common.annotations.VisibleForTesting;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.DoubleArrayColumnView;
import org.activityinfo.model.type.time.LocalDate;

import java.util.List;

/**
 * Calculates the fraction of the year represented by the number of whole days between two dates
 * (the start_date and the end_date) using the US 30/360 method.
 *
 * @see <a href="https://en.wikipedia.org/wiki/Day_count_convention#30.2F360_US">US 30/360 method</a>
 */
public class YearFracFunction extends FormulaFunction implements ColumnFunction {

    public static final YearFracFunction INSTANCE = new YearFracFunction();

    private YearFracFunction() {
    }

    @Override
    public String getId() {
        return "yearfrac";
    }

    @Override
    public FunctionCategory getCategory() {
        return FunctionCategory.DATE_TIME;
    }

    @Override
    public int getReferenceTopicId() {
        return 889846;
    }

    @Override
    public FormulaType resolveResultType(List<FormulaType> argumentTypes) {
        checkArity(argumentTypes, 2);
        checkArgumentType(argumentTypes, 0, FormulaType.LOCAL_DATE);
        checkArgumentType(argumentTypes, 1, FormulaType.LOCAL_DATE);

        return FormulaType.NUMBER;
    }

    @VisibleForTesting
    static double compute(LocalDate startDate, LocalDate endDate) {
        int y1 = startDate.getYear();
        int y2 = endDate.getYear();
        int m1 = startDate.getMonthOfYear();
        int m2 = endDate.getMonthOfYear();
        int d1 = startDate.getDayOfMonth();
        int d2 = endDate.getDayOfMonth();

        // If (Date1 is the last day of February) and (Date2 is the last day of February),
        // then change D2 to 30.
        if(lastDayOfFebruary(m1, d1) && lastDayOfFebruary(m2, d2)) {
            d2 = 30;
        }
        if (lastDayOfFebruary(m1, d1)) {
            d1 = 30;
        }
        if(d2 == 31 && (d1 == 30 || d1 == 31)) {
            d2 = 30;
        }
        if(d1 == 31) {
            d1 = 30;
        }

        return Math.abs((360d * (y2-y1) + 30d * (m2-m1) + (d2-d1)) / 360d);
    }

    private static boolean lastDayOfFebruary(int month, int day) {
        return month == 2 && (day == 28 || day == 29);
    }

    @Override
    public ColumnView apply(int numRows, List<ColumnView> arguments) {
        if(arguments.size() != 2) {
            throw new FormulaSyntaxException("YEARFRAC() requires two arguments");
        }
        ColumnView startView = arguments.get(0);
        ColumnView endView = arguments.get(1);

        double[] result = new double[numRows];

        for (int i = 0; i < numRows; i++) {
            String start = startView.getString(i);
            String end = endView.getString(i);
            if(start == null || end == null) {
                result[i] = Double.NaN;
            } else {
                result[i] = compute(LocalDate.parse(start), LocalDate.parse(end));
            }
        }

        return new DoubleArrayColumnView(result);
    }
}
