/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.resolved;

import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.*;
import org.activityinfo.model.formula.diagnostic.ArgumentException;
import org.activityinfo.model.formula.diagnostic.FormulaException;
import org.activityinfo.model.formula.diagnostic.FormulaSyntaxException;
import org.activityinfo.model.formula.functions.*;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.expr.CalculatedFieldType;
import org.activityinfo.model.type.subform.SubFormReferenceType;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.*;

public class FormulaResolver {
    private final FormTree formTree;
    private final NodeMatcher nodeMatcher;
    private final Optional<ResourceId> selfFieldId;

    private final ArrayDeque<FieldRef> evaluationStack;

    public FormulaResolver(FormTree formTree) {
        this(formTree, Optional.empty(), new ArrayDeque<>());
    }
    public FormulaResolver(FormTree formTree, Optional<ResourceId> selfFieldId) {
        this(formTree, selfFieldId, new ArrayDeque<>());
    }

    public FormulaResolver(FormTree formTree, Optional<ResourceId> selfFieldId, ArrayDeque<FieldRef> evaluationStack) {
        this.formTree = formTree;
        this.selfFieldId = selfFieldId;
        this.nodeMatcher = new NodeMatcher(formTree);
        this.evaluationStack = evaluationStack;
    }

    public static ResolvedNode resolve(FormTree formTree, String string) {
        return new FormulaResolver(formTree).resolve(string);
    }

    public static ResolvedNode resolve(FormTree formTree, FormulaNode node) {
        return new FormulaResolver(formTree).resolve(node);
    }

    public ResolvedNode resolve(String formula) {
        FormulaNode root;
        try {
            root = FormulaParser.parse(formula);
        } catch (FormulaSyntaxException e) {
            return new InvalidSyntaxNode(formula, e.getMessage(), e.getSourceRange());
        }

        return resolve(root);
    }

    public ResolvedNode resolve(FormulaNode formulaNode) {
        if(formulaNode instanceof ConstantNode) {
            return (ResolvedNode) formulaNode;

        } else if(formulaNode instanceof GroupNode) {
            return resolve(((GroupNode) formulaNode).getExpr());

        } else if(formulaNode instanceof SymbolNode) {
            return resolveReference(null, formulaNode);

        } else if(formulaNode instanceof CompoundExpr) {
            return resolveReference(null, formulaNode);

        } else if(formulaNode instanceof FunctionCallNode) {
            return resolveCall((FunctionCallNode) formulaNode);
        } else if(formulaNode instanceof ErrorFormulaNode) {
            ErrorFormulaNode errorFormulaNode = (ErrorFormulaNode) formulaNode;
            return new InvalidSyntaxNode(errorFormulaNode.getFormula(), "Invalid syntax", formulaNode.getSourceRange());
        } else {
            throw new UnsupportedOperationException("type: " + formulaNode.getClass().getSimpleName());
        }
    }

    private ResolvedNode resolveReference(@Nullable StatFunction statFunction, FormulaNode formulaNode) {
        Collection<NodeMatch> matches;
        try {
            if (formulaNode instanceof SymbolNode) {
                matches = nodeMatcher.resolveSymbol((SymbolNode) formulaNode);
            } else if (formulaNode instanceof CompoundExpr) {
                matches = nodeMatcher.resolveCompoundExpr((CompoundExpr) formulaNode);
            } else {
                throw new IllegalArgumentException();
            }
        } catch (FormulaException e){
            return new InvalidReferenceNode(formulaNode, e.getMessage());
        }

        if(matches.isEmpty()) {
            return new InvalidReferenceNode(formulaNode, "Invalid field reference");
        }

        List<ReferenceNode> nodes = new ArrayList<>();
        try {
            for (NodeMatch match : matches) {
                ReferenceNode referenceNode = fromNodeMatch(statFunction, formulaNode, match);
                if(isSelfReference(referenceNode)) {
                    return new InvalidReferenceNode(formulaNode, "Circular reference.");
                }
                nodes.add(fromNodeMatch(statFunction, formulaNode, match));
            }
        } catch (IllegalSubFormReferenceException e) {
            return new InvalidReferenceNode(formulaNode,
                "Fields in subforms can only be referenced as arguments to aggregate functions such as SUM(), COUNT(), etc");
        }

        return new ResolvedReference(formulaNode, nodes);
    }

    private boolean isSelfReference(ReferenceNode referenceNode) {
        if (!selfFieldId.isPresent()) {
            return false;
        }
        if(referenceNode instanceof FieldReferenceNode) {
            FieldReferenceNode fieldReferenceNode = (FieldReferenceNode) referenceNode;
            return fieldReferenceNode.getField().getId().equals(selfFieldId.get());
        }
        return false;
    }


    /**
     * TEmproary!!
     */
    private ReferenceNode fromNodeMatch(@Nullable StatFunction statFunction, FormulaNode formulaNode, NodeMatch match) {
        switch (match.getType()) {
            case RECORD_ID:
                return applyJoins(statFunction, match, new RecordIdNode(match.getFormClass().getId()));
            case LAST_EDIT_TIME:
                return applyJoins(statFunction, match, new LastEditTimeNode(match.getFormClass().getId()));
            case FORM_ID:
                return applyJoins(statFunction, match, new FormIdNode(match.getFormClass().getId()));

            case FIELD:
                FieldReferenceNode field;
                if (match.getFieldNode().isCalculated()) {
                    field = resolveCalculatedField(formulaNode, match.getFieldNode());
                } else {
                    field = new FieldReferenceNode(match.getFieldNode());
                }
                ReferenceNode result;
                if(field.getField().getType() instanceof SubFormReferenceType) {
                    SubFormReferenceType subFormType = (SubFormReferenceType) field.getField().getType();
                    result = new AggregateNode(field, CountFunction.INSTANCE, new RecordIdNode(subFormType.getClassId()));

                } else if(match.isEnumBoolean()) {
                    result = new EnumItemReference(formulaNode, field, match.getEnumItem());
                } else {
                    result = field;
                }

                return applyJoins(statFunction, match, result);
        }
        throw new IllegalStateException("" + match.getType());
    }

    private ReferenceNode applyJoins(@Nullable StatFunction statFunction, NodeMatch match, ReferenceNode result) {
        for (int i = match.getJoins().size() - 1; i >= 0; i--) {
            JoinNode join = match.getJoins().get(i);
            if (join.getType() == JoinType.REFERENCE) {
                FormClass leftForm = formTree.getFormClass(join.getLeftFormId());
                FormField referenceField = leftForm.getField(join.getLeftFieldId());
                FieldReferenceNode left = new FieldReferenceNode(leftForm, referenceField);

                result = new JoinQueryNode(left, join.getRightFormId(), result);
            } else {
                FormClass leftForm = formTree.getFormClass(join.getLeftFormId());
                FormField subFormField = leftForm.getField(join.getLeftFieldId());
                if(statFunction == null) {
                    throw new IllegalSubFormReferenceException(subFormField);
                }
                FieldReferenceNode subFormNode = new FieldReferenceNode(leftForm, subFormField);
                result = new AggregateNode(subFormNode, statFunction, result);
            }
        }
        return result;
    }

    private FieldReferenceNode resolveCalculatedField(FormulaNode referenceFormula, FormTree.Node fieldNode) {

        FieldRef fieldRef = new FieldRef(referenceFormula,
            fieldNode.getDefiningFormClass().getId(),
            fieldNode.getFieldId());

        if(evaluationStack.contains(fieldRef)) {
            return new CircularReferenceNode(fieldNode, circularErrorMessage(evaluationStack, fieldRef));
        }

        ResolvedNode resolved;
        evaluationStack.push(fieldRef);
        try {
            CalculatedFieldType fieldType = (CalculatedFieldType) fieldNode.getType();
            FormulaResolver resolver = new FormulaResolver(formTree.tree(fieldNode.getDefiningFormClass().getId()), selfFieldId, evaluationStack);
            resolved = resolver.resolve(fieldType.getFormula());
        } finally {
            evaluationStack.pop();
        }

        return new CalculatedFieldNode(fieldNode, resolved);
    }

    private String circularErrorMessage(ArrayDeque<FieldRef> evaluationStack, FieldRef fieldRef) {
        StringBuilder s = new StringBuilder();
        Iterator<FieldRef> it = evaluationStack.descendingIterator();

        s.append("Circular reference: ").append(it.next().getFormulaNode().deparse());
        while(it.hasNext()) {
            s.append(" -> ");
            s.append(it.next().getFormulaNode().deparse());
        }
        s.append(" -> ");
        s.append(fieldRef.getFormulaNode().deparse());
        return s.toString();
    }

    private ResolvedNode resolveCall(FunctionCallNode call) {

        /*
         * Is the function a valid function?
         */

        FormulaFunction function = FormulaFunctions.get(call.getFunctionName()).orElse(null);

        StatFunction statFunction = computeStatFunction(call, function);

        /*
         * Resolve the arguments, check if they are valid
         */
        List<FormulaNode> arguments = call.getArguments();
        List<ResolvedNode> resolvedArguments = new ArrayList<>();

        boolean validArguments = true;
        for (FormulaNode argument : arguments) {
            ResolvedNode resolvedArgument = resolveCallArgument(statFunction, argument);
            if(!resolvedArgument.isValid()) {
                validArguments = false;
            }

            resolvedArguments.add(resolvedArgument);
        }
        if (isCompoundExpr(function, resolvedArguments)) {
            FieldReferenceNode field = (FieldReferenceNode) ((ResolvedReference) resolvedArguments.get(0)).getRoot();
            ConstantNode constantNode = (ConstantNode) resolvedArguments.get(1);
            return getResolvedReference(call, field, constantNode);
        }

        /*
         * If the function is not valid or any of the arguments are not valid,
         * then stop here.
         */
        if(function == null || !validArguments) {
            return new InvalidCallNode(call, function != null, resolvedArguments);
        }

        /*
         * Now verify that the arguments are the right type for the
         * function.
         */
        List<FormulaType> argumentTypes = new ArrayList<>();
        for (ResolvedNode resolvedArgument : resolvedArguments) {
            argumentTypes.add(resolvedArgument.getResolvedType().get());
        }

        FormulaType resultType;
        try {
            resultType = function.resolveResultType(argumentTypes);
        } catch (ArgumentException e) {
            ResolvedNode invalidArgument = resolvedArguments.get(e.getArgumentIndex());
            resolvedArguments.set(e.getArgumentIndex(), new InvalidArgumentNode(invalidArgument, e.getMessage()));

            return new InvalidCallNode(call, true, resolvedArguments);

        } catch (FormulaException e) {
            return new InvalidCallNode(call, true, resolvedArguments, e.getMessage());
        }

        return new ResolvedCallNode(call, function, resolvedArguments, resultType);
    }

    private StatFunction computeStatFunction(FunctionCallNode call, FormulaFunction function) {
        StatFunction statFunction;
        if(function instanceof StatFunction) {
            statFunction = (StatFunction) function;
        } else if(function == AddFunction.INSTANCE && call.getArgumentCount() == 2) {
            statFunction = SumFunction.INSTANCE;
        } else {
            statFunction = null;
        }
        return statFunction;
    }

    private boolean isCompoundExpr(FormulaFunction function, List<ResolvedNode> resolvedArguments) {
        if (function != EqualFunction.INSTANCE && function != NotEqualFunction.INSTANCE) {
            return false;
        }
        if (resolvedArguments.size() != 2) {
            return false;
        }
        ResolvedNode firstArgument = resolvedArguments.get(0);
        if (!(firstArgument instanceof ResolvedReference)) {
            return false;
        }
        if (!(((ResolvedReference) firstArgument).getRoot() instanceof FieldReferenceNode)) {
            return false;
        }
        FieldReferenceNode field = (FieldReferenceNode) ((ResolvedReference) firstArgument).getRoot();
        if (!(field.getField().getType() instanceof EnumType)) {
            return false;
        }
        ResolvedNode secondArgument = resolvedArguments.get(1);
        if (!(secondArgument instanceof ConstantNode)) {
            return false;
        }
        if (((ConstantNode) secondArgument).getType() != FormulaType.STRING) {
            return false;
        }
        return true;
    }

    private ResolvedNode getResolvedReference(FunctionCallNode call, FieldReferenceNode field, ConstantNode secondArgument) {
        String option = secondArgument.getStringValue();
        EnumType enumType = (EnumType) field.getField().getType();
        for (EnumItem item : enumType.getItems()) {
            if (item.getLabel().equals(option)) {
                EnumItemReference enumItemReference = new EnumItemReference(call, field, item);
                List<ReferenceNode> nodes = new ArrayList<>();
                nodes.add(enumItemReference);
                ResolvedReference resolvedReference = new ResolvedReference(enumItemReference.toFormula(), nodes);
                if (call.getFunction() == EqualFunction.INSTANCE) {
                    return resolvedReference;
                }
                if (call.getFunction() == NotEqualFunction.INSTANCE) {
                    List<ResolvedNode> resolvedArguments = new ArrayList<>();
                    resolvedArguments.add(resolvedReference);
                    return new ResolvedCallNode(call, NotFunction.INSTANCE, resolvedArguments, FormulaType.BOOLEAN);
                }
            }
        }
        return new InvalidReferenceNode(call, "Invalid field reference");
    }


    private ResolvedNode resolveCallArgument(@Nullable StatFunction statFunction, FormulaNode argument) {
        if(argument instanceof SymbolNode || argument instanceof CompoundExpr) {
            return resolveReference(statFunction, argument);
        } else {
            return resolve(argument);
        }
    }
}
