/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.resolved;

import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formula.CompoundExpr;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.formula.functions.FormulaFunction;
import org.activityinfo.model.formula.functions.StatFunction;
import org.activityinfo.model.query.ColumnModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.geo.GeoPointType;

import java.util.*;

/**
 * Resolves symbols in queries to the fields on the base FormClass
 * or on related tables.
 *
 * TODO: Harmonize with FormSymbolTable. Do we need both?
 */
public class NodeMatcher {
    private final FormTree tree;

    @SuppressWarnings("NonJREEmulationClassesInClientCode") /* Supported in GWT 2.8+ */
    private Deque<StatFunction> aggregationContextStack = new ArrayDeque<>();

    public NodeMatcher(FormTree formTree) {
        this.tree = formTree;
    }

    /**
     * Called directly before the QueryEvaluator evaluates a function. If the function is
     * a {@code StatFunction}, it will be added to the context.
     *
     * For example, if we encounter the expression MAX(A), and A is resolved to a symbol
     * in a subform, then we know that we have to apply the MAX() function for each row in the parent form
     * to all of the values in the subform for that record.
     *
     */
    public void enterFunction(FormulaFunction function) {
        if(function instanceof StatFunction) {
            aggregationContextStack.push((StatFunction) function);
        }
    }

    public void exitFunction(FormulaFunction function) {
        if(function instanceof StatFunction) {
            aggregationContextStack.pop();
        }
    }

    private Optional<StatFunction> currentAggregation() {
        return Optional.ofNullable(aggregationContextStack.peek());
    }

    public Collection<NodeMatch> resolveSymbol(SymbolNode symbol) {
        return matchNodes(new QueryPath(symbol), tree.getRootFields());
    }

    /**
     * Resolves a compound expression like "province.name" to one or more {@code FormTree.Nodes}
     *
     * @return a binding to the corresponding {@code FormTree.Node}
     *
     */
    public Collection<NodeMatch> resolveCompoundExpr(CompoundExpr expr) {
        QueryPath queryPath = new QueryPath(expr);

        // Given an expression like Province.Name, see if we can match the
        // the first symbol against the root form's id or label
        if(queryPath.matches(tree.getRootFormClass())) {
            return sortByDistance(matchNodes(queryPath.next(), tree.getRootFields()));
        }
        return sortByDistance(matchNodes(queryPath, tree.getRootFields()));
    }


    private Collection<NodeMatch> matchNodes(QueryPath queryPath, Iterable<FormTree.Node> fields) {
        if(queryPath.isLeaf()) {
            return sortByDistance(matchTerminal(queryPath, fields));
        } else {
            return sortByDistance(matchReferenceField(queryPath, fields));
        }
    }


    private Collection<NodeMatch> sortByDistance(Collection<NodeMatch> matchNodes) {
        List<NodeMatch> sorted = new ArrayList<>(matchNodes);
        Collections.sort(sorted, (a, b) -> {
            return Integer.compare(a.getJoins().size(), b.getJoins().size());
        });
        return sorted;
    }

    private Collection<NodeMatch> matchReferenceField(QueryPath queryPath, Iterable<FormTree.Node> fields) {

        List<NodeMatch> matches = Lists.newArrayList();

        // Check first for component matches on the root form
        for (FormTree.Node field : fields) {
            if(field.getType() instanceof EnumType) {
                Optional<NodeMatch> result = matchEnum(queryPath, field);
                if (result.isPresent()) {
                    matches.add(result.get());
                }
            } else if(field.getType() instanceof GeoPointType) {
                matchCoordinate(queryPath, field).ifPresent(match -> {
                    matches.add(match);
                });
            }
        }
        if(matches.size() > 0) {
            return Collections.singleton(matches.get(0));
        }

        for (FormTree.Node field : fields) {
            if(field.isReference()) {
                Collection<NodeMatch> result = unionMatches(queryPath, field);
                if (!result.isEmpty()) {
                    matches.addAll(result);
                }
            }
        }

        if(!matches.isEmpty()) {
            return matches;
        }

        // If no results, check search at the next level
        List<FormTree.Node> children = childrenOf(fields);
        if(children.isEmpty()) {
            return Collections.emptyList();
        } else {
            return matchReferenceField(queryPath, children);
        }
    }

    private Optional<NodeMatch> matchEnum(QueryPath queryPath, FormTree.Node field) {
        if(queryPath.matches(field)) {
            QueryPath next = queryPath.next();
            if(next.isLeaf()) {
                EnumType type = (EnumType) field.getType();
                List<EnumItem> matchingItems = Lists.newArrayList();
                for (EnumItem enumItem : type.getItems()) {
                    String head = next.head();
                    if (head.equals(enumItem.getId().asString()) ||
                        Strings.nullToEmpty(enumItem.getLabel()).toLowerCase().startsWith(head.toLowerCase()) ||
                        head.equalsIgnoreCase(enumItem.getCode())) {
                        
                        matchingItems.add(enumItem);
                    }
                }
                if(matchingItems.size() == 1) {
                    return Optional.of(NodeMatch.forFieldComponent(field, matchingItems.get(0).getId().asString()));
                } 
            }
        }
        return Optional.empty();
    }

    private java.util.Optional<NodeMatch> matchCoordinate(QueryPath queryPath, FormTree.Node field) {
        // Look for a match in the format [field name].[latitude]
        if(queryPath.size() == 2) {
            String symbol = queryPath.peek().toLowerCase();

            if (queryPath.matches(field) &&
                (symbol.equals("latitude") || symbol.equals("longitude"))) {
                return java.util.Optional.of(NodeMatch.forFieldComponent(field, symbol));
            }
        }

        return java.util.Optional.empty();
    }


    /**
     * Matches a terminal symbol in a query path.
     */
    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    private Collection<NodeMatch> matchTerminal(QueryPath path, Iterable<FormTree.Node> fields) {

        List<NodeMatch> matches = Lists.newLinkedList();

        // Check for a reference to a form label
        if(path.isLeaf() && path.head().equals(ColumnModel.FORM_NAME_SYMBOL)) {
            matches.add(NodeMatch.forLabel(path.head(), tree.getRootFormClass()));
        }

        // Check for a reference to a metadata symbol
        if (path.isLeaf() && (path.head().equals(ColumnModel.RECORD_ID_SYMBOL) ||
            path.head().equals(ColumnModel.FORM_ID_SYMBOL) ||
            path.head().equals(ColumnModel.LAST_EDIT_TIME_SYMBOL))) {
            matches.add(NodeMatch.forMetadata(path.head(), tree.getRootFormClass()));
        }

        // check for a match on the form id of the current form tree
        if (path.isLeaf() && path.head().equals(tree.getRootFormId().asString())) {
            matches.add(NodeMatch.forMetadata(ColumnModel.RECORD_ID_SYMBOL, tree.getRootFormClass()));
        }

        // Check for a match of the query Path head to the set of fields
        for (FormTree.Node field : fields) {
            if(path.matches(field)) {
                matches.add(NodeMatch.forField(field, currentAggregation()));
            } else if (field.getType() instanceof ReferenceType) {
                // If this is a reference node, check if the any referenced forms match the query path head
                referencedFormIdMatches(matches, path, field);
            } else if(field.getType() instanceof GeoPointType) {
                String symbol = path.head().toLowerCase();
                if(symbol.equals("latitude") || symbol.equals("longitude")) {
                    matches.add(NodeMatch.forFieldComponent(field, symbol));
                }
            }
        }

        // If we have matched, then stop crawling.
        if(matches.size() > 0) {
            return matches;
        }

        // Otherwise, continue to the next level
        List<FormTree.Node> children = childrenOf(fields);
        if(children.isEmpty()) {
            return Collections.emptyList();
        } else {
            return matchTerminal(path, children);
        }
    }

    private void referencedFormIdMatches(List<NodeMatch> matches, QueryPath path, FormTree.Node field) {
        for (ResourceId referencedFormId : ((ReferenceType) field.getType()).getRange()) {
            java.util.Optional<FormClass> refFormClass = tree.getFormClassIfPresent(referencedFormId);
            if (!refFormClass.isPresent()) {
                continue;
            }
            if (path.matches(refFormClass.get())) {
                matches.add(NodeMatch.forId(field, refFormClass.get()));
            }
        }
    }

    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    private Collection<NodeMatch> unionMatches(QueryPath path, FormTree.Node parentField) {
        List<NodeMatch> results = Lists.newArrayList();
        for (ResourceId childFormId : parentField.getRange()) {
            java.util.Optional<FormClass> childForm = tree.getFormClassIfPresent(childFormId);
            if(childForm.isPresent()) {
                Iterable<FormTree.Node> childFields = parentField.getChildren(childFormId);

                if ((path.matches(childForm.get()) || path.matches(parentField)) && path.peek().equals(ColumnModel.RECORD_ID_SYMBOL)) {
                    results.add(NodeMatch.forId(parentField, childForm.get()));

                }else if ((path.matches(childForm.get()) || path.matches(parentField)) && path.peek().equals(ColumnModel.LAST_EDIT_TIME_SYMBOL)) {
                    results.add(NodeMatch.forLastEditTime(parentField, childForm.get()));

                } else if (path.matches(parentField) && path.peek().equals(ColumnModel.RECORD_ID_SYMBOL)) {
                    results.add(NodeMatch.forFormId(parentField, childForm.get()));

                } else if (path.matches(parentField) && path.peek().equals(ColumnModel.FORM_ID_SYMBOL)) {
                    results.add(NodeMatch.forFormId(parentField, childForm.get()));

                } else if (path.matches(parentField) && path.peek().equals(ColumnModel.FORM_NAME_SYMBOL)) {
                    results.add(NodeMatch.forLabel(parentField, childForm.get()));

                } else if (path.matches(childForm.get()) || path.matches(parentField)) {
                    results.addAll(matchNodes(path.next(), childFields));

                } else {
                    // Descend the next level
                    results.addAll(matchNodes(path, childFields));
                }
            }
        }
        return results;
    }


    private List<FormTree.Node> childrenOf(Iterable<FormTree.Node> fields) {
        List<FormTree.Node> children = Lists.newArrayList();
        for (FormTree.Node field : fields) {
            children.addAll(field.getChildren());
        }
        return children;
    }


}
