/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.formula.resolved;

import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.formula.SourceRange;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldType;

import java.util.Optional;

/**
 * A node in a resolved formula.
 *
 * <p>A formula string, even parsed into an abstract symbol tree of constants, function calls, and symbols,
 * is not ready to be evaluated, because those symbols need to be mapped to the fields they need to reference,
 * arguments to functions need to be evaluated, etc.</p>
 *
 * <p>A ResolveNode tree is the result of validating and resolving a formula against a {@link org.activityinfo.model.formTree.FormTree}.</p>
 */
public interface ResolvedNode {

    /**
     * Return true if this node is valid and can be evaluated.
     */
    boolean isValid();

    /**
     * Return a reference to the position in the original formula string to which this
     * node corresponds. Useful for providing feedback to the user.
     */
    SourceRange getSourceRange();

    /**
     * If the node is valid, return the resolved formula type.
     */
    Optional<FormulaType> getResolvedType();

    <T> T accept(ResolvedNodeVisitor<T> visitor);

    FormulaNode toFormula();

    Iterable<ResolvedNode> getChildren();

    /**
     * Constructs a human-readable label for this resolved node, relative
     * to the given {@code rootFormId}.
     */
    default String label(ResourceId rootFormId) {
        return toFormula().deparse();
    }

    /**
     * Return the resolved {@link FieldType}, if this node resolves to a single field type. Otherwise,
     * if this node resolves to a formula, or to multiple fields with different types, then return {@code Optional.empty()}.
     */
    default Optional<FieldType> getFieldType() {
        return Optional.empty();
    }

}
