/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.job;


import org.activityinfo.json.AutoJson;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.resource.ResourceId;

import java.util.Optional;

/**
 * Export of all forms in a database or folder to a CSV table
 */
@AutoJson
public class ExportDatabaseFormsJob implements JobDescriptor<ExportResult> {

    public static final String TYPE = "exportDatabaseForms";

    /**
     * The id of the database containing the forms to export.
     */
    private final ResourceId databaseId;

    /**
     * Optionally, the folder or form to include in the export. If ommitted, then all forms
     * in the database will be included.
     */
    private final Optional<ResourceId> folderId;

    /**
     * The format of the table structure.
     */
    private final ExportFormat format;

    /**
     * The file format.
     */
    private final ExportFileFormat fileFormat;

    /**
     * A formula that is applied to each form to filter the resulting records.
     */
    private final Optional<String> filter;

    /**
     * True if the export should include rows for blank quantities. Only applies to the LONG
     * format.
     */
    private final boolean includeBlanks;

    /**
     * The time zone to use for time values, as an offset
     * from Universal Coordinate Time (UTC) in minutes,
     */
    private final int utcOffset;

    @Builder
    public ExportDatabaseFormsJob(ResourceId databaseId, Optional<ResourceId> folderId, ExportFormat format, ExportFileFormat fileFormat, Optional<String> filter, boolean includeBlanks, int utcOffset) {
        this.databaseId = databaseId;
        this.folderId = folderId;
        this.format = format;
        this.fileFormat = fileFormat;
        this.filter = filter;
        this.includeBlanks = includeBlanks;
        this.utcOffset = utcOffset;
    }

    public ExportDatabaseFormsJob(ResourceId databaseId, Optional<ResourceId> folderId, ExportFormat format, ExportFileFormat fileFormat, int utcOffset) {
        this.databaseId = databaseId;
        this.folderId = folderId;
        this.format = format;
        this.fileFormat = fileFormat;
        this.filter = Optional.empty();
        this.utcOffset = utcOffset;
        this.includeBlanks = false;
    }


    public ResourceId getDatabaseId() {
        return databaseId;
    }

    public Optional<ResourceId> getFolderId() {
        return folderId;
    }

    public ExportFileFormat getFileFormat() {
        if(fileFormat == null) {
            return ExportFileFormat.XLSX;
        }
        return fileFormat;
    }

    @Override
    public String getType() {
        return TYPE;
    }

    public ExportFormat getFormat() {
        return format;
    }

    public Optional<String> getFilter() {
        return filter;
    }

    public int getUtcOffset(){
        return utcOffset;
    }

    @Override
    public ExportResult parseResult(JsonValue resultObject) {
        return ExportResult.fromJson(resultObject);
    }

    @Override
    public <R> R accept(JobVisitor<R> visitor) {
        return visitor.visit(this);
    }

    @Override
    public boolean isReadOnly() {
        return true;
    }

    public boolean isIncludeBlanks() {
        return includeBlanks;
    }

    @Override
    public JsonValue toJson() {
        return ExportDatabaseFormsJobJson.toJson(this);
    }

    public static ExportDatabaseFormsJob fromJson(JsonValue object) {
        return ExportDatabaseFormsJobJson.fromJson(object);
    }

}
