/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.job;

import org.activityinfo.json.JsonValue;
import org.activityinfo.model.analysis.TableAnalysisModel;
import org.activityinfo.model.analysis.TableAnalysisModelBuilder;
import org.activityinfo.model.analysis.TableAnalysisModelJson;
import org.activityinfo.model.resource.ResourceId;

import static org.activityinfo.json.Json.createObject;

/**
 * Exports a single form to a CSV table
 */
public class ExportFormJob implements JobDescriptor<ExportResult> {

    public static final String TYPE = "exportForm";


    private final TableAnalysisModel tableModel;
    private final ExportFileFormat format;

    /**
     * The time zone to use for time values, as an offset
     * from Universal Coordinate Time (UTC) in minutes,
     */
    private final int utcOffset;

    public ExportFormJob(TableAnalysisModel tableModel, ExportFileFormat format, int utcOffset) {
        this.tableModel = tableModel;
        this.format = format;
        this.utcOffset = utcOffset;
    }

    public ResourceId getFormId() {
        return tableModel.getFormId();
    }

    @Override
    public String getType() {
        return TYPE;
    }

    public int getUtcOffset() {
        return utcOffset;
    }

    @Override
    public ExportResult parseResult(JsonValue resultObject) {
        return ExportResult.fromJson(resultObject);
    }

    @Override
    public <R> R accept(JobVisitor<R> visitor) {
        return visitor.visit(this);
    }

    @Override
    public boolean isReadOnly() {
        return true;
    }

    public TableAnalysisModel getTableModel() {
        return tableModel;
    }

    public ExportFileFormat getFormat() {
        return format;
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = createObject();
        object.put("model", tableModel.toJson());
        object.put("format", format.name());
        object.put("utcOffset", this.getUtcOffset());
        return object;
    }

    public static ExportFormJob fromJson(JsonValue object) {
        TableAnalysisModel tableModel;
        if (object.hasKey("formId")) {
            tableModel = new TableAnalysisModelBuilder()
                .formId(ResourceId.valueOf(object.get("formId").asString()))
                .build();
        } else {
            tableModel = TableAnalysisModelJson.fromJson(object.get("model"));
        }
        String formatString = object.getString("format");
        ExportFileFormat format;
        if(formatString == null || formatString.toUpperCase().startsWith("XLS")) {
            format = ExportFileFormat.XLSX;
        } else {
            format = ExportFileFormat.valueOf(formatString);
        }

        return new ExportFormJob(tableModel, format, (int)object.getNumber("utcOffset"));
    }
}
