/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.job;

import org.activityinfo.i18n.shared.ApplicationLocale;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.annotation.Undocumented;
import org.activityinfo.model.resource.ResourceId;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.Optional;

import static org.activityinfo.json.Json.createObject;

/**
 * Shows the status of the job status
 */
public class JobStatus implements JsonSerializable {

    /**
     * The job's id
     */
    private final String id;

    /**
     * The id of the user who started the job
     */
    private final ResourceId userId;

    /**
     * The job descriptor used to start this job
     */
    private final JobDescriptor descriptor;

    /**
     * The job's current state
     */
    private final JobState state;

    /**
     * If the job state is COMPLETED, the result of the job
     */
    private final @Nullable JobResult result;

    /**
     * The locale
     */
    @Undocumented
    private final ApplicationLocale locale;

    private final double percentComplete;

    /**
     * If the job state is FAILED, the reason for the failure.
     */
    private final Optional<JobError> error;

    public JobStatus(String id, ResourceId userId, JobDescriptor descriptor, JobState state, @Nullable JobResult jobResult, ApplicationLocale locale, double percentComplete, Optional<JobError> error) {
        this.id = id;
        this.userId = userId;
        this.descriptor = descriptor;
        this.state = state;
        this.result = jobResult;
        this.locale = locale;
        this.percentComplete = percentComplete;
        this.error = error;
    }

    public String getId() {
        return id;
    }

    public ResourceId getUserId() {
        return userId;
    }

    public JobDescriptor getDescriptor() {
        return descriptor;
    }

    public @Nullable JobResult getResult() {
        return result;
    }

    public JobState getState() {
        return state;
    }

    public Optional<JobError> getError() {
        return error;
    }

    public ApplicationLocale getLocale() {
        return locale;
    }

    public double getPercentComplete() {
        return percentComplete;
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = createObject();
        object.put("id", id);
        object.put("userId", userId.toString());
        object.put("type", descriptor.getType());
        object.put("descriptor", descriptor.toJson());
        object.put("state", state.name().toLowerCase());
        object.put("locale", locale.name());
        object.put("percentComplete", percentComplete);
        if(result != null) {
            object.put("result", result.toJson());
        }
        error.ifPresent(e -> object.put("error", JobErrorJson.toJson(e)));
        return object;
    }

    public static JobStatus fromJson(JsonValue object) {
        String id = object.get("id").asString();
        String type = object.get("type").asString();
        JobDescriptor descriptor = JobRequest.parseDescriptor(type, object.get("descriptor"));
        JobState state = JobState.valueOf(object.get("state").asString().toUpperCase());
        JobResult result = null;
        if(object.hasKey("result")) {
            result = descriptor.parseResult(object.get("result"));
        }
        Optional<JobError> error = Optional.empty();
        if(object.hasKey("error")) {
            error = Optional.of(JobErrorJson.fromJson(object.get("error")));
        }
        ResourceId userId;
        if(object.hasKey("userId")) {
            userId = ResourceId.valueOf(object.getString("userId"));
        } else {
            userId = ResourceId.valueOf("0");
        }
        double percentComplete = 0;
        if(object.hasKey("percentComplete")) {
            percentComplete = object.getNumber("percentComplete");
        }
        ApplicationLocale locale = ApplicationLocale
            .validateLocale(object.getString("locale"))
            .orElse(ApplicationLocale.EN);

        return new JobStatus(id, userId, descriptor, state, result, locale, percentComplete, error);
    }

    public boolean isPending() {
        return state != JobState.COMPLETED && state != JobState.FAILED;
    }
}
