/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.permission;

import com.google.common.base.Strings;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.database.HasPermissions;
import org.activityinfo.model.database.RecordLockSet;
import org.activityinfo.model.database.Resource;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.resource.ResourceId;

import javax.annotation.Nullable;
import java.util.Collections;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

/**
 * Describes effective form-level permissions for a specific user. The principle permissions include:
 *
 * <ul>
 *     <li>create: the permission to create a new record.</li>
 *     <li>view: the permission to view this form</li>
 *     <li>edit: the permission to edit the field values of records belonging to this form</li>
 *     <li>delete: the permission to delete records belonging to the form</li>
 * </ul>
 *
 * In addition, the view, edit, and delete permissions may be further restricted by specifying
 * view, edit, and delete <em>filters</em>, using formulas that evaluate to a boolean value.</p>
 */
public final class FormPermissions implements JsonSerializable {

    private boolean view;
    private String viewFilter;

    private boolean addRecord;
    private String addFilter;

    private boolean editRecord;
    private String editFilter;
    private Set<ResourceId> securityCategories = new HashSet<>();

    private boolean deleteRecord;
    private String deleteFilter;

    private boolean bulkDelete;

    private boolean updateSchema;

    private boolean exportRecords;
    private boolean collectionLinks;

    private boolean publicForm;

    private RecordLockSet locks;

    private FormPermissions() {
    }

    public FormPermissions(Resource.Visibility visibility, HasPermissions permissions, RecordLockSet locks) {
        this.view = permissions.hasPermission(Operation.VIEW);
        if(visibility == Resource.Visibility.PRIVATE) {
            this.viewFilter = permissions.getPermission(Operation.VIEW).flatMap(p -> p.getFilter()).orElse(null);
        } else {
            this.viewFilter = null;
        }

        if(visibility == Resource.Visibility.REFERENCE) {
            this.addRecord =
                this.editRecord =
                    this.deleteRecord = permissions.hasPermission(Operation.MANAGE_REFERENCE_DATA);
            this.bulkDelete = permissions.hasPermission(Operation.MANAGE_REFERENCE_DATA) &&
                    permissions.hasPermission(Operation.BULK_DELETE);

        } else {
            this.addRecord = permissions.hasPermission(Operation.ADD_RECORD);
            this.addFilter = permissions.getPermission(Operation.ADD_RECORD).flatMap(p -> p.getFilter()).orElse(null);
            this.editRecord = permissions.hasPermission(Operation.EDIT_RECORD);
            this.editFilter = permissions.getPermission(Operation.EDIT_RECORD).flatMap(p -> p.getFilter()).orElse(null);

            this.securityCategories = permissions.getPermission(Operation.EDIT_RECORD)
                    .map(p -> p.getSecurityCategories())
                    .orElse(Collections.emptySet());

            this.deleteRecord = permissions.hasPermission(Operation.DELETE_RECORD);
            this.deleteFilter = permissions.getPermission(Operation.DELETE_RECORD).flatMap(p -> p.getFilter()).orElse(null);
            this.bulkDelete = permissions.hasPermission(Operation.BULK_DELETE);
        }

        this.exportRecords = permissions.hasPermission(Operation.EXPORT_RECORDS);
        this.updateSchema = permissions.hasPermission(Operation.EDIT_RESOURCE);
        this.collectionLinks = permissions.hasPermission(Operation.MANAGE_COLLECTION_LINKS);
        this.publicForm = visibility == Resource.Visibility.PUBLIC;
        this.locks = locks;
    }

    /**
     * 
     * @return true if this form is visible to the user.
     */
    public boolean isVisible() {
        return view;
    }

    public String getViewFilter() {
        return viewFilter;
    }

    public boolean isAddAllowed() {
        return addRecord;
    }

    public String getAddFilter() {
        return addFilter;
    }

    public boolean isEditAllowed() {
        return editRecord;
    }

    public String getEditFilter() {
        return editFilter;
    }

    public Set<ResourceId> getSecurityCategories() {
        return securityCategories;
    }

    public boolean isDeleteAllowed() {
        return deleteRecord;
    }

    public String getDeleteFilter() {
        return deleteFilter;
    }

    public boolean isBulkDeleteAllowed() {
        return bulkDelete;
    }

    public boolean isSchemaUpdateAllowed() {
        return updateSchema;
    }

    public boolean isExportRecordsAllowed() {
        return exportRecords;
    }

    public boolean isManageCollectionLinksAllowed() {
        return collectionLinks;
    }
    public RecordLockSet getLocks() {
        if (locks == null) {
            return RecordLockSet.EMPTY;
        }
        return locks;
    }

    public boolean hasLocks() {
        return locks != null;
    }

    public boolean isFieldEditable(FormField field) {
        return field.getSecurityCategoryId()
                .map(sc -> securityCategories.contains(sc))
                .orElse(true);
    }

    public static FormPermissions none() {
        return new FormPermissions();
    }

    public static FormPermissions readonly() {
        FormPermissions permissions = new FormPermissions();
        permissions.view = true;
        permissions.exportRecords = true;
        return permissions;
    }

    public static FormPermissions readWrite(@Nullable RecordLockSet locks) {
        FormPermissions permissions = new FormPermissions();
        permissions.view = true;
        permissions.addRecord = true;
        permissions.editRecord = true;
        permissions.deleteRecord = true;
        permissions.exportRecords = true;
        permissions.locks = locks;
        return permissions;
    }

    public static FormPermissions owner(@Nullable RecordLockSet locks, Set<ResourceId> securityCategories) {
        FormPermissions permissions = new FormPermissions();
        permissions.view = true;
        permissions.addRecord = true;
        permissions.editRecord = true;
        permissions.securityCategories = securityCategories;
        permissions.deleteRecord = true;
        permissions.updateSchema = true;
        permissions.exportRecords = true;
        permissions.locks = locks;
        return permissions;
    }

    public static FormPermissions fromJson(JsonValue object) {
        FormPermissions permissions = new FormPermissions();

        permissions.view = object.getBoolean("view");
        permissions.viewFilter = object.getString("viewFilter");

        permissions.addRecord = object.getBoolean("createRecord");
        permissions.addFilter = object.getString("createFilter");

        permissions.editRecord = object.getBoolean("editRecord");
        permissions.editFilter = object.getString("editFilter");

        if (object.hasKey("securityCategories") && object.get("securityCategories").isJsonArray()) {
            JsonValue scArray = object.get("securityCategories");
            for (int i = 0; i < scArray.length(); i++) {
                permissions.securityCategories.add(ResourceId.valueOf(scArray.getString(i)));
            }
        }

        permissions.deleteRecord = object.getBoolean("deleteRecord");
        permissions.deleteFilter = object.getString("deleteFilter");

        permissions.bulkDelete = object.getBoolean("bulkDelete");

        permissions.exportRecords = object.getBoolean("exportRecords");
        permissions.updateSchema = object.getBoolean("updateSchema");
        permissions.collectionLinks = object.getBoolean("collectionLinks");
        permissions.publicForm = object.getBoolean("publicForm");

        if (object.hasKey("locks") && object.get("locks").isJsonArray()) {
            permissions.locks = RecordLockSet.fromJson(object.get("locks"));
        }

        return permissions;
    }

    public JsonValue toJson() {
        JsonValue object = Json.createObject();

        object.put("view", view);
        object.put("viewFilter", viewFilter);

        object.put("createRecord", addRecord);
        object.put("createFilter", addFilter);

        object.put("editRecord", editRecord);
        object.put("editFilter", editFilter);

        if (securityCategories != null) {
            JsonValue scArray = Json.createArray();
            for (ResourceId sc : securityCategories) {
                scArray.add(Json.create(sc.asString()));
            }
            object.put("securityCategories", scArray);
        }

        object.put("deleteRecord", deleteRecord);
        object.put("deleteFilter", deleteFilter);

        object.put("bulkDelete", bulkDelete);

        object.put("exportRecords", exportRecords);

        object.put("updateSchema", updateSchema);

        object.put("collectionLinks", collectionLinks);

        object.put("publicForm", publicForm);

        if (hasLocks()) {
            object.put("locks", locks.toJson());
        }

        return object;
    }

    public static Builder builder() {
        return new Builder();
    }

    @Override
    public String toString() {
        return "CollectionPermissions{" +
                "visible=" + view +
                ", visibilityFilter='" + viewFilter + '\'' +
                ", createAllowed=" + addRecord +
                ", createFilter='" + addFilter + '\'' +
                ", editAllowed=" + editRecord +
                ", editFilter='" + editFilter + '\'' +
                ", securityCategories='" + securityCategories + '\'' +
                ", deleteAllowed=" + deleteRecord +
                ", deleteFilter='" + deleteFilter + '\'' +
                ", exportAllowed='" + exportRecords + '\'' +
                ", locks='" + locks +
                '}';
    }

    public boolean hasVisibilityFilter() {
        return !Strings.isNullOrEmpty(viewFilter);
    }

    public boolean hasCreateFilter() {
        return !Strings.isNullOrEmpty(addFilter);
    }

    public boolean hasEditFilter() {
        return !Strings.isNullOrEmpty(editFilter);
    }

    public boolean hasDeleteFilter() {
        return !Strings.isNullOrEmpty(deleteFilter);
    }

    public boolean isUnconditionallyAllowed(Operation operation) {
        return isAllowed(operation) && !isFiltered(operation);
    }

    public boolean isAllowed(Operation operation) {
        switch (operation) {
            case VIEW:
                return isVisible();
            case ADD_RECORD:
                return addRecord;
            case EDIT_RECORD:
                return editRecord;
            case DELETE_RECORD:
                return deleteRecord;
            case BULK_DELETE:
                return bulkDelete;
            case EXPORT_RECORDS:
                return exportRecords;
        }
        throw new IllegalArgumentException("operation: " + operation);
    }

    /**
     * Returns true if permission to execute the given permission depends on
     * the record's field values.
     */
    public boolean isFiltered(Operation operation) {
        switch (operation) {
            case VIEW:
            case EXPORT_RECORDS:
                return hasVisibilityFilter();
            case ADD_RECORD:
                return hasCreateFilter();
            case EDIT_RECORD:
                return hasEditFilter();
            case DELETE_RECORD:
                return hasDeleteFilter();
        }
        return false;
    }

    public String getFilter(Operation operation) {
        switch (operation) {
            case VIEW:
            case EXPORT_RECORDS:
                return viewFilter;
            case ADD_RECORD:
                return addFilter;
            case EDIT_RECORD:
                return editFilter;
            case DELETE_RECORD:
                return deleteFilter;
        }
        return null;
    }

    public boolean isPublicForm() {
        return publicForm;
    }

    public static class Builder {
        private FormPermissions permissions = new FormPermissions();

        /**
         * Allows viewing this form, its schema, and all of its records,
         */
        public Builder allowView() {
            permissions.view = true;
            return this;
        }

        public Builder forbidView() {
            permissions.view = false;
            return this;
        }

        public boolean isAllowedView() {
            return permissions.view;
        }

        public Builder allowSchemaUpdate() {
            permissions.updateSchema = true;
            return this;
        }

        /**
         * Allows viewing this form, its schema, and records that match the given formula.
         *
         * @param filter a boolean-valued formula
         */
        public Builder allowFilteredView(String filter) {
            permissions.view = true;
            permissions.viewFilter = filter;
            return this;
        }

        public void allowUnfilteredView() {
            permissions.view = true;
            permissions.viewFilter = null;
        }

        public Builder allowCreate(Optional<String> filter) {
            permissions.addRecord = true;
            permissions.addFilter = filter.orElse(null);
            return this;
        }

        public Builder allowEdit(Optional<String> filter) {
            permissions.editRecord = true;
            permissions.editFilter = filter.orElse(null);
            permissions.securityCategories = Collections.emptySet();
            return this;
        }

        public Builder allowEdit(Optional<String> filter, Set<ResourceId> securityCategories) {
            permissions.editRecord = true;
            permissions.editFilter = filter.orElse(null);
            permissions.securityCategories = securityCategories;
            return this;
        }

        public Builder allowDelete(Optional<String> filter) {
            permissions.deleteRecord = true;
            permissions.deleteFilter = filter.orElse(null);
            return this;
        }

        public Builder allowExport() {
            permissions.exportRecords = true;
            return this;
        }

        public Builder lock(RecordLockSet locks) {
            permissions.locks = locks;
            return this;
        }

        public FormPermissions build() {
            return permissions;
        }
    }
}
