/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.permission;

import com.google.common.collect.Sets;
import org.activityinfo.model.account.*;
import org.activityinfo.model.analysis.ReportRoleGrant;
import org.activityinfo.model.database.*;
import org.activityinfo.model.formula.ConstantNode;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FormulaParser;
import org.activityinfo.model.formula.Formulas;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.promise.Maybe;

import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

public class PermissionOracle {

    public static final ResourceId SUPPORT_ROLE_ID = ResourceId.valueOf("support");

    private PermissionOracle() {
    }

    public static Permission query(VisibleDatabaseTree db, Operation operation, ResourceId resourceId) {

        HasPermissions effectivePermission = findEffectivePermission(db, resourceId);

        return new Permission(operation,
            effectivePermission.hasPermission(operation),
            effectivePermission.getPermission(operation).flatMap(p -> p.getFilter()));
    }


    private static Optional<String> getFilter(Operation operation, ResourceId resourceId, VisibleDatabaseTree db) {
        Optional<String> filter = Optional.empty();
        if (db.hasGrant(resourceId) && db.getGrant(resourceId).get().hasOperation(operation)) {
            filter = db.getGrant(resourceId).get().getFilter(operation);
        }
        return filter;
    }

    ////////////////////////////////////////// RESOURCE DOMAIN CHECK METHODS ///////////////////////////////////////////

    private static boolean isDatabase(ResourceId resourceId, VisibleDatabaseTree db) {
        return resourceId.equals(db.getDatabaseId());
    }

    ///////////////////////////////////////////// FORM PERMISSION METHODS //////////////////////////////////////////////

    public static FormPermissions formPermissions(ResourceId formId, VisibleDatabaseTree db) {

        if(VisibleDatabaseTree.isUserForm(formId)) {
            return userFormPermissions(db);
        }

        return db.getResource(formId)
            .filter(resource -> resource.isFormOrSubform())
            .map(resource -> formPermissions(resource, db))
            .orElse(FormPermissions.none());
    }

    public static FormPermissions userFormPermissions(VisibleDatabaseTree tree) {
        // If the user can see the database, they can see the user form
        return FormPermissions.readonly();
    }

    private static FormPermissions formPermissions(Resource form, VisibleDatabaseTree db) {
        if(db.isSuspended()) {
            return FormPermissions.none();
        }
        HasPermissions effectivePermission = findEffectivePermission(db, form.getId());
        if(!effectivePermission.hasPermission(Operation.VIEW)) {
            return FormPermissions.none();
        } else {
            return new FormPermissions(form.getVisibility(), effectivePermission, db.getEffectiveLocks(form.getId()));
        }
    }


    ///////////////////////////////////////////// BASIC PERMISSION QUERIES /////////////////////////////////////////////

    public static Permission view(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.VIEW, resourceId);
    }

    public static Permission addRecord(ResourceId formId, VisibleDatabaseTree db) {
        return query(db, Operation.ADD_RECORD, formId);
    }

    public static Permission deleteRecord(ResourceId formId, VisibleDatabaseTree db) {
        return query(db, Operation.DELETE_RECORD, formId);
    }

    public static Permission editRecord(ResourceId formId, VisibleDatabaseTree db) {
        return query(db, Operation.EDIT_RECORD, formId);
    }

    public static Permission manageUsers(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.MANAGE_USERS, resourceId);
    }

    public static Permission createResource(ResourceId containerResourceId, VisibleDatabaseTree db) {
        return query(db, Operation.ADD_RESOURCE, containerResourceId);
    }

    public static Permission deleteResource(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.DELETE_RESOURCE, resourceId);
    }

    public static Permission editResource(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.EDIT_RESOURCE, resourceId);
    }

    public static Permission manageCollectionLinks(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.MANAGE_COLLECTION_LINKS, resourceId);
    }

    public static Permission lockRecords(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.LOCK_RECORDS, resourceId);
    }

    public static Permission exportRecords(ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, Operation.EXPORT_RECORDS, resourceId);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////// TASK METHODS ///////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    ///////////////////////////////////////////////// DATABASE Methods /////////////////////////////////////////////////

    /**
     * <p>The current User can transfer this Database.</p>
     */
    public static boolean canTransferDatabase(VisibleDatabaseTree db) {
        return db.isOwner();
    }

    /**
     * <p>The current User can delete this Database.</p>
     */
    public static boolean canDeleteDatabase(VisibleDatabaseTree db) {
        return db.isOwner();
    }

    public static boolean canDeleteDatabase(DatabaseHeader database, AuthenticatedUser user) {
        return database.getOwnerId().equals(user.getId());
    }

    /**
     * <p>The current User can edit this Database's metadata.</p>
     */
    public static boolean canEditDatabaseHeader(VisibleDatabaseTree db) {
        return db.isOwner();
    }

    /////////////////////////////////////////////////// SYNC Methods ///////////////////////////////////////////////////

    ////////////////////////////////////////////////// Generic Methods /////////////////////////////////////////////////

    /**
     * <p>The current User can perform the given Operation on the specified Resource in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean can(Operation operation, ResourceId resourceId, VisibleDatabaseTree db) {
        return query(db, operation, resourceId).isPermitted();
    }

    /////////////////////////////////////////////////// VIEW Methods ///////////////////////////////////////////////////

    /**
     * <p>The current User can view the root Resource(s) in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canView(VisibleDatabaseTree db) {
        return view(db.getDatabaseId(),db).isPermitted();
    }

    /**
     * <p>The current User can view the specified Resource in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canView(ResourceId resourceId, VisibleDatabaseTree db) {
        return view(resourceId,db).isPermitted();
    }

    ////////////////////////////////////////////// CREATE_RECORD Methods ///////////////////////////////////////////////

    /**
     * <p>The current User can create a Record within the specified Form in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canAddRecord(ResourceId formId, VisibleDatabaseTree db) {
        return formPermissions(formId, db).isAddAllowed();
    }

    ////////////////////////////////////////////// DELETE_RECORD Methods ///////////////////////////////////////////////

    /**
     * <p>The current User can delete a Record within the specified Form in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canDeleteRecord(ResourceId formId, VisibleDatabaseTree db) {
        return formPermissions(formId, db).isDeleteAllowed();
    }

    public static boolean canDeleteRecordsInBulk(ResourceId formId, VisibleDatabaseTree db) {
        return formPermissions(formId, db).isBulkDeleteAllowed();
    }

    /////////////////////////////////////////////// EDIT_RECORD Methods ////////////////////////////////////////////////

    /**
     * <p>The current User can edit a Record within the specified Form in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canEditRecord(ResourceId formId, VisibleDatabaseTree db) {
        return formPermissions(formId, db).isEditAllowed();
    }

    ///////////////////////////////////////////// CREATE_RESOURCE Methods //////////////////////////////////////////////

    /**
     * <p>The current User can create a new Resource <i>within</i> the specified Resource in this Database.</p>
     */
    public static boolean canAddResource(ResourceId containerResourceId, VisibleDatabaseTree db) {
        return createResource(containerResourceId,db).isPermitted();
    }

    /**
     * <p>The current User can create a new Form <i>within</i> the specified Resource in this Database.</p>
     */
    public static boolean canAddForm(ResourceId containerResourceId, VisibleDatabaseTree db) {
        return canAddResource(containerResourceId, db);
    }

    ///////////////////////////////////////////// DELETE_RESOURCE Methods //////////////////////////////////////////////

    /**
     * <p>The current User can delete the specified Resource in this Database.</p>
     */
    public static boolean canDeleteResource(ResourceId resourceId, VisibleDatabaseTree db) {
        return deleteResource(resourceId,db).isPermitted();
    }

    /**
     * <p>The current User can delete the specified Form in this Database.</p>
     */
    public static boolean canDeleteForm(ResourceId formId, VisibleDatabaseTree db) {
        return canDeleteResource(formId, db);
    }

    /**
     * <p>The current User can delete the specified Folder in this Database.</p>
     */
    public static boolean canDeleteFolder(ResourceId folderId, VisibleDatabaseTree db) {
        return canDeleteResource(folderId, db);
    }

    public static boolean canMoveResource(ResourceId resourceId, VisibleDatabaseTree db) {
        return canDeleteResource(resourceId, db) &&
               canEditResource(resourceId, db);
    }

    ////////////////////////////////////////////// EDIT_RESOURCE Methods ///////////////////////////////////////////////

    /**
     * <p>The current User can edit the specified Resource in this Database.</p>
     */
    public static boolean canEditResource(ResourceId resourceId, VisibleDatabaseTree db) {
        return editResource(resourceId,db).isPermitted();
    }

    /**
     * <p>The current User can edit the specified Form in this Database.</p>
     */
    public static boolean canEditForm(ResourceId formId, VisibleDatabaseTree db) {
        return canEditResource(formId, db);
    }

    /**
     * <p>The current User can edit the specified Folder in this Database.</p>
     */
    public static boolean canEditFolder(ResourceId folderId, VisibleDatabaseTree db) {
        return canEditResource(folderId, db);
    }


    //////////////////////////////////////////////// COLLECTION LINKS //////////////////////////////////////////////////
    /**
     * <p>The current User can manage collection links for the specified Form in this Database.</p>
     */
    public static boolean canManageCollectionLinks(ResourceId rootFormId, VisibleDatabaseTree database) {
        return manageCollectionLinks(rootFormId, database).isPermitted();
    }

    ////////////////////////////////////////////// LOCK_RECORDS Methods ////////////////////////////////////////////////

    /**
     * <p>The current User can lock records on the specified Resource (and all children) in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canLockRecords(ResourceId resourceId, VisibleDatabaseTree db) {
        return lockRecords(resourceId,db).isPermitted();
    }

    ///////////////////////////////////////////// EXPORT_RECORDS Methods ///////////////////////////////////////////////

    /**
     * <p>The current User can export records from the specified Resource in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */

    public static boolean canExportFormsIn(ResourceId parentId, VisibleDatabaseTree database) {
        Set<Resource> resources = new HashSet<>();
        if (parentId.equals(database.getDatabaseId())) {
            resources.addAll(database.getResources());
        } else {
            resources.addAll(
                    database.getResource(parentId)
                        .map(resource ->
                                database.getDescendants(resource.getId()))
                        .orElse(Collections.emptySet())
            );
        }
        for (Resource resource : resources) {
            if (resource.isFormOrSubform() && resource.getVisibility() != Resource.Visibility.REFERENCE) {
                if (!canExportRecords(resource.getId(), database)) {
                    return false;
                }
            }
        }
        return true;
    }

    public static boolean canExportRecords(ResourceId resourceId, VisibleDatabaseTree db) {
        return exportRecords(resourceId, db).isPermitted();
    }

    ////////////////////////////////////////////// MANAGE_USERS Methods ////////////////////////////////////////////////

    /**
     * <p>The current User can manage users for any Resource on this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canManageUsersOnWholeDatabase(VisibleDatabaseTree db) {
        return manageUsers(db.getDatabaseId(),db).isPermitted();
    }

    /**
     * <p>The current User can manage users on the specified Resource in this Database.</p>
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canManageUsersOnResource(ResourceId resourceId, VisibleDatabaseTree db) {
        return manageUsers(resourceId,db).isPermitted();
    }

    /**
     * <p>The current User can manage users for <b>one or more</b> Resources. If the User is <b>not</b> permitted to
     * manage users on <i>any</i> Resources on the database, then the User cannot manage users at all and this method
     * will return false.</p>
     *
     * <p>Does not account for any filters which may be applied/enforced.</p>
     */
    public static boolean canManageUsersForOneOrMoreResources(VisibleDatabaseTree db) {
        // If we have a root database grant, check if we can grant users at this level
        if (canManageUsersOnWholeDatabase(db)) {
            return true;
        }
        // If not, check if we have any resources for which we can manage users
        return !resourcesWithManageUserRights(db).isEmpty();
    }

    private static List<Resource> resourcesWithManageUserRights(VisibleDatabaseTree db) {
        return db.getResources().stream()
                .filter(r -> canManageUsersOnResource(r.getId(), db))
                .collect(Collectors.toList());
    }

    /**
     * <p>The legacy permission model restricts a Users ability to manage users to All Partners or a specified
     * Partner, across the set of Resources which the User has access to.</p>
     *
     * <p>The Partner restriction is consistent across all "manageable" Resources. Therefore, we find the first filter
     * defined on a manageable resource as it will be the same for all other manageable resources.</p>
     */
    public static Optional<String> legacyManageUserFilter(VisibleDatabaseTree db) {
        // If we have a root database grant, then return the filter defined there
        if (canManageUsersOnWholeDatabase(db)) {
            return manageUsers(db.getDatabaseId(),db).getOptionalFilter();
        }
        // If not, then check for the first filter we find on a manageable resource
        return db.getResources().stream()
                .map(r -> manageUsers(r.getId(), db))
                .filter(Permission::isPermitted)
                .map(Permission::getOptionalFilter)
                .findFirst().orElse(Optional.empty());
    }

    ///////////////////////////////////////////// MANAGE_PARTNER Methods ///////////////////////////////////////////////

    public static boolean canViewDatabaseSettings(VisibleDatabaseTree database) {
        if(database.isOwner()) {
            return true;
        }
        for (DatabaseUserGrant grant : database.getGrants()) {
            for (Operation operation : grant.getOperations()) {
                switch (operation) {
                    case MANAGE_USERS:
                    case ADD_RESOURCE:
                    case DELETE_RESOURCE:
                    case EDIT_RESOURCE:
                    case LOCK_RECORDS:
                        return true;
                }
            }
        }
        return false;
    }

    public static boolean canAddFolder(ResourceId containerId, VisibleDatabaseTree database) {
        return canAddForm(containerId, database);
    }


    /**
     * Performs a quick check to determine whether this user can assign the given role.
     */
    public static boolean canAssignRole(VisibleDatabaseTree tree, Role otherRole) {

        if(tree.isOwner()) {
            return true;
        }

        Set<Operation> permittedOperations = tree.getGrants()
            .stream()
            .flatMap(g -> g.getOperations().stream())
            .collect(Collectors.toSet());

        if(!permittedOperations.contains(Operation.MANAGE_USERS)) {
            return false;
        }

        for (FilteredPermission permission : otherRole.getPermissions()) {
            if (!permittedOperations.contains(permission.getOperation())) {
                return false;
            }
        }

        return true;
    }

    public static boolean canEditRoleAssignment(VisibleDatabaseTree tree, DatabaseUserGrantSet grantSet) {
        return !isSelf(tree, grantSet) && canEditRoleAssignment(tree, grantSet.getRole());
    }

    public static boolean isSelf(VisibleDatabaseTree tree, DatabaseUserGrantSet grantSet) {
        return tree.getUserId().map(uId -> uId.equals(grantSet.getUserId())).orElse(false);
    }

    /**
     * Return true if the user is allowed to edit the existing role assignment.
     *
     * <p>The user will be permitted if the role is not superior to the user's own role, and
     * if the would-be editor can change at least one resource assignment.
     * </p>
     */
    public static boolean canEditRoleAssignment(VisibleDatabaseTree tree, RoleAssignment assignment) {
        if(tree.isOwner()) {
            return true;
        }
        EffectiveRole effectiveRole = EffectiveRole.compute(tree, assignment);
        if(effectiveRole.isEmpty()) {
            // Invalid role assignment - role does not exist in the database
            return false;
        }
        if(!canAssignRole(tree, effectiveRole.getRole())) {
            return false;
        }

        for (DatabaseUserGrant grant : tree.getGrants()) {
            if (grant.hasPermission(Operation.MANAGE_USERS)) {
                if(canAssignResource(tree, assignment, grant.getResourceId())) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Returns true if the given role assignment includes resources for which we are not authorized to manage users.
     *
     */
    public static boolean assignedResourcesOutsidePurview(VisibleDatabaseTree tree, EffectiveRole assignment) {
        for (ResourceId roleResourceId : assignment.getResources()) {
            HasPermissions effectivePermission = findEffectivePermission(tree, roleResourceId);
            if(!effectivePermission.hasPermission(Operation.MANAGE_USERS)) {
                return true;
            }
        }
        return false;
    }


    /**
     * Return true if the user is allowed to add the given {@code resourceId} to an existing role
     * assignment.
     *
     * <p>If the user does not have manage users permission for this resource, adding will not be possible.</p>
     *
     * <p>If the user has been assigned a resource which is an ancestor of this resource, then adding
     * the resource from the assignment will not be permitted.</p>
     *
     * @param tree the database tree visible to the current user
     * @param resourceId the id of the resource to add or remove from the existing assignment
     */
    public static boolean canAddRoleResourceAssignment(VisibleDatabaseTree tree, RoleAssignment existingAssignment, ResourceId resourceId) {
        HasPermissions effectivePermission = findEffectivePermission(tree, resourceId);
        if(!effectivePermission.hasPermission(Operation.MANAGE_USERS)) {
            return false;
        }

        return canAssignResource(tree, existingAssignment, resourceId);
    }


    /**
     * Return true if the user is allowed to add or remove the given {@code resourceId} from an existing role
     * assignment.
     *
     * <p>If the user has been assigned a resource which is an ancestor of this resource, then adding or removing
     * the resource from the assignment will not be permitted.</p>
     *
     * @param tree the database tree visible to the current user
     * @param resourceId the id of the resource to add or remove from the existing assignment
     */
    private static boolean canAssignResource(VisibleDatabaseTree tree, RoleAssignment existingAssignment, ResourceId resourceId) {

        Set<ResourceId> assignedResources = existingAssignment.getRoleResources();
        if(assignedResources.contains(resourceId)) {
            return true;
        }

        Set<ResourceId> ancestors = new HashSet<>(tree.getAncestors(resourceId));

        Set<ResourceId> superior = Sets.intersection(ancestors, assignedResources);
        for (ResourceId superiorResourceId : superior) {
            HasPermissions effectivePermission = findEffectivePermission(tree, superiorResourceId);
            if(!effectivePermission.hasPermission(Operation.MANAGE_USERS)) {
                return false;
            }
        }

        return true;
    }

    public static HasPermissions findEffectivePermission(VisibleDatabaseTree tree, final ResourceId resourceId) {

        if(tree.isOwner()) {
            return HasPermissions.all(tree);
        }

        HasPermissions grant = findExplicitOrInheritedGrant(tree, resourceId);
        if (grant != null) {
            return grant;
        }

        // If we don't _inherit_ any grants, then we are here because one of our descendants
        // is visible.

        if(resourceId.equals(tree.getDatabaseId())) {
            return HasPermissions.VIEW_ONLY;
        }

        Resource resource = tree.getResource(resourceId).orElse(null);
        if(resource == null) {
            return HasPermissions.NONE;
        }

        // If the resource has Reference/Public visibility, then all records
        // are visible

        if(resource.getVisibility() == Resource.Visibility.REFERENCE ||
            resource.getVisibility() == Resource.Visibility.PUBLIC) {

            return HasPermissions.VIEW_ONLY;
        }

        // Record filtering is not relevant for folders, skip the search
        if(resource.getType() == ResourceType.FOLDER) {
            return HasPermissions.VIEW_ONLY;
        }

        // If this resource is a form or subform, then we don't automatically get to see all the records,
        // we should only be able to see the parents of the records we can see.
        Set<FormulaNode> filters = new HashSet<>();
        collectDescendantViewFilters(tree, resourceId, filters);

        // If any of the subforms is unfiltered, then the parent must also be unfiltered
        if(filters.isEmpty() || filters.contains(ConstantNode.TRUE)) {
            return HasPermissions.VIEW_ONLY;
        }

        // Otherwise the parent records are visible if any of the subrecords are visible
        return HasPermissions.viewOnly(Formulas.anyTrue(filters));
    }

    private static @Nullable HasPermissions findExplicitOrInheritedGrant(VisibleDatabaseTree tree, ResourceId resourceId) {
        // First look UP the tree to see if we inherit any grants
        ResourceId id = resourceId;
        while(true) {
            Optional<DatabaseUserGrant> grant = tree.getGrant(id);
            if(grant.isPresent()) {
                return combineGrantWithVisibility(grant.get(), tree.getResource(resourceId));
            }
            if(id.equals(tree.getDatabaseId())) {
                break;
            }
            Optional<ResourceId> parentId = tree.getResource(id).map(r -> r.getParentId());
            if(!parentId.isPresent()) {
                break;
            }

            id = parentId.get();
        }

        return null;
    }

    private static void collectDescendantViewFilters(VisibleDatabaseTree tree, ResourceId parentId, Set<FormulaNode> filters) {
        for (Resource resource : tree.getResources()) {
            if(resource.getParentId().equals(parentId)) {
                Optional<DatabaseUserGrant> grant = tree.getGrant(resource.getId());
                if(grant.isPresent()) {
                    Optional<String> viewFilter = grant.get().getFilter(Operation.VIEW);
                    if(!viewFilter.isPresent()) {
                        filters.add(ConstantNode.TRUE);
                    } else {
                        FormulaNode parsedViewFormula = FormulaParser.tryParse(viewFilter.get()).orElse(ConstantNode.FALSE);
                        filters.add(parsedViewFormula);
                    }
                } else {
                    // This node is not filtered, check ITS descendants
                    collectDescendantViewFilters(tree, resource.getId(), filters);
                }
            }
        }
    }

    public static HasPermissions combineGrantWithVisibility(DatabaseUserGrant grant, Optional<Resource> resource) {
        return resource
            .map(r -> combineGrantWithVisibility(r, grant))
            .orElse(grant);
    }

    private static HasPermissions combineGrantWithVisibility(Resource resource, DatabaseUserGrant grant) {
        // If the resource is present in the VisibleDatabaseTree, then it is visible, either
        // because it is public, or reference, or because it has a child which is visible.
        switch (resource.getVisibility()) {
            case PUBLIC:
            case REFERENCE:
                return PermissionSet.withUnfilteredView(grant);
            /*case PRIVATE:
                if(resource.getType() == ResourceType.FOLDER) {
                    return PermissionSet.withUnfilteredView(grant);
                }*/
        }
        return grant;
    }

    /**
     * Performs a check to determine whether this user can included in the given
     * resource in the role.
     */
    public static boolean canAssignRole(VisibleDatabaseTree tree, Role otherRole, ResourceId resourceId) {

        if(tree.isOwner()) {
            return true;
        }

        HasPermissions myPermissions = PermissionOracle.findEffectivePermission(tree, resourceId);
        if (myPermissions.hasPermission(Operation.MANAGE_USERS)) {
            return isSuperSet(myPermissions, otherRole);
        } else {
            return false;
        }
    }

    /**
     * Returns true if {@code a} is a superset of {@code b}. That is, {@code a} should
     * include at least all the permissions in {@code b}, and possibly more.
     */
    public static boolean isSuperSet(HasPermissions a, HasPermissions b) {
        for (FilteredPermission permission : b.getPermissions()) {
            if(!a.hasPermission(permission.getOperation())) {
                return false;
            }
        }

        return true;
    }

    public static boolean canDeleteGrant(VisibleDatabaseTree database, DatabaseUserGrant grant) {
        return canEditGrant(database, grant);
    }


    public static boolean canAddGrant(VisibleDatabaseTree tree, DatabaseUserGrant addedGrant) {
        return canEditGrant(tree, addedGrant);
    }

    public static boolean canEditGrant(VisibleDatabaseTree database, DatabaseUserGrant previousValue, DatabaseUserGrant newValue) {
        return canEditGrant(database, previousValue) &&
                    canAddGrant(database, newValue);
    }

    public static boolean canEditGrant(VisibleDatabaseTree database, DatabaseUserGrant grant) {
        if(database.isOwner()) {
            return true;
        }
        HasPermissions myPermissions = PermissionOracle.findEffectivePermission(database, grant.getResourceId());
        if(!myPermissions.hasPermission(Operation.MANAGE_USERS)) {
            return false;
        }

        return isSuperSet(myPermissions, grant);
    }

    public static boolean canPreflightUser(VisibleDatabaseTree tree) {
        if(tree.isOwner()) {
            return true;
        }
        for (DatabaseUserGrant grant : tree.getGrants()) {
            if(grant.hasPermission(Operation.MANAGE_USERS)) {
                return true;
            }
        }
        return false;
    }

    public static boolean canListAccounts(PlatformRole role) {
        switch (role) {
            case NONE:
                return false;
            default:
                return true;
        }
    }

    public static boolean canTransitionAccount(UserAccount executingUser, BillingAccount account, BillingAccountStatus newStatus) {
        if(canTransitionAccount(executingUser.getPlatformRole(), newStatus)) {
            return true;
        }
        return executingUser.getBillingRole()
            .filter(role -> role.getBillingAccountId() == account.getId())
            .map(role -> canTransitionBillingAccount(role, newStatus))
            .orElse(false);
    }

    public static boolean canTransitionBillingAccount(BillingRoleAssignment role, BillingAccountStatus newStatus) {
        // The actual owner of a billing account can close their own account,
        // but they can't reactivate it, or suspended etc.
        return
            newStatus == BillingAccountStatus.CLOSED &&
            role.getRole() == BillingAccountRole.OWNER;
    }

    public static boolean canTransitionAccount(PlatformRole platformRole, BillingAccountStatus newStatus) {

        if(newStatus == BillingAccountStatus.CLOSED) {
            // Only Platform owners or the billing account owner can close an account
            return platformRole == PlatformRole.OWNER;
        }

        switch (platformRole) {
            case OWNER:
            case SALES:
            case SUPPORT:
                return true;
            default:
                return false;
        }
    }

    public static boolean canViewDatabaseBillingAccount(VisibleDatabaseTree tree) {
        return canViewDatabaseSettings(tree);
    }

    /**
     * True if the user can add, edit, or delete roles in the database
     */
    public static boolean canManageRoles(VisibleDatabaseTree visibleTree) {
        return query(visibleTree, Operation.MANAGE_ROLES, visibleTree.getDatabaseId()).isPermitted();
    }

    public static boolean canViewAccount(AuthenticatedUser authenticatedUser) {
        switch (authenticatedUser.getPlatformRole()) {
            case NONE:
                return false;
            default:
                return true;
        }
    }

    public static boolean canUpdateAccount(AuthenticatedUser authenticatedUser) {
        switch (authenticatedUser.getPlatformRole()) {
            case SUPPORT:
            case OWNER:
            case SALES:
                return true;
            default:
                return false;
        }
    }

    public static boolean canTransferBillingAccountRole(PlatformRole platformRole) {
        switch (platformRole) {
            case OWNER:
            case SALES:
                return true;
            default:
                return false;
        }
    }

    public static boolean canAddBillingAccountUser(PlatformRole platformRole) {
        switch (platformRole) {
            case OWNER:
            case SALES:
            case SUPPORT:
                return true;
            default:
                return false;
        }
    }

    /**
     * Returns true if the user is allowed to include the given form in a published report.
     */
    public static boolean canPublishForm(VisibleDatabaseTree database, ResourceId formId) {
        if(database.isOwner()) {
            return true;
        }
        Resource.Visibility formVisibility = database
            .getResource(formId)
            .map(Resource::getVisibility)
            .orElse(Resource.Visibility.PRIVATE);

        if(formVisibility == Resource.Visibility.PUBLIC) {
            return true;
        }

        return findEffectivePermission(database, formId).hasPermission(Operation.PUBLISH_REPORTS);
    }

    public static boolean isReportGrantPermitted(VisibleDatabaseTree database, ReportRoleGrant roleGrant) {
        return canShareReports(database);
    }

    /**
     * Returns true if the user is allowed to share reports with users of the given database.
     */
    public static boolean canShareReports(VisibleDatabaseTree database) {
        if(database.isOwner()) {
            return true;
        }
        EffectiveRole effectiveRole = EffectiveRole.compute(database, database.getRole());
        return effectiveRole.getRole().hasPermission(Operation.SHARE_REPORTS);
    }

    public static boolean canAuditDatabase(VisibleDatabaseTree tree) {
        return tree.isOwner() || tree.getGrants().stream().anyMatch(g -> g.hasPermission(Operation.AUDIT));
    }


    public static boolean canAuditResource(VisibleDatabaseTree tree, ResourceId resourceId) {
        return findEffectivePermission(tree, resourceId).hasPermission(Operation.AUDIT);
    }

    public static boolean canAuditRecords(VisibleDatabaseTree tree, ResourceId resourceId) {
        // Special case for owner, because the resource does not need to be present in the
        // tree. It could have been deleted in the past, for example
        if(tree.isOwner()) {
            return true;
        }
        HasPermissions effectivePermission = findEffectivePermission(tree, resourceId);

        return effectivePermission.hasPermission(Operation.AUDIT) &&
            effectivePermission.hasUnconditionalPermission(Operation.VIEW);
    }

    public static boolean canDeleteUser(VisibleDatabaseTree visibleTree, DatabaseUserGrantSet grantSet) {

        if(!canManageUsersForOneOrMoreResources(visibleTree)) {
            return false;
        }

        EffectiveRole effectiveRole = EffectiveRole.compute(visibleTree, grantSet.getRole());
        if(!canAssignRole(visibleTree, effectiveRole.getRole())) {
            return false;
        }
        for (ResourceId resourceId : effectiveRole.getResources()) {
            if(!canManageUsersOnResource(resourceId, visibleTree)) {
                return false;
            }
        }

        for (DatabaseUserGrant grant : grantSet.getGrants()) {
            if (!PermissionOracle.canDeleteGrant(visibleTree, grant)) {
                return false;
            }
        }
        return true;
    }


    public static boolean canViewUser(AuthenticatedUser user, ResourceId userId) {
        return canViewUsers(user);
    }

    public static boolean canViewUsers(AuthenticatedUser user) {
        switch (user.getPlatformRole()) {
            case OWNER:
            case SUPPORT:
            case DEVELOPER:
            case SALES:
                return true;
            default:
                return false;
        }
    }

    public static boolean canSearchDatabases(AuthenticatedUser user) {
        switch (user.getPlatformRole()) {
            case OWNER:
            case SUPPORT:
            case DEVELOPER:
            case SALES:
                return true;
            default:
                return false;
        }
    }

    public static boolean canAuditUsers(AuthenticatedUser authUser) {
        return canViewUsers(authUser);
    }

    public static boolean canUniliaterallyTransferDatabase(AuthenticatedUser user) {
        switch (user.getPlatformRole()) {
            case OWNER:
            case SUPPORT:
            case DEVELOPER:
            case SALES:
                return true;
            default:
                return false;
        }
    }

    public static boolean canViewDeletedRecordHistory(VisibleDatabaseTree database, ResourceId formId) {
        return findEffectivePermission(database, formId).hasPermission(Operation.AUDIT);
    }

    public static boolean canUpdateMapTemplate(AuthenticatedUser authUser) {
        switch (authUser.getPlatformRole()) {
            case SUPPORT:
            case DEVELOPER:
            case OWNER:
                return true;
            default:
                return false;
        }
    }

    public static boolean canClaimSupportAccess(AuthenticatedUser authUser) {
        return canClaimSupportAccess(authUser.getPlatformRole());
    }

    public static boolean canClaimSupportAccess(PlatformRole platformRole) {
        switch (platformRole) {
            case SUPPORT:
            case DEVELOPER:
            case SALES:
            case OWNER:
                return true;
            default:
                return false;
        }

    }

    public static boolean canViewUserActions(AuthenticatedUser user) {
        switch (user.getPlatformRole()) {
            case DEVELOPER:
            case SUPPORT:
            case SALES:
            case OWNER:
                return true;
            default:
                return false;
        }
    }

    public static boolean canAddDatabase(AccountStatus status) {
        return status.getBillingAccount()
            .map(billingAccount -> billingAccount.getStatus() == BillingAccountStatus.ACTIVE)
            .orElse(false);
    }

    public static boolean canRecoverDeletedDatabase(AuthenticatedUser authenticatedUser, ResourceId databaseOwnerId) {
        if (databaseOwnerId.equals(authenticatedUser.getUserId())) {
            return true;
        }
        switch (authenticatedUser.getPlatformRole()) {
            case DEVELOPER:
            case SUPPORT:
            case SALES:
            case OWNER:
                return true;
            default:
                return false;
        }
    }

    public static boolean canExportBillingStats(UserAccount user, long billingAccountId) {

        // Platform support staff are authorized to export statistics over any billing account
        switch (user.getPlatformRole()) {
            case SUPPORT:
            case OWNER:
            case SALES:
                return true;
        }

        // Otherwise must be a billing account owner
        return user.getBillingRole()
            .filter(role -> role.getBillingAccountId() == billingAccountId)
            .map(role -> {
                switch (role.getRole()) {
                    case OWNER:
                        return true;
                    default:
                        return false;
                }
            })
            .orElse(false);
    }

    public static boolean canPublishTemplate(VisibleDatabaseTree tree, PlatformRole platformRole) {
        switch (platformRole) {
            case DEVELOPER:
            case OWNER:
            case SUPPORT:
            case SALES:
                return tree.isOwner();
            default:
                return false;
        }
    }

    public static boolean canRecoverDeletedRecord(Maybe<VisibleDatabaseTree> maybeTree, RecordRef recordRef) {
        HasPermissions permissions = maybeTree
                .transform(tree -> findEffectivePermission(tree, recordRef.getFormId()))
                .or(HasPermissions.NONE);

        return permissions.hasPermission(Operation.AUDIT) &&
                permissions.hasUnconditionalPermission(Operation.ADD_RECORD);
    }

    public static boolean canSetupServer(PlatformRole platformRole) {
        return platformRole == PlatformRole.OWNER;
    }
}