/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.query;

import com.google.common.base.Joiner;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;

import javax.annotation.Nonnull;
import java.util.Collections;
import java.util.Map;

/**
 * Contains the data
 */
public class ColumnSet implements JsonSerializable {

    public static final ColumnSet EMPTY = new ColumnSet();

    private final int numRows;
    private final Map<String, ColumnView> columnViews;

    private ColumnSet() {
        this.numRows = 0;
        this.columnViews = Collections.emptyMap();
    }

    public ColumnSet(int numRows, @Nonnull Map<String, ColumnView> columnViews) {
        this.numRows = numRows;
        this.columnViews = columnViews;
    }

    /**
     * @return the number of rows in this Table
     */
    public int getNumRows() {
        return numRows;
    }


    @Nonnull
    public Map<String, ColumnView> getColumns() {
        return columnViews;
    }

    /**
     *
     * @param columnModelId the {@code id} of the {@code ColumnModel}
     * @return the {@code ColumnView} generated from the given {@code ColumnModel}
     */
    public ColumnView getColumnView(String columnModelId) {
        return columnViews.get(columnModelId);
    }

    public ColumnView getColumnViewOrBlank(String columnId, ColumnType type) {
        ColumnView view = columnViews.get(columnId);
        if(view == null) {
            return new BlankColumnView(type, numRows);
        }
        return view;
    }

    @Override
    public String toString() {
        return "TableData{" +
               "numRows=" + numRows +
               ", columnViews=" + Joiner.on("\n").withKeyValueSeparator("=").join(columnViews) +
               '}';
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = Json.createObject();
        object.put("rows", numRows);

        JsonValue columns = Json.createObject();
        columnViews.forEach((id, columnView) -> {
            columnToJson(id, columnView, columns);
        });
        object.put("columns", columns);

        return object;
    }

    private void columnToJson(String id, ColumnView columnView, JsonValue columns) {
        JsonValue column = Json.createObject();
        column.put("type", columnView.getType().name());

        if (columnView instanceof BlankColumnView) {
            column.put("storage", "empty");
        } else if (columnView instanceof ConstantColumnView) {
            constantViewToJson(columnView, column);
        } else {
            arrayViewToJson(columnView, column);
        }
        columns.put(id, column);
    }

    private void constantViewToJson(ColumnView columnView, JsonValue column) {
        column.put("storage", "constant");

        switch (columnView.getType()) {
            case STRING:
                column.put("value", columnView.getString(0));
                break;
            case NUMBER:
                column.put("value", columnView.getDouble(0));
                break;
            case BOOLEAN:
                column.put("value", columnView.getBoolean(0));
                break;
        }
    }

    private void arrayViewToJson(ColumnView columnView, JsonValue column) {
        column.put("storage", "array");

        JsonValue values = Json.createArray();
        switch (columnView.getType()) {
            case STRING:
                for (int i = 0; i < columnView.numRows(); i++) {
                    String value = columnView.getString(i);
                    values.add(Json.createFromNullable(value));
                }
                break;
            case NUMBER:
                for (int i = 0; i < columnView.numRows(); i++) {
                    double value = columnView.getDouble(i);
                    if (Double.isNaN(value)) {
                        values.add(Json.create("null"));
                    } else {
                        values.add(Json.create(value));
                    }
                }
                break;
            case BOOLEAN:
                for (int i = 0; i < columnView.numRows(); i++) {
                    int value = columnView.getBoolean(i);
                    if (value == ColumnView.BLANK) {
                        values.add(Json.create("null"));
                    } else {
                        values.add(Json.create(value == ColumnView.TRUE));
                    }
                }
                break;
        }
        column.put("values", values);

    }

    public static ColumnSet fromJson(JsonValue jsonValue) {
        return ColumnSetParser.fromJson(jsonValue);
    }
}
