/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.query;

import com.google.common.collect.Maps;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonValue;

import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;


public class ColumnSetParser {

    private static final Logger LOGGER = Logger.getLogger(ColumnSetParser.class.getName());

    public static ColumnSet fromJson(String json) {
        try {
            JsonValue jsonObject = Json.parse(json);
            return fromJson(jsonObject);
        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, "Failed to parse column set: " + e.getMessage() + "\n" + json, e);
            throw e;
        }
    }

    public static ColumnSet fromJson(JsonValue jsonObject) {
        int rowsCount = (int) jsonObject.getNumber("rows");

        Map<String, ColumnView> views = Maps.newHashMap();

        JsonValue jsonColumns = jsonObject.get("columns");
        for (String columnId : jsonColumns.keys()) {
            views.put(columnId, parseColumn(rowsCount, jsonColumns.get(columnId)));
        }
        return new ColumnSet(rowsCount, views);
    }

    private static ColumnView parseColumn(int rowsCount, JsonValue jsonColumn) {
        String storage = jsonColumn.getString("storage");
        String typeName = jsonColumn.getString("type");
        ColumnType type = ColumnType.valueOf(typeName);

        switch (storage) {
            case "array":
                JsonValue jsonArray = jsonColumn.get("values");
                switch (type) {
                    case STRING:
                        return new JsonStringView(jsonArray);
                    case NUMBER:
                        return new JsonNumberView(jsonArray);
                    case BOOLEAN:
                        return new JsonBooleanView(jsonArray);
                }
                break;

            case "empty":
                return new BlankColumnView(type, rowsCount);

            case "constant":
                JsonValue jsonValue = jsonColumn.get("value");
                switch (type) {
                    case STRING:
                        return new ConstantColumnView(rowsCount, jsonValue.asString());
                    case NUMBER:
                        if(jsonValue.isNumber()) {
                            return new ConstantColumnView(rowsCount, jsonValue.asNumber());
                        } else {
                            return new ConstantColumnView(rowsCount, Double.NaN);
                        }
                    case BOOLEAN:
                        if(jsonValue.isBoolean() || jsonValue.isNumber()) {
                            return new ConstantColumnView(rowsCount, jsonValue.asBoolean());
                        } else {
                            return ConstantColumnView.nullBoolean(rowsCount);
                        }
                }
                break;
        }
        throw new UnsupportedOperationException("ColumnSetParser: storage is not supported, storage: " + storage);
    }

    private static class JsonStringView extends ColumnView {

        private JsonValue array;

        public JsonStringView(JsonValue array) {
            this.array = array;
        }

        @Override
        public ColumnType getType() {
            return ColumnType.STRING;
        }

        @Override
        public int numRows() {
            return array.length();
        }

        @Override
        public Object get(int row) {
            return getString(row);
        }

        @Override
        public double getDouble(int row) {
            return Double.NaN;
        }

        @Override
        public String getString(int row) {
            return array.getString(row);
        }

        @Override
        public boolean isBlank(int row) {
            return array.get(row).isJsonNull();
        }

        @Override
        public ColumnView select(int[] rows) {
            JsonValue selectedArray = Json.createArray();
            for (int i = 0; i < rows.length; i++) {
                selectedArray.add(array.get(rows[i]));
            }
            return new JsonStringView(selectedArray);
        }

        @Override
        public int[] order(int[] sortVector, SortDir direction, int[] range) {
            throw new UnsupportedOperationException();
        }
    }

    private static class JsonNumberView extends ColumnView {

        private JsonValue array;

        public JsonNumberView(JsonValue array) {
            this.array = array;
        }

        @Override
        public ColumnType getType() {
            return ColumnType.NUMBER;
        }

        @Override
        public int numRows() {
            return array.length();
        }

        @Override
        public Object get(int row) {
            JsonValue value = array.get(row);
            if (value.isJsonNull()) {
                return null;
            } else {
                return value.asNumber();
            }
        }

        @Override
        public double getDouble(int row) {
            JsonValue value = array.get(row);
            if (value.isJsonNull()) {
                return Double.NaN;
            } else {
                return value.asNumber();
            }
        }

        @Override
        public String getString(int row) {
            return null;
        }

        @Override
        public boolean isBlank(int row) {
            return array.get(row).isJsonNull();
        }

        @Override
        public ColumnView select(int[] rows) {
            JsonValue selectedArray = Json.createArray();
            for (int i = 0; i < rows.length; i++) {
                selectedArray.add(array.get(rows[i]));
            }
            return new JsonNumberView(selectedArray);
        }

        @Override
        public int[] order(int[] sortVector, SortDir direction, int[] range) {
            throw new UnsupportedOperationException();
        }
    }

    private static class JsonBooleanView extends ColumnView {

        private JsonValue array;

        private JsonBooleanView(JsonValue array) {
            this.array = array;
        }

        @Override
        public ColumnType getType() {
            return ColumnType.BOOLEAN;
        }

        @Override
        public int numRows() {
            return array.length();
        }

        @Override
        public Object get(int row) {
            JsonValue value = array.get(row);
            if(value.isJsonNull()) {
                return null;
            } else {
                return value.asBoolean();
            }
        }

        @Override
        public double getDouble(int row) {
            JsonValue value = array.get(row);
            if(value.isJsonNull()) {
                return Double.NaN;
            } else {
                return value.asNumber();
            }
        }

        @Override
        public String getString(int row) {
            return null;
        }

        @Override
        public boolean isBlank(int row) {
            return array.get(row).isJsonNull();
        }

        @Override
        public ColumnView select(int[] rows) {
            JsonValue selectedArray = Json.createArray();
            for (int i = 0; i < rows.length; i++) {
                selectedArray.add(array.get(rows[i]));
            }
            return new JsonBooleanView(selectedArray);
        }

        @Override
        public int[] order(int[] sortVector, SortDir direction, int[] range) {
            throw new UnsupportedOperationException();
        }
    }
}
