/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.query;

import org.activityinfo.model.type.time.LocalDate;

import java.io.Serializable;
import java.util.Objects;

public abstract class ColumnView implements Serializable {

    public static final char UNIT_SEPERATOR = 0x1F;

    public static final String UNIT_SEPERATOR_STRING = "" + UNIT_SEPERATOR;


    /**
     * The maximum length of a string in characters. If the string is truncated,
     * then it will be terminated with a
     */
    public static final int MAX_STRING_LENGTH = 128;

    public static final int TRUE = 1;
    public static final int FALSE = 0;
    public static final int BLANK = Integer.MAX_VALUE;

    public static boolean isTruncated(String string) {
        return string != null && string.endsWith(UNIT_SEPERATOR_STRING);
    }

    public abstract ColumnType getType();

    public abstract int numRows();

    public abstract Object get(int row);

    public abstract double getDouble(int row);

    public abstract String getString(int row);

    /**
     *
     * @param row
     * @return ColumnView#TRUE, ColumnView#FALSE, or ColumnView#NA if the value is not null or missing
     */
    public int getBoolean(int row) {
        double x = getDouble(row);
        if(Double.isNaN(x)) {
            return BLANK;
        } else if(x == 0) {
            return FALSE;
        } else {
            return TRUE;
        }
    }

    public abstract boolean isBlank(int row);

    public LocalDate getLocalDate(int i) {
        String s = getString(i);
        if(s == null) {
            return null;
        }
        try {
            return LocalDate.parse(s);
        } catch (NumberFormatException e) {
            return null;
        }
    }

    public final boolean isTrue(int i) {
        return getBoolean(i) == TRUE;
    }


    /**
     * Creates a new ColumnView with the rows listed in {@code rows}.
     *
     * @param rows an array of rows indexes into this ColumnView
     * @return a new ColumnView containing the selected rows.
     */
    public abstract ColumnView select(int[] rows);

    /**
     * Order the column rows in the given {@code direction},
     * within the given row {@code range}
     *
     * @param sortVector int array of sorted row indices
     * @param direction of ordering
     * @param range of rows to order
     * @return int array of the reordered row index
     */
    public abstract int[] order(int[] sortVector, SortDir direction, int[] range);


    @Override
    public final int hashCode() {
        return Objects.hash(getType(), numRows());
    }

    @Override
    public final boolean equals(Object obj) {
        if(!(obj instanceof ColumnView)) {
            return false;
        }
        ColumnView that = (ColumnView) obj;
        if(this.getType() != this.getType() || this.numRows() != that.numRows()) {
            return false;
        }
        switch (this.getType()) {
            case STRING:
                return stringsEqual(that);
            case NUMBER:
                return numbersEqual(that);
            default:
            case BOOLEAN:
                return booleanEquals(that);
        }
    }

    public boolean numbersEqual(ColumnView that) {
        for (int i = 0; i < numRows(); i++) {
            double thisi = this.getDouble(i);
            if(Double.isNaN(thisi)) {
                if(!Double.isNaN(that.getDouble(i))) {
                    return false;
                }
            } else {
                if(thisi != that.getDouble(i)) {
                    return false;
                }
            }
        }
        return true;
    }

    public boolean stringsEqual(ColumnView that) {
        for (int i = 0; i < numRows(); i++) {
            if(!Objects.equals(this.getString(i), that.getString(i))) {
                return false;
            }
        }
        return true;
    }

    public boolean booleanEquals(ColumnView that) {
        for (int i = 0; i < numRows(); i++) {
            if(this.getBoolean(i) != that.getBoolean(i)) {
                return false;
            }
        }
        return true;
    }

    /**
     * Truncates a string for inclusion in a string-based ColumnView.
     *
     * <p>Strings in the column view are meant to provide an efficient data source
     * for analysis and for display in a data table. In order to meet these goals, we
     * cannot include very large strings in the ColumnView, or else it will no longer be fast.</p>
     *
     * <p>The full string is always stored in the FormRecord, so when exporting or display record
     * details, we can fetch the full record from that data source.</p>
     * @param string
     * @return
     */
    public static String truncate(String string) {
        if(string == null) {
            return null;
        }
        int newLineIndex = string.indexOf('\n');
        if(newLineIndex == -1) {
            if (string.length() > MAX_STRING_LENGTH) {
                return string.substring(0, MAX_STRING_LENGTH - 1).trim() + UNIT_SEPERATOR;
            } else {
                return string;
            }
        } else {
            return string.substring(0, Math.min(newLineIndex, MAX_STRING_LENGTH - 1)).trim() + UNIT_SEPERATOR;
        }
    }

}
