/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.query;

import com.google.common.base.Strings;

public class ConstantColumnView extends ColumnView {

    private ColumnType type;
    private double doubleValue;
    private String stringValue;
    private int booleanValue;
    private int numRows;
    private boolean blank;

    protected ConstantColumnView() {
    }

    public ConstantColumnView(int numRows, double doubleValue) {
        this.type = ColumnType.NUMBER;
        this.doubleValue = doubleValue;
        this.stringValue = null;
        this.booleanValue = (doubleValue != 0) ? TRUE : FALSE;
        this.blank = Double.isNaN(doubleValue);
        this.numRows = numRows;
    }

    public ConstantColumnView(int numRows, String value) {
        this.type = ColumnType.STRING;
        this.doubleValue = Double.NaN;
        this.stringValue = value;
        this.booleanValue = BLANK;
        this.blank = Strings.isNullOrEmpty(value);
        this.numRows = numRows;
    }

    public ConstantColumnView(int numRows, boolean value) {
        this.type = ColumnType.BOOLEAN;
        this.doubleValue = (value ? 1 : 0);
        this.stringValue = null;
        this.booleanValue = value ? TRUE : FALSE;
        this.numRows = numRows;
        this.blank = false;
    }

    public ConstantColumnView(int numRows, ColumnView value) {
        this.type = value.getType();
        this.doubleValue = value.getDouble(0);
        this.stringValue = value.getString(0);
        this.booleanValue = value.getBoolean(0);
        this.numRows = numRows;
        this.blank = value.isBlank(0);
    }

    public static ConstantColumnView nullBoolean(int numRows) {
        ConstantColumnView view = new ConstantColumnView();
        view.type = ColumnType.BOOLEAN;
        view.doubleValue = Double.NaN;
        view.stringValue = null;
        view.booleanValue = ColumnView.BLANK;
        view.numRows = numRows;
        view.blank = true;
        return view;
    }

    public static ColumnView create(int numRows, ColumnView scalar) {
        if(scalar.isBlank(0)) {
            return new BlankColumnView(scalar.getType(), numRows);
        } else {
            switch (scalar.getType()) {
                default:
                case BOOLEAN:
                    return new ConstantColumnView(numRows, scalar.getBoolean(0));
                case STRING:
                    return new ConstantColumnView(numRows, scalar.getString(0));
                case NUMBER:
                    return new ConstantColumnView(numRows, scalar.getDouble(0));
            }
        }
    }



    @Override
    public ColumnType getType() {
        return type;
    }

    @Override
    public int numRows() {
        return numRows;
    }

    @Override
    public Object get(int row) {
        switch (type) {
            case BOOLEAN:
                return getBoolean(row);
            case NUMBER:
                return getDouble(row);
            case STRING:
                return getString(row);
        }
        throw new IllegalStateException("type: " + type);
    }

    @Override
    public double getDouble(int row) {
        return doubleValue;
    }

    @Override
    public String getString(int row) {
        return stringValue;
    }

    @Override
    public int getBoolean(int row) {
        return booleanValue;
    }

    @Override
    public boolean isBlank(int row) {
        return blank;
    }

    @Override
    public ColumnView select(int[] selectedRows) {
        if(blank) {
            return new BlankColumnView(this.type, selectedRows.length);
        } else {
            if(noneMissing(selectedRows)) {
                return copyOf(selectedRows.length);
            }
        }
        return new FilteredColumnView(this, selectedRows);
    }

    private boolean noneMissing(int[] selectedRows) {
        for (int i = 0; i < selectedRows.length; i++) {
            if(selectedRows[i] == -1) {
                return false;
            }
        }
        return true;
    }

    private ConstantColumnView copyOf(int length) {
        ConstantColumnView copy = new ConstantColumnView();
        copy.numRows = length;
        copy.type = type;
        copy.doubleValue = this.doubleValue;
        copy.stringValue = this.stringValue;
        copy.booleanValue = this.booleanValue;
        copy.blank = this.blank;
        return copy;
    }

    @Override
    public String toString() {
        return "[ " + get(0) + " x " + numRows + " rows ]";
    }

    @Override
    public int[] order(int[] sortVector, SortDir direction, int[] range) {
        return sortVector;
    }
}
