/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.query;

import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.annotation.Builder;
import org.activityinfo.model.formTree.FieldPath;
import org.activityinfo.model.formula.FormulaNode;
import org.activityinfo.model.formula.FormulaParser;
import org.activityinfo.model.formula.SymbolNode;
import org.activityinfo.model.resource.ResourceId;

import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import static org.activityinfo.json.Json.createObject;

/**
 * Describes a Table to be constructed from a
 * FormTree.
 */
public class QueryModel implements JsonSerializable {

    private final List<RowSource> rowSources = Lists.newArrayList();
    private final List<ColumnModel> columns = Lists.newArrayList();

    private List<SortModel> sortModels = Lists.newArrayList();

    /**
     * A boolean-valued formula to filter the records included in this query.
     */
    private FormulaNode filter;

    /**
     * Zero or more set-based filters to apply. Resulting rows must match both the {@link #filter} formula,
     * if provided, and all of the provided filterSets.
     */
    private List<FilterSet> filterSets = Lists.newArrayList();

    private boolean truncateStrings = true;

    public QueryModel() {
    }

    /**
     * Creates a new TableModel using the given {@code formId} as the
     * root FormClassId
     */
    public QueryModel(ResourceId formId) {
        rowSources.add(new RowSource(formId));
    }

    @Builder
    public QueryModel(ResourceId formId, List<ColumnModel> columns, Optional<String> filter, List<SortModel> sort) {
        this.rowSources.add(new RowSource(formId));
        this.columns.addAll(columns);
        filter.ifPresent(string -> setFilter(string));
    }

    public boolean isTruncateStrings() {
        return truncateStrings;
    }

    public void setTruncateStrings(boolean truncateStrings) {
        this.truncateStrings = truncateStrings;
    }

    public List<SortModel> getSortModels() {
        return sortModels;
    }

    public void setSortModels(List<SortModel> sortModels) {
        this.sortModels = sortModels;
    }

    public void addSortModel(SortModel sortModel) {
        this.sortModels.add(sortModel);
    }

    public List<RowSource> getRowSources() {
        return rowSources;
    }

    public List<ColumnModel> getColumns() {
        return columns;
    }
    
    public void addRowSource(ResourceId formId) {
        rowSources.add(new RowSource(formId));
    }

    /**
     * Adds a {@code ColumnModel} to this {@code TableModel} using the given
     * field's code or label as the column's id and value expression.
     */
    public ColumnModel selectField(String codeOrLabel) {
        ColumnModel column = new ColumnModel();
        column.setId(codeOrLabel);
        column.setFormula("[" + codeOrLabel + "]");
        columns.add(column);
        return column;
    }

    public ColumnModel selectExpr(String expression) {
        ColumnModel column = new ColumnModel();
        column.setId("_expr" + (columns.size()+1));
        column.setFormula(expression);
        columns.add(column);
        return column;
    }

    public ColumnModel selectExpr(FormulaNode expression) {
        ColumnModel column = new ColumnModel();
        column.setId("_expr" + (columns.size() + 1));
        column.setFormula(expression);
        columns.add(column);
        return column;
    }

    public ColumnModel selectField(FieldPath path) {
        ColumnModel column = new ColumnModel();
        column.setId(path.getLeafId().asString());
        column.setFormula(path);
        columns.add(column);
        return column;
    }

    /**
     * Adds a {@code ColumnModel} to this {@code TableModel} using the given
     * field's id as the column's id and value expression.
     */
    public ColumnModel selectField(ResourceId fieldId) {
        ColumnModel column = new ColumnModel();
        column.setId(fieldId.asString());
        column.setFormula(fieldId.asString());
        columns.add(column);
        return column;
    }

    public FormulaNode getFilter() {
        return filter;
    }

    public List<FilterSet> getFilterSets() {
        return filterSets;
    }

    public void setFilter(String filterExpression) {
        if(filterExpression == null) {
            this.filter = null;
        } else {
            this.filter = FormulaParser.parse(filterExpression);
        }
    }

    public void setFilter(FormulaNode filter) {
        this.filter = filter;
    }

    public void addFilterSet(FormulaNode formula, Set<String> set) {
        this.filterSets.add(new FilterSet(formula.deparse(), set));
    }

    public void setFilter(java.util.Optional<FormulaNode> filter) {
        this.filter = filter.orElse(null);
    }

    public QueryModel addColumn(ColumnModel criteriaColumn) {
        columns.add(criteriaColumn);
        return this;
    }

    public void addColumns(List<ColumnModel> requiredColumns) {
        columns.addAll(requiredColumns);
    }

    /**
     * Adds the {@code ResourceId} as a string column to the table model with
     * the given column id
     */
    public ColumnModel selectRecordId() {
        ColumnModel columnModel = new ColumnModel();
        columnModel.setFormula(new SymbolNode(ColumnModel.RECORD_ID_SYMBOL));
        columns.add(columnModel);
        return columnModel;
    }

    public ColumnModel selectLastEditTime() {
        ColumnModel columnModel = new ColumnModel();
        columnModel.setFormula(new SymbolNode(ColumnModel.LAST_EDIT_TIME_SYMBOL));
        columns.add(columnModel);
        return columnModel;
    }

    public ColumnModel selectClassId() {
        ColumnModel columnModel = new ColumnModel();
        columnModel.setFormula(new SymbolNode(ColumnModel.FORM_ID_SYMBOL));
        columns.add(columnModel);
        return columnModel;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("SELECT ");
        boolean needsComma = false;
        for(ColumnModel column : columns) {
            if(needsComma) {
                sb.append(", ");
            }
            sb.append("(").append(column.getFormula()).append(") as ").append(column.getId());
            needsComma = true;
        }
        return sb.toString();
    }
    
    public String toJsonString() {
        return toJson().toJson();
    }

    public static QueryModel fromJson(JsonValue jsonObject) {
        QueryModel queryModel = new QueryModel();
        JsonValue rowSources = jsonObject.get("rowSources");
        for (JsonValue rowSource : rowSources.values()) {
            queryModel.getRowSources().add(RowSource.fromJson(rowSource));
        }

        if(jsonObject.hasKey("truncateStrings")) {
            queryModel.truncateStrings = jsonObject.getBoolean("truncateStrings");
        }

        for (JsonValue column : jsonObject.get("columns").values()) {
            queryModel.getColumns().add(ColumnModel.fromJson(column));
        }
        String filter = jsonObject.getString("filter");
        if (!Strings.isNullOrEmpty(filter)) {
            queryModel.setFilter(filter);
        }

        for (JsonValue filterSet : jsonObject.get("filterSets").values()) {
            queryModel.filterSets.add(FilterSet.fromJson(filterSet));
        }

        for (JsonValue sort : jsonObject.get("sort").values()) {
            queryModel.addSortModel(SortModel.fromJson(sort));
        }

        return queryModel;
    }

    @Override
    public JsonValue toJson() {
        JsonValue object = createObject();
        object.put("rowSources", Json.toJson(rowSources));
        object.put("columns", Json.toJson(columns));
        object.put("truncateStrings", truncateStrings);

        if(filter != null) {
            object.put("filter", filter.deparse());
        }

        if(!filterSets.isEmpty()) {
            object.put("filterSets", Json.toJsonArray(filterSets));
        }

        if(!sortModels.isEmpty()) {
            object.put("sort", Json.toJson(sortModels));
        }

        return object;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        QueryModel that = (QueryModel) o;
        return truncateStrings == that.truncateStrings &&
            rowSources.equals(that.rowSources) &&
            columns.equals(that.columns) &&
            sortModels.equals(that.sortModels) &&
            Objects.equals(filter, that.filter) &&
            Objects.equals(this.filterSets, that.filterSets);
    }

    @Override
    public int hashCode() {
        return Objects.hash(rowSources, columns, sortModels, filter, filterSets, truncateStrings);
    }

}

