/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.query;

import com.google.gwt.core.shared.GwtIncompatible;
import org.activityinfo.model.util.HeapsortColumn;

import java.io.*;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * Simple Array of String values
 */
public class StringArrayColumnView extends ColumnView {

    private static final long serialVersionUID = 1L;

    private static final int FORMAT_UNCOMPRESSED = 0;
    private static final int FORMAT_COMPRESSED = 1;

    private String[] values;

    protected StringArrayColumnView() {
    }

    public StringArrayColumnView(String... values) {
        this.values = values;
    }

    public StringArrayColumnView(List<String> values) {
        this.values = values.toArray(new String[values.size()]);
    }

    @Override
    public ColumnType getType() {
        return ColumnType.STRING;
    }

    @Override
    public int numRows() {
        return values.length;
    }

    @Override
    public Object get(int row) {
        return values[row];
    }

    @Override
    public double getDouble(int row) {
        return Double.NaN;
    }

    @Override
    public String getString(int row) {
        return values[row];
    }

    @Override
    public int getBoolean(int row) {
        return BLANK;
    }

    @Override
    public boolean isBlank(int row) {
        return values[row] == null;
    }

    @Override
    public ColumnView select(int[] selectedRows) {
        String[] filteredValues = new String[selectedRows.length];
        for (int i = 0; i < filteredValues.length; i++) {
            int selectedRow = selectedRows[i];
            if(selectedRow != -1) {
                filteredValues[i] = this.values[selectedRow];
            }
        }
        return new StringArrayColumnView(filteredValues);
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder("[");
        for (int i = 0; i < Math.min(values.length, 5); i++) {
            if(i > 0) {
                sb.append(", ");
            }
            if(values[i] == null) {
                sb.append("NULL");
            } else {
                sb.append("'");
                sb.append(values[i]);
                sb.append("'");
            }
        }
        if(values.length > 5) {
            sb.append("... length=").append(values.length);
        }
        sb.append("]");
        return sb.toString();
    }

    @Override
    public int[] order(int[] sortVector, SortDir direction, int[] range) {
        int numRows = values.length;
        if (range == null || range.length == numRows) {
            HeapsortColumn.heapsortString(values, sortVector, numRows, direction == SortDir.ASC);
        } else {
            HeapsortColumn.heapsortString(values, sortVector, range.length, range, direction == SortDir.ASC);
        }
        return sortVector;
    }

    @GwtIncompatible
    private void writeObject(ObjectOutputStream out) throws IOException {
        writeCompressed(out);
    }

    @GwtIncompatible
    private void writeCompressed(ObjectOutputStream out) throws IOException {
        DataOutputStream daos = new DataOutputStream(out);
        daos.writeInt(FORMAT_COMPRESSED);

        GZIPOutputStream gzout = new GZIPOutputStream(out);
        DataOutputStream gzDataOut = new DataOutputStream(gzout);
        writeArray(gzDataOut);
        gzout.finish();
    }

    @GwtIncompatible
    private void writeArray(DataOutputStream daos) throws IOException {
        daos.writeInt(values.length);
        for (int i = 0; i < values.length; i++) {
            String value = values[i];
            if(value == null) {
                daos.writeUTF("");
            } else {
                daos.writeUTF(value);
            }
        }
    }

    @GwtIncompatible
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        DataInputStream dis = new DataInputStream(in);
        int format = dis.readInt();
        if(format == FORMAT_COMPRESSED) {
            readCompressed(dis);
        }
    }

    @GwtIncompatible
    private void readCompressed(DataInputStream in) throws IOException {

        GZIPInputStream gzin = new GZIPInputStream(in);
        DataInputStream gzDataIn = new DataInputStream(gzin);

        readArray(gzDataIn);
    }

    @GwtIncompatible
    private void readArray(DataInputStream input) throws IOException {
        int count = input.readInt();
        this.values = new String[count];

        for (int i = 0; i < count; i++) {
            String s = input.readUTF();
            if(!s.isEmpty()) {
                values[i] = s;
            }
        }
    }
}
