/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.resource;

import org.activityinfo.json.Json;
import org.activityinfo.json.JsonSerializable;

import javax.annotation.Nonnull;
import java.io.Serializable;
import java.util.Date;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Globally, universally unique and persistent identifier
 * for {@code Resources}
 *
 */
public final class ResourceId implements Serializable, JsonSerializable {

    public static final ResourceId ROOT_ID = ResourceId.valueOf("_root");

    public static final char GENERATED_ID_DOMAIN = 'c';
    public static final int RADIX = 10;
    public static AtomicLong COUNTER = new AtomicLong(1);

    private final String text;

    /**
     * Creates a new ResourceId from its string representation
     *
     * <p>Note: This method must be named {@code valueOf} in order to be
     * used as a Jersey {@code @PathParam}
     */
    public static ResourceId valueOf(@Nonnull String string) {
        assert string != null : "resourceid cannot be null";
        return new ResourceId(string);
    }

    /**
     * Generates a globally unique Id for a new collection
     */
    public static ResourceId generateId() {
        return valueOf(GENERATED_ID_DOMAIN + generateCuid());
    }

    public static String generateCuid() {
        return
            Integer.toString(Math.abs(ThreadLocalRandom.current().nextInt()), Character.MAX_RADIX) +
            Long.toString(new Date().getTime(), Character.MAX_RADIX) +
                Long.toString(COUNTER.incrementAndGet(), Character.MAX_RADIX);
    }

    private ResourceId(@Nonnull String text) {
        this.text = text;
    }

    public static ResourceId valueOf(long id) {
        return new ResourceId(Long.toString(id));
    }

    public String asString() {
        return this.text;
    }

    public char getDomain() {
        return text.charAt(0);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        ResourceId resourceId = (ResourceId) o;
        return text.equals(resourceId.text);
    }

    @Override
    public int hashCode() {
        return text.hashCode();
    }

    @Override
    public String toString() {
        return text;
    }

    public static ResourceId generateFieldId() {
        return ResourceId.generateId();
    }

    public static ResourceId fromJson(org.activityinfo.json.JsonValue value) {
        return ResourceId.valueOf(value.asString());
    }

    @Override
    public org.activityinfo.json.JsonValue toJson() {
        return Json.create(text);
    }

    public static void main(String[] args) {
        System.out.println(generateId());
    }

    public long toLong() {
        return Long.parseLong(toString());
    }
}
