/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.sync;

import org.activityinfo.json.Json;
import org.activityinfo.model.resource.ResourceId;

/**
 * A compact list of form's records, sent from the server in order to
 * populate the offline copy when syncing for the first time.
 *
 * <p>The format is line delimited as follows:</p>
 *
 * <pre>
 * InitialSyncSetv1
 * [Form Version]
 * [Record ID 1]
 * [Json-encoded RecordObject 1]
 * ...
 * ...
 * [Record ID n]
 * [Json-encoded RecordObject n]
 * </pre>
 */
public class InitialSyncSet {

    public static final String MIME_TYPE = "application/vnd.activityinfo.iss+text";

    public static final String V1_HEADER_LINE = "InitialSyncSetv1";

    private final String version;
    private final String header;

    public interface RecordConsumer {
        void accept(String recordId, RecordObject object);
    }

    private final ResourceId formId;
    private final String payload;
    private final int firstRecordIndex;

    public InitialSyncSet(ResourceId formId, String payload) {
        this.formId = formId;
        this.payload = payload;
        int headerEnd = payload.indexOf('\n');
        header = payload.substring(0, headerEnd);
        if (!header.equals(V1_HEADER_LINE)) {
            throw new IllegalArgumentException("Unexpected header: " + header);
        }
        int versionEnd = payload.indexOf('\n', headerEnd + 1);
        version = payload.substring(headerEnd + 1, versionEnd);
        firstRecordIndex = versionEnd + 1;
    }

    public String getVersion() {
        return version;
    }

    public ResourceId getFormId() {
        return formId;
    }

    public String getHeader() {
        return header;
    }

    public void forEach(RecordConsumer consumer) {
        int pos = firstRecordIndex;
        while (true) {
            int idEnd = payload.indexOf('\n', pos + 1);
            if (idEnd == -1) {
                break;
            }
            String id = payload.substring(pos, idEnd);
            int recordEnd = payload.indexOf('\n', idEnd + 1);
            if (recordEnd == -1) {
                throw new IllegalStateException("Expected additional line");
            }
            String record = payload.substring(idEnd + 1, recordEnd);
            consumer.accept(id, Json.fromJson(RecordObject.class, Json.parse(record)));
            pos = recordEnd + 1;
        }
    }
}
