/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.sync;

import org.activityinfo.json.JsonArrays;
import org.activityinfo.json.JsonSerializable;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.database.UpdateFormResponse;
import org.activityinfo.model.form.FormMetadata;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;
import java.util.stream.Collectors;

import static org.activityinfo.json.Json.createObject;

/**
 * Describes a snapshot of offline data. This stores the version
 * numbers we have downloaded.
 *
 * The OfflineSnapshot provides serialized consistency within a form,
 * but relaxed consistency between forms.
 *
 * For a specific form that is part of the snapshot, we provide serializable
 * consistency. The OfflineSnapshot is guaranteed to have a specific version
 * of the form.
 *
 * For all the forms that are part of the snapshot, we strive to have a set of
 * form versions that correspond to the server's state at a specific point in time.
 *
 */
public class OfflineSnapshot implements JsonSerializable {

    public static final OfflineSnapshot EMPTY = new OfflineSnapshot();

    private Date time = new Date();
    private final Map<ResourceId, String> databaseVersions;
    private final Map<ResourceId, String> formVersions;

    /**
     * Set of forms that are we are synchronizing for the first time, but are
     * not yet consistent with this snapshot. Normally we apply the delta between
     * two snapshots in one IndexedDB transaction to ensure that readers do not
     * get an inconsistent view of the data, and that we do not leave the local database
     * in an inconsistent state if we successfully update one form in the set, but the network
     * fails before we can apply the updates for a second form in the OfflineSet. This would
     * leave the database showing recent updates for one form, but potentially days old view
     * of another, potentially related form.
     *
     * <p>However, when we are syncing a form for the first time, we can afford to apply the
     * changes directly to the database as long as we do not allow readers to query it. If the
     * synchronization fails, it's ok because we keep the form in this set, preventing readers
     * from accessing it. Only once we successfully reach the version in the snapshot do we remove
     * the form from this staging set.</p>
     */
    private final Set<ResourceId> stagedForms;

    public OfflineSnapshot() {
        databaseVersions = new HashMap<>();
        formVersions = new HashMap<>();
        stagedForms = new HashSet<>();
    }

    public OfflineSnapshot(OfflineSnapshot previous, SyncDelta delta) {
        databaseVersions = new HashMap<>();
        databaseVersions.putAll(previous.databaseVersions);
        databaseVersions.keySet().removeAll(delta.getRemovedDatabases());
        databaseVersions.putAll(delta.getUpdatedDatabases()
                .stream()
                .collect(Collectors.toMap(UpdatedDatabase::getDatabaseId, UpdatedDatabase::getVersion)));

        formVersions = new HashMap<>();
        formVersions.putAll(previous.formVersions);
        formVersions.keySet().removeAll(delta.getRemovedForms());
        formVersions.putAll(delta.getUpdatedForms()
                .stream()
                .collect(Collectors.toMap(UpdatedForm::getFormId, UpdatedForm::getUpdatedVersion)));

        stagedForms = Collections.emptySet();
    }

    private OfflineSnapshot(Date time,
                            Map<ResourceId, String> databaseVersions,
                            Map<ResourceId, String> formVersions,
                            Set<ResourceId> stagedForms) {
        this.time = time;
        this.databaseVersions = databaseVersions;
        this.formVersions = formVersions;
        this.stagedForms = stagedForms;
    }

    public boolean isEmpty() {
        return formVersions.isEmpty();
    }

    public Date getSnapshotTime() {
        return time;
    }

    /**
     * Returns true if this form is downloaded and consistent with the OfflineSnapshot.
     */
    public boolean isFormCached(ResourceId formId) {
        return formVersions.containsKey(formId) && !stagedForms.contains(formId);
    }

    public boolean hasLocalVersion(ResourceId formId) {
        return formVersions.containsKey(formId);
    }

    public boolean isDatabaseCached(ResourceId databaseId) {
        return databaseVersions.containsKey(databaseId);
    }

    public String getLocalFormVersion(ResourceId formId) {
        return formVersions.getOrDefault(formId, "");
    }

    public String getLocalDatabaseVersion(ResourceId databaseId) {
        return databaseVersions.getOrDefault(databaseId, "");
    }

    public Set<ResourceId> getDatabases() {
        return databaseVersions.keySet();
    }

    public Set<ResourceId> getForms() {
        return formVersions.keySet();
    }

    /**
     * Update this snapshot, ensuring that it includes this (possibly) new form and all of its
     * (possibly) new subforms.
     *
     * <p>Note that the version stored in the {@code OfflineSnapshot} is the form's overall version,
     * NOT the schema version. For this reason, we don't update the version if the schema version is greater than
     * one, as we DO NOT have the corresponding records.</p>
     *
     */
    public OfflineSnapshot withUpdatedFormSchemas(UpdateFormResponse result) {
        Map<ResourceId, String> updatedFormVersions = new HashMap<>(formVersions);
        for (FormMetadata form : result.getForms()) {
            updatedFormVersions.putIfAbsent(form.getId(), "1");
        }
        return new OfflineSnapshot(time, databaseVersions, updatedFormVersions, stagedForms);
    }

    public OfflineSnapshot withInitialSync(InitialSyncSet initialSync) {

        // Be very sure that we are not applying an InitialSyncSet to a database
        // that already has been synced for this form.
        if(formVersions.containsKey(initialSync.getFormId())) {
            throw new IllegalStateException("InitialSyncSet for already synced form " + initialSync.getFormId());
        }

        Map<ResourceId, String> updatedFormVersions = new HashMap<>(formVersions);
        updatedFormVersions.putIfAbsent(initialSync.getFormId(), initialSync.getVersion());

        Set<ResourceId> updatedStagingSet = new HashSet<>(stagedForms);
        updatedStagingSet.add(initialSync.getFormId());

        return new OfflineSnapshot(time, databaseVersions, updatedFormVersions, updatedStagingSet);
    }

    public JsonValue toJson() {

        JsonValue databases = createObject();
        for (Map.Entry<ResourceId, String> entry : databaseVersions.entrySet()) {
            databases.put(entry.getKey().asString(), entry.getValue());
        }

        JsonValue forms = createObject();
        for (Map.Entry<ResourceId, String> entry : formVersions.entrySet()) {
            forms.put(entry.getKey().asString(), entry.getValue());
        }

        JsonValue object = createObject();
        object.put("time", (double)time.getTime());
        object.put("databases", databases);
        object.put("forms", forms);
        object.put("stagingForms", JsonArrays.toJsonArray(stagedForms));

        return object;
    }

    public static OfflineSnapshot fromJson(JsonValue object) {
        Date time = new Date((long) object.getNumber("time"));

        Map<ResourceId, String> databaseVersions = new HashMap<>();
        JsonValue databases = object.get("databases");
        String[] databaseKeys = databases.keys();

        for (String databaseKey : databaseKeys) {
            ResourceId databaseId = ResourceId.valueOf(databaseKey);
            String version = databases.getString(databaseKey);

            databaseVersions.put(databaseId, version);
        }

        Map<ResourceId, String> formVersions = new HashMap<>();
        JsonValue forms = object.get("forms");
        String[] formKeys = forms.keys();

        for (String form : formKeys) {
            ResourceId formId = ResourceId.valueOf(form);
            String version = forms.getString(form);

            formVersions.put(formId, version);
        }

        Set<ResourceId> stagingForms = JsonArrays.setFromJson(object.get("stagingForms"),
                ResourceId::fromJson);

        return new OfflineSnapshot(time, databaseVersions, formVersions, stagingForms);
    }

    @Override
    public String toString() {
        return "SnapshotStatus{" +
                "time=" + time +
                ", formVersions=" + formVersions +
                '}';
    }

    public boolean areAllCached(Set<ResourceId> formIds) {
        for (ResourceId formId : formIds) {
            if(!isFormCached(formId)) {
                return false;
            }
        }
        return true;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof OfflineSnapshot)) {
            return false;
        }
        OfflineSnapshot that = (OfflineSnapshot) o;
        return time.equals(that.time) &&
                databaseVersions.equals(that.databaseVersions) &&
                formVersions.equals(that.formVersions);
    }

    @Override
    public int hashCode() {
        return Objects.hash(time, databaseVersions, formVersions);
    }

}
