/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.sync;

import jsinterop.annotations.JsOverlay;
import jsinterop.annotations.JsPackage;
import jsinterop.annotations.JsType;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;

/**
 * Set of updates to apply to a local copy of the database
 */
@JsType(isNative = true, name = "Object", namespace = JsPackage.GLOBAL)
public final class RecordSyncSet {

    private String formId;
    private boolean reset;
    private String[] deleted;
    private UpdatedRecord[] updatedRecords;
    private String cursor;

    public RecordSyncSet() {
    }

    @JsOverlay
    public static RecordSyncSet incremental(String formId, Collection<String> deleted, Collection<UpdatedRecord> records, Optional<String> cursor) {
        RecordSyncSet syncSet = new RecordSyncSet();
        syncSet.formId = formId;
        syncSet.reset = false;
        syncSet.deleted = deleted.toArray(new String[0]);
        syncSet.updatedRecords = records.toArray(new UpdatedRecord[0]);
        syncSet.cursor = cursor.orElse(null);
        return syncSet;
    }

    @JsOverlay
    public static RecordSyncSet clear(ResourceId formId) {
        return initial(formId, Collections.emptyList(), Optional.empty());
    }

    @JsOverlay
    public static RecordSyncSet initial(ResourceId formId, List<UpdatedRecord> records, Optional<String> cursor) {
        RecordSyncSet syncSet = new RecordSyncSet();
        syncSet.formId = formId.asString();
        syncSet.reset = true;
        syncSet.deleted = new String[0];
        syncSet.updatedRecords = new UpdatedRecord[records.size()];
        syncSet.cursor = cursor.orElse(null);
        for (int i = 0; i < syncSet.updatedRecords.length; i++) {
            syncSet.updatedRecords[i] = records.get(i);
        }
        return syncSet;
    }

    @JsOverlay
    public static RecordSyncSet emptySet(ResourceId formId) {
        RecordSyncSet syncSet = new RecordSyncSet();
        syncSet.formId = formId.asString();
        syncSet.deleted = new String[0];
        syncSet.updatedRecords = new UpdatedRecord[0];
        return syncSet;
    }

    @JsOverlay
    public String getFormId() {
        return formId;
    }

    @JsOverlay
    public List<String> getDeleted() {
        return Arrays.asList(deleted);
    }

    @JsOverlay
    public List<UpdatedRecord> getUpdatedRecords() {
        return Arrays.asList(updatedRecords);
    }

    @JsOverlay
    public int getUpdatedRecordCount() {
        return updatedRecords.length;
    }

    @JsOverlay
    public boolean isEmpty() {
        return updatedRecords.length == 0 && deleted.length == 0;
    }

    @JsOverlay
    public UpdatedRecord getUpdatedRecord(int i) {
        return updatedRecords[i];
    }

    /**
     * @return true if there are no remaining chunks to retrieve.
     */
    @JsOverlay
    public boolean isComplete() {
        return cursor == null;
    }

    @JsOverlay
    public String getCursor() {
        return cursor;
    }

    @JsOverlay
    public boolean isReset() {
        return reset;
    }

    @JsOverlay
    public static RecordSyncSet foldLeft(RecordSyncSet a, RecordSyncSet b) {
        assert a.formId.equals(b.formId);

        RecordSyncSet merged = new RecordSyncSet();
        merged.formId = a.formId;
        merged.reset = a.reset;
        merged.deleted = combine(a.deleted, b.deleted);
        merged.updatedRecords = combine(a.updatedRecords, b.updatedRecords);
        merged.cursor = b.cursor;
        return merged;
    }

    @JsOverlay
    private static String[] combine(String[] a, String[] b) {
        if(a.length == 0) {
            return b;
        }
        if(b.length == 0) {
            return a;
        }
        String[] c = new String[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);
        return c;
    }

    @JsOverlay
    private static UpdatedRecord[] combine(UpdatedRecord[] a, UpdatedRecord[] b) {
        if(a.length == 0) {
            return b;
        }
        if(b.length == 0) {
            return a;
        }
        UpdatedRecord[] c = new UpdatedRecord[a.length + b.length];
        System.arraycopy(a, 0, c, 0, a.length);
        System.arraycopy(b, 0, c, a.length, b.length);
        return c;
    }
}
