/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.sync;

import com.google.common.hash.Hashing;
import org.activityinfo.model.permission.FormPermissions;
import org.activityinfo.model.permission.Permission;
import org.activityinfo.promise.Maybe;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.nio.charset.StandardCharsets;
import java.util.Objects;

public class VisibleFormVersion {

    public static final String UNFILTERED_HASH = "Z";

    /**
     * Bump or change this version string in order to force a resync for all forms.
     */
    public static final String FORMAT_SUFFIX = "2";

    private final long number;
    private final String filterHash;
    private final boolean empty;

    public VisibleFormVersion(long number, String filterHash, boolean empty) {
        this.number = number;
        this.filterHash = filterHash;
        this.empty = empty;
    }

    public VisibleFormVersion(long number, String filterHash) {
        this.number = number;
        this.filterHash = filterHash;
        this.empty = number == 0L;
    }

    public static VisibleFormVersion notFound() {
        return new VisibleFormVersion(0L, Maybe.State.NOT_FOUND.toString());
    }

    public static VisibleFormVersion forbidden() {
        return new VisibleFormVersion(0L, Maybe.State.FORBIDDEN.toString());
    }

    public static VisibleFormVersion computeFormVersion(long latestVersion, Permission permission) {
        return new VisibleFormVersion(latestVersion, computeFilterHash(permission));
    }

    public static String computeFilterHash(Permission permission) {
        return hashFormula(permission.getFilter());
    }

    public static String computeFilterHash(FormPermissions permission) {
        return hashFormula(permission.getViewFilter());
    }

    private static String hashFormula(@Nullable String formula) {
        String hash;
        if (formula == null) {
            hash = UNFILTERED_HASH;
        } else {
            hash = Hashing.sha256().hashString(formula, StandardCharsets.UTF_8).toString();
        }
        return hash + FORMAT_SUFFIX;
    }

    public long getNumber() {
        return number;
    }

    public String getFilterHash() {
        return filterHash;
    }

    public boolean isEmpty() {
        return empty;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (!(o instanceof VisibleFormVersion)) {
            return false;
        }
        VisibleFormVersion that = (VisibleFormVersion) o;
        return number == that.number &&
                empty == that.empty &&
                filterHash.equals(that.filterHash);
    }

    @Override
    public int hashCode() {
        return Objects.hash(number, filterHash, empty);
    }

    @Override
    public String toString() {
        return filterHash + "/" + number;
    }

    public static VisibleFormVersion parse(String versionString) {
        if(versionString.isEmpty()) {
            return new VisibleFormVersion(0, "", true);
        }
        int numberStart = versionString.lastIndexOf('/');
        if(numberStart == -1) {
            return new VisibleFormVersion(0, "", false);
        }
        try {
            long version = Long.parseLong(versionString.substring(numberStart + 1));
            String hash = versionString.substring(0, numberStart);
            return new VisibleFormVersion(version, hash, version == 0);
        } catch (NumberFormatException e) {
            return new VisibleFormVersion(0, "", false);
        }
    }

}
