/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.template;

import org.activityinfo.model.database.*;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.Cardinality;
import org.activityinfo.model.type.ReferenceType;
import org.activityinfo.model.type.SerialNumberType;

import java.util.Collections;


public class CaseManagement extends DatabaseTemplate {
    private final ResourceId databaseId;

    public CaseManagement(ResourceId databaseId) {
        this.databaseId = databaseId;
    }


    public UpdateDatabaseRequest build() {
        UpdateDatabaseRequestBuilder req = new UpdateDatabaseRequestBuilder();
        RoleBuilder admBuilder = new RoleBuilder()
                .id(ResourceId.valueOf("admin"))
                .label("Administrator");
        for (Operation operation : Operation.values()) {
            switch (operation) {
                case EDIT_RECORD:
                    admBuilder.addPermission(new FilteredPermission(Operation.EDIT_RECORD, Collections.singleton(SecurityCategory.REVIEWER.getId())));
                    break;
                case BULK_DELETE:
                case MANAGE_COLLECTION_LINKS:
                case MANAGE_ROLES:
                    // Be careful and do not even grant these to administrators
                    break;
                default:
                    admBuilder.addPermission(new FilteredPermission(operation));
            }
        }

        req.addRoleUpdate(admBuilder.build());
        ResourceId userFormId = VisibleDatabaseTree.getUserFormIdFromDatabaseId(databaseId);

        String caseWorkerFilter = String.format("%s == @user",
            userFormId.asString());

        String supervisorFilter = String.format("%s == @user || %s.supervisor == @user",
            userFormId.asString(),
            userFormId.asString());

        req.addRoleUpdate(new RoleBuilder()
            .id(ResourceId.valueOf("supervisor"))
            .label("Supervisor")
            .addPermission(new FilteredPermission(Operation.VIEW, supervisorFilter))
            .addPermission(new FilteredPermission(Operation.EDIT_RECORD, supervisorFilter))
            .addPermission(new FilteredPermission(Operation.ADD_RECORD, supervisorFilter))
            .addPermission(new FilteredPermission(Operation.DELETE_RECORD, supervisorFilter))
            .addPermission(new FilteredPermission(Operation.EXPORT_RECORDS))
            .addFilter(new RoleFilter(ResourceId.valueOf("supervise"), "case is owned or supervised by them", supervisorFilter))
            .build());

        req.addRoleUpdate(new RoleBuilder()
            .id(ResourceId.valueOf("caseworker"))
            .label("Case worker")
            .addParameter(new RoleParameter(ResourceId.valueOf("supervisor"), "Supervisor", userFormId))
            .addPermission(new FilteredPermission(Operation.VIEW, caseWorkerFilter))
            .addPermission(new FilteredPermission(Operation.EDIT_RECORD, caseWorkerFilter))
            .addPermission(new FilteredPermission(Operation.ADD_RECORD, caseWorkerFilter))
            .addPermission(new FilteredPermission(Operation.DELETE_RECORD, caseWorkerFilter))
            .addPermission(new FilteredPermission(Operation.EXPORT_RECORDS))
            .addFilter(new RoleFilter(ResourceId.valueOf("own"), "case is assigned to them", caseWorkerFilter))
            .build());


        return req.build();
    }


    public AddFormRequest buildCaseFormRequest() {

        ResourceId caseFormId = ResourceId.generateId();
        FormClass caseForm = new FormClass(caseFormId);
        caseForm.setLabel("Case");
        caseForm.setDatabaseId(databaseId);
        caseForm.addElement(new FormField(ResourceId.generateId())
            .setLabel("Case Number")
            .setType(new SerialNumberType())
            .setRequired(true)
            .setKey(true));

        caseForm.addElement(new FormField(ResourceId.valueOf("caseowner"))
            .setLabel("Case worker")
            .setType(new ReferenceType(Cardinality.SINGLE, VisibleDatabaseTree.getUserFormIdFromDatabaseId(databaseId))));

        AddFormRequest partner = new AddFormRequestBuilder()
            .formResource(new Resource.Builder()
                .setId(caseFormId)
                .setLabel(caseForm.getLabel())
                .setType(ResourceType.FORM)
                .setVisibility(Resource.Visibility.PRIVATE)
                .setParentId(databaseId)
                .build())
            .formClass(caseForm)
            .build();

        return partner;

    }

}
