/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.template;

import org.activityinfo.model.database.*;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.primitive.TextType;

import java.util.Collections;

public class PartnerReporting extends DatabaseTemplate {
    public static final ResourceId ADMIN_ROLE_ID = ResourceId.valueOf("admin");
    private final ResourceId partnerFormId;
    private final ResourceId referenceFolderId;
    private final ResourceId databaseId;

    public PartnerReporting(ResourceId databaseId) {
        this.databaseId = databaseId;
        this.partnerFormId = ResourceId.generateId();
        this.referenceFolderId = ResourceId.valueOf("reference");
    }

    public PartnerReporting(ResourceId partnerFormId, ResourceId referenceFolderId, ResourceId databaseId) {
        this.partnerFormId = partnerFormId;
        this.referenceFolderId = referenceFolderId;
        this.databaseId = databaseId;
    }

    public UpdateDatabaseRequest build() {

        UpdateDatabaseRequestBuilder req = new UpdateDatabaseRequestBuilder();
        req.addRoleUpdate(buildAdminRole());

        req.addRoleUpdate(new RoleBuilder()
            .id(ResourceId.valueOf("sectorlead"))
            .label("Sector Lead")
            .addPermission(new FilteredPermission(Operation.VIEW))
            .addPermission(new FilteredPermission(Operation.EDIT_RESOURCE))
            .addPermission(new FilteredPermission(Operation.EDIT_RECORD, Collections.singleton(SecurityCategory.REVIEWER.getId())))
            .addPermission(new FilteredPermission(Operation.ADD_RECORD))
            .addPermission(new FilteredPermission(Operation.DELETE_RECORD))
            .addPermission(new FilteredPermission(Operation.EXPORT_RECORDS))
            .addPermission(new FilteredPermission(Operation.LOCK_RECORDS))
            .addPermission(new FilteredPermission(Operation.MANAGE_USERS))
            .build());

        String partnerFilter = String.format("%s == @user.partner", partnerFormId.asString());

        req.addRoleUpdate(new RoleBuilder()
            .id(ResourceId.valueOf("rp"))
            .label("Reporting partner")
            .addPermission(new FilteredPermission(Operation.VIEW, partnerFilter))
            .addPermission(new FilteredPermission(Operation.EDIT_RECORD, partnerFilter))
            .addPermission(new FilteredPermission(Operation.ADD_RECORD, partnerFilter))
            .addPermission(new FilteredPermission(Operation.DELETE_RECORD, partnerFilter))
            .addPermission(new FilteredPermission(Operation.EXPORT_RECORDS))
            .addParameter(new RoleParameter(ResourceId.valueOf("partner"), "Partner", partnerFormId))
            .addFilter(new RoleFilter(ResourceId.valueOf("partner"), "partner is user's partner", partnerFilter))
            .build());

        req.addResourceUpdate(new Resource.Builder()
                .setId(referenceFolderId)
                .setLabel("Reference Data")
                .setVisibility(Resource.Visibility.REFERENCE)
                .setType(ResourceType.FOLDER)
                .setParentId(databaseId)
                .build());

        return req.build();
    }

    public static Role buildAdminRole() {
        RoleBuilder builder = new RoleBuilder()
            .id(ADMIN_ROLE_ID)
            .label("Administrator");

        for (Operation operation : Operation.values()) {
            if (operation == Operation.EDIT_RECORD) {
                builder.addPermission(new FilteredPermission(Operation.EDIT_RECORD, Collections.singleton(SecurityCategory.REVIEWER.getId())));
            } else {
                builder.addPermission(new FilteredPermission(operation));
            }
        }

        return builder.build();
    }

    public AddFormRequest buildAddPartnerRequest() {

        FormClass partnerForm = new FormClass(partnerFormId);
        partnerForm.setLabel("Partner");
        partnerForm.setDatabaseId(databaseId);
        partnerForm.addElement(new FormField(ResourceId.valueOf("name"))
            .setLabel("Name")
            .setType(TextType.SIMPLE)
            .setRequired(true)
            .setKey(true));

        AddFormRequest partner = new AddFormRequestBuilder()
            .formResource(new Resource.Builder()
                .setId(partnerFormId)
                .setLabel(partnerForm.getLabel())
                .setType(ResourceType.FORM)
                .setVisibility(Resource.Visibility.REFERENCE)
                .setParentId(referenceFolderId)
                .build())
            .formClass(partnerForm)
            .build();

        return partner;

    }

}
