/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.type;

import com.google.common.base.Strings;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;
import org.activityinfo.model.resource.ResourceId;

import java.util.*;

import static org.activityinfo.json.Json.createObject;

/**
 * A type that represents a link or reference to another {@code Resource}
 */
public class ReferenceType implements ParametrizedFieldType {

    public static class TypeClass implements ParametrizedFieldTypeClass, RecordFieldTypeClass {

        private TypeClass() {
        }

        @Override
        public String getId() {
            return "reference";
        }

        @Override
        public FieldType createType() {
            return new ReferenceType(Cardinality.SINGLE, Collections.emptySet());
        }

        @Override
        public FieldType deserializeType(JsonValue parametersObject) {
            
            List<ResourceId> range = new ArrayList<>();
            JsonValue rangeArray = parametersObject.get("range");
            for (JsonValue rangeElement : rangeArray.values()) {
                String formId;
                if(rangeElement.isJsonPrimitive()) {
                    formId = rangeElement.asString();
                } else {
                    formId = rangeElement.get("formId").asString();
                }
                range.add(ResourceId.valueOf(formId));
            }
            Cardinality cardinality = Cardinality.valueOf(parametersObject.get("cardinality"));
            Optional<String> lookupFormula = Optional.ofNullable(Strings.emptyToNull(parametersObject.getString("lookupFormula")));

            return new ReferenceType(cardinality, range);
        }
    }

    public static final TypeClass TYPE_CLASS = new TypeClass();

    private final Cardinality cardinality;
    private final List<ResourceId> range = Lists.newArrayList();

    public ReferenceType(Cardinality cardinality, ResourceId rangeFormId) {
        this.cardinality = cardinality;
        this.range.add(rangeFormId);
    }

    public ReferenceType(Cardinality cardinality, Collection<ResourceId> rangeFormIds) {
        this.cardinality = cardinality;
        this.range.addAll(rangeFormIds);
    }

    @Override
    public ParametrizedFieldTypeClass getTypeClass() {
        return TYPE_CLASS;
    }

    @Override
    public FieldValue parseJsonValue(JsonValue value) {
        if(value.isJsonNull()) {
            return new ReferenceValue();
        } else if(value.isJsonArray()) {
            JsonValue array = value;
            Set<RecordRef> refs = new HashSet<>();
            for (int i = 0; i < array.length(); i++) {
                refs.add(parseRef(array.getString(i)));
            }
            return new ReferenceValue(refs);
        } else {
            return new ReferenceValue(parseRef(value.asString()));
        }
    }

    @Override
    public <T> T accept(FieldTypeVisitor<T> visitor) {
        return visitor.visitReference(this);
    }

    @Override
    public boolean isUpdatable() {
        return true;
    }

    @Override
    public FormulaType getFormulaType() {
        if(cardinality == Cardinality.MULTIPLE) {
            return FormulaType.OPAQUE;
        } else {
            return FormulaType.STRING;
        }
    }

    @Override
    public ColumnView toFormulaValue(FieldValue fieldValue) {
        if(cardinality == Cardinality.SINGLE) {
            if(fieldValue instanceof ReferenceValue) {
                ReferenceValue refValue = (ReferenceValue) fieldValue;
                return ColumnViews.of(refValue.getOnlyReference().getRecordId().asString());
            }
        }
        return ColumnViews.BLANK_STRING;
    }

    private RecordRef parseRef(String ref) {
        int separator = ref.indexOf(':');
        if(separator == -1) {
            ResourceId formId = Iterables.getOnlyElement(range);
            return new RecordRef(formId, ResourceId.valueOf(ref));
        } else {
            return RecordRef.fromQualifiedString(ref);
        }
    }

    public Cardinality getCardinality() {
        return cardinality;
    }

    /**
     * @return the set of FormClasses to which fields of this type can refer.
     */
    public Collection<ResourceId> getRange() {
        return range;
    }

    @Override
    public JsonValue getParametersAsJson() {
        JsonValue object = createObject();
        object.put("cardinality", cardinality.name().toLowerCase());
        
        JsonValue rangeArray = Json.createArray();
        for (ResourceId formId : range) {
            JsonValue rangeObject = createObject();
            rangeObject.put("formId", formId.asString());
            rangeArray.add(rangeObject);
        }
        object.put("range", rangeArray);

        return object;
    }

    @Override
    public boolean isValid() {
        return true;
    }

    /**
     * Returns true if this is reference to a database user list.
     */
    public boolean isUserReference() {
        return range.size() == 1 &&
            VisibleDatabaseTree.isUserForm(range.iterator().next());
    }


    /**
     * Convenience constructor for ReferenceTypes with single cardinality
     * @param formClassId the id of the form class which is the range of this field
     * @return a new ReferenceType
     */
    public static ReferenceType single(ResourceId formClassId) {
        return new ReferenceType(Cardinality.SINGLE, Collections.singleton(formClassId));
    }

    public static FieldType multiple(Collection<ResourceId> formClassIds) {
        return new ReferenceType(Cardinality.MULTIPLE, formClassIds);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }
        ReferenceType that = (ReferenceType) o;
        return cardinality == that.cardinality && range.equals(that.range);
    }

    @Override
    public int hashCode() {
        return Objects.hash(cardinality, range);
    }

    @Override
    public String toString() {
        return "ReferenceType{" +
               "cardinality=" + cardinality +
               ", range=" + range +
               '}';
    }
}
