/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.type;

import com.google.common.base.Strings;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;

import static org.activityinfo.json.Json.createObject;

/**
 * Field to which a sequential number is assigned when the record is first saved.
 *
 * <p>The Serial number can include an optional "prefix" which is calculated from the other
 * fields when the record is first created.</p>
 *
 */
public class SerialNumberType implements ParametrizedFieldType {

    public static final FieldTypeClass TYPE_CLASS = new ParametrizedFieldTypeClass() {

        @Override
        public String getId() {
            return "serial";
        }

        @Override
        public FieldType createType() {
            return new SerialNumberType();
        }

        @Override
        public FieldType deserializeType(JsonValue parametersObject) {
            SerialNumberType type = new SerialNumberType();
            if(parametersObject.hasKey("prefixFormula")) {
                type.prefixFormula = parametersObject.get("prefixFormula").asString();
            }
            if(parametersObject.hasKey("digits")) {
                type.digits = parametersObject.get("digits").asInt();
            }
            return type;
        }
    };

    private String prefixFormula;
    private int digits = 5;

    public SerialNumberType() {
    }

    public SerialNumberType(String prefixFormula, int digits) {
        this.prefixFormula = prefixFormula;
        this.digits = digits;
    }

    @Override
    public FieldTypeClass getTypeClass() {
        return TYPE_CLASS;
    }

    @Override
    public SerialNumber parseJsonValue(JsonValue jsonElement) {
        if(jsonElement.isJsonPrimitive()) {
            JsonValue value = jsonElement;
            if (value.isNumber()) {
                return new SerialNumber(value.asInt());
            }
        } else if(jsonElement.isJsonObject()) {
            JsonValue jsonObject = jsonElement;
            return new SerialNumber(
                    jsonObject.get("prefix").asString(),
                    jsonObject.get("number").asInt());
        }

        throw new IllegalArgumentException();
    }

    /**
     * @return the formula used to compute the serial number prefix.
     */
    public String getPrefixFormula() {
        return prefixFormula;
    }

    public void setPrefixFormula(String prefixFormula) {
        this.prefixFormula = prefixFormula;
    }

    public SerialNumberType withPrefixFormula(String formula) {
        return new SerialNumberType(formula, this.digits);
    }


    public boolean hasPrefix() {
        return !Strings.isNullOrEmpty(prefixFormula);
    }

    /**
     *
     * @return the number of digits to use when displaying the record number.
     */
    public int getDigits() {
        return digits;
    }

    @Override
    public <T> T accept(FieldTypeVisitor<T> visitor) {
        return visitor.visitSerialNumber(this);
    }

    @Override
    public boolean isUpdatable() {
        return false;
    }

    @Override
    public FormulaType getFormulaType() {
        return FormulaType.STRING;
    }

    @Override
    public ColumnView toFormulaValue(FieldValue fieldValue) {
        if(fieldValue instanceof SerialNumber) {
            return ColumnViews.of(format(((SerialNumber) fieldValue)));
        } else {
            return ColumnViews.BLANK_STRING;
        }
    }

    @Override
    public JsonValue getParametersAsJson() {
        JsonValue object = createObject();
        if(!Strings.isNullOrEmpty(prefixFormula)) {
            object.put("prefixFormula", prefixFormula);
        }
        object.put("digits", digits);
        return object;
    }

    @Override
    public boolean isValid() {
        return true;
    }


    public String format(SerialNumber value) {
        return format(value.getPrefix(), value.getNumber());
    }

    public String format(String prefix, int number) {
        String formatted = Strings.padStart(Integer.toString(number), digits, '0');
        if(prefix != null) {
            formatted = prefix + "-" + formatted;
        }
        return formatted;
    }

}
