/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.type.attachment;

import org.activityinfo.json.JsonValue;
import org.activityinfo.model.form.JsonParsing;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;
import org.activityinfo.model.type.*;

import static org.activityinfo.json.Json.createObject;

/**
 * @author yuriyz on 8/6/14.
 */
public class AttachmentType implements ParametrizedFieldType {

    public static class TypeClass implements ParametrizedFieldTypeClass, RecordFieldTypeClass {

        private TypeClass() {
        }

        @Override
        public String getId() {
            return "attachment";
        }

        @Override
        public FieldType createType() {
            return new AttachmentType(Cardinality.SINGLE, Kind.ATTACHMENT);
        }


        @Override
        public FieldType deserializeType(JsonValue parametersObject) {
            // Explicit type parameter required by GWT's compiler
            Cardinality cardinality = Cardinality.valueOf(parametersObject.<JsonValue>get("cardinality"));
            Kind kind = Kind.valueOf(JsonParsing.fromEnumValue(parametersObject.get("kind")));
            return new AttachmentType(cardinality, kind);
        }

    }

    public static final TypeClass TYPE_CLASS = new TypeClass();

    public enum Kind {
        ATTACHMENT, IMAGE
    }

    private Cardinality cardinality;
    private Kind kind;

    public AttachmentType(Cardinality cardinality, Kind kind) {
        this.cardinality = cardinality;
        this.kind = kind;
    }

    @Override
    public ParametrizedFieldTypeClass getTypeClass() {
        return TYPE_CLASS;
    }

    @Override
    public AttachmentValue parseJsonValue(JsonValue value) {
        if(value.isJsonObject()) {
            value = value.get("values");
        }
        AttachmentValue fieldValue = new AttachmentValue();
        JsonValue array = value;
        for (JsonValue attachmentItem : array.values()) {
            JsonValue attachmentObject = (JsonValue) attachmentItem;
            String mimeType = attachmentObject.get("mimeType").asString();
            String filename = attachmentObject.get("filename").asString();
            String blobId = attachmentObject.get("blobId").asString();

            Attachment attachment = new Attachment(mimeType, filename, blobId);
            attachment.setWidth(attachmentObject.get("width").asInt());
            attachment.setHeight(attachmentObject.get("height").asInt());
            
            fieldValue.getValues().add(attachment);
        }
        return fieldValue;
    }

    @Override
    public <T> T accept(FieldTypeVisitor<T> visitor) {
        return visitor.visitAttachment(this);
    }

    @Override
    public boolean isUpdatable() {
        return true;
    }

    @Override
    public FormulaType getFormulaType() {
        return FormulaType.OPAQUE;
    }

    @Override
    public ColumnView toFormulaValue(FieldValue fieldValue) {
        if(fieldValue instanceof AttachmentValue) {
            return ColumnViews.of(((AttachmentValue) fieldValue).getValues().size() );
        } else {
            return ColumnViews.BLANK_NUMBER;
        }
    }

    public Cardinality getCardinality() {
        return cardinality;
    }

    public Kind getKind() {
        return kind;
    }

    public void setKind(Kind kind) {
        this.kind = kind;
    }

    @Override
    public JsonValue getParametersAsJson() {
        JsonValue object = createObject();
        object.put("cardinality", cardinality.name().toLowerCase());
        object.put("kind", kind.name().toLowerCase());
        return object;
    }

    @Override
    public boolean isValid() {
        return true;
    }

    @Override
    public String toString() {
        return "AttachmentType";
    }
}