/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.type.enumerated;

import com.google.common.collect.Lists;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.formula.FormulaType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.ColumnViews;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.*;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.activityinfo.json.Json.createObject;

public class EnumType implements ParametrizedFieldType {

    /**
     * The maximum number of items to show as checkboxes, when presentation is "Automatic". Above
     * this count, dropdown will be used.
     */
    public static final int MAX_CHECKBOX_ITEMS = 10;

    public interface EnumTypeClass extends ParametrizedFieldTypeClass, RecordFieldTypeClass {
        @Override
        EnumType deserializeType(JsonValue parametersObject);
    }

    public enum Presentation {
        AUTOMATIC,
        RADIO_BUTTON,
        DROPDOWN
    }

    public static final EnumTypeClass TYPE_CLASS = new EnumTypeClass() {

        @Override
        public String getId() {
            return "enumerated";
        }


        @Override
        public EnumType deserializeType(JsonValue parametersObject) {
            // Explicit type parameter required by GWT's compiler!
            Cardinality cardinality = Cardinality.valueOf(
                    parametersObject.<JsonValue>get("cardinality"));

            Presentation presentation = Presentation.AUTOMATIC;
            if(parametersObject.hasKey("presentation")) {
                String presentationType = parametersObject.get("presentation").asString().toUpperCase();
                switch (presentationType) {
                    case "CHECKBOX":
                    case "RADIO_BUTTON":
                        presentation = Presentation.RADIO_BUTTON;
                        break;
                    case "DROPDOWN":
                        presentation = Presentation.DROPDOWN;
                        break;
                    default:
                        presentation = Presentation.AUTOMATIC;
                        break;
                }
            }

            List<EnumItem> enumItems = Lists.newArrayList();
            JsonValue valuesArray = parametersObject.get("values");
            if(valuesArray != null) {
                if (valuesArray.isJsonArray()) {
                    JsonValue enumItemArray = valuesArray;
                    for (JsonValue record : enumItemArray.values()) {
                        enumItems.add(EnumItem.fromJsonObject(record));
                    }
                }
            }
            return new EnumType(cardinality, presentation, enumItems);
        }

        @Override
        public EnumType createType() {
            return new EnumType();
        }

    };

    private final Cardinality cardinality;
    private final List<EnumItem> items;
    private final List<EnumItem> defaultValues = Lists.newArrayList();
    private final Set<ResourceId> itemIds;
    private final Presentation presentation;

    public EnumType() {
        this(Cardinality.SINGLE, Collections.<EnumItem>emptyList());
    }

    public EnumType(Cardinality cardinality, List<EnumItem> items) {
        this(cardinality, Presentation.AUTOMATIC, items);
    }

    public EnumType(Cardinality cardinality, EnumItem... items) {
        this(cardinality, Arrays.asList(items));
    }

    public EnumType(Cardinality cardinality, Presentation presentation, List<EnumItem> items) {
        this.cardinality = cardinality;
        this.items = items != null ? items : new ArrayList<EnumItem>();
        this.itemIds = new HashSet<>();
        for (EnumItem item : this.items) {
            itemIds.add(item.getId());
        }
        this.presentation = presentation;
    }

    public EnumType withPresentation(Presentation presentation) {
        return new EnumType(cardinality, presentation, items);
    }

    public Cardinality getCardinality() {
        return cardinality;
    }

    public Presentation getPresentation() {
        return presentation;
    }

    public Presentation getEffectivePresentation() {
        if(presentation == Presentation.AUTOMATIC) {
            if(items.size() > MAX_CHECKBOX_ITEMS) {
                return Presentation.DROPDOWN;
            } else {
                return Presentation.RADIO_BUTTON;
            }
        } else {
            return presentation;
        }
    }

    public List<EnumItem> getItems() {
        return items;
    }

    public List<EnumItem> getDefaultValues() {
        return defaultValues;
    }

    @Override
    public ParametrizedFieldTypeClass getTypeClass() {
        return TYPE_CLASS;
    }

    @Override
    public EnumValue parseJsonValue(JsonValue value) {
        if(value.isString()) {
            ResourceId id = ResourceId.valueOf(value.asString());
            if(itemIds.contains(id)) {
                return new EnumValue(id);
            }
        } else if(value.isJsonArray()) {
            Set<ResourceId> ids = new HashSet<>();
            JsonValue array = value;
            for (int i = 0; i < array.length(); i++) {
                ResourceId id = ResourceId.valueOf(array.get(i).asString());
                if(itemIds.contains(id)) {
                    ids.add(id);
                }
            }
            return new EnumValue(ids);
        }

        return null;
    }

    @Override
    public <T> T accept(FieldTypeVisitor<T> visitor) {
        return visitor.visitEnum(this);
    }

    @Override
    public boolean isUpdatable() {
        return true;
    }

    @Override
    public FormulaType getFormulaType() {
        return FormulaType.STRING;
    }

    public Stream<EnumItem> getItemById(ResourceId id) {
        return items.stream().filter(i -> i.getId().equals(id));
    }

    @Override
    public ColumnView toFormulaValue(FieldValue fieldValue) {
        if(fieldValue instanceof EnumValue) {
            String joined =  ((EnumValue) fieldValue)
                .getResourceIds()
                .stream()
                .flatMap(id -> items.stream().filter(i -> i.getId().equals(id)))
                .map(item -> item.getLabel())
                .collect(Collectors.joining(", "));

            return ColumnViews.of(joined);

        } else {
            return ColumnViews.BLANK_STRING;
        }
    }

    @Override
    public JsonValue getParametersAsJson() {
        
        JsonValue enumValueArray = Json.createArray();
        for (EnumItem enumItem : getItems()) {
            enumValueArray.add(enumItem.toJsonObject());
        }

        JsonValue object = createObject();
        object.put("cardinality", cardinality.name().toLowerCase());
        object.put("presentation", presentation.name().toLowerCase());
        object.put("values", enumValueArray);
        return object;
    }

    @Override
    public boolean isValid() {
        return items.size() > 0;
    }


    public Map<ResourceId, String> labelMap() {
        Map<ResourceId, String> labelMap = new HashMap<>();
        for (EnumItem item : items) {
            labelMap.put(item.getId(), item.getLabel());
        }
        return labelMap;
    }

}
