/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.type.geo;

import org.activityinfo.json.JsonValue;

import java.io.Serializable;

import static org.activityinfo.json.Json.createObject;

/*
 * Bounding box for a map. 
 * 
 * This cannot be mapped 1:1 to a rectangle, since a lat/long combination is a coordinate on
 * a sphere as opposed to a coordinate on a 2D plane.
 */
public class Extents implements Serializable {

    public static final int LAT_MAX = 90;
    public static final int LNG_MAX = 180;
    public static final int LAT_MIN = -LAT_MAX;
    public static final int LNG_MIN = -180;

    private final double minLat;
    private final double maxLat;
    private final double minLon;
    private final double maxLon;

    public Extents(double minLat, double maxLat, double minLon, double maxLon) {
        this.minLat = minLat;
        this.maxLat = maxLat;
        this.minLon = minLon;
        this.maxLon = maxLon;
    }

    public Extents(Extents toCopy) {
        this.minLat = toCopy.minLat;
        this.maxLat = toCopy.maxLat;
        this.minLon = toCopy.minLon;
        this.maxLon = toCopy.maxLon;
    }

    public Extents(GeoPoint center) {
        this.minLat = this.maxLat = center.getLatitude();
        this.minLon = this.maxLon = center.getLongitude();
    }

    /**
     * @return maximum geographic bounds (-180, -90, 180, 90)s
     */
    public static Extents maxGeoBounds() {
        return new Extents(LAT_MIN, LAT_MAX, LNG_MIN, LNG_MAX);
    }

    public double getMinLat() {
        return minLat;
    }

    public double getMaxLat() {
        return maxLat;
    }

    public double getMinLon() {
        return minLon;
    }

    public double getMaxLon() {
        return maxLon;
    }

    public double getWest() {
        return minLon;
    }

    public double getEast() {
        return maxLon;
    }

    public double getNorth() {
        return maxLat;
    }

    public double getSouth() {
        return minLat;
    }

    /**
     * Calculates the intersection of this Extents with given Extents
     *
     * @param b another Extents with which to intersect this Extents
     * @return the intersection of the two Extentss
     */
    public Extents intersect(Extents b) {
        return new Extents(Math.max(minLat, b.minLat),
                Math.min(maxLat, b.maxLat),
                Math.max(minLon, b.minLon),
                Math.min(maxLon, b.maxLon));
    }

    /**
     * @return true if this Extents intersects with <code>b</code>
     */
    public boolean intersects(Extents b) {
        return !(b.maxLon < minLon || b.minLon > maxLon || b.maxLat < minLat || b.minLat > maxLat);
    }

    public static Extents emptyExtents() {
        return new Extents(+90.0, -90.0, +180.0, -180.0);
    }

    public static Extents empty() {
        return emptyExtents();
    }

    /**
     * @param b
     * @return true if this Extents contains <code>b</code>
     */
    public boolean contains(Extents b) {
        return b.minLon >= minLon && b.maxLon <= maxLon && b.minLat >= minLat && b.maxLat <= maxLat;
    }

    public static Extents create(double x1, double y1, double x2, double y2) {
        assert x1 <= x2 : x1 + " should be less than or equal to " + x2;
        assert y1 <= y2 : y1 + " should be less than or equal to " + y2;
        return new Extents(y1, y2, x1, x2);
    }
    

    public static Extents fromLatLng(double latitude, double longitude) {
        return new Extents(latitude, latitude, longitude, longitude);
    }
    
    /**
     * @return the x (longitude) coordinate of the Extents's centroid, (x1+x2)/2
     */
    public double getCenterX() {
        return (minLon + maxLon) / 2d;
    }

    /**
     * @return the y (latitudinal) coordinate of the Extents's centroid,
     * (y1+y2)/2
     */
    public double getCenterY() {
        return (minLat + maxLat) / 2d;
    }

    public GeoPoint getCenter() {
        return new GeoPoint(getCenterY(), getCenterX());
    }

    public boolean isEmpty() {
        return minLat > maxLat || minLon > maxLon;
    }

    public double area() {
        if(isEmpty()) {
            return 0d; 
        } else {
            return (maxLat - minLat) * (maxLon - minLon);
        }
    }

    public double getX1() {
        return getMinLon();
    }
    
    public double getY1() {
        return getMinLat();
    }
    
    public double getX2() {
        return getMaxLon();
    }
    
    public double getY2() {
        return getMaxLat();
    }
    
    @Override
    public int hashCode() {
        return (minLon + "").hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        Extents other = (Extents) obj;
        return minLat == other.minLat &&
               maxLat == other.maxLat &&
               minLon == other.minLon &&
               maxLon == other.maxLon;
    }
    
    

    @Override
    public String toString() {
        return "Extents{" +
               "minLat=" + minLat +
               ", maxLat=" + maxLat +
               ", minLon=" + minLon +
               ", maxLon=" + maxLon +
               '}';
    }

    public JsonValue toJsonElement() {
        JsonValue object = createObject();
        if (!Double.isNaN(minLat)) {
            object.put("minLat", minLat);
        }
        if (!Double.isNaN(maxLat)) {
            object.put("maxLat", maxLat);
        }
        if (!Double.isNaN(minLon)) {
            object.put("minLon", minLon);
        }
        if (!Double.isNaN(maxLon)) {
            object.put("maxLon", maxLon);
        }
        return object;
    }

    public static Extents fromJsonObject(JsonValue object) {
        return new Extents(
                object.getNumber("minLat"),
                object.getNumber("maxLat"),
                object.getNumber("minLon"),
                object.getNumber("maxLon"));
    }
}
