/*
 *  Adapted from format.c:
 *  https://github.com/wch/r-source/blob/trunk/src/main/format.c
 *
 *  R : A Computer Language for Statistical Data Analysis
 *  Copyright (C) 1997--2019  The R Core Team
 *  Copyright (C) 2003--2016  The R Foundation
 *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, a copy is available at
 *  https://www.R-project.org/Licenses/
 */

package org.activityinfo.model.type.number;

import org.activityinfo.model.query.ColumnView;

/**
 * Determines the best format for a set of quantities.
 *
 * <p>Adapted from the the format.info routine from GNU R, licensed under GPL v2.</p>
 */
public class QuantityFormatBuilder {

    public static final int MAX_FRACTION_DIGITS = 7;

    /**
     * When deciding between fixed notation and scientific notation, apply an additional
     * penalty to the width of the number in scientific notation.
     */
    private static final int SCIENTIFIC_NOTATION_PENALTY = 0;

    /**
     * Smallest decimal exponent
     */
    private static final int MIN_DECIMAL_EXPONENT = -308;

    private static final double[] POWERS_OF_TEN = {
        1e-1,
        1e00, 1e01, 1e02, 1e03, 1e04, 1e05, 1e06, 1e07, 1e08, 1e09,
        1e10, 1e11, 1e12, 1e13, 1e14, 1e15, 1e16, 1e17, 1e18, 1e19,
        1e20, 1e21, 1e22};

    private static final int KP_MAX = 22;

    private int negative = 0;

    private int maxSignificantDigits = Integer.MIN_VALUE;

    private int maxRightDigits = Integer.MIN_VALUE;

    /**
     * Determines the best way to format a {@code ColumnView} of numbers.
     */
    public static QuantityFormat from(ColumnView x, int startIndex, int endIndex) {
        QuantityFormatBuilder builder = new QuantityFormatBuilder();
        for (int i = startIndex; i < endIndex; i++) {
            double xi = x.getDouble(i);
            builder.add(xi);
        }
        return builder.build();
    }

    public static QuantityFormat from(Double value) {
        return new QuantityFormatBuilder().add(value).build();
    }


    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    public QuantityFormatBuilder add(double x) {
        if (!Double.isFinite(x) || x == 0.0) {
            return this;
        }

        /*
         * for a number x , determine
         *	neg    = 1_{x < 0}  {0/1}
         *	kpower = Exponent of 10;
         *	nsig   = min(R_print.digits, #{significant digits of alpha})
         *  roundingwidens = TRUE iff rounding causes x to increase in width
         *
         * where  |x| = alpha * 10^kpower	and	 1 <= alpha < 10
         */
        double r;
        int negative;
        if(x < 0.0) {
            negative = 1;
            r = -x;
        } else {
            negative = 0;
            r = x;
        }

        /* r = |x|; 10^(kp + digits - 1) <= r */
        int kp = (int) Math.floor(Math.log10(r)) - MAX_FRACTION_DIGITS + 1;

        double r_prec = r;

        /* use exact scaling factor in double precision, if possible */
        if (Math.abs(kp) <= KP_MAX) {
            if (kp >= 0) {
                r_prec /= POWERS_OF_TEN[ kp+1];
            } else {
                r_prec *= POWERS_OF_TEN[-kp+1];
            }
        }
        /* For IEC60559 1e-308 is not representable except by gradual underflow.
           Shifting by 303 allows for any potential denormalized numbers x,
           and makes the reasonable assumption that R_dec_min_exponent+303
           is in range. Representation of 1e+303 has low error.
         */
        else if (kp <= MIN_DECIMAL_EXPONENT) {
            r_prec = (r_prec * 1e+303)/ exp10((double)(kp+303));
        } else {
            r_prec /= exp10((double)kp);
        }
        if (r_prec < POWERS_OF_TEN[MAX_FRACTION_DIGITS]) {
            r_prec *= 10.0;
            kp--;
        }

        /* round alpha to integer, 10^(digits-1) <= alpha <= 10^digits */
            /* accuracy limited by double rounding problem,
               alpha already rounded to 53 bits */
        double alpha = Math.round(r_prec);

        int significantDigits = MAX_FRACTION_DIGITS;
        for (int j = 1; j <= MAX_FRACTION_DIGITS; j++) {
            alpha /= 10.0;
            if (alpha == Math.floor(alpha)) {
                significantDigits--;
            } else {
                break;
            }
        }
        if (significantDigits == 0 && MAX_FRACTION_DIGITS > 0) {
            significantDigits = 1;
            kp += 1;
        }
        int kpower = kp + MAX_FRACTION_DIGITS - 1;

        int left = kpower + 1;

        if (roundingWidens(r, kpower)) {
            left--;
        }

        /* >= 1 */
        int sleft = negative + ((left <= 0) ? 1 : left);

        /* #{digits} right of '.' ( > 0 often)*/
        int right = significantDigits - left;

        /* if any < 0, need extra space for sign */
        if (negative != 0) {
            this.negative = 1;
        }

        /* Infinite precision "F" Format : */
        if (right > maxRightDigits) {
            maxRightDigits = right;
        }

        if (significantDigits > maxSignificantDigits) {
            maxSignificantDigits = significantDigits;
        }

        return this;
    }

    /**
     * Scientific format may do more rounding than fixed format, e.g.
     * 9996 with 3 digits is 1e+04 in scientific, but 9996 in fixed.
     * This happens when the true value r is less than 10^(kpower+1)
     * and would not round up to it in fixed format.
     * Here rgt is the decimal place that will be cut off by rounding
     */
    private static boolean roundingWidens(double r, int kpower) {

        int rgt = MAX_FRACTION_DIGITS - kpower;

        /* bound rgt by 0 and KP_MAX */
        if (rgt < 0) {
            rgt = 0;
        } else {
            rgt = rgt > KP_MAX ? KP_MAX : rgt;
        }

        double fuzz = 0.5/ POWERS_OF_TEN[1 + rgt];

        // kpower can be bigger than the table.
        return kpower > 0 && kpower <= KP_MAX && r < POWERS_OF_TEN[kpower + 1] - fuzz;
    }


    private static double exp10(double v) {
        return Math.pow(v, 10);
    }

    public QuantityFormat build() {

        if (maxRightDigits < 0) {
            maxRightDigits = 0;
        }

        return new QuantityFormat(maxRightDigits);
    }
}
