/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.model.type.time;

import org.activityinfo.json.Json;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.type.FieldTypeClass;

import java.io.Serializable;

/**
 * The first epi week of the year ends, by definition, on the first Saturday of January,
 * as long as it falls at least four days into the month. Each epi week begins on a Sunday and ends on a Saturday.
 * <p/>
 * http://www.cmmcp.org/epiweek.htm
 * <p/>
 * MMWR week numbering is sequential beginning with 1 and incrementing with each week to a maximum of 52 or 53.
 * MMWR week #1 of an MMWR year is the first week of the year that has at least four days in the calendar year.
 * For example, if January 1 occurs on a Sunday, Monday, Tuesday or Wednesday, the calendar week that includes January 1 would be
 * MMWR week #1. If January 1 occurs on a Thursday, Friday, or Saturday, the calendar week that includes January 1 would be the last
 * MMWR week of the previous year (#52 or #53). Because of this rule, December 29, 30, and 31 could potentially fall into
 * MMWR week #1 of the following MMWR year.
 * <p/>
 * http://wwwn.cdc.gov/nndss/document/MMWR_week_overview.pdf
 *
 * @author yuriyz on 02/26/2015.
 */
public class EpiWeek implements Serializable, PeriodValue {

    private final int weekInYear;
    private final int year;

    public EpiWeek(EpiWeek week) {
        this(week.year, week.weekInYear);
    }

    private EpiWeek(int year, int weekInYear) {
        this.year = year;
        this.weekInYear = weekInYear;
    }

    /**
     * Creates a new {@code EpiWeek} instance from the given year and week.
     *
     * <p>If the week number is out of bounds for the given year, the year and week
     * are normalized. For example, {@code EpiWeek.of(2019, 53)} becomes {@code EpiWeek.of(2020,1)}</p>
     *
     * @param year Epiweek year
     * @param weekInYear Epiweek number, starting with 1
     * @return a new normalized {@code EpiWeek} instance.
     */
    public static EpiWeek of(int year, int weekInYear) {
        while (true) {
            int weeksInThisYear = weeksInYear(year);
            if(weekInYear > weeksInThisYear) {
                year++;
                weekInYear -= weeksInThisYear;
            } else if(weekInYear < 1){
                year--;
                weekInYear += weeksInYear(year);
            } else {
                break;
            }
        }
        return new EpiWeek(year, weekInYear);
    }



    public int getWeekInYear() {
        return weekInYear;
    }

    public int getYear() {
        return year;
    }

    public EpiWeek plus(int count) {
        return of(year, weekInYear + count);
    }

    public EpiWeek next() {
        return plus(+1);
    }

    public EpiWeek previous() {
        return plus(-1);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        EpiWeek epiWeek = (EpiWeek) o;
        return weekInYear == epiWeek.weekInYear && year == epiWeek.year;
    }

    @Override
    public int hashCode() {
        int result = weekInYear;
        result = 31 * result + year;
        return result;
    }

    @Override
    public String toString() {
        return toString(year, weekInYear);
    }

    public static String toString(int year, int weekInYear) {
        if(weekInYear < 10) {
            return year + "W0" + weekInYear;
        } else {
            return year + "W" + weekInYear;
        }
    }

    /**
     * @param epiWeekAsString (e.g. 2015W01, 2016W30)
     * @return epi week
     */
    public static EpiWeek parse(String epiWeekAsString) {
        //noinspection NonJREEmulationClassesInClientCode
        String[] tokens = epiWeekAsString.split("W");
        if (tokens.length != 2) {
            throw new NumberFormatException(epiWeekAsString);
        }
        String year = tokens[0];
        String weekInYear = tokens[1];
        return of(Integer.parseInt(year), Integer.parseInt(weekInYear));
    }

    public Month toMonth() {
        if(weekInYear == 1) {
            return new Month(year, 1);
        }
        return asInterval().getStartDate().getMonth();
    }


    /**
     * Calculates the date of the first epi-week of the year.
     */
    public static LocalDate dayOfFirstEpiWeek(int year) {
        // The first epi week of the year ends, by definition, on the first Saturday of January, as long as it falls
        // at least four days into the month.
        // Each epi week begins on a Sunday and ends on a Saturday.
        // http://www.cmmcp.org/epiweek.htm

        LocalDate jan1 = new LocalDate(year, 1, 1);
        switch (jan1.getDayOfWeek()) {
            case MONDAY:
                return new LocalDate(year - 1, 12, 31);
            case TUESDAY:
                return new LocalDate(year - 1, 12, 30);
            case WEDNESDAY:
                return new LocalDate(year - 1, 12, 29);
            case THURSDAY:
                return new LocalDate(year, 1, 4);
            case FRIDAY:
                return new LocalDate(year, 1, 3);
            case SATURDAY:
                return new LocalDate(year, 1, 2);
            case SUNDAY:
            default:
                return new LocalDate(year, 1, 1);
        }
    }


    public static EpiWeek weekOf(LocalDate date) {
        LocalDate firstDayOfYear = dayOfFirstEpiWeek(date.getYear());

        if(date.before(firstDayOfYear)) {
            return of(date.getYear() - 1, 52);
        } else {
            int daysBetween = LocalDate.daysBetween(firstDayOfYear, date);
            int weekNumber = daysBetween / 7;

            return of(date.getYear(), weekNumber + 1);
        }
    }

    public static int weeksInYear(int year) {
        LocalDate firstDayOfYear = dayOfFirstEpiWeek(year);
        LocalDate firstDayOfNextYear = dayOfFirstEpiWeek(year + 1);
        int numDays = LocalDate.daysBetween(firstDayOfYear, firstDayOfNextYear);
        if(numDays >= 53*7) {
            return 53;
        } else {
            return 52;
        }
    }

    public static boolean isValidWeek(int year, int weekNumber) {
        if(weekNumber < 1 || weekNumber > 53) {
            return false;
        }
        if(weekNumber <= 52) {
            return true;
        }
        return weekNumber <= weeksInYear(year);
    }


    @Override
    public LocalDateInterval asInterval() {
        LocalDate firstDayOfYear = dayOfFirstEpiWeek(this.year);
        LocalDate firstDayOfWeek = firstDayOfYear.plusDays((this.weekInYear - 1) * 7);
        LocalDate lastDayOfWeek = firstDayOfWeek.plusDays(6);
        return new LocalDateInterval(firstDayOfWeek, lastDayOfWeek);
    }

    @Override
    public FieldTypeClass getTypeClass() {
        return EpiWeekType.TYPE_CLASS;
    }

    @Override
    public JsonValue toJson() {
        return Json.create(toString());
    }


}