/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.observable;

import com.google.gwt.user.client.Timer;
import org.checkerframework.checker.nullness.qual.Nullable;

class OptimisticObservableWithTimeout<T> extends Observable<T> {
    private final Observable<T> observable;
    private final int timeoutMillis;

    private @Nullable Subscription subscription;

    private @Nullable Timer timer;

    OptimisticObservableWithTimeout(Observable<T> observable, int timeoutMillis, @Nullable T defaultValue) {
        this.observable = observable;
        this.timeoutMillis = timeoutMillis;
        this.cachedValue = defaultValue;
    }

    @Override
    protected void onConnect() {
        if(timer == null) {
            this.timer = new Timer() {
                @Override
                public void run() {
                    onTimeout();
                }
            };
        }

        subscription = observable.subscribe(new Observer<T>() {
            @Override
            public void onChange(@Nullable T newValue) {
                onValueChanged(newValue);
            }
        });
    }

    @Override
    protected void onDisconnect() {
        super.onDisconnect();

        assert subscription != null : "onConnect was not called";
        subscription.unsubscribe();
        subscription = null;
    }

    private void onValueChanged(@Nullable T newValue) {
        if(newValue != null) {
            if(newValue != cachedValue) {
                if(timer != null && timer.isRunning()) {
                    timer.cancel();
                }
                fireChange(newValue);
            }

        } else {
            // Start the timer... if we don't get a loaded value within the
            // provided timeout, THEN fire then alert observers that we are loading.
            assert timer != null;
            if(!timer.isRunning()) {
                timer.schedule(timeoutMillis);
            }
        }
    }

    private void onTimeout() {
        // We've been optimistic that the value would not be change...
        // but no we've lost our patience.
        fireChange(null);
    }
}
