/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.promise;

import org.checkerframework.checker.nullness.qual.EnsuresNonNullIf;
import org.checkerframework.checker.nullness.qual.NonNull;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.Optional;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * Holds either a value, or is empty because it is FORBIDDEN, DELETED, OR NOT_FOUND.
 */
public class Maybe<T> {

    public enum State {
        VISIBLE,
        FORBIDDEN,
        DELETED,
        NOT_FOUND
    }

    public interface Case<T, R> {
        R visible(T value);
        R forbidden();
        R deleted();
        R notFound();
    }

    private State state;

    private @Nullable T value;

    private static final Maybe<?> FORBIDDEN = new Maybe<>(State.FORBIDDEN, null);
    private static final Maybe<?> DELETED = new Maybe<>(State.DELETED, null);
    private static final Maybe<?> NOT_FOUND = new Maybe<>(State.NOT_FOUND, null);

    public static <T> Maybe<T> of(T value) {
        return new Maybe<T>(State.VISIBLE, value);
    }

    public static <T> Maybe<T> notFound() {
        @SuppressWarnings("unchecked")
        Maybe<T> m = (Maybe<T>) NOT_FOUND;
        return m;
    }

    public static <T> Maybe<T> deleted() {
        @SuppressWarnings("unchecked")
        Maybe<T> m = (Maybe<T>)DELETED;
        return m;
    }

    public static <T> Maybe<T> forbidden() {
        @SuppressWarnings("unchecked")
        Maybe<T> m = (Maybe<T>)FORBIDDEN;
        return m;
    }

    private Maybe(State state, @Nullable T value) {
        this.state = state;
        this.value = value;
    }

    public State getState() {
        return state;
    }

    @EnsuresNonNullIf(expression = "value", result = true)
    public boolean isVisible() {
        return value != null;
    }

    public @NonNull T get() {
        if(value == null) {
            throw new IllegalStateException();
        }
        return value;
    }

    public Optional<T> getIfVisible() {
        if(isVisible()) {
            return Optional.of(get());
        } else {
            return Optional.empty();
        }
    }

    public void ifVisible(Consumer<T> consumer) {
        if(isVisible()) {
            consumer.accept(get());
        }
    }

    /**
     * Return the contained value, if visible, otherwise throw an exception
     * to be created by the provided supplier.
     *
     * @apiNote A method reference to the exception constructor with an empty
     * argument list can be used as the supplier. For example,
     * {@code IllegalStateException::new}
     *
     * @param <X> Type of the exception to be thrown
     * @param exceptionSupplier The supplier which will return the exception to
     * be thrown
     * @return the present value
     * @throws X if there is no visible value
     * @throws NullPointerException if no value is present and
     * {@code exceptionSupplier} is null
     */
    public <X extends Throwable> T orElseThrow(Supplier<? extends X> exceptionSupplier) throws X {
        if (value != null) {
            return value;
        } else {
            throw exceptionSupplier.get();
        }
    }

    public <R> R switch_(Case<T, R> case_) {
        if(isVisible()) {
            return case_.visible(value);
        }
        switch (state) {
            case FORBIDDEN:
                return case_.forbidden();
            case DELETED:
                return case_.deleted();
            default:
            case NOT_FOUND:
                return case_.notFound();
        }
    }

    @SuppressWarnings("unchecked")
    public <R> Maybe<R> transform(Function<T, R> function) {
        if(isVisible()) {
            return Maybe.of(function.apply(value));
        } else {
            return (Maybe<R>) this;
        }
    }

    @SuppressWarnings("unchecked")
    public <R> Maybe<R> flatMap(Function<T, Maybe<R>> function) {
        if(isVisible()) {
            return function.apply(value);
        } else {
            return (Maybe<R>) this;
        }
    }

    public T or(T defaultValue) {
        if(isVisible()) {
            return get();
        } else {
            return defaultValue;
        }
    }

    public T orGet(Supplier<T> supplier) {
        if(isVisible()) {
            return get();
        } else {
            return supplier.get();
        }
    }

    public Result<T, State> toResult() {
        if(isVisible()) {
            return Result.ok(value);
        } else {
            return Result.error(state);
        }
    }

    public <E> Result<T, E> toResult(Function<State, E> errorHandler) {
        if(isVisible()) {
            return Result.ok(value);
        } else {
            return Result.error(errorHandler.apply(state));
        }
    }

    public static <T> Maybe<T> fromOptional(Optional<T> optional) {
        if(optional.isPresent()) {
            return Maybe.of(optional.get());
        } else {
            return Maybe.notFound();
        }
    }


    public static <T> Maybe<T> fromNullable(T value) {
        if(value == null) {
            return Maybe.notFound();
        } else {
            return Maybe.of(value);
        }
    }

    @Override
    public int hashCode() {
        if(value == null) {
            return 0;
        } else {
            return value.hashCode();
        }
    }

    @Override
    public boolean equals(@Nullable Object obj) {
        if(this == obj) {
            return true;
        }

        if(!(obj instanceof Maybe)) {
            return false;
        }
        Maybe other = (Maybe) obj;

        if(this.value != null) {
            return this.value.equals(other.value);
        } else {
            return this.state == other.state;
        }
    }

    @Override
    public String toString() {
        if(state == State.VISIBLE) {
            return "Maybe{" + value + "}";
        } else {
            return "Maybe{" + state + "}";
        }
    }
}
