/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.promise;

import org.checkerframework.checker.nullness.qual.EnsuresNonNullIf;
import org.checkerframework.checker.nullness.qual.Nullable;

import java.util.function.Function;
import java.util.function.Supplier;

/**
 * A Result is the result of a computation that may fail.
 */
public final class Result<T, E> {

    private final @Nullable T value;
    private final @Nullable E error;


    public static <T, E> Result<T, E> ok(T value) {
        return new Result<T, E>(value, null);
    }

    public static <T, E> Result<T, E> ok(T value, Class<E> errorType) {
        return new Result<T, E>(value, null);
    }

    public static <T, E> Result<T, E> error(E error) {
        return new Result<T, E>(null, error);
    }

    public static <T, E> Result<T, E> error(Class<T> resultType, E error) {
        return new Result<T, E>(null, error);
    }

    public Result(@Nullable T value, @Nullable E error) {
        this.value = value;
        this.error = error;
    }


    public @Nullable T getValue() {
        return value;
    }

    public @Nullable E getError() {
        return error;
    }

    public <U> Result<U, E> map(Function<T, U> function) {
        if (isOk()) {
            return Result.ok(function.apply(value));
        } else {
            // We can change the type from T to U because at this
            // point value must be null.
            @SuppressWarnings("unchecked")
            Result<U, E> newResult = (Result<U, E>) this;
            return newResult;
        }
    }

    public <U> Result<T, U> mapError(Function<E, U> function) {
        if(isOk()) {
            @SuppressWarnings("unchecked")
            Result<T, U> casted = (Result<T, U>) this;
            return casted;
        } else {
            return Result.error(function.apply(error));
        }
    }

    public <U> Result<U, E> flatMap(Function<T, Result<U, E>> successHandler) {
        if (isOk()) {
            return successHandler.apply(value);
        } else {
            @SuppressWarnings("unchecked")
            Result<U, E> casted = (Result<U, E>) this;
            return casted;
        }
    }


    public <U, Q> Result<U, Q> flatMap(Function<T, Result<U, Q>> successHandler, Function<E, Result<U, Q>> errorHandler) {
        if (isOk()) {
            return successHandler.apply(value);
        } else {
            return errorHandler.apply(error);
        }
    }

    public T orElse(Function<E, T> errorHandler) {
        if(!isOk()) {
            return errorHandler.apply(error);
        } else {
            return value;
        }
    }

    public T orElseGet(Supplier<T> supplier) {
        if(value == null) {
            return supplier.get();
        } else {
            return value;
        }
    }

    @EnsuresNonNullIf(expression = "value", result = true)
    @EnsuresNonNullIf(expression = "error", result = false)
    @SuppressWarnings("contracts.conditional.postcondition")
    public boolean isOk() {
        return value != null;
    }
}
