/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.server.api.common;

import org.activityinfo.model.analysis.Analysis;
import org.activityinfo.model.analysis.Report;
import org.activityinfo.model.analysis.ReportHeader;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.form.*;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formTree.FormTreeBuilder;
import org.activityinfo.model.formula.eval.FieldValueProvider;
import org.activityinfo.model.formula.eval.RecordBasedEvaluator;
import org.activityinfo.model.formula.resolved.FormulaResolver;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.permission.FormPermissions;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.permission.PermissionOracle;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.QueryModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.observable.Observable;
import org.activityinfo.promise.Maybe;
import org.activityinfo.store.query.shared.FormSource;
import org.activityinfo.store.spi.FormSupervisor;
import org.activityinfo.store.spi.FormVersion;
import org.activityinfo.store.spi.QueryTransaction;

import java.util.*;

/**
 * Wraps a {@link QueryTransaction} and adds logic related to permissions.
 */
public class SupervisedQueryTransaction {
    private QueryTransaction tx;
    private Map<ResourceId, Maybe<FormEntry>> formMap = new HashMap<>();
    private Map<ResourceId, FormTree> formTrees = new HashMap<>();


    private class FormEntry {

        private final FormMetadata metadata;
        private final FormClass formClass;
        private final FormPermissions permissions;

        public FormEntry(FormClass formClass, FormPermissions permissions) {
            this.formClass = formClass;
            this.permissions = permissions;
            this.metadata = FormMetadata.of(formClass, permissions);
        }

        private Maybe<TypedFormRecord> getRecord(RecordRef ref) {
            return tx
                .getRecord(ref)
                .map(record -> TypedFormRecord.toTypedFormRecord(metadata.getSchema(), record))
                .map(typedRecord -> checkVisible(typedRecord))
                .orElse(Maybe.notFound());
        }

        private Maybe<TypedFormRecord> checkVisible(TypedFormRecord typedRecord) {
            if(metadata.getPermissions().hasVisibilityFilter()) {
                FormTree formTree = getFormTree(formClass.getId());
                FormulaResolver resolver = new FormulaResolver(formTree);
                ResolvedNode resolvedFilter = resolver.resolve(metadata.getPermissions().getFilter(Operation.VIEW));
                PermissionsValueProvider valueProvider = new PermissionsValueProvider(typedRecord);
                RecordBasedEvaluator evaluator = new RecordBasedEvaluator(formTree.getRootFormId(), valueProvider);
                if (evaluator.evaluate(typedRecord.getRef(), resolvedFilter).waitFor().isTrue(0)) {
                    return Maybe.of(typedRecord);
                } else {
                    return Maybe.forbidden();
                }
            } else {
                return Maybe.of(typedRecord);
            }
        }

        public FormClass getFormClass() {
            return metadata.getSchema();
        }
    }

    public SupervisedQueryTransaction(QueryTransaction tx) {
        this.tx = tx;
    }

    public QueryTransaction getTransaction() {
        return tx;
    }

    private Maybe<FormEntry> form(ResourceId formId) {
        return formMap.computeIfAbsent(formId, id ->
            computeForm(formId));
    }

    private Maybe<FormEntry> computeForm(ResourceId formId) {
        if(VisibleDatabaseTree.isUserForm(formId)) {
            ResourceId databaseId = VisibleDatabaseTree.getDatabaseIdFromUserFormId(formId);
            return tx.getVisibleDatabaseTree(databaseId)
                .flatMap(tree -> tx.getUserFormVersion(databaseId)
                                .map(v -> maybeUserForm(tree, v))
                                .orElse(Maybe.notFound()));

        } else {

            return tx.getFormClass(formId)
                .map(formClass -> tx.getVisibleDatabaseTree(formClass.getDatabaseId())
                    .flatMap(tree -> maybeForm(tree, formClass)))
                .orElse(Maybe.notFound());
        }
    }

    private Maybe<FormEntry> maybeForm(VisibleDatabaseTree tree, FormClass formClass) {
        FormPermissions permissions = PermissionOracle.formPermissions(formClass.getId(), tree);
        if(!permissions.isVisible()) {
            return Maybe.forbidden();
        }
        return Maybe.of(new FormEntry(formClass, permissions));
    }

    private Maybe<FormEntry> maybeUserForm(VisibleDatabaseTree tree, FormVersion version) {
        ResourceId formId = VisibleDatabaseTree.getUserFormIdFromDatabaseId(tree.getDatabaseId());
        FormPermissions permissions = PermissionOracle.userFormPermissions(tree);
        FormClass formClass = UserFormBuilder.userForm(tree, version.getSchemaVersion());

        if(!permissions.isVisible()) {
            return Maybe.forbidden();
        }
        return Maybe.of(new SupervisedQueryTransaction.FormEntry(formClass, permissions));
    }


    public Maybe<VisibleDatabaseTree> getDatabaseTree(ResourceId databaseId) {
        return tx.getVisibleDatabaseTree(databaseId);
    }


    public FormTree getFormTree(ResourceId formId) {
        return formTrees.computeIfAbsent(formId, id -> {
            FormTreeBuilder builder = new FormTreeBuilder(this::formMetadataForTree);
            return builder.queryTree(id);
        });
    }

    public List<ReportHeader> getOwnedOrSharedReports() {
        return tx.getOwnedOrSharedReports();
    }


    /**
     * Returns the user's permissions for a form, if the form exists.
     */
    public Optional<FormPermissions> getFormPermissions(ResourceId formId) {
        return form(formId).switch_(new Maybe.Case<FormEntry, Optional<FormPermissions>>() {
            @Override
            public Optional<FormPermissions> visible(FormEntry value) {
                return Optional.of(value.permissions);
            }

            @Override
            public Optional<FormPermissions> forbidden() {
                return Optional.of(FormPermissions.none());
            }

            @Override
            public Optional<FormPermissions> deleted() {
                return Optional.empty();
            }

            @Override
            public Optional<FormPermissions> notFound() {
                return Optional.empty();
            }
        });
    }

    private FormMetadata formMetadataForTree(ResourceId formId) {
        return form(formId).switch_(new Maybe.Case<FormEntry, FormMetadata>() {
            @Override
            public FormMetadata visible(FormEntry value) {
                return value.metadata;
            }

            @Override
            public FormMetadata forbidden() {
                return FormMetadata.forbidden(formId);
            }

            @Override
            public FormMetadata deleted() {
                return FormMetadata.notFound(formId);
            }

            @Override
            public FormMetadata notFound() {
                return FormMetadata.notFound(formId);
            }
        });
    }

    public Maybe<FormClass> getFormClass(ResourceId formId) {
        return form(formId).transform(form -> form.getFormClass());
    }

    public Maybe<FormMetadata> getFormMetadata(ResourceId formId) {
        return form(formId).transform(form -> form.metadata);
    }

    /**
     * Returns the given record, if it exists, and the user has permission to view the record.
     */
    public Maybe<TypedFormRecord> getRecord(RecordRef ref) {
        return form(ref.getFormId()).flatMap(form -> form.getRecord(ref));
    }

    public FormSource asFormSource() {
        return new FormSourceAdapter();
    }

    public ColumnSet query(QueryModel queryModel) {
        return tx.query(queryModel, new FormSupervisorBandaid());
    }

    /**
     * Provides records values WITHOUT apply permissions for the purpose of evaluating
     * permissions themselves.
     */
    private class PermissionsValueProvider implements FieldValueProvider {

        private final TypedFormRecord record;

        public PermissionsValueProvider(TypedFormRecord record) {
            this.record = record;
        }

        @Override
        public Observable<Optional<RecordRef>> getParentRef(RecordRef recordRef) {
            if(recordRef.equals(record.getRef())) {
                return Observable.just(record.getParentRecordRef());
            } else {
                return Observable.just(getRecord(recordRef).flatMap(record -> record.getParentRecordRef()));
            }
        }

        @Override
        public Observable<Optional<FieldValue>> getFieldValue(RecordRef recordRef, ResourceId fieldId) {
            if(recordRef.equals(record.getRef())) {
                return Observable.just(Optional.ofNullable(record.get(fieldId)));

            } else {
                Optional<FieldValue> fieldValue = getRecord(recordRef)
                    .flatMap(r -> Optional.ofNullable(r.get(fieldId)));

                return Observable.just(fieldValue);
            }
        }

        @Override
        public Observable<Double> getLastEditTime(RecordRef recordRef) {
            if(recordRef.equals(record.getRef())) {
                return Observable.just(record.getLastEditTime());
            } else {
                return Observable.just(getRecord(recordRef)
                    .map(r -> r.getLastEditTime())
                    .orElse(Double.NaN));
            }
        }

        private Optional<TypedFormRecord> getRecord(RecordRef recordRef) {
            Optional<FormRecord> record = tx.getRecord(recordRef);
            Optional<FormClass> formClass = tx.getFormClass(recordRef.getFormId());

            return record.flatMap(r -> formClass.map(f -> TypedFormRecord.toTypedFormRecord(f, r)));
        }
    }

    private class FormSupervisorBandaid implements FormSupervisor {

        @Override
        public FormPermissions getFormPermissions(ResourceId formId) {
            return form(formId).transform(f -> f.permissions).or(FormPermissions.none());
        }

        @Override
        public Map<ResourceId, FormPermissions> getFormPermissions(Set<ResourceId> formIds) {
            Map<ResourceId, FormPermissions> map = new HashMap<>();
            for (ResourceId formId : formIds) {
                map.put(formId, getFormPermissions(formId));
            }
            return map;
        }
    }


    private class FormSourceAdapter implements FormSource {

        @Override
        public Observable<FormTree> getFormTree(ResourceId formId) {
            return Observable.just(SupervisedQueryTransaction.this.getFormTree(formId));
        }

        @Override
        public Observable<Maybe<VisibleDatabaseTree>> getDatabase(ResourceId databaseId) {
            return Observable.just(SupervisedQueryTransaction.this.getDatabaseTree(databaseId));

        }

        @Override
        public Observable<ColumnSet> query(QueryModel queryModel) {
            return Observable.just(SupervisedQueryTransaction.this.query(queryModel));
        }

        @Override
        public Observable<Maybe<Report>> getReport(String id) {
            return Observable.just(Maybe.fromOptional(tx.getReport(ResourceId.valueOf(id))));
        }

        @Override
        public Observable<Maybe<Analysis>> getAnalysis(String reportId, String analysisId) {
            return Observable.just(Maybe.fromOptional(tx.getAnalysis(ResourceId.valueOf(reportId), ResourceId.valueOf(analysisId))));
        }

        @Override
        public Observable<RecordHistory> getFormRecordHistory(RecordRef ref) {
            throw new UnsupportedOperationException("TODO");
        }

        @Override
        public Observable<Maybe<FormRecord>> getRecord(RecordRef recordRef) {
            return Observable.just(SupervisedQueryTransaction.this.getRecord(recordRef).transform(TypedFormRecord::toFormRecord));
        }
    }

}
