/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.server.api.common;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import org.activityinfo.json.Json;
import org.activityinfo.json.JsonMappingException;
import org.activityinfo.json.JsonValue;
import org.activityinfo.model.account.Action;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.form.FormRecord;
import org.activityinfo.model.form.TypedFormRecord;
import org.activityinfo.model.formTree.FormTree;
import org.activityinfo.model.formTree.FormTreeBuilder;
import org.activityinfo.model.formula.eval.FieldValueProvider;
import org.activityinfo.model.formula.eval.RecordBasedEvaluator;
import org.activityinfo.model.formula.resolved.FormulaResolver;
import org.activityinfo.model.formula.resolved.ResolvedNode;
import org.activityinfo.model.permission.FormPermissions;
import org.activityinfo.model.permission.Operation;
import org.activityinfo.model.permission.PermissionOracle;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.resource.*;
import org.activityinfo.model.type.*;
import org.activityinfo.model.type.attachment.Attachment;
import org.activityinfo.model.type.attachment.AttachmentType;
import org.activityinfo.model.type.attachment.AttachmentValue;
import org.activityinfo.model.type.enumerated.EnumItem;
import org.activityinfo.model.type.enumerated.EnumType;
import org.activityinfo.model.type.enumerated.EnumValue;
import org.activityinfo.promise.Maybe;
import org.activityinfo.store.query.server.FormStorage;
import org.activityinfo.store.query.server.InvalidUpdateException;
import org.activityinfo.store.spi.RecordTransaction;
import org.activityinfo.store.spi.TypedRecordUpdate;

import java.time.Instant;
import java.util.*;
import java.util.function.Function;
import java.util.logging.Level;
import java.util.logging.Logger;

import static java.lang.String.format;

/**
 * Validates and authorizes an update received from the client and
 * passes it on the correct {@link FormStorage} implementation.
 *
 * <p>This class enforces all logical permissions related to the form.</p>
 *
 */
public class Updater {

    private static final Logger LOGGER = Logger.getLogger(Updater.class.getName());

    private final RecordTransaction tx;
    private final TransactionMode mode;
    private final Optional<ResourceId> userId;


    public Updater(RecordTransaction tx, TransactionMode mode, ResourceId userId) {
        this.tx = tx;
        this.mode = mode;
        this.userId = Optional.of(userId);
    }

    public Updater(RecordTransaction tx, TransactionMode mode, Optional<ResourceId> userId) {
        this.tx = tx;
        this.mode = mode;
        this.userId = userId;
    }

    /**
     * Parses, validates and executes a {@link RecordUpdateSet} encoded as a JSON object. The object
     * must have a {@code changes} property that takes the value of an array.
     *
     * @throws InvalidUpdateException if the given update is not a validate update.
     */
    public void execute(JsonValue transactionObject) {
        try {
            execute(Json.fromJson(RecordUpdateSet.class, transactionObject));
        } catch (JsonMappingException e) {
            throw new InvalidUpdateException(e);
        }
    }

    /**
     * Executes a {@code RecordTransaction}.
     */
    public void execute(RecordUpdateSet updateSet) {
        try {
            List<RecordUpdate> sortedChanges = new ArrayList<>();
            List<RecordUpdate> leaves = new ArrayList<>();

            Deque<RecordUpdate> noIncomingEdgeNodes = new ArrayDeque<>();

            Set<String> recordIds = new HashSet<>();
            for (RecordUpdate update : updateSet.getChanges()) {
                recordIds.add(update.getRecordId().asString());
            }

            for (RecordUpdate update : updateSet.getChanges()) {
                if (Strings.isNullOrEmpty(update.getParentRecordId())) {
                    noIncomingEdgeNodes.add(update);
                } else if (!recordIds.contains(update.getParentRecordId())) {
                    noIncomingEdgeNodes.add(update);
                } else {
                    leaves.add(update);
                }
            }

            while (!noIncomingEdgeNodes.isEmpty()) {
                RecordUpdate next = noIncomingEdgeNodes.pop();
                sortedChanges.add(next);
                for (int k = 0; k < leaves.size(); k++) {
                    RecordUpdate record = leaves.get(k);
                    if (record.getParentRecordId().equals(next.getRecordId().asString())) {
                        noIncomingEdgeNodes.add(record);
                        leaves.remove(record);
                        k--;
                    }
                }
            }

            for (RecordUpdate change : sortedChanges) {
                executeChange(change);
            }
        } catch (Exception e) {
            throw e;
        }
    }

    public void executeChange(RecordUpdate update) {

        Optional<FormClass> formClass = tx.getFormClass(update.getFormId());

        if(!formClass.isPresent()) {
            throw new InvalidUpdateException(format("Form '%s' does not exist.", update.getFormId()));
        }

        TypedRecordUpdate typedUpdate = parseChange(formClass.get(), update);

        executeUpdate(formClass.get(), typedUpdate);
    }

    @VisibleForTesting
    TypedRecordUpdate parseChange(FormClass formClass, JsonValue changeObject) throws JsonMappingException {
        return parseChange(formClass, Json.fromJson(RecordUpdate.class, changeObject));
    }

    @VisibleForTesting
    TypedRecordUpdate parseChange(FormClass formClass, RecordUpdate changeObject) {

        // Build a lookup map to resolve the field name used in the JSON request.
        Map<String, FormField> idMap = new HashMap<>();
        Multimap<String, FormField> codeMap = HashMultimap.create();

        for (FormField formField : formClass.getFields()) {
            idMap.put(formField.getId().asString(), formField);
            codeMap.put(formField.getCode(), formField);
        }

        TypedRecordUpdate update = new TypedRecordUpdate();
        update.setUserId(userId);
        update.setFormId(formClass.getId());
        update.setRecordId(changeObject.getRecordId());
        update.setDeleted(changeObject.isDeleted());
        update.setLastEditTime(Instant.now());
        if(changeObject.getParentRecordId() != null) {
            update.setParentId(ResourceId.valueOf(changeObject.getParentRecordId()));
        }

        if(changeObject.getFields() != null) {
            for (Map.Entry<String, JsonValue> change : changeObject.getFields().entrySet()) {

                // Resolve what might be a human readable field name to the FormField
                // We accept FormField codes and ids
                findFieldByIdOrCode(idMap, codeMap, change.getKey()).ifPresent(field -> {

                    // Now use the type information to parse the JSON element
                    FieldValue fieldValue;
                    try {
                        fieldValue = parseFieldValue(field, change.getValue());
                    } catch (Exception e) {
                        throw new InvalidUpdateException(format("Invalid value for field '%s' (id: %s, code: %s): %s",
                            field.getLabel(),
                            field.getId(),
                            field.getCode(),
                            e.getMessage()), e);
                    }

                    update.set(field.getId(), validateType(field, fieldValue));
                });
            }
        }
        return update;
    }

    private Optional<FormField> findFieldByIdOrCode(Map<String, FormField> idMap, Multimap<String, FormField> codeMap, String fieldName) {
        if (idMap.containsKey(fieldName)) {
            return Optional.of(idMap.get(fieldName));
        }

        Collection<FormField> byCode = codeMap.get(fieldName);
        if (byCode.isEmpty() && mode == TransactionMode.STRICT) {
            throw new InvalidUpdateException("Unknown field '" + fieldName + "'.");
        }

        if (byCode.size() > 1) {
            throw new InvalidUpdateException("Ambiguous field code '" + fieldName + "'");
        }

        return byCode.stream().findFirst();
    }

    private static FieldValue parseFieldValue(FormField field, JsonValue jsonValue) {
        if(jsonValue.isJsonNull()) {
            return null;
        } else if(field.getType() instanceof EnumType) {
            return parseEnumValue((EnumType)field.getType(), jsonValue);
        } else {
            return field.getType().parseJsonValue(jsonValue);
        }
    }

    private static FieldValue parseEnumValue(EnumType type, JsonValue JsonValue) {

        Set<ResourceId> itemIds = new HashSet<>();

        if(JsonValue.isJsonPrimitive()) {
            itemIds.add(parseEnumId(type, JsonValue.asString()));
        } else if(JsonValue.isJsonArray()) {
            for (JsonValue element : JsonValue.values()) {
                itemIds.add(parseEnumId(type, element.asString()));
            }
        }
        if(type.getCardinality() == Cardinality.SINGLE && itemIds.size() > 1) {
            throw new InvalidUpdateException("Field with SINGLE enum type has multiple values.");
        }

        return new EnumValue(itemIds);
    }

    private static ResourceId parseEnumId(EnumType type, String item) {

        for (EnumItem enumItem : type.getItems()) {
            if(enumItem.getId().asString().equals(item)) {
                return enumItem.getId();
            }
        }
        for (EnumItem enumItem : type.getItems()) {
            if(enumItem.getLabel().equals(item)) {
                return enumItem.getId();
            }
        }

        throw new InvalidUpdateException(format("Invalid enum value '%s', expected one of: %s",
            item, Joiner.on(", ").join(type.getItems())));
    }


    public void execute(TypedRecordUpdate update, RecordUpdateSet updateSet) {
        if(update.getFormId() == null) {
            throw new IllegalArgumentException("No formId provided.");
        }
        Optional<FormClass> storage = tx.getFormClass(update.getFormId());
        if(!storage.isPresent()) {
            if(mode == TransactionMode.OFFLINE) {
                LOGGER.warning("Offline update to missing form " + update.getFormId());
                return;
            } else {
                throw new InvalidUpdateException("No such resource: " + update.getRecordId());
            }
        }

        executeUpdate(storage.get(), update);
    }

    public void executeUpdate(FormClass formClass, TypedRecordUpdate update) {

        Preconditions.checkNotNull(update.getFormId());

        Optional<FormRecord> existingRecord = findExistingRecord(update);

        if(update.isDeleted() && !existingRecord.isPresent() && mode == TransactionMode.OFFLINE) {
            LOGGER.warning("Offline deletion of non-existant record " + update.getRef());
            return;
        }

        validateUpdate(formClass, update, existingRecord);

        authorizeUpdate(formClass, existingRecord, update);

        generateSerialNumbers(formClass, existingRecord, update);

        applyUpdate(update, existingRecord);
    }

    private Optional<FormRecord> findExistingRecord(TypedRecordUpdate update) {
        if(update.isDefinitelyKnownToBeANewRecord()) {
            return Optional.empty();
        } else {
            return tx.getRecord(update.getRef());
        }
    }

    private void validateUpdate(FormClass formClass, TypedRecordUpdate update, Optional<FormRecord> existingRecord) {
        if(existingRecord.isPresent() && mode == TransactionMode.COLLECTION_LINK) {
            throw new InvalidUpdateException("Submissions via collection link may not update existing records.");
        }

        if(update.isDeleted() && update.getChangedFieldValues().size() > 0) {
            throw new InvalidUpdateException("A deletion may not include field value updates.");
        }
        if (update.isDeleted() && !existingRecord.isPresent()) {
            throw new InvalidUpdateException("Record " + update.getRef() + " does not exist.");
        }

        if(formClass.isSubForm() && !existingRecord.isPresent() &&
            !update.getParentId().isPresent()) {
            throw new InvalidUpdateException("Missing parentRecordId");
        }
    }

    private void applyUpdate(TypedRecordUpdate update, Optional<FormRecord> existingRecord) {
        if(update.isDeleted()) {
            tx.deleteRecord(update.getRef());
            tx.publishAction(Action.DELETE_RECORD, update.getFormId());

        } else if(existingRecord.isPresent()) {
            tx.updateRecord(update);
            tx.publishAction(Action.UPDATE_RECORD, update.getFormId());

        } else {
            tx.addRecord(update);
            tx.publishAction(Action.ADD_RECORD, update.getFormId());
        }
    }

    private void generateSerialNumbers(FormClass formClass,
                                       Optional<FormRecord> existingRecord,
                                       TypedRecordUpdate update) {

        TypedFormRecord effectiveRecord = computeEffectiveRecord(formClass, existingRecord, update);

        for (FormField formField : formClass.getFields()) {
            if(formField.getType() instanceof SerialNumberType) {
                if(update.getChangedFieldValues().containsKey(formField.getId())) {
                    updateSerialNumber(formClass, formField, update);
                } else if(!effectiveRecord.getFieldValueMap().containsKey(formField.getId())) {
                    generateSerialNumber(formClass, formField, effectiveRecord, update);
                }
            }
        }
    }

    @VisibleForTesting
    void generateSerialNumber(FormClass formClass,
                              FormField formField,
                              TypedFormRecord effectiveRecord,
                              TypedRecordUpdate update) {

        SerialNumberType type = (SerialNumberType) formField.getType();
        String prefix = computeSerialNumberPrefix(formClass, type, effectiveRecord);

        int serialNumber = tx.nextSerialNumber(formClass.getId(), formField.getId(), prefix);

        update.set(formField.getId(), new SerialNumber(prefix, serialNumber));
    }

    private void updateSerialNumber(FormClass formClass, FormField formField, TypedRecordUpdate update) {

        SerialNumber fieldValue = (SerialNumber) update.getChangedFieldValues().get(formField.getId());
        if(fieldValue == null) {
            throw new InvalidUpdateException("Serial number field " + formField.getId() + " cannot be set to blank.");
        }

        tx.updateSerialNumberSequence(formClass.getId(), formField.getId(), fieldValue.getPrefix(), fieldValue.getNumber());
    }

    private String computeSerialNumberPrefix(
            FormClass formClass,
            SerialNumberType type,
            TypedFormRecord effectiveRecord) {

        if(!type.hasPrefix()) {
            return null;
        }

        try {
            FormTreeBuilder builder = new FormTreeBuilder(new UnsupervisedFormClassProvider(tx));
            FormTree formTree = builder.queryTree(formClass.getId());

            FormulaResolver resolver = new FormulaResolver(formTree);
            ResolvedNode resolvedPrefix = resolver.resolve(type.getPrefixFormula());
            if(!resolvedPrefix.isValid()) {
                LOGGER.log(Level.SEVERE, "Invalid serial number prefix: " + type.getPrefixFormula());
                return null;
            }

            FieldValueProvider fieldValueProvider = new UpdatingValueProvider(tx, effectiveRecord);

            RecordBasedEvaluator rbe = new RecordBasedEvaluator(formTree.getRootFormId(), fieldValueProvider);

            return rbe.evaluate(effectiveRecord.getRef(), resolvedPrefix)
                .waitFor()
                .getString(0);

        } catch (Exception e) {
            LOGGER.log(Level.SEVERE, "Failed to compute prefix for serial number", e);
            return null;
        }
    }

    @VisibleForTesting
    TypedFormRecord computeEffectiveRecord(FormClass formClass, Optional<FormRecord> existingRecord, TypedRecordUpdate update) {
        TypedFormRecord record = new TypedFormRecord(update.getRecordId(), formClass.getId());
        if(existingRecord.isPresent()) {
            TypedFormRecord existingTypedRecord = TypedFormRecord.toTypedFormRecord(formClass, existingRecord.get());
            record.setAll(existingTypedRecord.getFieldValueMap());
            formClass.getParentFormId().map(parentFormId ->
                record.setParentRecordRef(new RecordRef(parentFormId, existingRecord.get().getParentRecordId())));
        } else {
            formClass.getParentFormId().map(parentFormId ->
                    update.getParentId().map(parentRecordId ->
                            record.setParentRecordRef(new RecordRef(parentFormId, parentRecordId))));
        }
        for (Map.Entry<ResourceId, FieldValue> entry : update.getChangedFieldValues().entrySet()) {
            record.set(entry.getKey(), entry.getValue());
        }

        return record;
    }


    /**
     * Verify that all required fields are provided for new resources
     */
    private static void validateRequiredFields(FormClass formClass, Optional<FormRecord> existingResource, TypedRecordUpdate update) {
        if(!existingResource.isPresent()) {
            for (FormField formField : formClass.getFields()) {
                if (formField.isRequired() &&
                    formField.isDataEntryVisible() &&
                    canBeUpdated(formField) && !isProvided(formField, existingResource, update)) {
                    throw new InvalidUpdateException("Required field '%s' [%s] is missing from record with schema %s",
                        formField.getCode(), formField.getId(), formClass.getId().asString());
                }
            }
        }
    }

    private static boolean canBeUpdated(FormField formField) {
        return formField.getType().isUpdatable() || formField.getType() instanceof SerialNumberType;
    }

    private static boolean isProvided(FormField formField, Optional<FormRecord> existingResource, TypedRecordUpdate update) {

        if(update.getChangedFieldValues().containsKey(formField.getId())) {

            // This update includes an explicit update for the given field.
            // The updated value must not be null.

            FieldValue updatedValue = update.getChangedFieldValues().get(formField.getId());
            return updatedValue != null;

        } else {
            // This update does *not* include an updated value for this required field.
            // This is only possible if this is indeed and update and not a new record.

            return existingResource.isPresent();
        }
    }

    private void authorizeUpdate(FormClass formClass,
                                 Optional<FormRecord> existingResource,
                                 TypedRecordUpdate update) {


        Preconditions.checkNotNull(update.getFormId());
        if (mode == TransactionMode.COLLECTION_LINK) {
            return;
        }

        // Check form-level permissions
        Maybe<VisibleDatabaseTree> visible = tx.getVisibleDatabaseTree(formClass.getDatabaseId());
        if (!visible.isVisible()) {
            throw new InvalidUpdateException("Database is not found or you do not have permission to access it.");
        }

        FormPermissions formPermissions = PermissionOracle.formPermissions(formClass.getId(), visible.get());

        // Verify that the user has the right to modify the *existing* record

        Optional<TypedFormRecord> existingTypedRecord = existingResource.map(new Function<FormRecord, TypedFormRecord>() {
            @Override
            public TypedFormRecord apply(FormRecord record) {
                return TypedFormRecord.toTypedFormRecord(formClass, record);
            }
        });

        if (existingTypedRecord.isPresent()) {
            if (update.isDeleted()) {
                authorize(existingTypedRecord.get(), formPermissions, formClass, update, Operation.DELETE_RECORD);
            } else {
                // Evaluate rule for both the *old* record, as well as the *proposed* updated.
                // For example, if the user is only allowed to edit records where AGE > 18, we have to very if that
                // AGE > 18 for the existing record, because otherwise we wouldn't be allowed to edit that record, AND
                // that the user is not trying to set the AGE to 14, for example, because then they are not allowed to
                // edit records where AGE <= 18.
                authorize(existingTypedRecord.get(), formPermissions, formClass, update, Operation.EDIT_RECORD);
                authorize(applyUpdates(formClass, existingTypedRecord, update), formPermissions, formClass,update, Operation.EDIT_RECORD);
            }
        } else {
            authorize(applyUpdates(formClass, existingTypedRecord, update), formPermissions, formClass, update, Operation.ADD_RECORD);
        }

        // Check field-level permissions
        for (Map.Entry<ResourceId, FieldValue> change : update.getChangedFieldValues().entrySet()) {
            if(change.getValue() instanceof AttachmentValue) {
                FormField field = formClass.getField(change.getKey());
                checkBlobPermissions(field, existingResource, (AttachmentValue) change.getValue());
            }
        }
    }

    private void authorize(TypedFormRecord record, FormPermissions formPermissions, FormClass formClass, TypedRecordUpdate update, Operation operation) {

        if (mode == TransactionMode.COLLECTION_LINK) {
            return;
        }

        if (!formPermissions.isAllowed(operation)) {
            throw new InvalidUpdateException("Unauthorized " + operation.toString());
        }

        authorizeFieldEditing(formPermissions, formClass, update, operation);

        if (!formPermissions.isFiltered(operation)) {
            return;
        }

        FormTreeBuilder builder = new FormTreeBuilder(new UnsupervisedFormClassProvider(tx));
        FormTree formTree = builder.queryTree(formClass.getId());

        String filter = formPermissions.getFilter(operation);

        LOGGER.info("Evaluating permission filter: " + filter);

        ResolvedNode resolvedFilter = FormulaResolver.resolve(formTree, filter);

        LOGGER.info("Resolved permission filter: " + resolvedFilter);

        if(!resolvedFilter.isValid()) {
            LOGGER.severe("Invalid permission filter: " + filter);
            throw new InvalidUpdateException("Permission filter is invalid; record cannot be updated. Contact the database owner");
        }

        UpdatingValueProvider valueProvider = new UpdatingValueProvider(tx, record);

        RecordBasedEvaluator evaluator = new RecordBasedEvaluator(formTree.getRootFormId(), valueProvider);
        ColumnView result = evaluator.evaluate(record.getRef(), resolvedFilter).waitFor();

        if(result.getBoolean(0) != ColumnView.TRUE) {
            throw new InvalidUpdateException("Unauthorized " + operation.toString());
        }
    }

    private void authorizeFieldEditing(FormPermissions formPermissions, FormClass formClass, TypedRecordUpdate update, Operation operation) {
        Set<ResourceId> securityCategories = formPermissions.getSecurityCategories();
        for (FormField field: formClass.getFields()) {
            if (!formPermissions.isFieldEditable(field) &&
                    update.getChangedFieldValues().containsKey(field.getId())) {
                throw new InvalidUpdateException("You do not have permission to edit the value of \""
                        + field.getLabel() + "\" field");
            }
        }
    }

    private TypedFormRecord applyUpdates(FormClass formClass, Optional<TypedFormRecord> existingRecord, TypedRecordUpdate update) {
        TypedFormRecord updated = new TypedFormRecord(update.getRecordId(), update.getFormId());
        updated.setParentRecordRef(existingRecord.flatMap(TypedFormRecord::getParentRecordRef));

        update.getParentId().ifPresent(newParentRecordId -> {
            assertUnchanged(existingRecord, newParentRecordId);
            updated.setParentRecordRef(new RecordRef(formClass.getParentFormId().get(), newParentRecordId));
        });

        if(existingRecord.isPresent()) {
            updated.setAll(existingRecord.get().getFieldValueMap());
        }

        updated.setAll(update.getChangedFieldValues());

        return updated;
    }

    private void assertUnchanged(Optional<TypedFormRecord> existingRecord, ResourceId newParentRecordId) {
        existingRecord.ifPresent(existing -> {
            // if there was an existing record, then the parent id cannot be different
            if (!existing.getParentRecordId().equals(java.util.Optional.of(newParentRecordId))) {
                throw new InvalidUpdateException("A record's parentId cannot be changed");
            }
        });
    }


    private static FieldValue validateType(FormField field, FieldValue updatedValue) {
        Preconditions.checkNotNull(field);

        if(updatedValue != null) {
            if ( !canBeUpdated(field)) {
                throw new InvalidUpdateException(
                    format("Field %s ('%s') is a field of type '%s' and its value cannot be set. Found %s",
                        field.getId(),
                        field.getLabel(),
                        field.getType().getTypeClass().getId(),
                        updatedValue));
            }
            if (!field.getType().getTypeClass().equals(updatedValue.getTypeClass())) {
                throw new InvalidUpdateException(
                    format("Updated value for field %s ('%s') has invalid type. Expected %s, found %s.",
                        field.getId(),
                        field.getLabel(),
                        field.getType().getTypeClass().getId(),
                        updatedValue.getTypeClass().getId()));
            }
        }

        return updatedValue;
    }

    /**
     * Verifies that the user has permission to associate the given blob with this record.
     *
     * <p>Updating blob-valued fields is done by the user in two steps. First, the user uploads a file and
     * receives a unique id for the blob. Then, the user updates a record's field with the blob's unique id. </p>
     *
     * <p>Once the blob is associated with the record, then any user with permission to view the record is extended
     * permission to view the blob. This opens an avenue of attack where by an attacker with seeks to obtain access
     * to a blob with some id by assigning it to an unrelated record to which they have access.</p>
     *
     * <p>For this reason, only users who originally uploaded the blob may assign the blob to a record's field value.</p>
     */
    private void checkBlobPermissions(
        FormField field,
        Optional<FormRecord> existingResource,
        AttachmentValue updatedValue) {

        AttachmentType fieldType = (AttachmentType) field.getType();

        // Identity the blob ids that are already associated with this record
        Set<String> existingBlobIds = new HashSet<>();
        if(existingResource.isPresent()) {
            JsonValue existingFieldValue = existingResource.get().getFields().get(field.getId().asString());
            if(!existingFieldValue.isJsonNull()) {
                AttachmentValue existingValue = fieldType.parseJsonValue(existingFieldValue);
                for (Attachment attachment : existingValue.getValues()) {
                    existingBlobIds.add(attachment.getBlobId());
                }
            }
        }
    }


    public void execute(TypedFormRecord typedFormRecord) {
        RecordUpdateSet tx = new RecordUpdateSetBuilder()
            .create(typedFormRecord)
            .build();

        execute(tx);
    }


    public void create(ResourceId formId, JsonValue jsonObject) {
        String id = jsonObject.get("id").asString();
        createOrUpdate(formId, ResourceId.valueOf(id), jsonObject, true);
    }

    public void execute(ResourceId formId, ResourceId recordId, JsonValue jsonObject) {
        createOrUpdate(formId, recordId, jsonObject, false);
    }

    @SuppressWarnings({"checkstyle:cyclomaticcomplexity", "OverlyComplexMethod"})
    private void createOrUpdate(ResourceId formId, ResourceId recordId, JsonValue jsonObject, boolean create) {

        FormClass formClass = tx.getFormClass(formId)
            .orElseThrow(() -> new InvalidUpdateException("No such formId: " + formId));

        TypedRecordUpdate update = new TypedRecordUpdate();
        update.setUserId(userId);
        update.setFormId(formId);
        update.setRecordId(recordId);

        if(jsonObject.hasKey("deleted") && !jsonObject.get("deleted").isJsonNull()) {
            update.setDeleted(jsonObject.get("deleted").asBoolean());
        }

        if (jsonObject.hasKey("parentRecordId") && !jsonObject.get("parentRecordId").isJsonNull()) {
            update.setParentId(ResourceId.valueOf(jsonObject.getString("parentRecordId")));
        }

        JsonValue fieldValues;
        if(jsonObject.hasKey("fieldValues")) {
            fieldValues = jsonObject.get("fieldValues");
        } else if(jsonObject.hasKey("fields")) {
            fieldValues = jsonObject.get("fields");
        } else {
            fieldValues = Json.createObject();
        }
        for (FormField formField : formClass.getFields()) {
            if(canBeUpdated(formField)) {
                if (fieldValues.hasKey(formField.getName())) {
                    JsonValue updatedValueElement = fieldValues.get(formField.getName());
                    FieldValue updatedValue;
                    if(updatedValueElement.isJsonNull()) {
                        updatedValue = null;
                    } else {
                        try {
                            updatedValue = formField.getType().parseJsonValue(updatedValueElement);
                        } catch(Exception e) {
                            e.printStackTrace();
                            throw new InvalidUpdateException("Could not parse updated value for field " +
                                formField.getId() + ": " + e.getMessage());
                        }
                    }
                    update.getChangedFieldValues().put(formField.getId(), updatedValue);

                } else if (create) {
                    update.getChangedFieldValues().put(formField.getId(), null);
                }
            }
        }

        executeUpdate(formClass, update);
    }
}
