/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.server.api.common;

import org.activityinfo.model.form.TypedFormRecord;
import org.activityinfo.model.formula.eval.FieldValueProvider;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.observable.Observable;
import org.activityinfo.store.spi.RecordTransaction;

import java.util.Optional;

/**
 * Provides field values from records within the context of an update transaction.
 *
 * <p>Depending on the type of the rule, this may be evaluated based on the existing values of the record, for
 * example, permission rules need to be evaluated to see if the user is authorized to edit an existing record, or based
 * on the record that is about to be updated.
 */
class UpdatingValueProvider implements FieldValueProvider {

    private final RecordTransaction tx;
    private final TypedFormRecord currentRecord;

    public UpdatingValueProvider(RecordTransaction tx, TypedFormRecord currentRecord) {
        this.tx = tx;
        this.currentRecord = currentRecord;
    }

    @Override
    public Observable<Optional<RecordRef>> getParentRef(RecordRef recordRef) {
        if(recordRef.equals(currentRecord.getRef())) {
            return Observable.just(currentRecord.getParentRecordRef());
        } else {
            return Observable.just(getRecord(recordRef).flatMap(record -> record.getParentRecordRef()));
        }
    }

    @Override
    public Observable<Optional<FieldValue>> getFieldValue(RecordRef recordRef, ResourceId fieldId) {
        if(recordRef.equals(currentRecord.getRef())) {
            return Observable.just(Optional.ofNullable(currentRecord.get(fieldId)));

        } else {
            Optional<FieldValue> fieldValue = getRecord(recordRef)
                .flatMap(r -> Optional.ofNullable(r.get(fieldId)));

            return Observable.just(fieldValue);
        }
    }

    @Override
    public Observable<Double> getLastEditTime(RecordRef recordRef) {
        if(recordRef.equals(currentRecord.getRef())) {
            return Observable.just(currentRecord.getLastEditTime());
        } else {
            return Observable.just(getRecord(recordRef)
                .map(r -> r.getLastEditTime())
                .orElse(Double.NaN));
        }
    }

    private Optional<TypedFormRecord> getRecord(RecordRef recordRef) {
        return tx.getFormClass(recordRef.getFormId())
            .flatMap(schema ->
                tx.getRecord(recordRef)
                    .map(r -> TypedFormRecord.toTypedFormRecord(schema, r)));
    }
}
