/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.query.server.columns;

import com.google.common.primitives.UnsignedBytes;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.SortDir;
import org.activityinfo.model.util.HeapsortColumn;

/**
 * Compact ColumnView for numbers are all integers and have a range of less than 255
 */
public class IntColumnView8 extends AbstractNumberColumn {

    static final int MAX_RANGE = 255;

    private byte[] values;
    private int delta;

    public IntColumnView8(double doubleValues[], int numRows, int minValue) {
        this.values = new byte[numRows];

        // Reserve 0 for missing values
        this.delta = minValue - 1;

        for (int i = 0; i < numRows; i++) {
            double doubleValue = doubleValues[i];
            if(!Double.isNaN(doubleValue)) {
                values[i] = (byte)(doubleValue - this.delta);
            }
        }
    }

    private IntColumnView8(byte[] values, int delta) {
        this.values = values;
        this.delta = delta;
    }

    @Override
    public int numRows() {
        return values.length;
    }

    @Override
    public double getDouble(int row) {
        byte byteValue = values[row];
        if(byteValue == 0) {
            return Double.NaN;
        } else {
            return delta + UnsignedBytes.toInt(byteValue);
        }
    }

    @Override
    public boolean isBlank(int row) {
        return values[row] == 0;
    }

    @Override
    public ColumnView select(int[] selectedRows) {
        byte[] selectedValues = new byte[selectedRows.length];
        for (int i = 0; i < selectedRows.length; i++) {
            int selectedRow = selectedRows[i];
            if(selectedRow != -1) {
                selectedValues[i] = this.values[selectedRow];
            }
        }
        return new IntColumnView8(selectedValues, delta);
    }

    @Override
    public int[] order(int[] sortVector, SortDir direction, int[] range) {
        int numRows = values.length;
        if (range == null || range.length == numRows) {
            HeapsortColumn.heapsortByte(values, sortVector, numRows,
                    HeapsortColumn.withDirection(IntColumnView8::isLessThan, direction));
        } else {
            HeapsortColumn.heapsortByte(values, sortVector, range.length, range,
                    HeapsortColumn.withDirection(IntColumnView8::isLessThan, direction));
        }
        return sortVector;
    }

    /**
     * Given two numbers encoded as unsigned bytes from 0x01-0xFF, with missing values encoded as zeroes
     *
     */
    private static boolean isLessThan(byte bx, byte by) {
        // Treat as unsigned
        int x = UnsignedBytes.toInt(bx);
        int y = UnsignedBytes.toInt(by);

        // Missing values encoded as zeroes
        boolean xMissing = (x == 0);
        boolean yMissing = (y == 0);

        if(xMissing && !yMissing) {
            return true;
        }
        return x < y;
    }


}
