/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.query.shared;

import org.activityinfo.i18n.shared.I18N;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormField;
import org.activityinfo.model.form.SubFormKind;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.type.FieldValue;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.model.type.time.PeriodType;

import java.util.ListIterator;
import java.util.Optional;
import java.util.logging.Logger;

public class SubFormPatch {

    private static final Logger LOGGER = Logger.getLogger(SubFormPatch.class.getName());

    public static final ResourceId PERIOD_FIELD_ID = ResourceId.valueOf("period");

    public static FormClass patch(FormClass formClass) {
        if(formClass.isSubForm()) {
            if(formClass.getSubFormKind().isPeriod()) {
                return ensureSubFormHasPeriodField(
                    removeExtraSubFormPeriodField(formClass));
            }
        }
        return formClass;
    }

    private static FormClass ensureSubFormHasPeriodField(FormClass formClass) {
        Optional<FormField> period = formClass.getFieldIfPresent(PERIOD_FIELD_ID);
        if(period.isPresent()) {
            return formClass;
        }

        // Add period field
        FormField periodField = new FormField(ResourceId.valueOf("period"));
        periodField.setLabel(fieldName(formClass.getSubFormKind()));
        periodField.setType(formClass.getSubFormKind().getPeriodType());
        periodField.setKey(true);
        periodField.setRequired(true);

        formClass.getElements().add(0, periodField);


        return formClass;
    }

    /**
     * Previously, when forms were cloned, the 'period' field was cloned along with it but
     * given a new id in the form 'p34243423'. The routine above then added another field with the
     * id 'period'. This removes the extra field.
     *
     */
    private static FormClass removeExtraSubFormPeriodField(FormClass formClass) {
        ListIterator<FormField> it = formClass.getElements().listIterator();
        while(it.hasNext()) {
            FormField field = it.next();
            if (field.getType().equals(formClass.getSubFormKind().getPeriodType()) &&
                field.isKey() &&
                !field.getId().equals(PERIOD_FIELD_ID) &&
                 field.getId().asString().startsWith("p")) {

                it.remove();
            }
        }
        return formClass;
    }

    private static String fieldName(SubFormKind kind) {
        switch (kind) {
            case MONTHLY:
                return I18N.CONSTANTS.month();
            case WEEKLY:
                return I18N.CONSTANTS.weekFieldLabel();
            case BIWEEKLY:
                return I18N.CONSTANTS.fortnight();
            case DAILY:
            default:
                return I18N.CONSTANTS.date();
        }
    }

    public static CursorObserver<ResourceId> fromRecordId(FormClass formClass, final CursorObserver<FieldValue> observer) {
        final ResourceId formId = formClass.getId();
        final PeriodType periodType = formClass.getSubFormKind().getPeriodType();

        return new CursorObserver<ResourceId>() {
            @Override
            public void onNext(ResourceId value) {
                observer.onNext(periodType.fromSubFormKey(new RecordRef(formId, value)));
            }

            @Override
            public void done() {
                observer.done();
            }
        };
    }
}
