/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.query.shared.columns;

import com.google.common.base.Strings;
import org.activityinfo.model.query.ColumnType;
import org.activityinfo.model.query.ColumnView;
import org.activityinfo.model.query.SortDir;
import org.activityinfo.model.util.HeapsortColumn;

import java.util.BitSet;

public class MultiDiscreteStringColumnView extends ColumnView {

    public static final String SEPARATOR = ", ";

    private String[] labels;
    private BitSet[] selected;
    private int numRows;

    public MultiDiscreteStringColumnView(int numRows, String[] labels, BitSet[] selected) {
        this.numRows = numRows;
        this.labels = labels;
        this.selected = selected;
    }


    @Override
    public ColumnType getType() {
        return ColumnType.STRING;
    }

    @Override
    public int numRows() {
        return numRows;
    }

    @Override
    public String get(int row) {
        return getString(row);
    }

    @Override
    public double getDouble(int row) {
        return Double.NaN;
    }

    @Override
    public String getString(int row) {
        return concatenateValues(labels, row);
    }

    private String concatenateValues(String[] array, int row) {
        StringBuilder value = new StringBuilder();
        for(int i=0; i<array.length; i++) {
            if (selected[i].get(row)) {
                add(value, array[i]);
            }
        }
        return Strings.emptyToNull(value.toString());
    }

    private StringBuilder add(StringBuilder array, String value) {
        if (array.length() > 0) {
            array.append(SEPARATOR);
        }
        return array.append(value);
    }

    @Override
    public int getBoolean(int row) {
        return BLANK;
    }

    @Override
    public boolean isBlank(int row) {
        for (BitSet selection : selected) {
            if(selection.get(row)) {
                return false;
            }
        }
        return true;
    }

    @Override
    public ColumnView select(int[] selectedRows) {
        BitSet[] filtered = new BitSet[labels.length];
        for (int i=0; i < filtered.length; i++) {
            filtered[i] = new BitSet();
        }

        for (int i = 0; i < selectedRows.length; i++) {
            int selectedRow = selectedRows[i];
            if(selectedRow == -1) {
                for (int j=0; j < filtered.length; j++) {
                    filtered[j].clear(i);
                }
            } else {
                for (int j=0; j < filtered.length; j++) {
                    filtered[j].set(i, selected[j].get(selectedRow));
                }
            }
        }

        return new MultiDiscreteStringColumnView(selectedRows.length, labels, filtered);
    }

    @Override
    public String toString() {  
        return getClass().getSimpleName() + "{numRows=" + numRows() + "}";
    }

    @Override
    public int[] order(int[] sortVector, SortDir direction, int[] range) {
        String[] concatenatedSelections = new String[numRows];
        for (int i=0; i<concatenatedSelections.length; i++) {
            concatenatedSelections[i] = getString(i);
        }

        if (range == null || range.length == numRows) {
            HeapsortColumn.heapsortString(concatenatedSelections, sortVector, numRows, direction == SortDir.ASC);
        } else {
            HeapsortColumn.heapsortString(concatenatedSelections, sortVector, range.length, range, direction == SortDir.ASC);
        }

        return sortVector;
    }
}
