/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.spi;

import org.activityinfo.model.account.Action;
import org.activityinfo.model.account.AuthenticatedUser;
import org.activityinfo.model.account.BillingAccount;
import org.activityinfo.model.account.UserAccount;
import org.activityinfo.model.collectionLink.CollectionLinkStatus;
import org.activityinfo.model.database.*;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.promise.Maybe;

import java.time.Instant;
import java.util.*;
import java.util.function.Function;

/**
 * Queries and modifications that must be applied atomically to a database.
 */
public interface DatabaseTransaction {

    ResourceId getDatabaseId();

    DatabaseHeader getDatabaseHeader();

    DatabaseTree getDatabaseTree();

    Maybe<VisibleDatabaseTree> getVisibleDatabaseTree(ResourceId userId);

    void updateLabel(String label);

    void updateDescription(String description);

    void updateResources(Collection<Resource> resources);

    void updateLocks(Collection<RecordLock> resources);

    void deleteDatabase();

    /**
     * Increments the version of the DatabaseTree and returns the new version number
     */
    long incrementDatabaseVersion();

    List<DatabaseUser> getDatabaseUsers();

    Optional<DatabaseUser> getDatabaseUser(ResourceId userId);

    Optional<DatabaseUserGrantSet> getDatabaseUserGrantSet(ResourceId userId);

    List<DatabaseUserGrantSet> getUsersWithGrants(List<ResourceId> resources);

    void addUser(AuthenticatedUser authenticatedUser, DatabaseUserGrantSet newUser);

    void addInvite(String token, AuthenticatedUser invitingUser, ResourceId userId, Instant time);

    DatabaseUserGrantSet updateGrants(AuthenticatedUser authenticatedUser, ResourceId userId, Collection<DatabaseUserGrant> grants);

    void deleteUser(AuthenticatedUser authenticatedUser, ResourceId userId);

    /**
     * Adds a new form belonging to this database.
     *
     * <p>Note that this only stores a form's metadata, callers must also add the form to the database tree.</p>
     *
     * @param schema the form's schema
     * @throws IllegalStateException if a form with this id already exists.
     */
    void addForm(FormClass schema);

    Map<ResourceId, FormClass> getFormSchemas(Set<ResourceId> formIds);

    /**
     * Updates the form's schema, and increments the form's schema version number.
     * @param updatedSchema the updated form schema
     * @return a copy of the {@code updatedSchema} with updated schema version
     */
    FormClass updateForm(FormClass updatedSchema);

    void updateRoles(Collection<Role> updatedRoles);

    void updateSecurityCategories(Collection<SecurityCategory> securityCategories);

    void updateOwner(AuthenticatedUser executingUser, ResourceId newOwnerId);

    DatabaseUserGrantSet updateUserRole(AuthenticatedUser authenticatedUser, ResourceId userId, RoleAssignment newAssignment);

    void addDatabaseUpdateLogEntry(ResourceId userId, long version, UpdateDatabaseRequest request, DatabaseTree originalTree);

    void addUserUpdateLogEntry(ResourceId executingUserId, DatabaseUserGrantSet newUserGrantSet);

    void addUserUpdateLogEntry(AuthenticatedUser executingUser, DatabaseUserGrantSet previousGrantSet, DatabaseUserGrantSet newGrantSet);

    void addDeleteUserLogEntry(AuthenticatedUser executingUser, DatabaseUserGrantSet deletedUser);

    void markDatabaseUpdateReverted(long version);

    void markUserUpdateReverted(long changeId);


    /**
     * Retrieves the update that led to the given {@code version}
     */
    Optional<UpdateDatabaseRequest> getUpdate(long version);

    Optional<DatabaseTree> getDatabaseTree(long version);

    /**
     * Try to recover forms that have been previously deleted.
     */
    void recoverDeletedForms(Set<ResourceId> addedForms);

    boolean isUserAssignedWithRole(Collection<ResourceId> roleIds);

    <T> T updateOrAddUser(String userEmail, Function<UpdateUserTransaction, T> mutator);

    <T> T updateUser(String email, Function<UpdateUserTransaction, T> mutator, Function<UpdateUserError, T> errorHandler);

    Optional<UserAccount> getUserAccount(ResourceId userId);

    BillingAccount getDatabaseBillingAccount();

    void addCollectionLink(ResourceId formId, String linkId, CollectionLinkStatus status, String locale, String label);

    void publishEvent(AuthenticatedUser executingUser, Action action);

    void updatePublishedTemplate(boolean published);

    List<UserUpdate> getUserUpdates(ResourceId userId);

    void enqueueMail(Mail mail);


}
