/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.spi;

import org.activityinfo.model.account.*;
import org.activityinfo.model.analysis.Analysis;
import org.activityinfo.model.analysis.Report;
import org.activityinfo.model.database.AddDatabaseError;
import org.activityinfo.model.database.DatabaseHeader;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.job.JobRequest;
import org.activityinfo.model.job.JobStatus;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.resource.UploadAttachmentResult;
import org.activityinfo.model.type.attachment.AttachmentRef;
import org.activityinfo.promise.Result;
import org.glassfish.jersey.media.multipart.FormDataBodyPart;

import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.io.IOException;
import java.nio.channels.ReadableByteChannel;
import java.time.Instant;
import java.util.*;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 * PlatformStorage encapsulates all platform-specific functionality, such as storage, caching, and the
 * execution of long-running batch jobs.
 *
 * It offers no validation or permission enforcement, and simply executes the queries
 * and modifications requested.
 *
 * This in principle should allow ActivityInfo to run in very different environments, from our
 * SaaS offering in Google Cloud, to a single node
 *
 * All database state mutations must be executed within a transaction (for example {@link PlatformStorage#updateDatabase(ResourceId, Function, Function)}).
 * Any and all methods executed as part of a transaction are guaranteed to be atomic, applied to a consistent
 * version of the database, and return data from a consistent version of the database at the time of execution.
 *
 * Due to varying storage implementations, the store cannot be assumed to have strong consistency outside of
 * transactional methods. Some of the query methods defined on this interface which cannot be performed in a transaction
 * (see {@link PlatformStorage#queryOwnedOrSharedDatabases}) may therefore return data which does not represent an
 * entirely consistent version of the database at the time of execution, as updates may have been applied by other
 * requests during execution of the query.
 */
public interface PlatformStorage {


    /**
     * Starts a transaction to add a new database, if a database with this id does not already exist.
     *
     */
    Result<VisibleDatabaseTree, AddDatabaseError> addDatabase(DatabaseHeader newDatabase, Consumer<DatabaseTransaction> template);

    /**
     * Recovers the database if possible, and if the given user was the previous owner.
     */
    <T> T recoverDatabase(ResourceId databaseId, Function<RecoverDatabaseTransaction, T> mutator, Supplier<T> notFoundHandler);

    /**
     * <p>Query the databases that the given user owns or which have been explicitly shared with the user.</p>
     *
     * <p>This method provides only eventual consistency. It may not (yet) take time for the results of this query
     * to reflect a successfully completed transaction started with {@link #updateDatabase(ResourceId, Function, Function)} } or
     * {@link #addDatabase(DatabaseHeader, Consumer)}.</p>
     *
     * @param userId the system-assigned id of the user.
     * @return A list of DatabaseHeaders for databases the user owns, or has been assigned to by another user.
     */
    List<DatabaseHeader> queryOwnedOrSharedDatabases(ResourceId userId);

    /**
     * Starts a read-only transaction.
     *
     * @param user the authenticated user on whose behalf this query is being executed. This is used only
     *             as a hint to the implementation to improve performance with pre-fetching.
     * @param transactor a function which receives the transaction. This function may be invoked multiple times if the
     *                   the underlying datastore uses optimistic concurrency and the first attempt fails.
     * @param <T> the type of the transactor's result
     * @return the transactor's result.
     */
    default <T> T query(AuthenticatedUser user, Function<QueryTransaction, T> transactor) {
        return query(Optional.of(user), transactor);
    }

    /**
     * Starts a read-only transaction.
     *
     * @param user the authenticated user on whose behalf this query is being executed, if one is present. This is used only
     *             as a hint to the implementation to improve performance with pre-fetching.
     * @param transactor a function which receives the transaction. This function may be invoked multiple times if the
     *                   the underlying datastore uses optimistic concurrency and the first attempt fails.
     * @param <T> the type of the transactor's result
     * @return the transactor's result.
     */
    <T> T query(Optional<AuthenticatedUser> user, Function<QueryTransaction, T> transactor);

    <T> T updateDatabase(ResourceId databaseId,
                                       Function<DatabaseTransaction, T> mutator,
                                       Function<DatabaseTransactionError, T> errorHandler);


    default <T> T updateDatabase(ResourceId databaseId, DatabaseOperation<T> operation) {
        return this.updateDatabase(databaseId,
            tx -> operation.update(tx),
            error -> operation.handleError(error));
    }


    /**
     * Starts a new transaction to update an existing user, identified by the user's email. This transaction must be
     * isolated from any other updates to the user with serializable consistency.
     *
     * @param userEmail the user's email
     * @param mutator a function which accepts the newly-started transaction, if the user exists.
     *                 The underlying storage may use optimistic
     *                concurrency to ensure consistency, so the {@code mutator} function may be invoked more than once.
     * @param errorHandler a function which handles an error condition, such as the possiblity that no user exists
     *                     with this email address.
     * @param <T> the type of the result of the {@code mutator} and {@code errorHandler} functions.
     * @return the result of the {@code mutator} or the {@code errorHandler} function.
     */
    <T> T updateUser(String userEmail, Function<UpdateUserTransaction, T> mutator, Function<UpdateUserError, T> errorHandler);

    /**
     * Starts a new transaction to update an existing user, identified by an invitation token. This transaction must be
     * isolated from any other updates to the user with serializable consistency.
     *
     * @param <T> the type of the result of the {@code mutator} and {@code errorHandler} functions.
     * @param token the invitation token
     * @param mutator a function which accepts the newly-started transaction, if the user exists.
     *                 The underlying storage may use optimistic
     *                concurrency to ensure consistency, so the {@code mutator} function may be invoked more than once.
     * @param errorHandler a function which supplies a value if no such invitation exists.
     * @return the result of the {@code mutator} or the {@code errorHandler} function.
     */
    <T> T updateInvitedUser(String token, Function<UpdateInvitedUserTransaction, T> mutator, Supplier<T> errorHandler);

    /**
     * Starts a new transaction to update an existing user, identified by the user's id. This transaction must be
     * isolated from any other updates to the user with serializable consistency.
     *
     * @param userId the user's platform-assigned id
     * @param mutator a function which accepts the newly-started transaction, if the user exists.
     *                 The underlying storage may use optimistic
     *                concurrency to ensure consistency, so the {@code mutator} function may be invoked more than once.
     *
     * @param errorHandler a function which handles an error condition, such as the possiblity that no user exists
     *                     with this email address.
     *
     * @param <T> the type of the result of the {@code mutator} and {@code errorHandler} functions.
     * @return the result of the {@code mutator} or the {@code errorHandler} function.
     */
    <T> T updateUserById(ResourceId userId, Function<UpdateUserTransaction, T> mutator, Function<UpdateUserError, T> errorHandler);

    /**
     * Starts a new transaction to update an existing user with the given {@code email}, or if no such user yet exists,
     * add a user with this email to the database. This transaction must be
     * isolated from any other updates to the user with serializable consistency.
     *
     * @param email the user's email
     * @param mutator a function which accepts the newly-started transaction, if the user exists.
     *                 The underlying storage may use optimistic
     *                concurrency to ensure consistency, so the {@code mutator} function may be invoked more than once.
     *
     * @param <T> the type of the result of the {@code mutator} function.
     *
     * @return the result of the {@code mutator} or the {@code errorHandler} function.
     */

    <T> T updateOrAddUser(String email, Function<UpdateUserTransaction, T> mutator);

    /**
     * Queries the storage for an exising user, by their system-assigned id.
     *
     * <p>The result of this method must reflect the effects of any successfully-completed transactions initiated
     * with {@link #updateUser(String, Function, Function)} or {@link #updateOrAddUser(String, Function)}</p>
     * @param userId the user's system-assigned id
     * @return the current state of the user's account, if an account with this id exists
     */
    Optional<UserAccount> findUserById(ResourceId userId);

    /**
     * Retrieves a user by the randomly-generated password reset token associated with their account.
     *
     * <p>This method should not check whether the token has expired or otherwise filter the results.</p>
     * @param token a secret token previously associated with the user's account
     * @return the user associated with this token, if one exists.
     */
    Optional<UserAccount> findUserByPasswordResetToken(String token);

    /**
     * Retrieves a user invitation using the invitation's secret token.
     */
    Optional<Invitation> getInvite(String token);

    /**
     * Queries the storage for an exising user with the given {@code email}
     *
     * <p>The result of this method must reflect the effects of any successfully-completed transactions initiated
     * with {@link #updateUser(String, Function, Function)} or {@link #updateOrAddUser(String, Function)}</p>
     * @param email the user's email
     * @return the current state of the user's account, if an account with this id exists
     */
    Optional<UserAccount> findUserByEmail(String email);

    /**
     * Queries the storage for the users with the given system-assigned ids.
     *
     * <p>The result of this method must reflect the effects of any successfully-completed transactions initiated
     * with {@link #updateUser(String, Function, Function)} or {@link #updateOrAddUser(String, Function)}.</p>
     *
     * @param users the set of user ids to retrieve
     * @return a map with all the
     */
    Map<ResourceId, UserAccount> queryUsers(Set<ResourceId> users);

    /**
     * Queries the storage for any database with names that match the given {@code searchString}.
     *
     * <p>Search results are required only to be <em>eventually-consistent</em>, they are not guaranteed to reflect
     * the effect of a successfully completed transaction started by {@link #addDatabase(DatabaseHeader, Consumer)}} or
     * {@link #updateDatabase(ResourceId, Function, Function)}.</p>
     *
     * <p>The implementation should not apply permission logic or otherwise filter the results.</p>
     *
     * @param maxResults the maximum number of results to return
     * @return an un-ordered list of databases matching the search string.
     */
    Iterable<DatabaseHeader> searchDatabases(String searchString, int maxResults);

    /**
     * Start a new transaction to update records in the given the forms.
     *
     * <p>The transaction provides SERIALIZABLE consistency for records within the requested forms. Database
     * metadata and form metadata for other forms are gauranteed only to have repeatable-read consistency.</p>
     *
     * @param <T> the return type of the mutator function
     * @param user the user on whose behalf this transaction is executed. Used only as a hint by the implementation to
     *             improve performance with pre-fetching.
     * @param formIds the forms whose records will be updated during this transaction. The transaction may include
     *                a maximum of 20 forms.
     * @param limitTries the maximum number of attempts to make, if this datastore uses optimistic concurrency.
     * @param mutator the function to receive the new transaction
     * @return the mutator function's result
     */
    <T> T updateRecords(Optional<AuthenticatedUser> user, Collection<ResourceId> formIds, int limitTries, Function<RecordTransaction, T> mutator);

    /**
     * Start a new transaction to transfer ownership of a database from one user to another.
     *
     * @param mutator the function that will receive the transaction. Can be executed multiple times.
     * @param <T> the return type of the {@code mutator} function
     * @return the result of the {@code mutator} function
     */
    <T> T transfer(Function<TransferTransaction, T> mutator);

    /**
     * Start a new transaction to update the billing account identified by {@code id}, if one exists.
     * @param id the id of the billing account to update
     * @param mutator the function that will receive the transaction and perform the changes. May be invoked more than once.
     * @param noBillingAccountResult the value to return if no billing account with the given id exists.
     * @param <T> the return type of the {@code mutator} function
     * @return the result of {@code mutator} function, or {@code noBillingAccountResult} if no such billing account with
     * the given {@code id} exists.
     */
    <T> T updateBillingAccount(long id, Function<UpdateBillingAccountTransaction, T> mutator, Supplier<T> noBillingAccountResult);


    /**
     * Publishes a usage event for the given user.
     *
     * <p>Implementations are not required to implement this method, usage analytics are optional.</p>
     */
    void publishEvent(ResourceId userId, Action action);

    /**
     * Publishes a usage event for the given user and billing account
     *
     * <p>Implementations are not required to implement this method, usage analytics are optional.</p>
     */
    void publishEvent(ResourceId userId, long billingAccountId, Action action);


    /**
     * Starts a long-running batch job on behalf of the user.
     *
     * @param user the user executing the job
     * @param request the job request
     * @return the initial {@code JobStatus}
     */
    JobStatus startUserJob(AuthenticatedUser user, JobRequest request);

    /**
     * Retrieves the status for a user's job, if a record exists.
     */
    Optional<JobStatus> getJobStatus(String jobId);

    /**
     * Returns a handle that can be used to write and later serve a file for a batch job.
     *
     * @param jobId the id of the associated job
     * @param filename the filename. Must be safe to use <em>without</em> URL encoding.
     * @param mimeType the mimetype of the file.
     */
    GeneratedResource createExport(String jobId, String filename, String mimeType);

    /**
     * Serves the exported file previously created with {@link #createExport(String, String, String)}}
     */
    Response serveExport(String jobId, String exportId);

    /**
     * Retrieves the {@link Report} with the given {@code reportId}, if one exists.
     */
    Optional<Report> getReport(ResourceId reportId);

    /**
     * Retrieves the {@link Analysis} with the given {@code reportId} and {@code analysisId}, if one exists.
     */
    Optional<Analysis> getAnalysis(ResourceId reportId, ResourceId analysisId);

    /**
     * Starts a transaction to add or update a report with the given {@code reportId}. If a report with this id does not
     * exist, it will be added.
     *
     * @param user the user adding or updating the report
     * @param reportId the id of the report to add or update
     * @param mutator a function which will receive the transaction. May be invoked multiple times.
     * @param <T> the return type of the {@code mutator}
     * @return the result of the {@code mutator}
     */
    <T> T updateOrAddReport(AuthenticatedUser user, ResourceId reportId, Function<UpdateReportTransaction, T> mutator);

    /**
     * Stores an uploaded attachment.
     */
    UploadAttachmentResult storeAttachment(AttachmentRef attachmentRef, FormDataBodyPart file);

    /**
     * Serves an attachment previously stored with {@link #storeAttachment(AttachmentRef, FormDataBodyPart)}
     */
    Response serveAttachment(AttachmentRef attachmentRef, String mimeType);

    /**
     * Retrieves a handle for an attachment that can be used to read the contents of an attachment previously stored
     * using the {@link #storeAttachment(AttachmentRef, FormDataBodyPart)} method.
     */
    Optional<AttachmentSource> getAttachment(AttachmentRef attachmentRef);


    /**
     * Starts a new session for a user.
     * 
     * @param token a secret token identifying the session
     * @param userId the user's id
     * @param email the user's email
     * @param method the authentication method
     * @param headers the http headers associated with the start of this request.
     * @param expireTime the moment at which this session should expire.
     */
    void addSession(String token, ResourceId userId, String email, AuthMethod method, HttpHeaders headers, Instant expireTime);

    /**
     * Tries to authenticate a user using either a session token, or a personal API token.
     */
    Optional<AuthenticatedUser> authenticateUser(String token);

    /**
     * Retrieves a list of all users who have a billing role in the billing account with the given
     * {@code billingAccountId}
     */
    Iterable<UserAccount> queryBillingUsers(long billingAccountId);

    /**
     * Retrieves the {@link DomainPolicy} for the given domain.
     */
    DomainPolicy getDomainPolicy(String domain);

    /**
     * Updates or creates a domain policy for the given domain.
     * @param domain the name of the domain, in lower case.
     * @param mutator a function which will receive the transaction. May be invoked multiple times.
     * @param <T> the return type of the mutator.
     * @return the result of the mutator function.
     */
    <T> T updateDomainPolicy(String domain, Function<DomainPolicyTransaction, T> mutator);


    Iterable<DatabaseHeader> queryBillingAccountDatabases(long billingAccountId);

    <T> T updateCollectionLink(ResourceId formId, String linkId, Function<CollectionLinkTransaction, T> updater, Function<CollectionLinkTransactionError, T> errorHandler);

    RequestGeoLocation geoLocateRequest(HttpHeaders headers);

    void addOpenIdRequest(OpenIdRequest request);

    Optional<OpenIdRequest> getOpenIdRequestByState(String state);

    ClientIpAddressProfile profileIpAddress(ClientIpAddress address);

    MailResult sendMail(Mail mail);

    void enqueueMail(Mail mail);

    void addApiToken(String tokenId, AuthenticatedUser authUser, String token, Instant timeAdded, String label, AuthenticationScope scope);

    boolean deleteToken(AuthenticatedUser authUser, String tokenId);

    List<ApiToken> getApiTokens(AuthenticatedUser user);

    ReadableByteChannel openUserUpload(ResourceId userId, String uploadId) throws IOException;

    /**
     * Triggers an alert for the system administrator.
     */
    void alert(String message);

    ServingDomain getServingDomain(UriInfo requestUriInfo);

    Optional<Avatar> getUserAvatar(ResourceId userId);

    Optional<String> getCachedPublishedReport(String reportId, String analysisId);

    void cachePublishedReport(String reportId, String analysisId, String serializedJson);
}
