/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.spi;

import org.activityinfo.model.account.UserAccount;
import org.activityinfo.model.analysis.Analysis;
import org.activityinfo.model.analysis.Report;
import org.activityinfo.model.analysis.ReportHeader;
import org.activityinfo.model.collectionLink.CollectionLink;
import org.activityinfo.model.database.*;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormRecord;
import org.activityinfo.model.query.ColumnSet;
import org.activityinfo.model.query.QueryModel;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.sync.RecordSyncSet;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.promise.Maybe;

import javax.ws.rs.core.Response;
import java.util.*;
import java.util.function.Predicate;

/**
 * A read-only transaction.
 */
public interface QueryTransaction {

    /**
     * The list of databases which the user owns or that have been explicitly shared with the user.
     *
     * <p>Should not return public databases to which the user has not been explicitly invited.</p>
     */
    List<DatabaseHeader> getOwnedOrSharedDatabases();

    default Map<ResourceId, Maybe<VisibleDatabaseTree>> getVisibleDatabaseTrees(Collection<ResourceId> databaseIds) {
        Map<ResourceId, Maybe<VisibleDatabaseTree>> map = new HashMap<>();
        for (ResourceId databaseId : databaseIds) {
            map.put(databaseId, getVisibleDatabaseTree(databaseId));
        }
        return map;
    }

    Maybe<VisibleDatabaseTree> getVisibleDatabaseTree(ResourceId databaseId);

    Iterable<DatabaseUserGrantSet> getDatabaseUsers(ResourceId databaseId);

    Optional<FormClass> getFormClass(ResourceId formId);

    Optional<FormRecord> getRecord(RecordRef ref);

    Iterable<FormRecord> getRecords(ResourceId formId);

    /**
     * Returns all versions of a given record, or an empty list if the form or record does not exist.
     */
    Iterable<RecordVersion> getRecordVersions(RecordRef ref);

    RecordSyncSet getVersionRange(ResourceId formId, long localVersion, long toVersion, Predicate<String> visibilityPredicate, Optional<String> cursor);

    Map<ResourceId, UserAccount> queryUsers(Set<ResourceId> userSet);

    Optional<FormVersion> getUserFormVersion(ResourceId databaseId);

    /**
     * Returns the requested version of the form, if it exists and has not been deleted.
     */
    Optional<FormVersion> getFormVersion(ResourceId formId);

    Optional<DatabaseUserGrantSet> getDatabaseUser(ResourceId databaseId, ResourceId userId);

    ColumnSet query(QueryModel queryModel, FormSupervisor supervisor);

    Optional<DatabaseTree> getDatabaseTree(ResourceId databaseId);

    Optional<Report> getReport(ResourceId id);

    Optional<Analysis> getAnalysis(ResourceId reportId, ResourceId analysisId);

    /**
     * Retrieves all collection links for a given {@code formId}
     */
    List<CollectionLink> getCollectionLinks(ResourceId formId);

    AuditEventList audit(VisibleDatabaseTree tree, AuditRequest request);


    /**
     * Directly serves an InitialSyncSet for the given form. The implementation
     * may choose to serve a cached version of this response that might not be completely
     * up to date. The InitialSyncSet must be consistent with a specific Form version however.
     */
    Response serveInitialSyncSet(ResourceId formId, Predicate<String> visibilityPredicate, String filterHash);

    List<ReportHeader> getOwnedOrSharedReports();

    /**
     * "Stages" a form, before it is added to a database.
     *
     * <p>This method <strong>SHOULD NOT</strong> add a form to a databases' tree, and should ONLY be used
     * when significant work needs to be done before the form should be visible to the user.</p>
     *
     * <p>Implementations <strong>MAY</strong> stage this form outside of the current transaction.</p>
     *
     * @param schema the schema to add
     * @param records Records to add
     */
    void stageForm(FormClass schema, Iterable<TypedRecordUpdate> records);
}
