/*
 * ActivityInfo 4.0
 * Copyright © 2014-2021 BeDataDriven Groep B.V. - All Rights Reserved
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited
 * Proprietary and confidential
 */
package org.activityinfo.store.spi;

import org.activityinfo.json.JsonValue;
import org.activityinfo.model.account.Action;
import org.activityinfo.model.collectionLink.CollectionLink;
import org.activityinfo.model.database.VisibleDatabaseTree;
import org.activityinfo.model.form.FormClass;
import org.activityinfo.model.form.FormRecord;
import org.activityinfo.model.resource.RecoverRecordResult;
import org.activityinfo.model.resource.ResourceId;
import org.activityinfo.model.resource.UpdateRecordResult;
import org.activityinfo.model.type.RecordRef;
import org.activityinfo.promise.Maybe;

import java.util.Optional;

public interface RecordTransaction {

    /**
     * Finds the database to which a form belongs, if it exists.
     */
    Optional<ResourceId> getFormDatabaseId(ResourceId formId);

    /**
     * Fetches a form's schema, if the form exists.
     */
    Optional<FormClass> getFormClass(ResourceId formId);

    /**
     * Retrieves the {@link VisibleDatabaseTree} for the author of this transaction, if one is present.
     */
    Maybe<VisibleDatabaseTree> getVisibleDatabaseTree(ResourceId databaseId);

    /**
     * Retrieves the current version of the referenced record, if it exists. The return value should
     * reflect any changes made during this transaction.
     */
    Optional<FormRecord> getRecord(RecordRef recordRef);

    void publishAction(Action updateRecord, ResourceId formId);

    /**
     * Store a validated record update.
     *
     * This should have the following effects:
     * <ul>
     *     <li>The new record should be added to the data storage</li>
     *     <li>If this form's version has not already been incremented during this transaction,
     *     increment it now.</li>
     *     <li>A record snapshot should be stored with the new version</li>
     * </ul>
     *
     * The new record should be visible to callers of {@link #getRecord(RecordRef)}
     */
    UpdateRecordResult addRecord(TypedRecordUpdate update);

    UpdateRecordResult updateRecord(TypedRecordUpdate update);

    UpdateRecordResult deleteRecord(RecordRef recordRef);

    RecoverRecordResult recoverDeletedRecord(ResourceId userId, RecordRef recordRef);

    /**
     * Retrieves the next serial number for the given form, field, and prefix, and increments
     * the version number.
     */
    int nextSerialNumber(ResourceId formId, ResourceId fieldId, String prefix);

    /**
     * Updates the *current* number of a serial number sequence.
     */
    void updateSerialNumberSequence(ResourceId formId, ResourceId fieldId, String prefix, int number);

    Optional<CollectionLink> getCollectionLink(ResourceId formId, String linkId);

    /**
     * Updates the checkpoint data for a long-running job, within this transaction.
     */
    void updateJobCheckpoint(String jobId, JsonValue json);

    /**
     * Retrieves the checkpoint data for a long-running job.
     */
    Optional<JsonValue> getJobCheckpoint(String jobId);
}
